
// DIRE includes.
#include "include/SplittingLibrary.h"
#include "include/DireTimes.h"
#include "include/DireSpace.h"
#include "include/WeightContainer.h"

// Pythia includes.
#include "Pythia8/Pythia.h"

using namespace Pythia8;

// Use UserHooks as trick to initialse the Weight container.

class WeightHooks : public UserHooks {

public:  

  // Constructor and destructor.
  WeightHooks(WeightContainer* weightsIn) : weights(weightsIn) {}
 ~WeightHooks() {}

  bool canVetoProcessLevel() { return true;}
  bool doVetoProcessLevel( Event&) {
    weights->init();
    return false;
  }

  WeightContainer* weights;

};

//==========================================================================

// Example main programm to illustrate merging.

int main( int argc, char* argv[] ){

  // Check that correct number of command-line arguments
  if (argc != 2) {
    cerr << " Unexpected number of command-line arguments ("<<argc<<"). \n"
         << " You are expected to provide the arguments" << endl
         << " 1. Input file for settings" << endl
         << " Program stopped. " << endl;
    return 1;
  }

  Pythia pythia;

  // Jet finder for some of the histograms.
  SlowJet* slowJet = new SlowJet(-1, 1.0, 20., 4.4, 2, 2, NULL, false);
  Hist njets("njets",7,-0.5,6.5);

  // Teach Pythia the additional DIRE input settings.
  pythia.settings.addFlag("ShowerPDF:usePDFalphas",false);
  pythia.settings.addFlag("ShowerPDF:usePDFmasses",true);
  pythia.settings.addFlag("ShowerPDF:useSummedPDF",true);
  pythia.settings.addFlag("DireSpace:useGlobalMapIF",false);
  pythia.settings.addFlag("DireSpace:forceMassiveMap",false);
  pythia.settings.addMode("DireTimes:nFinalMax",-10,true,false,-10,10000000);
  pythia.settings.addMode("DireSpace:nFinalMax",-10,true,false,-10,10000000);

  // Construct showers.
  WeightContainer weights;
  DireTimes times(&pythia);
  DireSpace space(&pythia);
  DireTimes timesDec(&pythia);

  // Initialise weight book-keeping.
  UserHooks* wts = new WeightHooks(&weights);
  times.setWeightContainerPtr(&weights);
  space.setWeightContainerPtr(&weights);
  timesDec.setWeightContainerPtr(&weights);

  // Feed new DIRE showers to Pythia.
  pythia.setShowerPtr( &timesDec, &times, &space);
  pythia.setUserHooksPtr(wts);

  // Read Pythia settings from file and initialise.
  pythia.readFile(argv[1]);
  pythia.init();

  // Initialise library of splitting functions.
  SplittingLibrary* splittings = new SplittingLibrary();

  // Reinitialise showers to ensure that pointers are
  // correctly set.
  times.reinitPtr(&pythia.info,
    &pythia.settings,
    &pythia.particleData,
    &pythia.rndm,
    &pythia.partonSystems,
     wts, NULL,
     splittings);
  space.reinitPtr(&pythia.info,
    &pythia.settings,
    &pythia.particleData,
    &pythia.rndm,
    &pythia.partonSystems,
     wts, NULL,
     splittings);
  timesDec.reinitPtr(&pythia.info,
    &pythia.settings,
    &pythia.particleData,
    &pythia.rndm,
    &pythia.partonSystems,
     wts, NULL,
     splittings);

  // Initialise splitting function library here so that beam pointers
  // are already correctly initialised.
  splittings->init(&pythia.settings, &pythia.particleData, &pythia.rndm,
    space.beamAPtr, space.beamBPtr);

  // Feed the splitting functions to the showers.
  splittings->setTimesPtr(&times);
  splittings->setSpacePtr(&space);

  // Reset Pythia masses if necessary.
  if ( pythia.settings.flag("ShowerPDF:usePDFmasses") ) {
    for (int i=1; i <= 5; ++i) {
      double mPDF = (abs(space.beamAPtr->id()) > 30)
                  ? space.beamAPtr->mQuarkPDF(i)
                  : (abs(space.beamBPtr->id()) > 30) 
                    ? space.beamBPtr->mQuarkPDF(i) : -1.0;
      if (mPDF > -1.) {
        stringstream resetMass;
        resetMass << i << ":m0 = " << mPDF;
        pythia.readString(resetMass.str());
      }
    }
  }

  // Start generation loop
  int nEvent = pythia.settings.mode("Main:numberOfEvents");
  for( int iEvent=0; iEvent<nEvent; ++iEvent ){

    // Generate next event
    if( !pythia.next() ) {
      if( pythia.info.atEndOfFile() )
        break;
      else continue;
    }

    // Get event weight(s).
    double evtweight         = pythia.info.weight();

    // Do not print zero-weight events.
    if ( evtweight == 0. ) continue;

    // Retrieve the shower weight.
    weights.calcWeight(0.);
    weights.reset();
    double wt = weights.getShowerWeight();
    evtweight *= wt;

    // Construct input for the jet algorithm.
    Event jetInput;
    jetInput.init("jet input",&pythia.particleData);
    jetInput.clear();
    for (int i =0; i < pythia.event.size(); ++i)
      if ( pythia.event[i].isFinal()
        && (pythia.event[i].colType() != 0 || pythia.event[i].isHadron()) )
        jetInput.append(pythia.event[i]);
    // Run jet algorithm.
    slowJet->analyze(jetInput);
    // Histogram number of jets.
    int nJet = slowJet->sizeJet();
    for (int i =0; i <= nJet; ++i) njets.fill(i, evtweight); 

  } // end loop over events to generate

  // print cross section, errors
  pythia.stat();

  // Normalise histograms
  njets *= pythia.info.sigmaGen() / pythia.info.nAccepted();

  // Print histograms.
  cout << njets;

  // Done
  return 0;

}
