
// DIRE includes.
#include "include/SplittingLibrary.h"
#include "include/DireTimes.h"
#include "include/DireSpace.h"
#include "include/WeightContainer.h"

// Pythia includes.
#include "Pythia8/Pythia.h"
#include "Pythia8Plugins/HepMC2.h"
#include <unistd.h>

using namespace Pythia8;

// Use UserHooks as trick to initialse the Weight container.

class WeightHooks : public UserHooks {

public:  

  // Constructor and destructor.
  WeightHooks(WeightContainer* weightsIn) : weights(weightsIn) {}
 ~WeightHooks() {}

  bool canVetoProcessLevel() { return true;}
  bool doVetoProcessLevel( Event&) {
    weights->init();
    return false;
  }

  WeightContainer* weights;

};
//==========================================================================

// Example main programm to illustrate merging.

int main( int argc, char* argv[] ){

  // Check that correct number of command-line arguments
  if (argc != 3) {
    cerr << " Unexpected number of command-line arguments ("<<argc<<"). \n"
         << " You are expected to provide the arguments" << endl
         << " 1. Input file for settings" << endl
         << " 2. Name of the input LHE file (with path), up to the '_tree'"
         << " identifier" << endl
         << " 3. Output hepmc file name" << endl
         << " Program stopped. " << endl;
    return 1;
  }

  Pythia pythia;

  // Teach Pythia the additional DIRE input settings.
  pythia.settings.addFlag("ShowerPDF:usePDFalphas",false);
  pythia.settings.addFlag("ShowerPDF:usePDFmasses",true);
  pythia.settings.addFlag("ShowerPDF:useSummedPDF",true);
  pythia.settings.addFlag("DireSpace:useGlobalMapIF",false);
  pythia.settings.addFlag("DireSpace:forceMassiveMap",false);
  pythia.settings.addMode("DireTimes:nFinalMax",-1,true,false,-1,10000000);
  pythia.settings.addMode("DireSpace:nFinalMax",-1,true,false,-1,10000000);

  pythia.settings.forceParm("SpaceShower:pT0Ref",0.0);

  // Construct showers.
  WeightContainer weights;
  DireTimes times(&pythia);
  DireSpace space(&pythia);
  DireTimes timesDec(&pythia);

  // Initialise weight book-keeping.
  UserHooks* wts = new WeightHooks(&weights);
  times.setWeightContainerPtr(&weights);
  space.setWeightContainerPtr(&weights);
  timesDec.setWeightContainerPtr(&weights);

  // Feed new DIRE showers to Pythia.
  pythia.setShowerPtr( &timesDec, &times, &space);
  pythia.setUserHooksPtr(wts);

  // Read Pythia settings from file.
  pythia.readFile(argv[1]);

  int nEventEst = pythia.settings.mode("Main:numberOfEvents");

  // Switch off all showering and MPI when extimating the cross section.
  bool fsr = pythia.flag("PartonLevel:FSR");
  bool isr = pythia.flag("PartonLevel:ISR");
  bool mpi = pythia.flag("PartonLevel:MPI");
  bool had = pythia.flag("HadronLevel:all");
  pythia.settings.flag("PartonLevel:FSR",false);
  pythia.settings.flag("PartonLevel:ISR",false);
  pythia.settings.flag("PartonLevel:MPI",false);
  pythia.settings.flag("HadronLevel:all",false);

  pythia.init();

  // Initialise library of splitting functions.
  SplittingLibrary* splittings = new SplittingLibrary();

  // Reinitialise showers to ensure that pointers are
  // correctly set.
  times.reinitPtr(&pythia.info,
    &pythia.settings,
    &pythia.particleData,
    &pythia.rndm,
    &pythia.partonSystems,
     wts, NULL,
     splittings);
  space.reinitPtr(&pythia.info,
    &pythia.settings,
    &pythia.particleData,
    &pythia.rndm,
    &pythia.partonSystems,
     wts, NULL,
     splittings);
  timesDec.reinitPtr(&pythia.info,
    &pythia.settings,
    &pythia.particleData,
    &pythia.rndm,
    &pythia.partonSystems,
     wts, NULL,
     splittings);

  // Initialise splitting function library here so that beam pointers
  // are already correctly initialised.
  splittings->init(&pythia.settings, &pythia.particleData, &pythia.rndm,
    space.beamAPtr, space.beamBPtr);

  // Feed the splitting functions to the showers.
  splittings->setTimesPtr(&times);
  splittings->setSpacePtr(&space);

  // Reset Pythia masses if necessary.
  if ( pythia.settings.flag("ShowerPDF:usePDFmasses") ) {
    for (int i=1; i <= 5; ++i) {
      double mPDF = (abs(space.beamAPtr->id()) > 30)
                  ? space.beamAPtr->mQuarkPDF(i)
                  : (abs(space.beamBPtr->id()) > 30) 
                    ? space.beamBPtr->mQuarkPDF(i) : -1.0;
      if (mPDF > -1.) {
        stringstream resetMass;
        resetMass << i << ":m0 = " << mPDF;
        pythia.readString(resetMass.str());
      }
    }
  }

  // Cross section estimate run.
  double sumSH = 0.;
  double nAcceptSH = 0.;
  for( int iEvent=0; iEvent<nEventEst; ++iEvent ){
    // Generate next event
    if( !pythia.next() ) {
      if( pythia.info.atEndOfFile() )
        break;
      else continue;
    }
    sumSH     += pythia.info.weight();
    string trials = pythia.info.getEventAttribute("trials",true);
    if (trials != "") nAcceptSH += atof(trials.c_str());
  }
  pythia.stat();
  double xs = pythia.info.sigmaGen();
  int nA    = pythia.info.nAccepted();

  // Read Pythia settings from file.
  pythia.readFile(argv[1]);

  // Histogram the weight.
  Hist histWT("weight",10000,-500.,500.);

  // Interface for conversion from Pythia8::Event to HepMC one. 
  HepMC::Pythia8ToHepMC ToHepMC;
  // Specify file where HepMC events will be stored.
  HepMC::IO_GenEvent ascii_io(argv[2], std::ios::out);
  // Switch off warnings for parton-level events.
  ToHepMC.set_print_inconsistency(false);
  ToHepMC.set_free_parton_warnings(false);
  // Do not store cross section information, as this will be done manually.
  ToHepMC.set_store_pdf(false);
  ToHepMC.set_store_proc(false);
  ToHepMC.set_store_xsec(false);

  int nEvent = pythia.settings.mode("Main:numberOfEvents");

  // Cross section an error.
  double sigmaTotal  = 0.;
  double errorTotal  = 0.;

  cout << endl << endl << endl;
  cout << "Start generating events" << endl;

  double sigmaSample = 0., errorSample = 0.;

  // Switch showering and multiple interaction back on.
  pythia.settings.flag("PartonLevel:FSR",fsr);
  pythia.settings.flag("PartonLevel:ISR",isr);
  pythia.settings.flag("HadronLevel:all",had);
  pythia.settings.flag("PartonLevel:MPI",mpi);

  pythia.init();

  double wmax =-1e15;
  double wmin = 1e15;
  double sumwt = 0.;
  double sumwtsq = 0.;

  // Start generation loop
  for( int iEvent=0; iEvent<nEvent; ++iEvent ){

    // Generate next event
    if( !pythia.next() ) {
      if( pythia.info.atEndOfFile() )
        break;
      else continue;
    }

    // Get event weight(s).
    double evtweight         = pythia.info.weight();

    // Do not print zero-weight events.
    if ( evtweight == 0. ) continue;

    // Retrieve the shower weight.
    weights.calcWeight(0.);
    weights.reset();
    double wt = weights.getShowerWeight();

    evtweight *= wt;

    wmin = min(wmin,wt);
    wmax = max(wmax,wt);
    sumwt += wt;
    sumwtsq+=pow2(wt);
    histWT.fill( wt, 1.0);

    // Construct new empty HepMC event.
    HepMC::GenEvent* hepmcevt = new HepMC::GenEvent();

    double normhepmc = xs / double(nA);

    // sherpa weighted events
    if( pythia.info.lhaStrategy() != 3)
      normhepmc = 1. / (1e9*nAcceptSH);

    // Set event weight
    hepmcevt->weights().push_back(evtweight*normhepmc);
    // Fill HepMC event
    ToHepMC.fill_next_event( pythia, hepmcevt );
 
    // Add the weight of the current event to the cross section.
    sigmaTotal  += evtweight*normhepmc;
    sigmaSample += evtweight*normhepmc;
    errorTotal  += pow2(evtweight*normhepmc);
    errorSample += pow2(evtweight*normhepmc);
    // Report cross section to hepmc
    HepMC::GenCrossSection xsec;
    xsec.set_cross_section( sigmaTotal*1e9, pythia.info.sigmaErr()*1e9 );
    hepmcevt->set_cross_section( xsec );
    // Write the HepMC event to file. Done with it.
    ascii_io << hepmcevt;
    delete hepmcevt;

  } // end loop over events to generate

  // print cross section, errors
  pythia.stat();

  cout << "\t Minimal shower weight=" << wmin
       << "\n\t Maximal shower weight=" << wmax
       << "\n\t Mean shower weight=" << sumwt/double(nEvent)
       << "\n\t Variance of shower weight="
       << sqrt(1/double(nEvent)*(sumwtsq - pow(sumwt,2)/double(nEvent)))
       << endl;

  ofstream write;
  // Write histograms to file
  write.open("wt.dat");
  histWT.table(write);
  write.close();

  // Done
  return 0;

}
