
// DIRE includes.
#include "include/SplittingLibrary.h"
#include "include/DireTimes.h"
#include "include/DireSpace.h"
#include "include/WeightContainer.h"

// Pythia includes.
#include "Pythia8/Pythia.h"
#include "Pythia8Plugins/HepMC2.h"
#include <unistd.h>

using namespace Pythia8;

// Use UserHooks as trick to initialse the Weight container.

class WeightHooks : public UserHooks {

public:  

  // Constructor and destructor.
  WeightHooks(WeightContainer* weightsIn) : weights(weightsIn) {}
 ~WeightHooks() {}

  bool canVetoProcessLevel() { return true;}
  bool doVetoProcessLevel( Event&) {
    weights->init();
    return false;
  }

  WeightContainer* weights;

};

//==========================================================================

// Example main programm to illustrate merging.

int main( int argc, char* argv[] ){

  // Check that correct number of command-line arguments
  if (argc != 4) {
    cerr << " Unexpected number of command-line arguments ("<<argc<<"). \n"
         << " You are expected to provide the arguments" << endl
         << " 1. Input file for settings" << endl
         << " 2. Output HepMC file" << endl
         << " Program stopped. " << endl;
    return 1;
  }

  Pythia pythia;

  // New setting to allow processing of multiple input LHEFs.
  pythia.settings.addMode("LHEFInputs:nSubruns",0,true,false,0,100);

  // Teach Pythia the additional DIRE input settings.
  pythia.settings.addFlag("ShowerPDF:usePDFalphas",false);
  pythia.settings.addFlag("ShowerPDF:usePDFmasses",true);
  pythia.settings.addFlag("ShowerPDF:useSummedPDF",true);
  pythia.settings.addFlag("DireSpace:useGlobalMapIF",false);
  pythia.settings.addFlag("DireSpace:forceMassiveMap",false);
  pythia.settings.addMode("DireTimes:nFinalMax",-10,true,false,-1,10000000);
  pythia.settings.addMode("DireSpace:nFinalMax",-10,true,false,-1,10000000);

  // Construct showers.
  WeightContainer weights;
  DireTimes times(&pythia);
  DireSpace space(&pythia);
  DireTimes timesDec(&pythia);

  // Initialise weight book-keeping.
  UserHooks* wts = new WeightHooks(&weights);
  times.setWeightContainerPtr(&weights);
  space.setWeightContainerPtr(&weights);
  timesDec.setWeightContainerPtr(&weights);

  // Feed new DIRE showers to Pythia.
  pythia.setShowerPtr( &timesDec, &times, &space);
  pythia.setUserHooksPtr(wts);

  // Read Pythia settings from file and initialise.
  pythia.readFile(argv[1],0);

  // Check if internal merging should be applied.
  bool doMerge   = !(pythia.settings.word("Merging:Process").compare("void")
    == 0);

  // Interface for conversion from Pythia8::Event to HepMC one. 
  HepMC::Pythia8ToHepMC ToHepMC;
  // Specify file where HepMC events will be stored.
  HepMC::IO_GenEvent ascii_io(argv[2], std::ios::out);
  // Switch off warnings for parton-level events.
  ToHepMC.set_print_inconsistency(false);
  ToHepMC.set_free_parton_warnings(false);
  // Do not store cross section information, as this will be done manually.
  ToHepMC.set_store_pdf(false);
  ToHepMC.set_store_proc(false);
  ToHepMC.set_store_xsec(false);

  // Get number of subruns.
  int nMerge = pythia.mode("LHEFInputs:nSubruns");
  int nEvent = pythia.settings.mode("Main:numberOfEvents");
  double sigmaTotal = 0.0;

  // Histogram the weight.
  Hist histWT("weight",10000,-500.,500.);

  // Loop over subruns with varying number of jets.
  for (int iMerge = 0; iMerge < nMerge; ++iMerge) {

    // Read in name of LHE file for current subrun and initialize.
    pythia.readFile(argv[1], iMerge);
    // Initialise.
    pythia.init();

    double wmax =-1e15;
    double wmin = 1e15;
    double sumwt = 0.;
    double sumwtsq = 0.;

    // Get the inclusive x-section by summing over all process x-sections.
    double xs = 0.;
    for (int i=0; i < pythia.info.nProcessesLHEF(); ++i)
      xs += pythia.info.sigmaLHEF(i);

    // Initialise library of splitting functions.
    SplittingLibrary* splittings = new SplittingLibrary();

    // Reinitialise showers to ensure that pointers are
    // correctly set.
    times.reinitPtr(&pythia.info,
      &pythia.settings,
      &pythia.particleData,
      &pythia.rndm,
      &pythia.partonSystems,
       wts, NULL,
       splittings);
    space.reinitPtr(&pythia.info,
      &pythia.settings,
      &pythia.particleData,
      &pythia.rndm,
      &pythia.partonSystems,
       wts, NULL,
       splittings);
    timesDec.reinitPtr(&pythia.info,
      &pythia.settings,
      &pythia.particleData,
      &pythia.rndm,
      &pythia.partonSystems,
       wts, NULL,
       splittings);

    // Initialise splitting function library here so that beam pointers
    // are already correctly initialised.
    splittings->init(&pythia.settings, &pythia.particleData, &pythia.rndm,
      space.beamAPtr, space.beamBPtr);

    // Feed the splitting functions to the showers.
    splittings->setTimesPtr(&times);
    splittings->setSpacePtr(&space);

    // Reset Pythia masses if necessary.
    if ( pythia.settings.flag("ShowerPDF:usePDFmasses") ) {
      for (int i=1; i <= 5; ++i) {
        double mPDF = (abs(space.beamAPtr->id()) > 30)
                    ? space.beamAPtr->mQuarkPDF(i)
                    : (abs(space.beamBPtr->id()) > 30) 
                      ? space.beamBPtr->mQuarkPDF(i) : -1.0;
        if (mPDF > -1.) {
          stringstream resetMass;
          resetMass << i << ":m0 = " << mPDF;
          pythia.readString(resetMass.str());
        }
      }
    }

    // Start generation loop
    for( int iEvent=0; iEvent<nEvent; ++iEvent ){

      // Generate next event
      if( !pythia.next() ) {
        if( pythia.info.atEndOfFile() )
          break;
        else continue;
      }

      // Get event weight(s).
      double evtweight         = pythia.info.weight();

      // Do not print zero-weight events.
      if ( evtweight == 0. ) continue;

      // Retrieve the shower weight.
      weights.calcWeight(0.);
      weights.reset();
      double wt = weights.getShowerWeight();
      evtweight *= wt;

      wmin = min(wmin,wt);
      wmax = max(wmax,wt);

      sumwt += wt;
      sumwtsq+=pow2(wt);
      histWT.fill( wt, 1.0);

      // Additional PDF/alphaS weight for internal merging.
      if (doMerge) evtweight  *= pythia.info.mergingWeightNLO();

      // Work with weighted (LHA strategy=-4) events.
      double normhepmc = 1.;
      if (abs(pythia.info.lhaStrategy()) == 4)
        normhepmc = 1. / double(1e9*nEvent);
      // Work with unweighted events.
      else
        normhepmc = xs / double(1e9*nEvent);

      // Construct new empty HepMC event.
      HepMC::GenEvent* hepmcevt = new HepMC::GenEvent();
      // Set event weight
      hepmcevt->weights().push_back(evtweight*normhepmc);
      // Fill HepMC event
      ToHepMC.fill_next_event( pythia, hepmcevt );
 
      // Add the weight of the current event to the cross section.
      sigmaTotal  += evtweight*normhepmc;
      // Report cross section to hepmc
      HepMC::GenCrossSection xsec;
      xsec.set_cross_section( sigmaTotal*1e9, pythia.info.sigmaErr()*1e9 );
      hepmcevt->set_cross_section( xsec );
      // Write the HepMC event to file. Done with it.
      ascii_io << hepmcevt;
      delete hepmcevt;

    } // end loop over events to generate

    // print cross section, errors
    pythia.stat();

    cout << "\t Minimal shower weight=" << wmin
       << "\n\t Maximal shower weight=" << wmax
       << "\n\t Mean shower weight=" << sumwt/double(nEvent)
       << "\n\t Variance of shower weight="
       << sqrt(1/double(nEvent)*(sumwtsq - pow(sumwt,2)/double(nEvent)))
       << endl;

  }

  ofstream write;
  // Write histograms to file
  write.open(  (char*)( string(argv[3]) + "wt.dat" ).c_str() );
  histWT.table(write);
  write.close();

  // Done
  return 0;

}
