// DireSpace.h is a part of the PYTHIA event generator.
// Copyright (C) 2014 Torbjorn Sjostrand.
// PYTHIA is licenced under the GNU GPL version 2, see COPYING for details.
// Please respect the MCnet Guidelines, see GUIDELINES for details.

// Header file for the spacelike initial-state showers.
// DireSpaceEnd: radiating dipole end in ISR.
// DireSpace: handles the showering description.

#ifndef Pythia8_DireSpace_H
#define Pythia8_DireSpace_H

#define DIRE_SPACE_VERSION "0.900"

#include "Pythia8/Basics.h"
#include "Pythia8/Pythia.h"
#include "Pythia8/SpaceShower.h"
#include "Pythia8/BeamParticle.h"
#include "Pythia8/Event.h"
#include "Pythia8/Info.h"
#include "Pythia8/ParticleData.h"
#include "Pythia8/PartonSystems.h"
#include "Pythia8/PythiaStdlib.h"
#include "Pythia8/Settings.h"
#include "Pythia8/StandardModel.h"
#include "Pythia8/UserHooks.h"
#include "Pythia8/MergingHooks.h"
#include "Pythia8/WeakShowerMEs.h"
#include "SplittingLibrary.h"
#include "WeightContainer.h"

namespace Pythia8 {

//==========================================================================

// Data on radiating dipole ends, only used inside DireSpace.

class DireSpaceEnd {
  
public:

  // Constructor.
  DireSpaceEnd( int systemIn = 0, int sideIn = 0, int iRadiatorIn = 0,
    int iRecoilerIn = 0, double pTmaxIn = 0., int colTypeIn = 0,
    int chgTypeIn = 0, int weakTypeIn = 0,  int MEtypeIn = 0,
    bool normalRecoilIn = true, int weakPolIn = 0,
    vector<int> iSpectatorIn = vector<int>()) :
    system(systemIn), side(sideIn), iRadiator(iRadiatorIn),
    iRecoiler(iRecoilerIn), pTmax(pTmaxIn), colType(colTypeIn),
    chgType(chgTypeIn), weakType(weakTypeIn), MEtype(MEtypeIn),
    normalRecoil(normalRecoilIn), weakPol(weakPolIn), nBranch(0),
    pT2Old(0.), zOld(0.5), iSpectator(iSpectatorIn) { }
 
  // Store values for trial emission.
  void store( int idDaughterIn, int idMotherIn, int idSisterIn,
    double x1In, double x2In, double m2DipIn, double pT2In, double zIn,
    double xMoIn, double Q2In, double mSisterIn, double m2SisterIn,
    double pT2corrIn) {idDaughter = idDaughterIn; idMother = idMotherIn;
    idSister = idSisterIn; x1 = x1In; x2 = x2In; m2Dip = m2DipIn;
    pT2 = pT2In; z = zIn; xMo = xMoIn; Q2 = Q2In; mSister = mSisterIn;
    m2Sister = m2SisterIn; pT2corr = pT2corrIn;}
 
  // Basic properties related to evolution and matrix element corrections.
  int    system, side, iRadiator, iRecoiler;
  double pTmax;
  int    colType, chgType, weakType, MEtype;
  bool   normalRecoil;
  int    weakPol;

  // Properties specific to current trial emission.
  int    nBranch, idDaughter, idMother, idSister, iFinPol;
  double x1, x2, m2Dip, pT2, z, xMo, Q2, mSister, m2Sister, pT2corr,
         pT2Old, zOld, asymPol;

  // Extended list of recoilers.
  vector<int> iSpectator;

} ;
 
//==========================================================================

// The DireSpace class does spacelike showers.

class DireSpace : public Pythia8::SpaceShower {

public:

  // Constructor.
  DireSpace() {}

  DireSpace(Pythia8::Pythia* pythiaPtr) {
    beamOffset        = 0;
    pTdampFudge       = 0.;
    infoPtr           = &pythiaPtr->info;
    particleDataPtr   = &pythiaPtr->particleData;
    partonSystemsPtr  = &pythiaPtr->partonSystems;
    rndmPtr           = &pythiaPtr->rndm;
    settingsPtr       = &pythiaPtr->settings;
    //userHooksPtr      =  pythiaPtr->userHooksPtr;
    userHooksPtr      = 0;
    mergingHooksPtr   = pythiaPtr->mergingHooksPtr;
    splittingsPtr     = 0;
    weights           = 0;
    printBanner       = true;
  }

  // Destructor.
  virtual ~DireSpace() {}

  // Initialize generation. Possibility to force re-initialization by hand.
  virtual void init(BeamParticle* beamAPtrIn, BeamParticle* beamBPtrIn);

  // Initialize various pointers.
  // (Separated from rest of init since not virtual.)
  void reinitPtr(Info* infoPtrIn, Settings* settingsPtrIn,
       ParticleData* particleDataPtrIn, Rndm* rndmPtrIn,
       PartonSystems* partonSystemsPtrIn, UserHooks* userHooksPtrIn,
       MergingHooks* mergingHooksPtrIn, SplittingLibrary* splittingsPtrIn) {
       infoPtr = infoPtrIn;
       settingsPtr = settingsPtrIn;
       particleDataPtr = particleDataPtrIn;
       rndmPtr = rndmPtrIn;
       partonSystemsPtr = partonSystemsPtrIn;
       userHooksPtr = userHooksPtrIn;
       mergingHooksPtr = mergingHooksPtrIn;
       splittingsPtr = splittingsPtrIn;
  }

  void setWeightContainerPtr(WeightContainer* weightsIn) { weights = weightsIn;}

  // Find whether to limit maximum scale of emissions, and whether to dampen.
  virtual bool limitPTmax( Event& event, double Q2Fac = 0.,
    double Q2Ren = 0.);

  // Potential enhancement factor of pTmax scale for hardest emission.
  virtual double enhancePTmax() const {return pTmaxFudge;}

  // Prepare system for evolution; identify ME.
  virtual void prepare( int iSys, Event& event, bool limitPTmaxIn = true);

  // Update dipole list after each FSR emission.
  // Usage: update( iSys, event).
  virtual void update( int , Event&, bool = false);

  // Select next pT in downwards evolution.
  virtual double pTnext( Event& event, double pTbegAll, double pTendAll,
    int nRadIn = -1, bool = false);

  // Setup branching kinematics.
  virtual bool branch( Event& event);
  bool branch_II( Event& event);
  bool branch_IF( Event& event);

  // Setup splitting kinematics.
  virtual Event branched( const Event& event, int iRadBef, int iRecBef, int idEmt,
    int idMother, double pT2, double z, double RN1, vector<double> aux);
  vector<Vec4> split_II( Event& event, int iRadBef, int iRecBef, double pT2,
    double z, double RN);
  vector<Vec4> split_IF( Event& event, int iRadBef, int iRecBef, double pT2,
    double z, double RN);

  // For new Pythia version superseeding 8.212
  //Event branched( const Event& event, int iRadBef, int iRecBef, int idEmt,
  //  int idMother, double pT2, double z, double RN1, vector<double> aux);

  // Setup clustering kinematics.
  virtual Event clustered( const Event& state, string name, int iRad,
    int iEmt, int iRecAft);
  bool cluster_II( const Event& state, int iRad,
    int iEmt, int iRecAft, int idRadBef, Particle& radBef, Particle& recBef,
    Event& partialState);
  bool cluster_IF( const Event& state, int iRad,
    int iEmt, int iRecAft, int idRadBef, Particle& radBef, Particle& recBef,
    Event& partialState);

  // Return ordering variable.
  virtual double pT2Space ( const Particle& rad, const Particle& emt,
    const Particle& rec) {
    if (rec.isFinal()) return pT2_IF(rad,emt,rec);
    return pT2_II(rad,emt,rec);
  }

  // For new Pythia version superseeding 8.212
  //// Return ordering variable.
  //double pT2Space ( const Particle& rad, const Particle& emt,
  //  const Particle& rec) {
  //  if (rec.isFinal()) return pT2_IF(rad,emt,rec);
  //  return pT2_II(rad,emt,rec);
  //}

  double pT2_II ( const Particle& rad, const Particle& emt,
    const Particle& rec);
  double pT2_IF ( const Particle& rad, const Particle& emt,
    const Particle& rec);

  // Return auxiliary variable.
  virtual double zSpace ( const Particle& rad, const Particle& emt,
    const Particle& rec) {
    if (rec.isFinal()) return z_IF(rad,emt,rec);
    return z_II(rad,emt,rec);
  }

  // For new Pythia version superseeding 8.212
  //// Return auxiliary variable.
  //double zSpace ( const Particle& rad, const Particle& emt,
  //  const Particle& rec) {
  //  if (rec.isFinal()) return z_IF(rad,emt,rec);
  //  return z_II(rad,emt,rec);
  //}

  double z_II ( const Particle& rad, const Particle& emt,
    const Particle& rec);
  double z_IF ( const Particle& rad, const Particle& emt,
    const Particle& rec);

  // Return a string identifier of a splitting.
  // Usage: getSplittingName( const Event& event, int iRad, int iEmt, int iRec)
  virtual string getSplittingName( const Event& state, int iRad, int iEmt)
    { return splittingsPtr->getSplittingName(state,iRad,iEmt); }

  // Return the splitting probability.
  // Usage: getSplittingProb( const Event& event, int iRad, int iEmt, int iRec)
  virtual double getSplittingProb( const Event& state, int iRad,
    int iEmt, int iRecAft);

  // For new Pythia version superseeding 8.212
  //// Return the evolution variable.
  //// Usage: getStateVariables( const Event& event,  int iRad, int iEmt, 
  ////                   int iRec, string name)
  //// Important note:
  ////  - The first element of the return vector *must* be the value of the
  ////    shower evolution variable corresponding to the branching defined by
  ////    the integers.
  ////  - The second element of the return vector *must* be the value of the
  ////    shower evolution variable from which the shower would restart after
  ////    the branching (two values will often be identical).
  //virtual vector<double> getStateVariables (const Event& state, int rad,
  //  int emt, int rec, string ) {
  //  vector<double> ret;
  //  double pT2 = pT2Space ( state[rad], state[emt], state[rec]);
  //  double z   = zSpace ( state[rad], state[emt], state[rec]);
  //  ret.push_back(pT2);
  //  ret.push_back(pT2);
  //  ret.push_back(z);
  //  return ret; 
  //}

  // For new Pythia version superseeding 8.212
  //// Check if attempted clustering is handled by timelike shower
  //// Usage: isTimelike( const Event& event,  int iRad, int iEmt, 
  ////                   int iRec, string name)
  //virtual bool isSpacelike(const Event& state, int iRad, int, int)
  //  { return !state[iRad].isFinal(); }

  // For new Pythia version superseeding 8.212
  //// Return a string identifier of a splitting.
  //// Usage: getSplittingName( const Event& event, int iRad, int iEmt, int iRec)
  //virtual string getSplittingName( const Event& state, int iRad, int iEmt,int)
  //  { return splittingsPtr->getSplittingName(state,iRad,iEmt); }

  // For new Pythia version superseeding 8.212
  //// Return the splitting probability.
  //// Usage: getSplittingProb( const Event& event, int iRad, int iEmt, int iRec)
  //virtual double getSplittingProb( const Event& state, int iRad,
  //  int iEmt, int iRecAft, string);

  // Return Jacobian for initial-initial phase space factorisation.
  double jacobian_II(double z, double pT2, double m2dip, double = 0.,
    double = 0., double = 0., double = 0., double = 0.) {
    // Calculate CS variables.
    double kappa2 = pT2 / m2dip;
    double xCS    = (z*(1-z)- kappa2)/(1-z);
    double vCS    = kappa2/(1-z); 
    return 1./xCS * ( 1. - xCS - vCS) / (1.- xCS - 2.*vCS);
  }

  // Return Jacobian for initial-final phase space factorisation.
  double jacobian_IF(double z, double pT2, double m2dip, double = 0.,
    double = 0., double = 0., double = 0., double = 0.) {
    // Calculate CS variables.
    double xCS = z;
    double uCS = (pT2/m2dip)/(1-z);
    return 1./xCS * ( 1. - uCS) / (1. - 2.*uCS);
  }

  // Auxiliary function to return the position of a particle.
  // Should go int Event class eventually!
  int FindParticle( const Particle& particle, const Event& event,
    bool checkStatus = true );

  // Print dipole list; for debug mainly.
  virtual void list(ostream& os = cout) const;

  // Check colour/flavour correctness of state.
  bool validEvent( const Event& state );

  // Find index colour partner for input colour.
  int FindCol(int col, int iExclude, const Event& event, int type);

  // Pointers to the two incoming beams.
  BeamParticle*  beamAPtr;
  BeamParticle*  beamBPtr;

private:

  // Number of times the same error message is repeated, unless overridden.
  static const int TIMESTOPRINT;

  // Allow conversion from mb to pb.
  static const double CONVERTMB2PB;

  // Store common beam quantities.
  int    idASave, idBSave;

protected:

  // Pointer to various information on the generation.
  Info*          infoPtr;

  // Pointer to the settings database.
  Settings*      settingsPtr;

  // Pointer to the particle data table.
  ParticleData*  particleDataPtr;

  // Pointer to the random number generator.
  Rndm*          rndmPtr;

  // Pointer to Standard Model couplings.
  CoupSM*        coupSMPtr;

  // Offset of beam location in event.
  int            beamOffset;

  // Pointer to information on subcollision parton locations.
  PartonSystems* partonSystemsPtr;

  // Pointer to userHooks object for user interaction with program.
  UserHooks*     userHooksPtr;

  // Store properties to be returned by methods.
  int    iSysSel;
  double pTmaxFudge;

private:

  // Constants: could only be changed in the code itself.
  static const int    MAXLOOPTINYPDF;
  static const double MCMIN, MBMIN, CTHRESHOLD, BTHRESHOLD, EVALPDFSTEP, 
         TINYPDF, TINYKERNELPDF, TINYPT2, HEAVYPT2EVOL, HEAVYXEVOL, 
         EXTRASPACEQ, LAMBDA3MARGIN, PT2MINWARN, LEPTONXMIN, LEPTONXMAX, 
         LEPTONPT2MIN, LEPTONFUDGE, HEADROOMQ2Q, HEADROOMQ2G, 
         HEADROOMG2G, HEADROOMG2Q, TINYMASS;
  static const double G2QQPDFPOW1, G2QQPDFPOW2; 

  // Initialization data, normally only set once.
  bool   doQCDshower, useSamePTasMPI,
         doMEcorrections, doMEafterFirst, doPhiPolAsym,
         doPhiIntAsym, doRapidityOrder, useFixedFacScale, doSecondHard,
         canVetoEmission, hasUserHooks, alphaSuseCMW, printBanner;
  int    pTmaxMatch, pTdampMatch, alphaSorder, alphaSnfmax,
         nQuarkIn, enhanceScreening, nFinalMax;
  double pTdampFudge, mc, mb, m2c, m2b, renormMultFac, factorMultFac,
         fixedFacScale2, alphaSvalue, alphaS2pi, Lambda3flav, Lambda4flav,
         Lambda5flav, Lambda3flav2, Lambda4flav2, Lambda5flav2, pT0Ref,
         ecmRef, ecmPow, pTmin, sCM, eCM, pT0, pT20,
         pT2min, pTmaxFudgeMPI, strengthIntAsym;
  double alphaS2piOverestimate;
  bool usePDFalphas, usePDFmasses, useSummedPDF;
  bool useGlobalMapIF, forceMassiveMap;

  // alphaStrong and alphaEM calculations.
  AlphaStrong alphaS;

  // Some current values.
  bool   sideA, dopTlimit1, dopTlimit2, dopTdamp;
  int    iNow, iRec, idDaughter, nRad, idResFirst, idResSecond;
  double xDaughter, x1Now, x2Now, m2Dip, m2Rec, pT2damp, pTbegRef, pdfScale2;

  // List of emissions in different sides in different systems:
  vector<int> nRadA,nRadB;

  // All dipole ends
  vector<DireSpaceEnd> dipEnd;

  // Pointers to the current and hardest (so far) dipole ends.
  int iDipNow, iSysNow;
  DireSpaceEnd* dipEndNow;
  int iDipSel;
  DireSpaceEnd* dipEndSel;

  void setupQCDdip( int iSys, int side, int colTag, int colSign,
    Event& event, int MEtype, bool limitPTmaxIn);

  virtual int system() const { return iSysSel;}
 
  // Evolve a QCD dipole end.
  void pT2nextQCD( double pT2begDip, double pT2endDip,
    DireSpaceEnd& dip, Event& event);
  void pT2nextQCD_II( double pT2begDip, double pT2endDip,
    DireSpaceEnd& dip, Event& event);
  void pT2nextQCD_IF( double pT2begDip, double pT2endDip,
    DireSpaceEnd& dip, Event& event);

  // Function to determine how often the integrated overestimate should be
  // recalculated.
  double evalpdfstep(int idRad, double pT2, double m2cPhys = -1.,
    double m2bPhys = -1.) {

    double ret = 0.1;
    if (m2cPhys < 0.) m2cPhys = particleDataPtr->m0(4);
    if (m2bPhys < 0.) m2bPhys = particleDataPtr->m0(5);
    // More steps close to the thresholds.
    if ( abs(idRad) == 4 && pT2 < 1.2*m2cPhys && pT2 > m2cPhys)
      ret = 0.1*(pT2-m2cPhys);
    if ( abs(idRad) == 5 && pT2 < 1.2*m2bPhys && pT2 > m2bPhys)
      ret = 0.1*(pT2-m2bPhys);
    return ret;
  }

  // Evolve a QCD dipole end near heavy quark threshold region.
  void pT2nearQCDthreshold( BeamParticle& beam, double m2Massive,
    double m2Threshold, double xMaxAbs, double zMinAbs,
    double zMaxMassive);

  // Pointer to MergingHooks object for NLO merging.
  MergingHooks* mergingHooksPtr;

  SplittingLibrary* splittingsPtr;

  // Number of proposed splittings in hard scattering systems.
  map<int,int> nProposedPT;

  // Return headroom factors for integrated/differential overestimates.
  double enhanceFactors(string, bool, double, double);

  // Function to fill map of integrated overestimates.
  void getNewOverestimates( int, int, DireSpaceEnd*, double, double, double,
    double, multimap<double,string>& );

  // Function to sum all integrated overestimates.
  void addNewOverestimates( multimap<double,string>, double&);

  // Function to evaluate the accept-probability, including picking of z.
  void getNewSplitting( const Event&, DireSpaceEnd*, double, double, double,
    double, double, int, string, int&, int&, double&, double&, double&,
    double&);

  // Get particle masses.
  double getMass(int id, int strategy, double mass = 0.) {
    BeamParticle& beam = (abs(beamAPtr->id()) == 2212) ? *beamAPtr : *beamBPtr;
    double mRet = 0.;
    if (strategy == 1) mRet = particleDataPtr->m0(id);
    if (strategy == 2 &&  usePDFmasses) mRet = beam.mQuarkPDF(id);
    if (strategy == 2 && !usePDFmasses) mRet = particleDataPtr->m0(id);
    if (strategy == 3) mRet = mass;
    if (mRet < TINYMASS) mRet = 0.;
    return pow2(max(0.,mRet));
  }

  // Check if variables are in allowed phase space.
  bool inAllowedPhasespace(double z, double pT2, double m2dip,
    double xOld, int splitType = 0, double m2RadBef = 0., double m2r = 0.,
    double m2s = 0., double m2e = 0.);

  // Kallen function and derived quantities. Helpers for massive phase
  // space mapping.
  double lABC(double a, double b, double c) { return pow2(a-b-c) - 4.*b*c;}
  double bABC(double a, double b, double c) { 
    double ret = 0.;
    if      ((a-b-c) > 0.) ret = sqrt(lABC(a,b,c));
    else if ((a-b-c) < 0.) ret =-sqrt(lABC(a,b,c));
    else                   ret = 0.;
    return ret; }
  double gABC(double a, double b, double c) { return 0.5*(a-b-c+bABC(a,b,c));}

  // Identifier of the splitting
  string splittingNowName, splittingSelName;

  // Weighted shower book-keeping.
  map<double,double> acceptProbability;
  multimap<double,double> rejectProbability;
  WeightContainer* weights;

};
 
//==========================================================================

} // end namespace Pythia8

#endif
