// DireTimes.h is a part of the PYTHIA event generator.
// Copyright (C) 2015 Torbjorn Sjostrand.
// PYTHIA is licenced under the GNU GPL version 2, see COPYING for details.
// Please respect the MCnet Guidelines, see GUIDELINES for details.

// Header file for the timelike final-state showers.
// DireTimesEnd: data on a radiating dipole end.
// DireTimes: handles the showering description.

#ifndef Pythia8_DireTimes_H
#define Pythia8_DireTimes_H

#define DIRE_TIMES_VERSION "0.900"

#include "Pythia8/Basics.h"
#include "Pythia8/TimeShower.h"
#include "Pythia8/Pythia.h"
#include "Pythia8/BeamParticle.h"
#include "Pythia8/Event.h"
#include "Pythia8/Info.h"
#include "Pythia8/ParticleData.h"
#include "Pythia8/PartonSystems.h"
#include "Pythia8/PythiaStdlib.h"
#include "Pythia8/Settings.h"
#include "Pythia8/StandardModel.h"
#include "Pythia8/UserHooks.h"
#include "Pythia8/MergingHooks.h"
#include "SplittingLibrary.h"
#include "WeightContainer.h"

namespace Pythia8 {

//==========================================================================

// Data on radiating dipole ends; only used inside DireTimes class.

class DireTimesEnd {

public:

  // Constructors.
  DireTimesEnd() : iRadiator(-1), iRecoiler(-1), pTmax(0.), colType(0),
    chgType(0), gamType(0), weakType(0), isrType(0), system(0), systemRec(0),
    MEtype(0), iMEpartner(-1), weakPol(0), isOctetOnium(false),
    isHiddenValley(false), colvType(0), MEmix(0.), MEorder(true),
    MEsplit(true), MEgluinoRec(false), isFlexible(false) { }
 DireTimesEnd(int iRadiatorIn, int iRecoilerIn, double pTmaxIn = 0.,
    int colIn = 0, int chgIn = 0, int gamIn = 0, int weakTypeIn = 0,
    int isrIn = 0, int systemIn = 0, int MEtypeIn = 0, int iMEpartnerIn = -1,
    int weakPolIn = 0, bool isOctetOniumIn = false,
    bool isHiddenValleyIn = false, int colvTypeIn = 0, double MEmixIn = 0.,
    bool MEorderIn = true, bool MEsplitIn = true, bool MEgluinoRecIn = false,
    bool isFlexibleIn = false, vector<int> iSpectatorIn = vector<int>()) :
    iRadiator(iRadiatorIn), iRecoiler(iRecoilerIn), pTmax(pTmaxIn),
    colType(colIn), chgType(chgIn), gamType(gamIn), weakType(weakTypeIn),
    isrType(isrIn), system(systemIn), systemRec(systemIn), MEtype(MEtypeIn),
    iMEpartner(iMEpartnerIn), weakPol(weakPolIn), isOctetOnium(isOctetOniumIn),
    isHiddenValley(isHiddenValleyIn), colvType(colvTypeIn), MEmix(MEmixIn),
    MEorder (MEorderIn), MEsplit(MEsplitIn), MEgluinoRec(MEgluinoRecIn),
    isFlexible(isFlexibleIn), iSpectator(iSpectatorIn) { }

  // Basic properties related to dipole and matrix element corrections.
  int    iRadiator, iRecoiler;
  double pTmax;
  int    colType, chgType, gamType, weakType, isrType, system, systemRec,
         MEtype, iMEpartner, weakPol;
  bool   isOctetOnium, isHiddenValley;
  int    colvType;
  double MEmix;
  bool   MEorder, MEsplit, MEgluinoRec, isFlexible;

  // Properties specific to current trial emission.
  int    flavour, iAunt;
  double mRad, m2Rad, mRec, m2Rec, mDip, m2Dip, m2DipCorr,
         pT2, m2, z, mFlavour, asymPol, flexFactor;

  // Extended list of recoilers.
  vector<int> iSpectator;

};

//==========================================================================

// The DireTimes class does timelike showers.

class DireTimes : public Pythia8::TimeShower {

public:

  // Constructor.
  DireTimes() {}

  DireTimes(Pythia8::Pythia* pythiaPtr) {
    beamOffset = 0;
    infoPtr           = &pythiaPtr->info;
    particleDataPtr   = &pythiaPtr->particleData;
    partonSystemsPtr  = &pythiaPtr->partonSystems;
    rndmPtr           = &pythiaPtr->rndm;
    settingsPtr       = &pythiaPtr->settings;
    userHooksPtr      = 0;
    mergingHooksPtr   = pythiaPtr->mergingHooksPtr;
    splittingsPtr     = 0;
    weights           = 0;
    printBanner       = true;
  }

  // Destructor.
  virtual ~DireTimes() {}

  // Initialize alphaStrong and related pTmin parameters.
  virtual void init( BeamParticle* beamAPtrIn = 0,
    BeamParticle* beamBPtrIn = 0);

  // Initialize various pointers.
  // (Separated from rest of init since not virtual.)
  void reinitPtr(Info* infoPtrIn, Settings* settingsPtrIn,
       ParticleData* particleDataPtrIn, Rndm* rndmPtrIn,
       PartonSystems* partonSystemsPtrIn, UserHooks* userHooksPtrIn,
       MergingHooks* mergingHooksPtrIn, SplittingLibrary* splittingsPtrIn) {
       infoPtr = infoPtrIn;
       settingsPtr = settingsPtrIn;
       particleDataPtr = particleDataPtrIn;
       rndmPtr = rndmPtrIn;
       partonSystemsPtr = partonSystemsPtrIn;
       userHooksPtr = userHooksPtrIn;
       mergingHooksPtr = mergingHooksPtrIn;
       splittingsPtr = splittingsPtrIn;
  }

  void setWeightContainerPtr(WeightContainer* weightsIn) { weights = weightsIn;}

  // Find whether to limit maximum scale of emissions, and whether to dampen.
  virtual bool limitPTmax( Event& event, double Q2Fac = 0.,
    double Q2Ren = 0.);

  // Potential enhancement factor of pTmax scale for hardest emission.
  virtual double enhancePTmax() { return pTmaxFudge;}

  // Top-level routine to do a full time-like shower in resonance decay.
  virtual int shower( int iBeg, int iEnd, Event& event, double pTmax,
    int nBranchMax = 0);

  // Global recoil: reset counters and store locations of outgoing partons.
  virtual void prepareGlobal( Event&);

  // Prepare system for evolution after each new interaction; identify ME.
  virtual void prepare( int iSys, Event& event, bool limitPTmaxIn = true);

  // Update dipole list after a multiparton interactions rescattering.
  virtual void rescatterUpdate( int iSys, Event& event);

  // Update dipole list after each ISR emission.
  virtual void update( int iSys, Event& event, bool = false);

  // Select next pT in downwards evolution.
  virtual double pTnext( Event& event, double pTbegAll, double pTendAll,
    bool = false, bool = false);

  // Setup branching kinematics.
  virtual bool branch( Event& event, bool = false);
  bool branch_FF( Event& event, bool = false);
  bool branch_FI( Event& event, bool = false);

  // Tell which system was the last processed one.
  virtual int system() const {return iSysSel;};

  // Setup splitting kinematics.
  virtual Event branched( const Event& event, int iRadBef, int iRecBef,
    int idEmt, double pT2, double z, double RNphi, vector<double>);

  // For new Pythia version superseeding 8.212
  //Event branched( const Event& event, int iRadBef, int iRecBef,
  //  int idEmt, double pT2, double z, double RNphi, vector<double>);

  vector<Vec4> split_FF( const Event& event, int iRadBef, int iRecBef,
    double pT2, double z, double RNphi, int flavour, int idRad, int idEmt);
  vector<Vec4> split_FI( const Event& event, int iRadBef, int iRecBef,
    double pT2, double z, double RNphi, int flavour, int idRad, int idEmt);

  // Setup clustering kinematics.
  virtual Event clustered( const Event& state, string name, int iRad,
    int iEmt, int iRec);
  bool cluster_FF( const Event& state, int iRad,
    int iEmt, int iRec, int idRadBef, Particle& radBef, Particle& recBef);
  bool cluster_FI( const Event& state, int iRad,
    int iEmt, int iRec, int idRadBef, Particle& radBef, Particle& recBef);

  // Return ordering variable.
  virtual double pT2Times ( const Particle& rad, const Particle& emt,
    const Particle& rec) {
    if (rec.isFinal()) return pT2_FF(rad,emt,rec);
    return pT2_FI(rad,emt,rec);
  }

  //// For new Pythia version superseeding 8.212
  //double pT2Times ( const Particle& rad, const Particle& emt,
  //  const Particle& rec) {
  //  if (rec.isFinal()) return pT2_FF(rad,emt,rec);
  //  return pT2_FI(rad,emt,rec);
  //}

  double pT2_FF ( const Particle& rad, const Particle& emt,
    const Particle& rec);
  double pT2_FI ( const Particle& rad, const Particle& emt,
    const Particle& rec);

  // Return auxiliary variable.
  virtual double zTimes ( const Particle& rad, const Particle& emt,
    const Particle& rec) {
    if (rec.isFinal()) return z_FF(rad,emt,rec);
    return z_FI(rad,emt,rec);
  }

  // For new Pythia version superseeding 8.212
  //double zTimes ( const Particle& rad, const Particle& emt,
  //  const Particle& rec) {
  //  if (rec.isFinal()) return z_FF(rad,emt,rec);
  //  return z_FI(rad,emt,rec);
  //}

  double z_FF ( const Particle& rad, const Particle& emt,
    const Particle& rec);
  double z_FI ( const Particle& rad, const Particle& emt,
    const Particle& rec);
  //double z_FF_fromVec ( const Vec4 rad, const Vec4 emt,
  //  const Vec4 rec);


  // Return a string identifier of a splitting.
  // Usage: getSplittingName( const Event& event, int iRad, int iEmt, int iRec)
  virtual string getSplittingName( const Event& state, int iRad, int iEmt)
    { return splittingsPtr->getSplittingName(state,iRad,iEmt); }

  // Return the splitting probability.
  // Usage: getSplittingProb( const Event& event, int iRad, int iEmt, int iRec)
  virtual double getSplittingProb( const Event& state, int iRad,
    int iEmt, int iRec);

  // For new Pythia version superseeding 8.212
  //// Return the evolution variable.
  //// Usage: getStateVariables( const Event& event,  int iRad, int iEmt, 
  ////                   int iRec, string name)
  //// Important note:
  ////  - The first element of the return vector *must* be the value of the
  ////    shower evolution variable corresponding to the branching defined by
  ////    the integers.
  ////  - The second element of the return vector *must* be the value of the
  ////    shower evolution variable from which the shower would restart after
  ////    the branching (two values will often be identical).
  //virtual vector<double> getStateVariables (const Event& state, int rad,
  //  int emt, int rec, string ) {
  //  vector<double> ret;
  //  double pT2 = pT2Times ( state[rad], state[emt], state[rec]);
  //  double z   = zTimes ( state[rad], state[emt], state[rec]);
  //  ret.push_back(pT2);
  //  ret.push_back(pT2);
  //  ret.push_back(z);
  //  return ret; 
  //}

  // For new Pythia version superseeding 8.212
  //// Check if attempted clustering is handled by timelike shower
  //// Usage: isTimelike( const Event& event,  int iRad, int iEmt, 
  ////                   int iRec, string name)
  //virtual bool isTimelike(const Event& state, int iRad, int, int)
  //  { return state[iRad].isFinal(); }

  // For new Pythia version superseeding 8.212
  //// Return a string identifier of a splitting.
  //// Usage: getSplittingName( const Event& event, int iRad, int iEmt, int iRec)
  //virtual string getSplittingName( const Event& state, int iRad, int iEmt, int)
  //  { return splittingsPtr->getSplittingName(state,iRad,iEmt); }

  // For new Pythia version superseeding 8.212
  //// Return the splitting probability.
  //// Usage: getSplittingProb( const Event& event, int iRad, int iEmt, int iRec)
  //virtual double getSplittingProb( const Event& state, int iRad,
  //  int iEmt, int iRec, string);

  // Return Jacobian for final-final phase space factorisation.
  double jacobian_FF(double z, double pT2, double m2dip, double q2,
    double m2RadBef = 0., double m2r = 0., double m2s = 0.,
    double m2e = 0.) {

    // Calculate CS variables.
    double kappa2 = pT2/m2dip;
    double yCS    = kappa2 / (1.-z);

    // Get scaled masses.
    double mu2RadBef = m2RadBef/q2;
    double mu2Rad = m2r/q2;
    double mu2Rec = m2s/q2;
    double mu2Emt = m2e/q2;

    // Get Jacobian.
    double jac1 = (1. - mu2Rad - mu2Rec - mu2Emt)
                / sqrt(lABC(1.,mu2RadBef,mu2Rec));
    double jac2 = 1. + ( mu2Rad + mu2Emt - mu2RadBef)
                     / ( yCS * ( 1. - mu2Rad - mu2Emt - mu2Rec ) );
    return jac1 / jac2 * (1-yCS);

  }

  // Return Jacobian for final-initial phase space factorisation.
  double jacobian_FI(double = 0., double = 0., double = 0., double = 0.,
    double = 0., double = 0., double = 0., double = 0.) { return 1.;}

  // Auxiliary function to return the position of a particle.
  // Should go int Event class eventually!
  int FindParticle( const Particle& particle, const Event& event,
    bool checkStatus = true );

  // Print dipole list; for debug mainly.
  virtual void list( ostream& os = cout) const;

  // Check colour/flavour correctness of state.
  bool validEvent( const Event& state );

  // Find index of colour partner for input colour.
  int FindCol(int col, int iExclude, const Event& event, int type,
    int iSys = -1);

private:

  // Number of times the same error message is repeated, unless overridden.
  static const int TIMESTOPRINT;

  // Allow conversion from mb to pb.
  static const double CONVERTMB2PB;

  // Store common beam quantities.
  int    idASave, idBSave;

protected:

  // Pointer to various information on the generation.
  Info*          infoPtr;

  // Pointer to the settings database.
  Settings*      settingsPtr;

  // Pointer to the particle data table.
  ParticleData*  particleDataPtr;

  // Pointer to the random number generator.
  Rndm*          rndmPtr;

  // Pointers to the two incoming beams. Offset their location in event.
  BeamParticle*  beamAPtr;
  BeamParticle*  beamBPtr;
  int            beamOffset;

  // Pointer to information on subcollision parton locations.
  PartonSystems* partonSystemsPtr;

  // Pointer to userHooks object for user interaction with program.
  UserHooks*     userHooksPtr;

  // Store properties to be returned by methods.
  int    iSysSel;
  double pTmaxFudge, pTLastBranch;

private:

  // Constants: could only be changed in the code itself.
  static const double MCMIN, MBMIN, SIMPLIFYROOT, XMARGIN, XMARGINCOMB,
         TINYPDF, LARGEM2, THRESHM2, LAMBDA3MARGIN, WG2QEXTRA, TINYMASS;

  // Initialization data, normally only set once.
  bool   doQCDshower, doMEcorrections, doMEafterFirst, doPhiPolAsym,
         doInterleave, allowBeamRecoil, dampenBeamRecoil, recoilToColoured,
         useFixedFacScale, allowRescatter, canVetoEmission, hasUserHooks,
         doSecondHard, alphaSuseCMW, printBanner;
  int    pTmaxMatch, pTdampMatch, alphaSorder, alphaSnfmax, nGluonToQuark,
         weightGluonToQuark, nFinalMax;
  double pTdampFudge, mc, mb, m2c, m2b, renormMultFac, factorMultFac,
         fixedFacScale2, alphaSvalue, alphaS2pi, Lambda3flav, Lambda4flav,
         Lambda5flav, Lambda3flav2, Lambda4flav2, Lambda5flav2,
         scaleGluonToQuark, extraGluonToQuark, pTcolCutMin, pTcolCut,
         pT2colCut, mZ, gammaZ, thetaWRat, mW, gammaW, pTmaxFudgeMPI;
  double alphaS2piOverestimate;
  bool usePDFalphas, usePDFmasses, useSummedPDF;

  // alphaStrong and alphaEM calculations.
  AlphaStrong alphaS;

  // Some current values.
  bool   dopTlimit1, dopTlimit2, dopTdamp;
  double pT2damp, kRad, kEmt, pdfScale2;

  // All dipole ends and a pointer to the selected hardest dipole end.
  vector<DireTimesEnd> dipEnd;
  DireTimesEnd* dipSel;
  int iDipSel;

  // Setup a dipole end, either QCD, QED/photon, weak or Hidden Valley one.
  void setupQCDdip( int iSys, int i, int colTag,  int colSign, Event& event,
    bool isOctetOnium = false, bool limitPTmaxIn = true);

  // Evolve a QCD dipole end.
  void pT2nextQCD( double pT2begDip, double pT2sel, DireTimesEnd& dip,
    Event& event);
  void pT2nextQCD_FF( double pT2begDip, double pT2sel, DireTimesEnd& dip,
    Event& event);
  void pT2nextQCD_FI( double pT2begDip, double pT2sel, DireTimesEnd& dip,
    Event& event);

  // Pointer to MergingHooks object for NLO merging.
  MergingHooks* mergingHooksPtr;
  SplittingLibrary* splittingsPtr;

  // Number of proposed splittings in hard scattering systems.
  map<int,int> nProposedPT;

  // Return headroom factors for integrated/differential overestimates.
  double enhanceFactors(string, double, double);

  // Function to fill map of integrated overestimates.
  void getNewOverestimates( int, DireTimesEnd*, double, double, double,
    multimap<double,string>&);

  // Function to sum all integrated overestimates.
  void addNewOverestimates( multimap<double,string>, double&);

  // Function to evaluate the accept-probability, including picking of z.
  void getNewSplitting( const Event&, DireTimesEnd*, double, double, double,
    double, double, int, string, int&, int&, double&, double&, double&,
    double&);

  // Get particle masses.
  double getMass(int id, int strategy, double mass = 0.) {
    BeamParticle& beam = (abs(beamAPtr->id()) == 2212) ? *beamAPtr : *beamBPtr;
    double mRet = 0.;
    if (strategy == 1) mRet = particleDataPtr->m0(id);
    if (strategy == 2 &&  usePDFmasses) mRet = beam.mQuarkPDF(id);
    if (strategy == 2 && !usePDFmasses) mRet = particleDataPtr->m0(id);
    if (strategy == 3) mRet = mass;
    if (mRet < TINYMASS) mRet = 0.;
    return pow2(max(0.,mRet));
  }

  // Check if variables are in allowed phase space.
  bool inAllowedPhasespace(double z, double pT2, double m2dip, double q2,
    double xOld, int splitType = 0, double m2RadBef = 0., double m2r = 0.,
    double m2s = 0., double m2e = 0.);

  // Kallen function and derived quantities. Helpers for massive phase
  // space mapping.
  double lABC(double a, double b, double c) { return pow2(a-b-c) - 4.*b*c;}
  double bABC(double a, double b, double c) { 
    double ret = 0.;
    if      ((a-b-c) > 0.) ret = sqrt(lABC(a,b,c));
    else if ((a-b-c) < 0.) ret =-sqrt(lABC(a,b,c));
    else                   ret = 0.;
    return ret; }
  double gABC(double a, double b, double c) { return 0.5*(a-b-c+bABC(a,b,c));}

  // Identifier of the splitting
  string splittingNowName, splittingSelName;

  // Weighted shower book-keeping.
  map<double,double> acceptProbability;
  multimap<double,double> rejectProbability;
  WeightContainer* weights;

  double octetOniumColFac;
  bool useLocalRecoilNow;

};

//==========================================================================

} // end namespace Pythia8

#endif
