
#ifndef Pythia8_Splittings_H
#define Pythia8_Splittings_H

#define ZETA3 1.202056903159594

#define DIRE_SPLITTINGS_VERSION "0.900"

extern "C" {
  double polylog(int n, double x);
}

#include "Pythia8/Basics.h"
#include "Pythia8/Pythia.h"
#include "Pythia8/BeamParticle.h"
#include "Pythia8/ParticleData.h"
#include "Pythia8/PythiaStdlib.h"
#include "Pythia8/Settings.h"
#include "Pythia8/StandardModel.h"

namespace Pythia8 {

class DireSpace;
class DireTimes;
 
//==========================================================================

class Splitting {

public:  

  // Constructor and destructor.
  Splitting(Settings* settings, ParticleData* particleData, Rndm* rndm,
    BeamParticle* beamA, BeamParticle* beamB) :
      settingsPtr(settings),
      particleDataPtr(particleData), rndmPtr(rndm), beamAPtr(beamA),
      beamBPtr(beamB) { init(); }
  virtual ~Splitting() {}

  void init();

  // VARIABLES
  double CA, TR, CF;
  int NF_fsr;
  bool usePDFalphaS;
  double alphaS2pi;

  Settings* settingsPtr;
  ParticleData* particleDataPtr;
  Rndm* rndmPtr;
  BeamParticle* beamAPtr;
  BeamParticle* beamBPtr;

  AlphaStrong alphaS;
  AlphaEM     alphaEM;

  // AUXILIARY FUNCTIONS
  double Gamma2(double NF=5.);
  double Gamma3(double NF=5.);
  double beta0(double NF=5.);
  double beta1(double NF=5.);
  double softRescaleInt(int order);
  double softRescaleDiff(int order, double pT2);

  virtual string name () {return "";}
  virtual int type () {return 0;}

  // Return id of mother after splitting.
  virtual int motherID(int) {return 0;}

  // Return id of emission.
  virtual int sisterID(int) {return 0;}

  // Return id of recombined radiator (before splitting!)
  virtual int radID(int, int) {return 0;}

  // Return colours of recombined radiator (before splitting!)
  virtual pair<int,int> radCols(int, int, int, int)
    {return make_pair(0,0);}

  // Return color factor for splitting.
  virtual double gaugeFactor () {return 1.;}

  // Return color factor for splitting.
  virtual double symmetryFactor () {return 1.;}

  // Pick z for new splitting.
  virtual double zSplit(double, double, double) {return 0.5;}

  // New overestimates, z-integrated versions.
  virtual double overestimateInt(double, double, double, double) { return 0.;}

  // Return kernel for new splitting.
  virtual double overestimateDiff(double, double) {return 1.;}

  // Return kernel for new splitting.
  virtual double kernel(double, double, double, int = 0, double = 0.,
    double = 0., double = 0., double = 0.) { return 0.; }
  virtual double kernelME(const Event&, int, int, double, double, double,
    vector<double>) { return 0.; }

  DireSpace* isr;
  DireTimes* fsr;
  void setTimesPtr(DireTimes* fsrIn) { fsr=fsrIn;}
  void setSpacePtr(DireSpace* isrIn) { isr=isrIn;}

};

class fsr_Q2QG : public Splitting {

public:  

  fsr_Q2QG(string idIn, int softRS, Settings* settings,
    ParticleData* particleData, Rndm* rndm, BeamParticle* beamA,
    BeamParticle* beamB) :
    Splitting(settings, particleData, rndm, beamA, beamB),
    id(idIn), correctionOrder(softRS) {}

  string id;
  int correctionOrder;
  string name();
  int type();

  // Return id of mother after splitting.
  int motherID(int idDaughter);

  // Return id of emission.
  int sisterID(int idDaughter);

  // Return id of recombined radiator (before splitting!)
  int radID(int idRadAfter, int idEmtAfter);

  // Return colours of recombined radiator (before splitting!)
  pair<int,int> radCols(int colRadAfter, int acolRadAfter, 
    int colEmtAfter, int acolEmtAfter);

  double gaugeFactor ();
  double symmetryFactor ();

  // Pick z for new splitting.
  double zSplit(double zMinAbs, double zMaxAbs, double m2dip);

  // New overestimates, z-integrated versions.
  double overestimateInt(double zMinAbs,double zMaxAbs,
    double pT2Old, double m2dip);

  // Return kernel for new splitting.
  double overestimateDiff(double z, double m2dip);

  // Return kernel for new splitting.
  double kernel(double z, double pT2, double m2dip, int splitType = 0,
    double m2RadBef = 0., double m2Rad = 0., double m2Rec = 0.,
    double m2Emt = 0.);

};

class fsr_G2GG : public Splitting {

public:  

  fsr_G2GG(string idIn, int softRS, Settings* settings,
    ParticleData* particleData, Rndm* rndm, BeamParticle* beamA,
    BeamParticle* beamB) :
    Splitting(settings, particleData, rndm, beamA, beamB),
    id(idIn), correctionOrder(softRS) {}

  string id;
  int correctionOrder;
  string name();
  int type();

  // Return id of mother after splitting.
  int motherID(int idDaughter);

  // Return id of emission.
  int sisterID(int idDaughter);

  // Return id of recombined radiator (before splitting!)
  int radID(int idRadAfter, int idEmtAfter);

  // Return colours of recombined radiator (before splitting!)
  pair<int,int> radCols(int colRadAfter, int acolRadAfter, 
    int colEmtAfter, int acolEmtAfter);

  double gaugeFactor ();
  double symmetryFactor ();

  // Pick z for new splitting.
  double zSplit(double zMinAbs, double zMaxAbs, double m2dip);

  // New overestimates, z-integrated versions.
  double overestimateInt(double zMinAbs,double zMaxAbs,
    double pT2Old, double m2dip);

  // Return kernel for new splitting.
  double overestimateDiff(double z, double m2dip);

  // Return kernel for new splitting.
  double kernel(double z, double pT2, double m2dip, int splitType = 0,
    double m2RadBef = 0., double m2Rad = 0., double m2Rec = 0.,
    double m2Emt = 0.);

};

class fsr_G2QQ1 : public Splitting {

public:  

  fsr_G2QQ1(string idIn, int softRS, Settings* settings,
    ParticleData* particleData, Rndm* rndm, BeamParticle* beamA,
    BeamParticle* beamB) :
    Splitting(settings, particleData, rndm, beamA, beamB),
    id(idIn), correctionOrder(softRS) {}

  string id;
  int correctionOrder;
  string name();
  int type();

  // Return id of mother after splitting.
  int motherID(int idDaughter);

  // Return id of emission.
  int sisterID(int idDaughter);

  // Return id of recombined radiator (before splitting!)
  int radID(int idRadAfter, int idEmtAfter);

  // Return colours of recombined radiator (before splitting!)
  pair<int,int> radCols(int colRadAfter, int acolRadAfter, 
    int colEmtAfter, int acolEmtAfter);

  double gaugeFactor ();
  double symmetryFactor ();

  // Pick z for new splitting.
  double zSplit(double zMinAbs, double zMaxAbs, double m2dip);

  // New overestimates, z-integrated versions.
  double overestimateInt(double zMinAbs,double zMaxAbs,
    double pT2Old, double m2dip);

  // Return kernel for new splitting.
  double overestimateDiff(double z, double m2dip);

  // Return kernel for new splitting.
  double kernel(double z, double pT2, double m2dip, int splitType = 0,
    double m2RadBef = 0., double m2Rad = 0., double m2Rec = 0.,
    double m2Emt = 0.);

};

class fsr_G2QQ2 : public Splitting {

public:  

  fsr_G2QQ2(string idIn, int softRS, Settings* settings,
    ParticleData* particleData, Rndm* rndm, BeamParticle* beamA,
    BeamParticle* beamB) :
    Splitting(settings, particleData, rndm, beamA, beamB),
    id(idIn), correctionOrder(softRS) {}

  string id;
  int correctionOrder;
  string name();
  int type();

  // Return id of mother after splitting.
  int motherID(int idDaughter);

  // Return id of emission.
  int sisterID(int idDaughter);

  // Return id of recombined radiator (before splitting!)
  int radID(int idRadAfter, int idEmtAfter);

  // Return colours of recombined radiator (before splitting!)
  pair<int,int> radCols(int colRadAfter, int acolRadAfter, 
    int colEmtAfter, int acolEmtAfter);

  double gaugeFactor ();
  double symmetryFactor ();

  // Pick z for new splitting.
  double zSplit(double zMinAbs, double zMaxAbs, double m2dip);

  // New overestimates, z-integrated versions.
  double overestimateInt(double zMinAbs,double zMaxAbs,
    double pT2Old, double m2dip);

  // Return kernel for new splitting.
  double overestimateDiff(double z, double m2dip);

  // Return kernel for new splitting.
  double kernel(double z, double pT2, double m2dip, int splitType = 0,
    double m2RadBef = 0., double m2Rad = 0., double m2Rec = 0.,
    double m2Emt = 0.);

};

class isr_Q2QG : public Splitting {

public:  

  isr_Q2QG(string idIn, int softRS, Settings* settings,
    ParticleData* particleData, Rndm* rndm, BeamParticle* beamA,
    BeamParticle* beamB) :
    Splitting(settings, particleData, rndm, beamA, beamB),
    id(idIn), correctionOrder(softRS) {}

  string id;
  int correctionOrder;
  string name();
  int type();

  // Return id of mother after splitting.
  int motherID(int idDaughter);

  // Return id of emission.
  int sisterID(int idDaughter);

  // Return id of recombined radiator (before splitting!)
  int radID(int idRadAfter, int idEmtAfter);

  // Return colours of recombined radiator (before splitting!)
  pair<int,int> radCols(int colRadAfter, int acolRadAfter, 
    int colEmtAfter, int acolEmtAfter);

  double gaugeFactor ();
  double symmetryFactor ();

  // Pick z for new splitting.
  double zSplit(double zMinAbs, double zMaxAbs, double m2dip);

  // New overestimates, z-integrated versions.
  double overestimateInt(double zMinAbs,double zMaxAbs,
    double pT2Old, double m2dip);

  // Return kernel for new splitting.
  double overestimateDiff(double z, double m2dip);

  // Return kernel for new splitting.
  double kernel(double z, double pT2, double m2dip, int,
    double, double, double, double);

};

class isr_G2GG1 : public Splitting {

public:  

  isr_G2GG1(string idIn, int softRS, Settings* settings,
    ParticleData* particleData, Rndm* rndm, BeamParticle* beamA,
    BeamParticle* beamB) :
    Splitting(settings, particleData, rndm, beamA, beamB),
    id(idIn), correctionOrder(softRS) {}

  string id;
  int correctionOrder;
  string name();
  int type();

  // Return id of mother after splitting.
  int motherID(int idDaughter);

  // Return id of emission.
  int sisterID(int idDaughter);

  // Return id of recombined radiator (before splitting!)
  int radID(int idRadAfter, int idEmtAfter);

  // Return colours of recombined radiator (before splitting!)
  pair<int,int> radCols(int colRadAfter, int acolRadAfter, 
    int colEmtAfter, int acolEmtAfter);

  double gaugeFactor ();
  double symmetryFactor ();

  // Pick z for new splitting.
  double zSplit(double zMinAbs, double zMaxAbs, double m2dip);

  // New overestimates, z-integrated versions.
  double overestimateInt(double zMinAbs,double zMaxAbs,
    double pT2Old, double m2dip);

  // Return kernel for new splitting.
  double overestimateDiff(double z, double m2dip);

  // Return kernel for new splitting.
  double kernel(double z, double pT2, double m2dip, int splitType = 0,
    double m2RadBef = 0., double m2Rad = 0., double m2Rec = 0.,
    double m2Emt = 0.);

};

class isr_G2GG2 : public Splitting {

public:  

  isr_G2GG2(string idIn, int softRS, Settings* settings,
    ParticleData* particleData, Rndm* rndm, BeamParticle* beamA,
    BeamParticle* beamB) :
    Splitting(settings, particleData, rndm, beamA, beamB),
    id(idIn), correctionOrder(softRS) {}

  string id;
  int correctionOrder;
  string name();
  int type();

  // Return id of mother after splitting.
  int motherID(int idDaughter);

  // Return id of emission.
  int sisterID(int idDaughter);

  // Return id of recombined radiator (before splitting!)
  int radID(int idRadAfter, int idEmtAfter);

  // Return colours of recombined radiator (before splitting!)
  pair<int,int> radCols(int colRadAfter, int acolRadAfter, 
    int colEmtAfter, int acolEmtAfter);

  double gaugeFactor ();
  double symmetryFactor ();

  // Pick z for new splitting.
  double zSplit(double zMinAbs, double zMaxAbs, double m2dip);

  // New overestimates, z-integrated versions.
  double overestimateInt(double zMinAbs,double zMaxAbs,
    double pT2Old, double m2dip);

  // Return kernel for new splitting.
  double overestimateDiff(double z, double m2dip);

  // Return kernel for new splitting.
  double kernel(double z, double pT2, double m2dip, int splitType = 0, 
    double m2RadBef = 0., double m2Rad = 0., double m2Rec = 0.,
    double m2Emt = 0.);

};

class isr_G2QQ : public Splitting {

public:  

  isr_G2QQ(string idIn, int softRS, Settings* settings,
    ParticleData* particleData, Rndm* rndm, BeamParticle* beamA,
    BeamParticle* beamB) :
    Splitting(settings, particleData, rndm, beamA, beamB),
    id(idIn), correctionOrder(softRS) {}

  string id;
  int correctionOrder;
  string name();
  int type();

  // Return id of mother after splitting.
  int motherID(int idDaughter);

  // Return id of emission.
  int sisterID(int idDaughter);

  // Return id of recombined radiator (before splitting!)
  int radID(int idRadAfter, int idEmtAfter);

  // Return colours of recombined radiator (before splitting!)
  pair<int,int> radCols(int colRadAfter, int acolRadAfter, 
    int colEmtAfter, int acolEmtAfter);

  double gaugeFactor ();
  double symmetryFactor ();

  // Pick z for new splitting.
  double zSplit(double zMinAbs, double zMaxAbs, double m2dip);

  // New overestimates, z-integrated versions.
  double overestimateInt(double zMinAbs,double zMaxAbs,
    double pT2Old, double m2dip);

  // Return kernel for new splitting.
  double overestimateDiff(double z, double m2dip);

  // Return kernel for new splitting.
  double kernel(double z, double pT2, double m2dip, int,
    double, double, double, double);

};

class isr_Q2GQ : public Splitting {

public:  

  isr_Q2GQ(string idIn, int softRS, Settings* settings,
    ParticleData* particleData, Rndm* rndm, BeamParticle* beamA,
    BeamParticle* beamB) :
    Splitting(settings, particleData, rndm, beamA, beamB),
    id(idIn), correctionOrder(softRS) {}

  string id;
  int correctionOrder;
  string name();
  int type();

  // Return id of mother after splitting.
  int motherID(int idDaughter);

  // Return id of emission.
  int sisterID(int idDaughter);

  // Return id of recombined radiator (before splitting!)
  int radID(int idRadAfter, int idEmtAfter);

  // Return colours of recombined radiator (before splitting!)
  pair<int,int> radCols(int colRadAfter, int acolRadAfter, 
    int colEmtAfter, int acolEmtAfter);

  double gaugeFactor ();
  double symmetryFactor ();

  // Pick z for new splitting.
  double zSplit(double zMinAbs, double zMaxAbs, double m2dip);

  // New overestimates, z-integrated versions.
  double overestimateInt(double zMinAbs,double zMaxAbs,
    double pT2Old, double m2dip);

  // Return kernel for new splitting.
  double overestimateDiff(double z, double m2dip);

  // Return kernel for new splitting.
  double kernel(double z, double pT2, double m2dip, int splitType = 0,
    double m2RadBef = 0., double m2Rad = 0., double m2Rec = 0.,
    double m2Emt = 0.);

};

class X2XA : public Splitting {

public:  

  X2XA(string idIn, int softRS, Settings* settings,
    ParticleData* particleData, Rndm* rndm, BeamParticle* beamA,
    BeamParticle* beamB) :
    Splitting(settings, particleData, rndm, beamA, beamB),
    id(idIn), correctionOrder(softRS) {}

  string id;
  int correctionOrder;
  string name();
  int type();

  // Return id of mother after splitting.
  int motherID(int idDaughter);

  // Return id of emission.
  int sisterID(int idDaughter);

  // Return id of recombined radiator (before splitting!)
  int radID(int idRadAfter, int idEmtAfter);

  // Return colours of recombined radiator (before splitting!)
  pair<int,int> radCols(int colRadAfter, int acolRadAfter, 
    int colEmtAfter, int acolEmtAfter);

  double gaugeFactor ();
  double symmetryFactor ();

  // Pick z for new splitting.
  double zSplit(double zMinAbs, double zMaxAbs, double m2dip);

  // New overestimates, z-integrated versions.
  double overestimateInt(double zMinAbs,double zMaxAbs,
    double pT2Old, double m2dip);

  // Return kernel for new splitting.
  double overestimateDiff(double z, double m2dip);

  // Return kernel for new splitting.
  double kernelME(const Event& state, int iRadBef, int iRecBef,
    double z, double pT2, double m2dip, vector<double> aux);

};

} // end namespace Pythia8

#endif
