
#ifndef Pythia8_WeightContainer_H
#define Pythia8_WeightContainer_H

#define DIRE_WEIGHTCONTAINER_VERSION "0.900"

#include "Pythia8/Pythia.h"
#include "Pythia8/PythiaStdlib.h"

namespace Pythia8 {

//==========================================================================

// The container for shower weights.

class WeightContainer {

public:

  // Constructor.
  WeightContainer() { init(); }

  // Destructor.
  virtual ~WeightContainer() {}

  // Initialize weights.
  void init() { reset(); showerWeight = 1.; }

  // Reset accept/reject probabilities.
  void reset() {
    acceptProb.clear();
    rejectProb.clear();
  }

  // Attach accept/reject probabilities for a proposed shower step.
  void insertWeights( map<double,double> aProb,
    multimap<double,double> rProb ) {
    acceptProb.insert(aProb.begin(), aProb.end());
    rejectProb.insert(rProb.begin(), rProb.end());
  }

  // Function to calculate the weight of the shower evolution step.
  void calcWeight(double pT2) {

    // Set shower weight.
    bool hasAccept      = (acceptProb.lower_bound(pT2) != acceptProb.end());
    double acceptWeight = (hasAccept) ? acceptProb.lower_bound(pT2)->second
                        : 1.;

    // Now multiply rejection weights.
    double rejectWeight = 1.;
    for ( multimap<double,double>::reverse_iterator it = rejectProb.rbegin();
      it != rejectProb.rend(); ++it ){
        if ( it->first > pT2) rejectWeight *= it->second;
        if ( it->first < pT2) break;
        if ( abs(it->first-pT2) < 1e-5) break;
    }

    // Remember weights
    showerWeight *= acceptWeight*rejectWeight;
    // Reset accept / reject weights.
    reset();

  }

  // Function to return weight of the shower evolution.
  double getShowerWeight() { return showerWeight; }

private:

  map<double,double> acceptProb;
  multimap<double,double> rejectProb;
  double showerWeight;

};

//==========================================================================

} // end namespace Pythia8

#endif
