// DireSpace.cc is a part of the PYTHIA event generator.
// Copyright (C) 2015 Torbjorn Sjostrand.
// PYTHIA is licenced under the GNU GPL version 2, see COPYING for details.
// Please respect the MCnet Guidelines, see GUIDELINES for details.

// Function definitions (not found in the header) for the
// DireSpace class.

#include "../include/DireSpace.h"

namespace Pythia8 {

//==========================================================================

// The DireSpace class.

//--------------------------------------------------------------------------

// Constants: could be changed here if desired, but normally should not.
// These are of technical nature, as described for each.

// Leftover companion can give PDF > 0 at small Q2 where other PDF's = 0,
// and then one can end in infinite loop of impossible kinematics.
const int    DireSpace::MAXLOOPTINYPDF = 10;

// Minimal allowed c and b quark masses, for flavour thresholds.
const double DireSpace::MCMIN          = 1.2;
const double DireSpace::MBMIN          = 4.0;

// Switch to alternative (but equivalent) backwards evolution for
// g -> Q Qbar (Q = c or b) when below QTHRESHOLD * mQ2.
const double DireSpace::CTHRESHOLD     = 2.0;
const double DireSpace::BTHRESHOLD     = 2.0;

// Renew evaluation of PDF's when the pT2 step is bigger than this
// (in addition to initial scale and c and b thresholds.)
const double DireSpace::EVALPDFSTEP    = 0.1;

// Lower limit on PDF value in order to avoid division by zero.
const double DireSpace::TINYPDF        = 1e-15;

// Lower limit on estimated evolution rate, below which stop.
const double DireSpace::TINYKERNELPDF  = 1e-6;

// Lower limit on pT2, below which branching is rejected.
const double DireSpace::TINYPT2        = 0.25e-6;

// No attempt to do backwards evolution of a heavy (c or b) quark
// if evolution starts at a scale pT2 < HEAVYPT2EVOL * mQ2.
const double DireSpace::HEAVYPT2EVOL   = 1.1;

// No attempt to do backwards evolution of a heavy (c or b) quark
// if evolution starts at a  x > HEAVYXEVOL * x_max, where
// x_max is the largest possible x value for a g -> Q Qbar branching.
const double DireSpace::HEAVYXEVOL     = 0.9;

// When backwards evolution Q -> g + Q creates a heavy quark Q,
// an earlier branching g -> Q + Qbar will restrict kinematics
// to  M_{Q Qbar}^2 > EXTRASPACEQ * 4 m_Q^2. (Smarter to be found??)
const double DireSpace::EXTRASPACEQ    = 2.0;

// Never pick pT so low that alphaS is evaluated too close to Lambda_3.
const double DireSpace::LAMBDA3MARGIN  = 1.1;

// Do not warn for large PDF ratios at small pT2 scales.
const double DireSpace::PT2MINWARN = 1.;

// Cutoff for f_e^e at x < 1 - 10^{-10} to be used in z selection.
// Note: the x_min quantity come from 1 - x_max.
const double DireSpace::LEPTONXMIN     = 1e-10;
const double DireSpace::LEPTONXMAX     = 1. - 1e-10;

// Stop l -> l gamma evolution slightly above m2l.
const double DireSpace::LEPTONPT2MIN   = 1.2;

// Enhancement of l -> l gamma trial rate to compensate imperfect modelling.
const double DireSpace::LEPTONFUDGE    = 10.;

// Overestimation extra factors by branching type
const double DireSpace::HEADROOMQ2G = 1.35;
const double DireSpace::HEADROOMQ2Q = 1.15;
const double DireSpace::HEADROOMG2Q = 1.35;
const double DireSpace::HEADROOMG2G = 1.35;

const double DireSpace::TINYMASS = 1e-3;

//--------------------------------------------------------------------------

// Initialize alphaStrong, alphaEM and related pTmin parameters.

void DireSpace::init( BeamParticle* beamAPtrIn,
  BeamParticle* beamBPtrIn) {

  // Header.
  if (printBanner) {
  cout << "\n"
       << " *-----------------  Welcome to DIRE version " << DIRE_SPACE_VERSION
       << "  ----------------*\n"
       << " |                                                "
       << "                  |\n"
       << " | Please consider citing Eur.Phys.J. C75 (2015)"
       << " 9, 461             |\n"
       << " | if you use this program for scientific purposes."
       << "                 |\n";
  cout << " |                                                "
       << "                  |\n";
  cout << " | You are using DIRE as spacelike shower.        "
       << "                  |\n";
  cout << " |                                                "
       << "                  |\n"
       << " *-----------------  Have a nice day!  -----------"
       << "------------------*" << endl;
  printBanner = false;}

  // Store input pointers for future use.
  beamAPtr        = beamAPtrIn;
  beamBPtr        = beamBPtrIn;

  // Main flags to switch on and off branchings.
  doQCDshower     = settingsPtr->flag("SpaceShower:QCDshower");

  // Matching in pT of hard interaction to shower evolution.
  pTmaxMatch      = settingsPtr->mode("SpaceShower:pTmaxMatch");
  pTdampMatch     = settingsPtr->mode("SpaceShower:pTdampMatch");
  pTmaxFudge      = settingsPtr->parm("SpaceShower:pTmaxFudge");
  pTmaxFudgeMPI   = settingsPtr->parm("SpaceShower:pTmaxFudgeMPI");
  pTdampFudge     = settingsPtr->parm("SpaceShower:pTdampFudge");

  // Optionally force emissions to be ordered in rapidity/angle.
  doRapidityOrder = settingsPtr->flag("SpaceShower:rapidityOrder");

  // Charm, bottom and lepton mass thresholds.
  mc              = max( MCMIN, particleDataPtr->m0(4));
  mb              = max( MBMIN, particleDataPtr->m0(5));
  m2c             = pow2(mc);
  m2b             = pow2(mb);

  // Parameters of scale choices.
  renormMultFac     = settingsPtr->parm("SpaceShower:renormMultFac");
  factorMultFac     = settingsPtr->parm("SpaceShower:factorMultFac");
  useFixedFacScale  = settingsPtr->flag("SpaceShower:useFixedFacScale");
  fixedFacScale2    = pow2(settingsPtr->parm("SpaceShower:fixedFacScale"));

  // Parameters of alphaStrong generation.
  alphaSvalue     = settingsPtr->parm("SpaceShower:alphaSvalue");
  alphaSorder     = settingsPtr->mode("SpaceShower:alphaSorder");
  alphaSnfmax     = settingsPtr->mode("StandardModel:alphaSnfmax");
  alphaSuseCMW    = settingsPtr->flag("SpaceShower:alphaSuseCMW");
  alphaS2pi       = 0.5 * alphaSvalue / M_PI;

  // Initialize alpha_strong generation.
  alphaS.init( alphaSvalue, alphaSorder, alphaSnfmax, alphaSuseCMW);

  // Set flavour thresholds by default Pythia masses, unless zero.
  double mcpy = particleDataPtr->m0(4);
  double mbpy = particleDataPtr->m0(5);
  double mtpy = particleDataPtr->m0(6);
  if (mcpy > 0.0 && mbpy > 0.0 && mtpy > 0.0)
    alphaS.setThresholds(mcpy, mbpy, mtpy);

  // Lambda for 5, 4 and 3 flavours.
  Lambda5flav     = alphaS.Lambda5();
  Lambda4flav     = alphaS.Lambda4();
  Lambda3flav     = alphaS.Lambda3();
  Lambda5flav2    = pow2(Lambda5flav);
  Lambda4flav2    = pow2(Lambda4flav);
  Lambda3flav2    = pow2(Lambda3flav);

  // Regularization of QCD evolution for pT -> 0. Can be taken
  // same as for multiparton interactions, or be set separately.
  useSamePTasMPI  = settingsPtr->flag("SpaceShower:samePTasMPI");
  if (useSamePTasMPI) {
    pT0Ref        = settingsPtr->parm("MultipartonInteractions:pT0Ref");
    ecmRef        = settingsPtr->parm("MultipartonInteractions:ecmRef");
    ecmPow        = settingsPtr->parm("MultipartonInteractions:ecmPow");
    pTmin         = settingsPtr->parm("MultipartonInteractions:pTmin");
  } else {
    pT0Ref        = settingsPtr->parm("SpaceShower:pT0Ref");
    ecmRef        = settingsPtr->parm("SpaceShower:ecmRef");
    ecmPow        = settingsPtr->parm("SpaceShower:ecmPow");
    pTmin         = settingsPtr->parm("SpaceShower:pTmin");
  }

  // Calculate nominal invariant mass of events. Set current pT0 scale.
  sCM             = m2( beamAPtr->p(), beamBPtr->p());
  eCM             = sqrt(sCM);
  pT0             = pT0Ref * pow(eCM / ecmRef, ecmPow);

  // Set regularisation parameter to zero.
  pT0 = 0.;

  // Restrict pTmin to ensure that alpha_s(pTmin^2 + pT_0^2) does not blow up.
  double pTminAbs = sqrtpos(pow2(LAMBDA3MARGIN) * Lambda3flav2 / renormMultFac
                  - pT0*pT0);
  if (pTmin < pTminAbs) {
    pTmin         = pTminAbs;
    ostringstream newPTmin;
    newPTmin << fixed << setprecision(3) << pTmin;
    infoPtr->errorMsg("Warning in DireSpace::init: pTmin too low",
                      ", raised to " + newPTmin.str() );
    infoPtr->setTooLowPTmin(true);
  }

  // Derived parameters of QCD evolution.
  pT20            = pow2(pT0);
  pT2min          = pow2(pTmin);

  usePDFalphas       = settingsPtr->flag("ShowerPDF:usePDFalphas");
  useSummedPDF       = settingsPtr->flag("ShowerPDF:useSummedPDF");
  BeamParticle& beam = (abs(beamAPtr->id()) == 2212) ? *beamAPtr : *beamBPtr;
  alphaS2piOverestimate = (usePDFalphas) ? beam.alphaS(pT2min) * 0.5/M_PI
                        : (alphaSorder > 0) ? alphaS.alphaS(pT2min) * 0.5/M_PI
                                            :  0.5 * 0.5/M_PI;

  usePDFmasses       = settingsPtr->flag("ShowerPDF:usePDFmasses");

  rejectProbability.clear();
  acceptProbability.clear();
  splittingSelName="";
  splittingNowName="";

  nFinalMax          = settingsPtr->mode("DireSpace:nFinalMax");
  useGlobalMapIF     = settingsPtr->flag("DireSpace:useGlobalMapIF");
  forceMassiveMap    = settingsPtr->flag("DireSpace:forceMassiveMap");

  // Various other parameters.
  doMEcorrections    = settingsPtr->flag("SpaceShower:MEcorrections");
  doMEafterFirst     = settingsPtr->flag("SpaceShower:MEafterFirst");
  doPhiPolAsym       = settingsPtr->flag("SpaceShower:phiPolAsym");
  doPhiIntAsym       = settingsPtr->flag("SpaceShower:phiIntAsym");
  strengthIntAsym    = settingsPtr->parm("SpaceShower:strengthIntAsym");
  nQuarkIn           = settingsPtr->mode("SpaceShower:nQuarkIn");

  // Possibility of two predetermined hard emissions in event.
  doSecondHard       = settingsPtr->flag("SecondHard:generate");

  // Optional dampening at small pT's when large multiplicities.
  enhanceScreening
    = settingsPtr->mode("MultipartonInteractions:enhanceScreening");
  if (!useSamePTasMPI) enhanceScreening = 0;

  // Possibility to allow user veto of emission step.
  hasUserHooks       = (userHooksPtr != 0);
  canVetoEmission    = (userHooksPtr != 0)
                     ? userHooksPtr->canVetoISREmission() : false;

}

//--------------------------------------------------------------------------

// Find whether to limit maximum scale of emissions.
// Also allow for dampening at factorization or renormalization scale.

bool DireSpace::limitPTmax( Event& event, double Q2Fac, double Q2Ren) {

  // Find whether to limit pT. Begin by user-set cases.
  bool dopTlimit = false;
  dopTlimit1 = dopTlimit2 = false;
  int nHeavyCol = 0;
  if      (pTmaxMatch == 1) dopTlimit = dopTlimit1 = dopTlimit2 = true;
  else if (pTmaxMatch == 2) dopTlimit = dopTlimit1 = dopTlimit2 = false;

  // Always restrict SoftQCD processes.
  else if (infoPtr->isNonDiffractive() || infoPtr->isDiffractiveA()
    || infoPtr->isDiffractiveB() || infoPtr->isDiffractiveC() )
    dopTlimit = dopTlimit1 = dopTlimit2 = true;

  // Look if any quark (u, d, s, c, b), gluon or photon in final state.
  // Also count number of heavy coloured particles, like top. 
  else {
    int n21 = 0;
    for (int i = 5; i < event.size(); ++i) {
      if (event[i].status() == -21) ++n21;
      else if (n21 == 0) {
        int idAbs = event[i].idAbs();
        if (idAbs <= 5 || idAbs == 21 || idAbs == 22) dopTlimit1 = true;
        if ( (event[i].col() != 0 || event[i].acol() != 0)
          && idAbs > 5 && idAbs != 21 ) ++nHeavyCol; 
      } else if (n21 == 2) {
        int idAbs = event[i].idAbs();
        if (idAbs <= 5 || idAbs == 21 || idAbs == 22) dopTlimit2 = true;
      }
    }
    dopTlimit = (doSecondHard) ? (dopTlimit1 && dopTlimit2) : dopTlimit1;
  }

  // Dampening at factorization or renormalization scale; only for hardest.
  dopTdamp   = false;
  pT2damp    = 0.;
  if (!dopTlimit1 && (pTdampMatch == 1 || pTdampMatch == 2)) {
    dopTdamp = true;
    pT2damp  = pow2(pTdampFudge) * ((pTdampMatch == 1) ? Q2Fac : Q2Ren);
  }
  if (!dopTlimit1 && nHeavyCol > 1 && (pTdampMatch == 3 || pTdampMatch == 4)) {
    dopTdamp = true;
    pT2damp  = pow2(pTdampFudge) * ((pTdampMatch == 3) ? Q2Fac : Q2Ren);
  }

  // Done.
  return dopTlimit;

}

//--------------------------------------------------------------------------

// Prepare system for evolution; identify ME.
// Routine may be called after multiparton interactions, for a new subystem.

void DireSpace::prepare( int iSys, Event& event, bool limitPTmaxIn) {

  // Find positions of incoming colliding partons.
  int in1 = partonSystemsPtr->getInA(iSys);
  int in2 = partonSystemsPtr->getInB(iSys);

  // Rescattered partons cannot radiate.
  bool canRadiate1 = !(event[in1].isRescatteredIncoming());
  bool canRadiate2 = !(event[in2].isRescatteredIncoming());

  // Reset dipole-ends list for first interaction. Also resonances.
  if (iSys == 0) dipEnd.resize(0);
  if (iSys == 0) idResFirst  = 0;
  if (iSys == 1) idResSecond = 0;

  // Find matrix element corrections for system.
  int MEtype = 0;

  // In case of DPS overwrite limitPTmaxIn by saved value.
  if (doSecondHard && iSys == 0) limitPTmaxIn = dopTlimit1;
  if (doSecondHard && iSys == 1) limitPTmaxIn = dopTlimit2;

  // Maximum pT scale for dipole ends.
  double pTmax1 = (limitPTmaxIn) ? event[in1].scale() : eCM;
  double pTmax2 = (limitPTmaxIn) ? event[in2].scale() : eCM;
  if ( limitPTmaxIn && (iSys == 0 || (iSys == 1 && doSecondHard)) ) {
    pTmax1 *= pTmaxFudge;
    pTmax2 *= pTmaxFudge;
  } else if (limitPTmaxIn && iSys > 0) {
    pTmax1 *= pTmaxFudgeMPI;
    pTmax2 *= pTmaxFudgeMPI;
  }

  // Find dipole ends for QCD radiation.
  // Note: colour type can change during evolution, so book also if zero.
  if (doQCDshower) {

    // Find dipole end formed by colour index.
    int colTag = event[in1].col();
    if (canRadiate1 && colTag > 0)  setupQCDdip( iSys, 1,  colTag,  1, event,
                                                 MEtype, limitPTmaxIn);
    // Find dipole end formed by anticolour index.
    int acolTag = event[in1].acol();
    if (canRadiate1 && acolTag > 0) setupQCDdip( iSys, 1, acolTag, -1, event,
                                                 MEtype, limitPTmaxIn);
    // Find dipole end formed by colour index.
    colTag = event[in2].col();
    if (canRadiate2 && colTag > 0)  setupQCDdip( iSys, 2,  colTag,  1, event,
                                                 MEtype, limitPTmaxIn);
    // Find dipole end formed by anticolour index.
    acolTag = event[in2].acol();
    if (canRadiate2 && acolTag > 0) setupQCDdip( iSys, 2, acolTag, -1, event,
                                                 MEtype, limitPTmaxIn);
  }

  // Store the z and pT2 values of the last previous splitting
  // when an event history has already been constructed.
  if (iSys == 0 && infoPtr->hasHistory()) {
    double zNow   = infoPtr->zNowISR();
    double pT2Now = infoPtr->pT2NowISR();
    for (int iDipEnd = 0; iDipEnd < int(dipEnd.size()); ++iDipEnd) {
      dipEnd[iDipEnd].zOld = zNow;
      dipEnd[iDipEnd].pT2Old = pT2Now;
      ++dipEnd[iDipEnd].nBranch;
    }
  }

  // Counter of proposed emissions.
  nProposedPT.clear();
  if ( nProposedPT.find(iSys) == nProposedPT.end() )
    nProposedPT.insert(make_pair(iSys,0));

  splittingSelName="";
  splittingNowName="";
  rejectProbability.clear();
  acceptProbability.clear();

}

//--------------------------------------------------------------------------

// Setup a dipole end for a QCD colour charge.

void DireSpace::setupQCDdip( int iSys, int side, int colTag, int colSign,
  Event& event, int MEtype, bool limitPTmaxIn) {

  // Initial values. Find if allowed to hook up beams.
  int iRad     = (side == 1) ? partonSystemsPtr->getInA(iSys)
                             : partonSystemsPtr->getInB(iSys);
  int iPartner = 0;
  int sizeAllA = partonSystemsPtr->sizeAll(iSys);
  int sizeOut  = partonSystemsPtr->sizeOut(iSys);
  int sizeAll  = sizeAllA;
  int sizeIn   = sizeAll - sizeOut;
  int sizeInA  = sizeAllA - sizeIn - sizeOut;

  // Colour: other end by same index in final state or opposite in beam.
  if (colSign > 0)
  for (int j = 0; j < sizeAll; ++j) {
    int iRecNow = partonSystemsPtr->getAll(iSys, j + sizeInA);
    if (iRecNow == iRad) continue;
    if ( ( j >= sizeIn && event[iRecNow].col() == colTag
      && event[iRecNow].isFinal() )
      || ( j <  sizeIn && event[iRecNow].acol() == colTag
      && !event[iRecNow].isRescatteredIncoming() ) ) {
      iPartner = iRecNow;
      break;
    }
  }

  // Anticolour: other end by same index in final state or opposite in beam.
  if (colSign < 0)
  for (int j = 0; j < sizeAll; ++j) {
    int iRecNow = partonSystemsPtr->getAll(iSys, j + sizeInA);
    if (iRecNow == iRad) continue;
    if ( ( j >= sizeIn && event[iRecNow].acol()  == colTag
      && event[iRecNow].isFinal() )
      || ( j <  sizeIn && event[iRecNow].col() == colTag
      && !event[iRecNow].isRescatteredIncoming() ) ) {
      iPartner = iRecNow;
      break;
    }
  }

  // Check for failure to locate any recoiler
  if ( iPartner == 0 ) {
    infoPtr->errorMsg("Error in DireSpace::setupQCDdip: "
                      "failed to locate any recoiling partner");
    return;
  }

  // Store dipole colour end(s).
  // Max scale either by parton scale or by half dipole mass.
  double pTmax = event[iRad].scale();
  if (limitPTmaxIn) {
    if (iSys == 0 || (iSys == 1 && doSecondHard)) pTmax *= pTmaxFudge;
    else if (sizeIn > 0) pTmax *= pTmaxFudgeMPI;
  } else pTmax = 0.5 * m( event[iRad], event[iPartner]);
  int colType  = (event[iRad].id() == 21) ? 2 * colSign : colSign;
  dipEnd.push_back( DireSpaceEnd( iSys, side, iRad, iPartner, pTmax, colType,
                                  0, 0, MEtype, true) );

}

//--------------------------------------------------------------------------

// Update dipole list after each ISR emission (so not used for resonances).

void DireSpace::update( int iSys, Event& event, bool) {

  // Find positions of incoming colliding partons.
  int in1 = partonSystemsPtr->getInA(iSys);
  int in2 = partonSystemsPtr->getInB(iSys);

  // Rescattered partons cannot radiate.
  bool canRadiate1 = !(event[in1].isRescatteredIncoming()) && doQCDshower;
  bool canRadiate2 = !(event[in2].isRescatteredIncoming()) && doQCDshower;

  // Find matrix element corrections for system.
  int MEtype = 0;

  dipEnd.resize(0);
  // Find dipole end formed by colour index.
  int colTag = event[in1].col();
  if (canRadiate1 && colTag > 0)  setupQCDdip( iSys, 1,  colTag,  1, event,
                                               MEtype, true);
  // Find dipole end formed by anticolour index.
  int acolTag = event[in1].acol();
  if (canRadiate1 && acolTag > 0) setupQCDdip( iSys, 1, acolTag, -1, event,
                                               MEtype, true);

  // Find dipole end formed by colour index.
  colTag = event[in2].col();
  if (canRadiate2 && colTag > 0)  setupQCDdip( iSys, 2,  colTag,  1, event,
                                               MEtype, true);
  // Find dipole end formed by anticolour index.
  acolTag = event[in2].acol();
  if (canRadiate2 && acolTag > 0) setupQCDdip( iSys, 2, acolTag, -1, event,
                                               MEtype, true);

}

//--------------------------------------------------------------------------

// Select next pT in downwards evolution of the existing dipoles.

double DireSpace::pTnext( Event& event, double pTbegAll, double pTendAll,
  int nRadIn, bool) {

  // Current cm energy, in case it varies between events.
  sCM           = m2( beamAPtr->p(), beamBPtr->p());
  eCM           = sqrt(sCM);
  pTbegRef      = pTbegAll;

  // Starting values: no radiating dipole found.
  nRad          = nRadIn;
  double pT2sel = pow2(pTendAll);
  iDipSel       = 0;
  iSysSel       = 0;
  dipEndSel     = 0;
  splittingNowName="";
  splittingSelName="";

  // Loop over all possible dipole ends.
  for (int iDipEnd = 0; iDipEnd < int(dipEnd.size()); ++iDipEnd) {
    iDipNow        = iDipEnd;
    dipEndNow      = &dipEnd[iDipEnd];
    iSysNow        = dipEndNow->system;
    dipEndNow->pT2 = 0.;
    double pTbegDip = min( pTbegAll, dipEndNow->pTmax );

    // Limit final state multiplicity. For debugging only
    int nFinal = 0;
    for (int i=0; i < event.size(); ++i)
      if (event[i].isFinal()) nFinal++;
    if (nFinalMax > -10 && nFinal > nFinalMax) continue;

    // Check whether dipole end should be allowed to shower.
    double pT2begDip = pow2(pTbegDip);
    if (pT2begDip > pT2sel && dipEndNow->colType != 0 ) {
      double pT2endDip = 0.;

      // Determine lower cut for evolution for QCD
//      if (dipEndNow->colType != 0)
//        pT2endDip = max( pT2sel, pT2min );
      pT2endDip = pT2min;

      // Find properties of dipole and radiating dipole end.
      sideA        = ( abs(dipEndNow->side) == 1 );

      bool finalRec = event[dipEndNow->iRecoiler].isFinal(); 
      BeamParticle& beamNow = (sideA) ? *beamAPtr : *beamBPtr;
      BeamParticle& beamRec = (sideA) ? *beamBPtr : *beamAPtr;
      iNow         = beamNow[iSysNow].iPos();
      iRec         = (finalRec) ? dipEndNow->iRecoiler
                                : beamRec[iSysNow].iPos();
      idDaughter   = beamNow[iSysNow].id();
      xDaughter    = beamNow[iSysNow].x();
      x1Now        = (sideA) ? xDaughter : beamRec[iSysNow].x();
      x2Now        = (sideA) ? beamRec[iSysNow].x() : xDaughter;

      // Note dipole mass correction when recoiler is a rescatter.
      m2Rec        = (dipEndNow->normalRecoil) ? 0. : event[iRec].m2();
      m2Dip       = abs(2.*event[iNow].p()*event[iRec].p());

      // Now do evolution in pT2, for QCD
      if (pT2begDip > pT2endDip) {

        if (dipEndNow->colType != 0){
          // Regular shower.
          pT2nextQCD( pT2begDip, pT2endDip, *dipEndNow, event);
        }

        // Update if found larger pT than current maximum.
        if (dipEndNow->pT2 > pT2sel) {
          pT2sel    = dipEndNow->pT2;
          iDipSel   = iDipNow;
          iSysSel   = iSysNow;
          dipEndSel = dipEndNow;
          splittingSelName = splittingNowName;
        }

      }
    }
  // End loop over dipole ends.
  }

  weights->insertWeights(acceptProbability, rejectProbability);
  rejectProbability.clear();
  acceptProbability.clear();

  // Return nonvanishing value if found pT is bigger than already found.
  return (dipEndSel == 0) ? 0. : sqrt(pT2sel);

}

//--------------------------------------------------------------------------

double DireSpace::enhanceFactors( string name, bool isValence, double m2dip,
  double pT2Old ) {

  double factor = 1.;

  // Additional weight to smooth out valence bump.
  if (isValence && name.find("isr_1->1&21") != string::npos)
    factor *= log(max(2.71828,16/(pT2Old/m2dip))); 

  // Additional enhancement for G->QQ, to smooth out PDF factors.
  if (name.find("isr_21->1&1") != string::npos)
    factor *= log(max(2.71828,log(max(2.71828,m2dip/pT2Old))
                    + pow(m2dip/pT2Old,3./2.)));

  // Artificial constant increase of overestimate.
  double MARGIN = 1.;
  if (name.find("isr_1->1&21") != string::npos && !isValence)
    MARGIN = 1.15;
  if (name.find("isr_21->1&1") != string::npos)
    MARGIN = 1.65; 
  if (name.find("isr_21->21&21a") != string::npos && pT2Old < 2.0)
    MARGIN = 1.25; 
  if (name.find("isr_21->21&21b") != string::npos && pT2Old < 2.0)
    MARGIN = 1.25; 
  factor *= MARGIN;

  return factor;

}

//--------------------------------------------------------------------------

// Function to generate new user-defined overestimates to evolution.

void DireSpace::getNewOverestimates( int type, int idDau, DireSpaceEnd* dip,
  double pT2Old, double xDau, double zMinAbs, double zMaxAbs,
  multimap<double,string>& newOverestimates ) {

  // Get beam for correction factors.
  BeamParticle& beam = (sideA) ? *beamAPtr : *beamBPtr;
  bool   isValence   = beam[iSysNow].isValence();

  double sum=0.;

  // Loop over splitting names and get overestimates.
  vector<Splitting* > splits = splittingsPtr->getSplittingNames();

  for (int i = 0; i < int(splits.size()); ++i) {

    int typeNow = splits[i]->type();

    if ( !(typeNow >= type && typeNow/100 == type/100) ) continue;
    if ( !(typeNow >= type && typeNow/1000 == type/1000) ) continue;
    string name          = splits[i]->name(); 

    // Get overestimate (of splitting kernel only)
    double wt = splits[i]->overestimateInt(zMinAbs,zMaxAbs,pT2Old,dip->m2Dip);

    // Get current PDF value.
    double scale2 = (useFixedFacScale) ? fixedFacScale2 : factorMultFac*pT2Old;
    bool inD = beam.insideBounds(xDau, scale2); 
    double xPDFdaughter = (useSummedPDF) ? beam.xf(idDau, xDau, scale2)
                        : beam.xfISR(iSysNow,idDau, xDau, scale2);

    // PDF factors for Q -> GQ.
    double pdfRatio = 1.;
    if (name.find("isr_1->21&1") != string::npos) {
      // Sum over all potential quark mothers to a g.
      double xPDFmother = 0.;
      for (int iQuark = -nQuarkIn; iQuark <= nQuarkIn; ++iQuark) {
        if (iQuark != 0) xPDFmother += (useSummedPDF)
                                    ? beam.xf(iQuark, xDau, scale2)
                                    : beam.xfISR(iSysNow,iQuark, xDau, scale2);
      }
      pdfRatio = xPDFmother / xPDFdaughter;

    // PDF factors for G -> QQ
    } else if (name.find("isr_21->1&1") != string::npos) {
      double xPDFmother = (useSummedPDF)
                        ? beam.xf(21, xDau, scale2)
                        : beam.xfISR(iSysNow,21, xDau, scale2);
      if ( abs(xPDFmother) < TINYPDF) {
        int sign   = (xPDFmother > 0.) ? 1 : -1;
        xPDFmother = sign*TINYPDF;
      }
      pdfRatio = xPDFmother / xPDFdaughter;
    }

    if (pdfRatio < 0. || abs(xPDFdaughter) < TINYPDF) pdfRatio = 0.;
    if (!inD) pdfRatio = 0.;

    // Include PDF ratio for Q->GQ or G->QQ.
    wt *= pdfRatio;

    // Include artificial enhancements.
    wt *= enhanceFactors(name, isValence, dip->m2Dip, pT2Old);

    // Save this overestimate.
    // Do not include zeros (could lead to trouble with lower_bound?)
    if (wt != 0.) {
      sum += abs(wt);
      newOverestimates.insert(make_pair(sum,name));
    }

  }

}

//--------------------------------------------------------------------------

// Function to generate new user-defined overestimates to evolution.

void DireSpace::getNewSplitting( const Event& state, DireSpaceEnd* dip,
  double pT2Old, double xDau, double pT2, double zMinAbs, double zMaxAbs,
  int idDau, string name, int& idMother, int& idSister, double& z, double& wt,
  double& full, double& over ) {

  BeamParticle& beam = (sideA) ? *beamAPtr : *beamBPtr;
  bool   isValence   = beam[iSysNow].isValence();

  // Return auxiliary variable, overestimate, mother and sister ids.
  z           = (*splittingsPtr)[name]->zSplit(zMinAbs, zMaxAbs, dip->m2Dip);
  over        = (*splittingsPtr)[name]->overestimateDiff(z, dip->m2Dip);
  idMother    = (*splittingsPtr)[name]->motherID(idDau);
  idSister    = (*splittingsPtr)[name]->sisterID(idDau);

  // Get old PDF for PDF weights.
  double scale2 = (useFixedFacScale) ? fixedFacScale2 : factorMultFac * pT2Old;
  double xPDFdaughter = (useSummedPDF) ? beam.xf(idDau, xDau, scale2)
                      : beam.xfISR(iSysNow,idDau, xDau, scale2);
  bool inD = beam.insideBounds(xDau, scale2); 

  // PDF factors for Q -> GQ.
  double pdfRatio = 1.;
  if (name.find("isr_1->21&1") != string::npos) {

    // Get sum of PDFs for all potential quark mothers to a g.
    double xPDFmotherSum = 0.;
    double xPDFmother[21] = {0.};
    for (int iQuark = -nQuarkIn; iQuark <= nQuarkIn; ++iQuark) {
      if (iQuark == 0) {
        xPDFmother[10] = 0.;
      } else {
        xPDFmother[iQuark+10] = (useSummedPDF)
                              ? beam.xf(iQuark, xDau, scale2)
                              : beam.xfISR(iSysNow,iQuark, xDau, scale2);
        xPDFmotherSum += xPDFmother[iQuark+10];
      }
    }
    // Pick flavour for Q -> GQ (picked here, as kernel might depend on
    // flavour)
    double temp = xPDFmotherSum * rndmPtr->flat();
    idMother = -nQuarkIn - 1;
    do { temp -= xPDFmother[(++idMother) + 10]; }
    while (temp > 0. && idMother < nQuarkIn);
    idSister = idMother;
    pdfRatio = xPDFmother[idMother + 10]/xPDFdaughter;
  // PDF factors for G -> QQ
  } else if (name.find("isr_21->1&1") != string::npos) {
    double xPDFmother = (useSummedPDF)
                      ? beam.xf(21, xDau, scale2)
                      : beam.xfISR(iSysNow,21, xDau, scale2);
    if ( abs(xPDFmother) < TINYPDF) {
      int sign   = (xPDFmother > 0.) ? 1 : -1;
      xPDFmother = sign*TINYPDF;
    }
    pdfRatio = xPDFmother / xPDFdaughter;
  }

  // Get particle masses.
  double m2Bef = 0.0;
  double m2r   = 0.0;
  double m2s   = 0.0;
  int type     = (state[dip->iRecoiler].isFinal()) ? 1 : -1;
  if ( state[dip->iRecoiler].idAbs() < 6 && type == 1 )
    m2s = getMass(state[dip->iRecoiler].id(),2);
  else if ( type == 1 )
    m2s = getMass(state[dip->iRecoiler].id(),3,
          state[dip->iRecoiler].mCalc());

  // Force emission massless for now.
  double m2e   = 0.0;

  // Upate type if this is a massive splitting.
  if (type == 1 && (m2Bef > TINYMASS || m2r > TINYMASS || m2e > TINYMASS
    || m2s > TINYMASS)) type = 2;
  if (type ==-1 && (m2Bef > TINYMASS || m2r > TINYMASS || m2e > TINYMASS
    || m2s > TINYMASS)) type =-2;

  // Adjust the dipole kinematical mass to accomodate masses after branching.
  //double m2dipCorr  = dip->m2Dip + m2Bef - m2r - m2e;
  double m2dipCorr  = dip->m2Dip - m2Bef + m2r + m2e;
  // Check phase space contraints.
  if ( !inAllowedPhasespace( z, pT2, m2dipCorr, xDau, type, m2Bef, m2r, m2s,
    m2e) ) {
    wt = full = over = 0.; return;
  }

  // Return overestimate.
  over        = (*splittingsPtr)[name]->overestimateDiff(z, dip->m2Dip);

  // Get complete kernel.
  full        = (*splittingsPtr)[name]->kernel(z, pT2, m2dipCorr, type,
                  m2Bef, m2r, m2s, m2e);

  // Calculate accept probability.
  wt          = full/over;

  if (pdfRatio < 0. || abs(xPDFdaughter) < TINYPDF) pdfRatio = 0.;
  if (!inD) pdfRatio = 0.;

  // Divide out PDF ratio used in overestimate.
  if (pdfRatio != 0.) wt /= pdfRatio;
  else wt = 0.;
  over *= pdfRatio;

  // Divide out artificial enhancements.
  double headRoom = enhanceFactors(name, isValence, dip->m2Dip, pT2Old);
  wt   /= headRoom;
  over *= headRoom;

  // Ensure positive weight.
  wt = abs(wt);

}

//--------------------------------------------------------------------------

// Check if variables are in allowed phase space.

bool DireSpace::inAllowedPhasespace( double z, double pT2, double m2dip,
  double xOld, int splitType, double m2RadBef, double m2r, double m2s, double m2e) {

  // splitType == 1 -> Massless IF
  if (splitType == 1) {

    // Calculate CS variables.
    double xCS = z;
    double uCS = (pT2/m2dip)/(1-z);

    // Forbidden emission if outside allowed z range for given pT2.
    if ( xCS < xOld || xCS > 1. || uCS < 0. || uCS > 1. ) return false;

  // splitType == 2 -> Massive IF
  } else if (splitType == 2) {

    // Calculate CS variables.
    double xCS = z;
    double uCS = (pT2/m2dip)/(1-z);

    // Construct massive phase space boundary for uCS
    double pijpa  = m2dip - m2r - m2e + m2RadBef;
    double mu2Rec = m2s / pijpa * xCS;
    double uCSmax = (1. - xCS) / (1. - xCS + mu2Rec );

    // Forbidden emission if outside allowed z range for given pT2.
    if ( xCS < xOld || xCS > 1. || uCS < 0. || uCS > uCSmax ) return false;

  // splitType ==-1 -> Massless II
  } else if (splitType < 0) {

    // Calculate CS variables.
    double kappa2 = pT2 / m2dip;
    double xCS    = (z*(1-z)- kappa2)/(1-z);
    double vCS    = kappa2/(1-z); 

    // Forbidden emission if outside allowed z range for given pT2.
    if (xCS < xOld || xCS > 1. || vCS < 0. || vCS > 1.) return false;

  }

  return true;

}

//--------------------------------------------------------------------------

// Function to add user-defined overestimates to old overestimate.

void DireSpace::addNewOverestimates( multimap<double,string> newOverestimates,
  double& oldOverestimate ) {

  // No other tricks necessary at the moment.
  if (!newOverestimates.empty())
    oldOverestimate += newOverestimates.rbegin()->first;

  // Done.

}

//--------------------------------------------------------------------------

// Evolve a QCD dipole end.

void DireSpace::pT2nextQCD( double pT2begDip, double pT2endDip,
  DireSpaceEnd& dip, Event& event) {

  if (event[dip.iRecoiler].isFinal()) { 
    pT2nextQCD_IF(pT2begDip, pT2endDip, dip, event);
  } else { 
    pT2nextQCD_II(pT2begDip, pT2endDip, dip, event);
  }

  // Done
}

//--------------------------------------------------------------------------

// Evolve a QCD dipole end.

void DireSpace::pT2nextQCD_II( double pT2begDip, double pT2endDip,
  DireSpaceEnd& dip, Event& event) {

  // Reset dipole mass.
  int iRadi  = dip.iRadiator; 
  int iReco  = dip.iRecoiler;
  m2Dip      = abs(2.*event[iRadi].p()*event[iReco].p());
  dip.m2Dip  = m2Dip;

  // Some properties and kinematical starting values.
  BeamParticle& beam = (sideA) ? *beamAPtr : *beamBPtr;
  bool   isGluon     = (idDaughter == 21);
  double pT2         = pT2begDip;
  double xMaxAbs     = beam.xMax(iSysNow);
  double zMinAbs     = xDaughter;

  if (xMaxAbs < 0.) {
    infoPtr->errorMsg("Warning in DireSpace::pT2nextQCD: "
    "xMaxAbs negative");
    return;
  }

  // Variables used inside evolution loop. (Mainly dummy starting values.)
  int    nFlavour       = 3;
  double b0             = 4.5;
  double Lambda2        = Lambda3flav2;
  double pT2minNow      = pT2endDip;
  int    idMother       = 0;
  int    idSister       = 0;
  double z              = 0.;
  double zMaxAbs        = 0.;
  double xPDFdaughter   = 0.;
  double kernelPDF      = 0.;
  double xMother        = 0.;
  double wt             = 0.;
  double Q2             = 0.;
  double mSister        = 0.;
  double m2Sister       = 0.;
  double pT2corr        = 0.;
  double pT2PDF         = pT2;
  bool   needNewPDF     = true;

  multimap<double,string> newOverestimates;
  int type = -1;
  if (isGluon) type -= 200;
  else type -= 100;

  double fullWeightNow = 0.;
  double overWeightNow = 0.;
  double auxWeightNow = 0.;

  double m2cPhys = (usePDFalphas) ? pow2(max(0.,beam.mQuarkPDF(4)))
                 : alphaS.muThres2(4);
  double m2bPhys = (usePDFalphas) ? pow2(max(0.,beam.mQuarkPDF(5)))
                 : alphaS.muThres2(5);

  // Begin evolution loop towards smaller pT values.
  int    loopTinyPDFdau = 0;
  bool   hasTinyPDFdau  = false;
  do {
    wt = 0.;

    splittingNowName="";

    // Update event weight after one step.  
    if ( fullWeightNow != 0. && overWeightNow != 0. ) {
      double w = auxWeightNow/overWeightNow
               * (overWeightNow-fullWeightNow)
               / (auxWeightNow - fullWeightNow);
      rejectProbability.insert( make_pair(pT2,w));
    }

    fullWeightNow = overWeightNow = auxWeightNow = 0.;

    // Leave unconverted for now.
    if ( abs(idDaughter) == 4 && pT2 <= m2cPhys ) return;
    if ( abs(idDaughter) == 5 && pT2 <= m2bPhys ) return;
 
    // Bad sign if repeated looping with small daughter PDF, so fail.
    // (Example: if all PDF's = 0 below Q_0, except for c/b companion.)
    if (hasTinyPDFdau) ++loopTinyPDFdau;
    if (loopTinyPDFdau > MAXLOOPTINYPDF) {
      infoPtr->errorMsg("Warning in DireSpace::pT2nextQCD: "
      "small daughter PDF");
      return;
    }

    // Initialize integrals of splitting kernels and evaluate parton
    // densities at the beginning. Reinitialize after long evolution
    // in pT2 or when crossing c and b flavour thresholds.
    if (needNewPDF 
      || pT2 < evalpdfstep(event[iRadi].id(),pT2,m2cPhys,m2bPhys) * pT2PDF) {
      pT2PDF        = pT2;
      hasTinyPDFdau = false;

      newOverestimates.clear();
      kernelPDF = 0.;

      // Determine overestimated z range; switch at c and b masses.
      if (pT2 > m2b) {
        nFlavour  = 5;
        pT2minNow = max( m2b, pT2endDip);
        b0        = 23./6.;
        Lambda2   = Lambda5flav2;
      } else if (pT2 > m2c) {
        nFlavour  = 4;
        pT2minNow = max( m2c, pT2endDip);
        b0        = 25./6.;
        Lambda2   = Lambda4flav2;
      } else {
        nFlavour  = 3;
        pT2minNow = pT2endDip;
        b0        = 27./6.;
        Lambda2   = Lambda3flav2;
      }

      // A change of renormalization scale expressed by a change of Lambda.
      Lambda2    /= renormMultFac;
      zMaxAbs = 1.;

      // Parton density of daughter at current scale.
      pdfScale2    = (useFixedFacScale) ? fixedFacScale2 : factorMultFac*pT2;
      xPDFdaughter = (useSummedPDF)
                   ? beam.xf(idDaughter, xDaughter, pdfScale2)
                   : beam.xfISR(iSysNow, idDaughter, xDaughter, pdfScale2);
      if (abs(xPDFdaughter) < TINYPDF) {
        int sign      = (xPDFdaughter > 0.) ? 1 : -1;
        xPDFdaughter  = sign*TINYPDF;
        hasTinyPDFdau = true;
      }

      // Calculate and add user-defined overestimates.
      getNewOverestimates( type, idDaughter, dipEndNow, pT2PDF, xDaughter, 
        zMinAbs, zMaxAbs, newOverestimates);
      addNewOverestimates(newOverestimates, kernelPDF);

      // End evaluation of splitting kernels and parton densities.
      needNewPDF = false;
    }

    if (kernelPDF < TINYKERNELPDF) return;

    // Pick pT2 (in overestimated z range), for one of three different cases.
    // Assume form alphas(pT0^2 + pT^2) * dpT^2/(pT0^2 + pT^2).
    double Q2alphaS;

    // Fixed alpha_strong, reweighted later to PDF running alpha_s.
    if (usePDFalphas) {
      pT2 = (pT2 + pT20) * pow( rndmPtr->flat(),
        1. / (alphaS2piOverestimate * kernelPDF)) - pT20;

    // Fixed alpha_strong.
    } else if (alphaSorder == 0) {
      pT2 = (pT2 + pT20) * pow( rndmPtr->flat(),
        1. / (alphaS2pi * kernelPDF)) - pT20;

    // First-order alpha_strong.
    } else if (alphaSorder == 1) {
      pT2 = Lambda2 * pow( (pT2 + pT20) / Lambda2,
        pow(rndmPtr->flat(), b0 / kernelPDF) ) - pT20;

    // For second order reject by second term in alpha_strong expression.
    } else {
      do {
        pT2 = Lambda2 * pow( (pT2 + pT20) / Lambda2,
          pow(rndmPtr->flat(), b0 / kernelPDF) ) - pT20;
        Q2alphaS = renormMultFac * max( pT2 + pT20,
          pow2(LAMBDA3MARGIN) * Lambda3flav2);
      } while (alphaS.alphaS2OrdCorr(Q2alphaS) < rndmPtr->flat()
        && pT2 > pT2minNow);
    }

    // Abort evolution if not allowed.
    if (pT2 < pT2endDip) return;

    // Check for pT2 values that prompt special action.
    // If crossed b threshold, continue evolution from this threshold.
    if (nFlavour == 5 && pT2 < m2bPhys) {
      needNewPDF = true;
    // If crossed c threshold, continue evolution from this threshold.
    } else if (nFlavour == 4 && pT2 < m2cPhys) {
      needNewPDF = true;
    }

    // Leave heavy quarks below threshold unconverted for now.
    if (abs(idDaughter) == 5 && pT2 <= m2bPhys) {
      return;
    } else if (abs(idDaughter) == 4 && pT2 <= m2cPhys) {
      return;
    }

    // Select z value of branching to g, and corrective weight.
    // User-defined splittings.
    double R = kernelPDF*rndmPtr->flat();
    if (!newOverestimates.empty()) {
      if (newOverestimates.lower_bound(R) == newOverestimates.end())
        splittingNowName = newOverestimates.rbegin()->second;
      else
        splittingNowName = newOverestimates.lower_bound(R)->second;
      getNewSplitting( event, dipEndNow, pT2PDF, xDaughter, pT2, zMinAbs,
        zMaxAbs, idDaughter, splittingNowName, idMother, idSister, z, wt,
        fullWeightNow, overWeightNow);
    }

    // Impossible emission (e.g. if outside allowed z range for given pT2).
    if ( wt == 0.) {
      wt = fullWeightNow = overWeightNow = auxWeightNow = 0.;
      continue;
    }

    // Disallow gluon -> heavy quarks if pT has fallen below 2*mQuark.
    if ( abs(idDaughter) == 5 && abs(idSister) == 5
      && pT2 <= 4.*m2bPhys) {
      wt = fullWeightNow = overWeightNow = auxWeightNow = 0.;
      continue;
    } else if ( abs(idDaughter) == 4 && abs(idSister) == 4
      && pT2 <= 4.*m2cPhys) {
      wt = fullWeightNow = overWeightNow = auxWeightNow = 0.;
      continue;
    }

    // Get particle masses.
    double m2Bef = 0.0;
    double m2r   = 0.0;
    // Force emission massless for now.
    double m2e   = 0.0;

    // Adjust the dipole kinematical mass to accomodate masses after branching.
    double m2dipCorr  = m2Dip - m2Bef + m2r + m2e;
    // Calculate CS variables.
    double kappa2 = pT2 / m2dipCorr;
    double xCS    = (z*(1-z)- kappa2)/(1-z);
    xMother = xDaughter/xCS;

    // Evaluation of new daughter and mother PDF's.
    pdfScale2 = (useFixedFacScale) ? fixedFacScale2 : factorMultFac * pT2;
    bool inD = beam.insideBounds(xDaughter, pdfScale2); 
    bool inM = beam.insideBounds(xMother,   pdfScale2); 
    double xPDFdaughterNew = 
      (useSummedPDF) ? beam.xf(idDaughter, xDaughter, pdfScale2)
                     : beam.xfISR(iSysNow, idDaughter, xDaughter, pdfScale2);
    double xPDFmotherNew =
      (useSummedPDF) ? beam.xf(idMother, xMother, pdfScale2)
                     : beam.xfISR(iSysNow, idMother, xMother, pdfScale2);
    if ( abs(xPDFdaughterNew) < TINYPDF ) {
      hasTinyPDFdau = true;
      needNewPDF    = true;
      wt = fullWeightNow = overWeightNow = auxWeightNow = 0.;
      continue;
    }

    // Set PDF ratio to zero if x-/pT-values are out-of-bounds.
    double pdfRatio = (inD && inM) ? xPDFmotherNew/xPDFdaughterNew : 0.;

    wt             *= pdfRatio;
    fullWeightNow  *= pdfRatio;

    // Reweight to PDF alpha_s, if necessary.
    wt            *= (usePDFalphas)
                   ? beam.alphaS(pT2)/(alphaS2piOverestimate*2.*M_PI)
                   : 1.; 
    fullWeightNow *= (usePDFalphas)
                   ? beam.alphaS(pT2)/(2.*M_PI)
                   : (alphaSorder==0) ? alphaS2pi
                                      : alphaS.alphaS(pT2)/(2.*M_PI); 
    overWeightNow *= (usePDFalphas)
                   ? alphaS2piOverestimate
                   : (alphaSorder==0) ? alphaS2pi
                                      : alphaS.alphaS(pT2)/(2.*M_PI);
    auxWeightNow   = overWeightNow;

    // Ensure that accept probability is positive.
    if (fullWeightNow < 0.) auxWeightNow *= -1.;

    if (wt == 0. ) {
      needNewPDF = true;
      wt = fullWeightNow = overWeightNow = auxWeightNow = 0.;
      continue;
    }

    // Reset overestimate if necessary.
    if ( fullWeightNow/auxWeightNow > 1.) {
      double rescale = fullWeightNow/auxWeightNow * 1.5;
      auxWeightNow *= rescale;
      wt /= rescale;
      needNewPDF = true;
      infoPtr->errorMsg("Info in DireSpace::pT2nextQCD_II: Found large "
                        "acceptance weight for " + splittingNowName);
    }

    // Check that valence step does not cause problem.
    if (wt > 1. && pT2 > PT2MINWARN) { infoPtr->errorMsg("Warning in "
      "SpaceShower::pT2nextQCD: weight above unity");
    }

  // Iterate until acceptable pT (or have fallen below pTmin).
  } while (wt < rndmPtr->flat()) ;

  // Update accepted event weight.  
  if ( fullWeightNow != 0. && overWeightNow != 0. ) {
    acceptProbability.insert( make_pair(pT2,auxWeightNow/overWeightNow));
    double w = auxWeightNow/overWeightNow
             * (overWeightNow- fullWeightNow)
             / (auxWeightNow - fullWeightNow);
    rejectProbability.insert( make_pair(pT2, w));
  }

  // Save values for (so far) acceptable branching.
  dipEndNow->store( idDaughter,idMother, idSister, x1Now, x2Now, m2Dip,
    pT2, z, xMother, Q2, mSister, m2Sister, pT2corr);

}

//--------------------------------------------------------------------------

// Evolve a QCD dipole end.

void DireSpace::pT2nextQCD_IF( double pT2begDip, double pT2endDip,
  DireSpaceEnd& dip, Event& event) {

  // Reset dipole mass.
  int iRadi  = dip.iRadiator; 
  int iReco  = dip.iRecoiler;
  m2Dip      = abs(2.*event[iRadi].p()*event[iReco].p());
  dip.m2Dip  = m2Dip;

  // Some properties and kinematical starting values.
  //BeamParticle& beam = (sideA) ? *beamAPtr : *beamBPtr;
  BeamParticle& beam = (sideA && abs(beamAPtr->id()) == 2212)
                     ? *beamAPtr
                     : ((abs(beamBPtr->id()) == 2212) ? *beamBPtr
                                                      : *beamAPtr );
  bool   isGluon     = (idDaughter == 21);
  double pT2         = pT2begDip;
  double xMaxAbs     = beam.xMax(iSysNow);
  double zMinAbs     = xDaughter;

  if (xMaxAbs < 0.) {
    infoPtr->errorMsg("Warning in DireSpace::pT2nextQCD: "
    "xMaxAbs negative");
    return;
  }

  // Variables used inside evolution loop. (Mainly dummy starting values.)
  int    nFlavour       = 3;
  double b0             = 4.5;
  double Lambda2        = Lambda3flav2;
  double pT2minNow      = pT2endDip;
  int    idMother       = 0;
  int    idSister       = 0;
  double z              = 0.;
  double zMaxAbs        = 0.;
  double kernelPDF      = 0.;
  double xMother        = 0.;
  double wt             = 0.;
  double Q2             = 0.;
  double mSister        = 0.;
  double m2Sister       = 0.;
  double pT2corr        = 0.;
  double pT2PDF         = pT2;
  bool   needNewPDF     = true;

  multimap<double,string> newOverestimates;
  int type = -1;
  if (isGluon) type -= 200;
  else type -= 100;

  double fullWeightNow = 0.;
  double overWeightNow = 0.;
  double auxWeightNow = 0.;

  double m2cPhys = (usePDFalphas) ? pow2(max(0.,beam.mQuarkPDF(4)))
                 : alphaS.muThres2(4);
  double m2bPhys = (usePDFalphas) ? pow2(max(0.,beam.mQuarkPDF(5)))
                 : alphaS.muThres2(5);

  // Begin evolution loop towards smaller pT values.
  int    loopTinyPDFdau = 0;
  bool   hasTinyPDFdau  = false;
  do {
    wt = 0.;

    splittingNowName="";

    // Update event weight after one step.  
    if ( fullWeightNow != 0. && overWeightNow != 0. ) {
      double w = auxWeightNow/overWeightNow
               * (overWeightNow-fullWeightNow)
               / (auxWeightNow - fullWeightNow);
      rejectProbability.insert( make_pair(pT2,w));
    }

    fullWeightNow = overWeightNow = auxWeightNow = 0.;

    // Leave unconverted for now.
    if ( abs(idDaughter) == 4 && pT2 <= m2cPhys ) return;
    if ( abs(idDaughter) == 5 && pT2 <= m2bPhys ) return;
 
    // Bad sign if repeated looping with small daughter PDF, so fail.
    // (Example: if all PDF's = 0 below Q_0, except for c/b companion.)
    if (hasTinyPDFdau) ++loopTinyPDFdau;
    if (loopTinyPDFdau > MAXLOOPTINYPDF) {
      infoPtr->errorMsg("Warning in DireSpace::pT2nextQCD: "
      "small daughter PDF");
      return;
    }

    // Initialize integrals of splitting kernels and evaluate parton
    // densities at the beginning. Reinitialize after long evolution
    // in pT2 or when crossing c and b flavour thresholds.

    if ( needNewPDF
      || pT2 < evalpdfstep(event[iRadi].id(),pT2,m2cPhys,m2bPhys) * pT2PDF) {
      pT2PDF        = pT2;
      hasTinyPDFdau = false;

      newOverestimates.clear();
      kernelPDF = 0.;

      // Determine overestimated z range; switch at c and b masses.
      if (pT2 > m2b) {
        nFlavour  = 5;
        pT2minNow = max( m2b, pT2endDip);
        b0        = 23./6.;
        Lambda2   = Lambda5flav2;
      } else if (pT2 > m2c) {
        nFlavour  = 4;
        pT2minNow = max( m2c, pT2endDip);
        b0        = 25./6.;
        Lambda2   = Lambda4flav2;
      } else {
        nFlavour  = 3;
        pT2minNow = pT2endDip;
        b0        = 27./6.;
        Lambda2   = Lambda3flav2;
      }

      // A change of renormalization scale expressed by a change of Lambda.
      Lambda2    /= renormMultFac;

      zMinAbs     = xDaughter;
      zMaxAbs     = 1.;

      // Calculate and add user-defined overestimates.
      getNewOverestimates( type, idDaughter, dipEndNow, pT2PDF, xDaughter, 
        zMinAbs, zMaxAbs, newOverestimates );
      addNewOverestimates(newOverestimates, kernelPDF);

      // End evaluation of splitting kernels and parton densities.
      needNewPDF = false;
    }

    if (kernelPDF < TINYKERNELPDF) return;

    if (newOverestimates.empty()) return;

    // Pick pT2 (in overestimated z range), for one of three different cases.
    // Assume form alphas(pT0^2 + pT^2) * dpT^2/(pT0^2 + pT^2).
    double Q2alphaS;

    // Fixed alpha_strong, reweighted later to PDF running alpha_s.
    if (usePDFalphas) {
      pT2 = (pT2 + pT20) * pow( rndmPtr->flat(),
        1. / (alphaS2piOverestimate * kernelPDF)) - pT20;

    // Fixed alpha_strong.
    } else if (alphaSorder == 0) {
      pT2 = (pT2 + pT20) * pow( rndmPtr->flat(),
        1. / (alphaS2pi * kernelPDF)) - pT20;

    // First-order alpha_strong.
    } else if (alphaSorder == 1) {
      pT2 = Lambda2 * pow( (pT2 + pT20) / Lambda2,
        pow(rndmPtr->flat(), b0 / kernelPDF) ) - pT20;

    // For second order reject by second term in alpha_strong expression.
    } else {
      do {
        pT2 = Lambda2 * pow( (pT2 + pT20) / Lambda2,
          pow(rndmPtr->flat(), b0 / kernelPDF) ) - pT20;
        Q2alphaS = renormMultFac * max( pT2 + pT20,
          pow2(LAMBDA3MARGIN) * Lambda3flav2);
      } while (alphaS.alphaS2OrdCorr(Q2alphaS) < rndmPtr->flat()
        && pT2 > pT2minNow);
    }

    // Abort evolution if not allowed.
    if (pT2 < 0.) return;

    // Check for pT2 values that prompt special action.

    if (pT2 < pT2endDip) return;

    // If crossed b threshold, continue evolution from this threshold.
    if (nFlavour == 5 && pT2 < m2bPhys) {
      needNewPDF = true;
    // If crossed c threshold, continue evolution from this threshold.
    } else if (nFlavour == 4 && pT2 < m2cPhys) {
      needNewPDF = true;
    }

    // Leave heavy quarks below threshold unconverted for now.
    if (abs(idDaughter) == 5 && pT2 <= m2bPhys) {
      return;
    } else if (abs(idDaughter) == 4 && pT2 <= m2cPhys) {
      return;
    }

    // Select z value of branching, and corrective weight.
    double R = kernelPDF*rndmPtr->flat();
    if (!newOverestimates.empty()) {
      if (newOverestimates.lower_bound(R) == newOverestimates.end())
        splittingNowName = newOverestimates.rbegin()->second;
      else
        splittingNowName = newOverestimates.lower_bound(R)->second;
      getNewSplitting( event, dipEndNow, pT2PDF, xDaughter, pT2, zMinAbs,
        zMaxAbs, idDaughter, splittingNowName, idMother, idSister, z, wt,
        fullWeightNow, overWeightNow);
    }

    // Impossible emission (e.g. if outside allowed z range for given pT2).
    if ( wt == 0.) {
      wt = fullWeightNow = overWeightNow = auxWeightNow = 0.;
      continue;
    }

    // Disallow gluon -> heavy quarks if pT has fallen below 2*mQuark.
    if ( abs(idDaughter) == 5 && abs(idSister) == 5
      && pT2 <= 4.*m2bPhys) {
      wt = fullWeightNow = overWeightNow = auxWeightNow = 0.;
      continue;
    } else if (abs(idDaughter) == 4 && abs(idSister) == 4
      && pT2 <= 4.*m2cPhys) {
      wt = fullWeightNow = overWeightNow = auxWeightNow = 0.;
      continue;
    }

    // Calculate CS variables.
    double xCS = z;
    xMother = xDaughter/xCS;

    // Evaluation of new daughter and mother PDF's.
    double pdfRatio = 1.;
    pdfScale2 = (useFixedFacScale) ? fixedFacScale2 : factorMultFac * pT2;
    bool inD = beam.insideBounds(xDaughter, pdfScale2); 
    bool inM = beam.insideBounds(xMother,   pdfScale2); 
    double xPDFdaughterNew =
      (useSummedPDF) ? beam.xf(idDaughter, xDaughter, pdfScale2)
                     : beam.xfISR(iSysNow, idDaughter, xDaughter, pdfScale2);
    double xPDFmotherNew =
      (useSummedPDF) ? beam.xf(idMother, xMother, pdfScale2)
                     : beam.xfISR(iSysNow, idMother, xMother, pdfScale2);

    if ( abs(xPDFdaughterNew) < TINYPDF ) { 
      hasTinyPDFdau = true;
      needNewPDF = true;
      wt = fullWeightNow = overWeightNow = auxWeightNow = 0.;
      continue;
    }

    // Set PDF ratio to zero if x-/pT-values are out-of-bounds.
    pdfRatio = (inD && inM) ? xPDFmotherNew/xPDFdaughterNew : 0.;

    wt             *= pdfRatio;
    fullWeightNow  *= pdfRatio;

    // Reweight to PDF alpha_s, if necessary.
    wt            *= (usePDFalphas)
                   ? beam.alphaS(pT2)/(alphaS2piOverestimate*2.*M_PI)
                   : 1.; 
    fullWeightNow *= (usePDFalphas)
                   ? beam.alphaS(pT2)/(2.*M_PI)
                   : (alphaSorder==0) ? alphaS2pi
                                      : alphaS.alphaS(pT2)/(2.*M_PI); 
    overWeightNow *= (usePDFalphas)
                   ? alphaS2piOverestimate
                   : (alphaSorder==0) ? alphaS2pi
                                      : alphaS.alphaS(pT2)/(2.*M_PI);
    auxWeightNow   = overWeightNow;

    // Ensure that accept probability is positive.
    if (fullWeightNow < 0.) auxWeightNow *= -1.; 

    if (wt == 0. ) {
      needNewPDF = true;
      wt = fullWeightNow = overWeightNow = auxWeightNow = 0.;
      continue;
    }

    // Reset overestimate if necessary.
    if ( fullWeightNow/auxWeightNow > 1.) {
      double rescale = fullWeightNow/auxWeightNow * 1.5;
      auxWeightNow *= rescale;
      wt /= rescale;
      needNewPDF = true;
      infoPtr->errorMsg("Info in DireSpace::pT2nextQCD_IF: Found large "
                        "acceptance weight for " + splittingNowName);
    }

  // Iterate until acceptable pT (or have fallen below pTmin).
  } while (wt < rndmPtr->flat()) ;

  // Update accepted event weight.  
  if ( fullWeightNow != 0. && overWeightNow != 0. ){
    acceptProbability.insert( make_pair(pT2,auxWeightNow/overWeightNow));
    double w = auxWeightNow/overWeightNow
             * (overWeightNow- fullWeightNow)
             / (auxWeightNow - fullWeightNow);
    rejectProbability.insert( make_pair(pT2, w));
  }

  // Save values for (so far) acceptable branching.
  dipEndNow->store( idDaughter,idMother, idSister, x1Now, x2Now, m2Dip,
    pT2, z, xMother, Q2, mSister, m2Sister, pT2corr);

}

//--------------------------------------------------------------------------

// Evolve a QCD dipole end near threshold, with g -> Q + Qbar enforced.
// Note: No explicit Sudakov factor formalism here. Instead use that
// df_Q(x, pT2) = (alpha_s/2pi) * (dT2/pT2) * ((gluon) * (splitting)).
// This implies that effects of Q -> Q + g are neglected in this range.

void DireSpace::pT2nearQCDthreshold( BeamParticle& beam,
  double m2Massive, double m2Threshold, double xMaxAbs,
  double zMinAbs, double zMaxMassive) {

  // Initial values, to be used in kinematics and weighting.
  double Lambda2       = (abs(idDaughter) == 4) ? Lambda4flav2 : Lambda5flav2;
  Lambda2             /= renormMultFac;
  double logM2Lambda2  = (alphaSorder > 0) ? log( m2Massive / Lambda2 ) : 1.;
  pdfScale2 = (useFixedFacScale) ? fixedFacScale2
    : factorMultFac * m2Threshold;
  double xPDFmotherOld = beam.xfISR(iSysNow, 21, xDaughter, pdfScale2);

  // Variables used inside evolution loop. (Mainly dummy start values.)
  int    loop    = 0;
  double wt      = 0.;
  double pT2     = 0.;
  double z       = 0.;
  double Q2      = 0.;
  double pT2corr = 0.;
  double xMother = 0.;

  // Begin loop over tries to find acceptable g -> Q + Qbar branching.
  do {
    wt = 0.;

    // Check that not caught in infinite loop with impossible kinematics.
    if (++loop > 100) {
      infoPtr->errorMsg("Error in DireSpace::pT2nearQCDthreshold: "
        "stuck in loop");
      return;
    }

    // Pick dpT2/pT2 in range [m2Massive,thresholdRatio * m2Massive].
    pT2 = m2Massive * pow( m2Threshold / m2Massive, rndmPtr->flat() );

    // Pick z flat in allowed range.
    z = zMinAbs + rndmPtr->flat() * (zMaxMassive - zMinAbs);

    // Check that kinematically possible choice.
    Q2 = pT2 / (1.-z) - m2Massive;
    pT2corr = Q2 - z * (m2Dip + Q2) * (Q2 + m2Massive) / m2Dip;
    if (pT2corr < TINYPT2) continue;

    // Correction factor for running alpha_s. (Only first order for now.)
    wt = (alphaSorder > 0) ? logM2Lambda2 / log( pT2 / Lambda2 ) : 1.;

    // Correction factor for splitting kernel.
    wt *= pow2(z) + pow2(1.-z) + 2. * z * (1.-z) * m2Massive / pT2;

    // x, including correction for massive recoiler from rescattering.
    xMother = xDaughter / z;
    if (!dipEndNow->normalRecoil) {
      if (sideA) xMother += (m2Rec / (x2Now * sCM)) * (1. / z - 1.);
      else       xMother += (m2Rec / (x1Now * sCM)) * (1. / z - 1.);
    }
    if (xMother > xMaxAbs) { wt = 0.; continue; }

    // Correction factor for gluon density.
    pdfScale2 = (useFixedFacScale) ? fixedFacScale2 : factorMultFac * pT2;
    double xPDFmotherNew = beam.xfISR(iSysNow, 21, xMother, pdfScale2);
    wt *= xPDFmotherNew / xPDFmotherOld;

  // Iterate until acceptable pT and z.
  } while (wt < rndmPtr->flat()) ;

  // Save values for (so far) acceptable branching.
  double mSister = (abs(idDaughter) == 4) ? mc : mb;
  dipEndNow->store( idDaughter, 21, -idDaughter, x1Now, x2Now, m2Dip,
    pT2, z, xMother, Q2, mSister, pow2(mSister), pT2corr);

}

//--------------------------------------------------------------------------

// Kinematics of branching.
// Construct mother -> daughter + sister, with recoiler on other side.

bool DireSpace::branch( Event& event) {

  if ( event[dipEndSel->iRecoiler].isFinal() )
    return branch_IF(event);
  else
    return branch_II(event);

  return false;

}

//--------------------------------------------------------------------------

// Kinematics of branching.
// Construct mother -> daughter + sister, with recoiler on other side.

bool DireSpace::branch_II( Event& event) {

  // Side on which branching occured.
  int side          = abs(dipEndSel->side);

  // Read in flavour and colour variables.
  int iDaughter     = partonSystemsPtr->getInA(iSysSel);
  int iRecoiler     = partonSystemsPtr->getInB(iSysSel);
  if (side == 2) swap(iDaughter, iRecoiler);
  int idDaughterNow = dipEndSel->idDaughter;
  int idMother      = dipEndSel->idMother;
  int idSister      = dipEndSel->idSister;
  int colDaughter   = event[iDaughter].col();
  int acolDaughter  = event[iDaughter].acol();
  int colRecBef     = event[iRecoiler].col();
  int acolRecBef    = event[iRecoiler].acol();
  bool colMatch     = (acolRecBef == colDaughter);
  bool acolMatch    = (colRecBef  == acolDaughter);

  // Read in kinematical variables.
  double pT2        = dipEndSel->pT2;
  double z          = dipEndSel->z;

  // Get particle masses.
  double m2Bef = 0.0, m2r = 0.0, m2s = 0.0;
  // Emission.
  double m2e = (abs(idSister) < 6 ) ? getMass(idSister,2) : getMass(idSister,1);
  // Force emission massless by default.
  if (!forceMassiveMap) m2e = 0.0;

  double m2dipCorr  = dipEndSel->m2Dip - m2Bef + m2r + m2e;

  //double kappa2 = pT2 / m2;
  double kappa2 = pT2 / m2dipCorr;
  double xCS    = (z*(1-z)- kappa2)/(1-z);
  double vCS    = kappa2/(1-z); 

  // Current event and subsystem size.
  int eventSizeOld  = event.size();
  int systemSizeOld = partonSystemsPtr->sizeAll(iSysSel);

  // Save properties to be restored in case of user-hook veto of emission.
  int beamOff1 = 1 + beamOffset;
  int beamOff2 = 2 + beamOffset;
  int ev1Dau1V = event[beamOff1].daughter1();
  int ev2Dau1V = event[beamOff2].daughter1();
  vector<int> statusV, mother1V, mother2V, daughter1V, daughter2V;

  // Check if the first emission shoild be checked for removal
  bool canMergeFirst = (mergingHooksPtr != 0)
                     ? mergingHooksPtr->canVetoEmission() : false;
  if (canVetoEmission || canMergeFirst) {
    for ( int iCopy = 0; iCopy < systemSizeOld; ++iCopy) {
      int iOldCopy    = partonSystemsPtr->getAll(iSysSel, iCopy);
      statusV.push_back( event[iOldCopy].status());
      mother1V.push_back( event[iOldCopy].mother1());
      mother2V.push_back( event[iOldCopy].mother2());
      daughter1V.push_back( event[iOldCopy].daughter1());
      daughter2V.push_back( event[iOldCopy].daughter2());
    }
  }

  // Take copy of existing system, to be given modified kinematics.
  // Incoming negative status. Rescattered also negative, but after copy.
  for ( int iCopy = 0; iCopy < systemSizeOld; ++iCopy) {
    int iOldCopy    = partonSystemsPtr->getAll(iSysSel, iCopy);
    int statusOld   = event[iOldCopy].status();
    int statusNew   = (iOldCopy == iDaughter
      || iOldCopy == iRecoiler) ? statusOld : 44;
    int iNewCopy    = event.copy(iOldCopy, statusNew);
    if (statusOld < 0) event[iNewCopy].statusNeg();
  }

  // Define colour flow in branching.
  // Default corresponds to f -> f + gamma.
  int colMother     = colDaughter;
  int acolMother    = acolDaughter;
  int colSister     = 0;
  int acolSister    = 0;
  // q -> q + g and 50% of g -> g + g; need new colour.
  double RN2 = rndmPtr->flat();
  if (idSister == 21 && ( (idMother > 0 && idMother < 9)
    || ( idMother == 21 && colMatch && acolMatch && RN2 < 0.5) ) ) {
    colMother       = event.nextColTag();
    colSister       = colMother;
    acolSister      = colDaughter;
  // qbar -> qbar + g and other 50% of g -> g + g; need new colour.
  } else if (idSister == 21 && ( (idMother < 0 && idMother > -9)
    || ( idMother == 21 && colMatch && acolMatch) ) ) {
    acolMother      = event.nextColTag();
    acolSister      = acolMother;
    colSister       = acolDaughter;
  } else if (idMother == 21 && idSister == 21 && colMatch && !acolMatch) {
    colMother       = event.nextColTag();
    acolMother      = acolDaughter;
    colSister       = colMother;
    acolSister      = colDaughter;
  } else if (idMother == 21 && idSister == 21 && !colMatch && acolMatch) {
    colMother       = colDaughter;
    acolMother      = event.nextColTag();
    acolSister      = acolMother;
    colSister       = acolDaughter;
  // q -> g + q.
  } else if (idDaughterNow == 21 && idMother > 0) {
    colMother       = colDaughter;
    acolMother      = 0;
    colSister       = acolDaughter;
  // qbar -> g + qbar
  } else if (idDaughterNow == 21) {
    acolMother      = acolDaughter;
    colMother       = 0;
    acolSister      = colDaughter;
  // g -> q + qbar.
  } else if (idDaughterNow > 0 && idDaughterNow < 9) {
    acolMother      = event.nextColTag();
    acolSister      = acolMother;
  // g -> qbar + q.
  } else if (idDaughterNow < 0 && idDaughterNow > -9) {
    colMother       = event.nextColTag();
    colSister       = colMother;
  }

  // Swap colours if radiated gluon carries momentum fraction z
  int colSave = colSister, acolSave = acolSister;
  if ( splittingSelName.compare("isr_21->21&21b_CS") == 0) {
    colSister  = acolMother;
    acolSister = colMother;
    colMother  = acolSave;
    acolMother = colSave; 
  }

  // Indices of partons involved. Add new sister.
  int iMother       = eventSizeOld + side - 1;
  int iNewRecoiler  = eventSizeOld + 2 - side;
  int iSister       = event.append( idSister, 43, iMother, 0, 0, 0,
     colSister, acolSister, Vec4(0.,0.,0.,0.), 0.0, sqrt(pT2) );

  // References to the partons involved.
  Particle& daughter    = event[iDaughter];
  Particle& mother      = event[iMother];
  Particle& newRecoiler = event[iNewRecoiler];
  Particle& sister      = event.back();

  // Replace old by new mother; update new recoiler.
  mother.id( idMother );
  mother.status( -41);
  mother.cols( colMother, acolMother);
  mother.p(0.,0., event[iDaughter].pz()/xCS, event[iDaughter].e()/xCS);
  if (mother.idAbs() == 21 || mother.idAbs() == 22) mother.pol(9);
  newRecoiler.status(-42);
  newRecoiler.p(0.,0., event[iRecoiler].pz(), event[iRecoiler].e());

  // Update mother and daughter pointers; also for beams.
  daughter.mothers( iMother, 0);
  mother.daughters( iSister, iDaughter);
  if (iSysSel == 0) {
    event[beamOff1].daughter1( (side == 1) ? iMother : iNewRecoiler );
    event[beamOff2].daughter1( (side == 2) ? iMother : iNewRecoiler );
  }

  // Massive kinematics

  // Get dipole 4-momentum.
  Vec4 paj_tilde(event[iDaughter].p());
  Vec4 pb_tilde(event[iRecoiler].p());
  Vec4 q(pb_tilde+paj_tilde);

  // Calculate derived variables.
  double q2   = q.m2Calc();
  double sab  = (q2 - m2e)/xCS + (m2r+m2s) * (1-1/xCS);
  double saj  = -vCS*(sab - m2r-m2s) + m2r + m2e;
  double zbar = (sab - m2r - m2s) / bABC(sab,m2r,m2s)
              *( (xCS + vCS)  - m2s / gABC(sab,m2r,m2s)
                     * (saj + m2r - m2e) / (sab - m2r - m2s));
  double kT2  = zbar*(1.-zbar)*m2r - (1-zbar)*saj - zbar*m2e;

  // Not possible to construct kinematics if kT2 < 0.0
  if (kT2 < 0.) {
    infoPtr->errorMsg("Warning in DireSpace::branch_II: Reject state "
                      "with kinematically forbidden kT^2.");
    return false;
  }   

  // Now construct radiator in lab frame.
  Vec4 pRad = (paj_tilde - m2Bef/gABC(q2,m2Bef,m2s)*pb_tilde)
             *sqrt(lABC(sab,m2r,m2s)/lABC(q2,m2Bef,m2s))
           + m2r / gABC(sab,m2r,m2s)*pb_tilde;
    
  // Now construct the transverse momentum vector in the dipole CM
  // frame.
  double RNphi = rndmPtr->flat();
  double phi_kt = 2.*M_PI*RNphi;
  //double phi_kt = 2.*M_PI*RN1;
  Vec4 kTmom(sqrt(kT2)*cos(phi_kt), sqrt(kT2)*sin(phi_kt),0.,0.);

  // Boost the transverse momentum vector into the lab frame.
  RotBstMatrix fromDipCM;
  if (side == 1) fromDipCM.fromCMframe( paj_tilde, pb_tilde);
  else           fromDipCM.fromCMframe( pb_tilde, paj_tilde);
  kTmom.rotbst(fromDipCM);

  // Construct new emission momentum.
  Vec4 pEmt = (1-zbar) * (gABC(sab,m2r,m2s)*pRad - m2r*pb_tilde) / bABC(sab,m2r,m2s)
            + (m2e + kT2) / ((1-zbar)*bABC(sab,m2r,m2s))
            * (pb_tilde - m2s/gABC(sab,m2r,m2s)*pRad)
            + kTmom;

  // Ensure that radiator is on mass-shell
  double mTolErr = settingsPtr->parm("Check:mTolErr");
  double errMass = abs(pRad.mCalc() - sqrt(m2r)) / max( 1.0, pRad.e());
  if ( errMass > mTolErr*1e-2 ) {
    double deltam2 = pRad.m2Calc() - m2r;
    pRad.e(sqrtpos(pow2(pRad.e()) - deltam2));
  }
  // Ensure that emission is on mass-shell
  errMass = abs(pEmt.mCalc() - sqrt(m2e)) / max( 1.0, pEmt.e());
  if ( errMass > mTolErr*1e-2 ) {
    double deltam2 = pEmt.m2Calc() - m2e;
    pEmt.e(sqrtpos(pow2(pEmt.e()) - deltam2));
  }

  // Set all momenta.
  sister.p(pEmt);
  mother.p(pRad);
  newRecoiler.p(pb_tilde);
  sister.m(sqrtpos(m2e));
  mother.m(sqrtpos(m2r));
  newRecoiler.m(sqrtpos(m2s));

  Vec4 kTilde(paj_tilde + pb_tilde);
  Vec4 k(mother.p() + newRecoiler.p()  - sister.p());

  Vec4 kSum(kTilde + k);
  for ( int i = eventSizeOld + 2; i < eventSizeOld + systemSizeOld; ++i) {
    Vec4 pIn = event[i].p();
    double kSum2    = kSum.m2Calc();
    double k2       = k.m2Calc();
    double kTildeXp = kTilde*pIn;
    double kSumXp   = kSum*pIn;
    Vec4 res = pIn - kSum * 2.0*( kSumXp / kSum2 ) + k * 2.0 *( kTildeXp/k2);  
    event[i].p(res);
  }

  // Allow veto of branching. If so restore event record to before emission.
  if ( (canVetoEmission
    && userHooksPtr->doVetoISREmission(eventSizeOld, event, iSysSel))
    || (canMergeFirst
    && mergingHooksPtr->doVetoEmission( event )) ) {
    event.popBack( event.size() - eventSizeOld);
    event[beamOff1].daughter1( ev1Dau1V);
    event[beamOff2].daughter1( ev2Dau1V);
    for ( int iCopy = 0; iCopy < systemSizeOld; ++iCopy) {
      int iOldCopy = partonSystemsPtr->getAll(iSysSel, iCopy);
      event[iOldCopy].status( statusV[iCopy]);
      event[iOldCopy].mothers( mother1V[iCopy], mother2V[iCopy]);
      event[iOldCopy].daughters( daughter1V[iCopy], daughter2V[iCopy]);
    }
    return false;
  }

  // Update the number of proposed emissions.
  if (dipEndSel != 0 && nProposedPT.find(dipEndSel->system) != nProposedPT.end())
    ++nProposedPT[dipEndSel->system];

  // Update list of partons in system; adding newly produced one.
  partonSystemsPtr->setInA(iSysSel, eventSizeOld);
  partonSystemsPtr->setInB(iSysSel, eventSizeOld + 1);
  for (int iCopy = 2; iCopy < systemSizeOld; ++iCopy)
    partonSystemsPtr->setOut(iSysSel, iCopy - 2, eventSizeOld + iCopy);
  partonSystemsPtr->addOut(iSysSel, eventSizeOld + systemSizeOld);
  // Get new center-of-mass energy
  int iA      = partonSystemsPtr->getInA(iSysSel);
  int iB      = partonSystemsPtr->getInB(iSysSel);
  double shat = (event[iA].p() + event[iB].p()).m2Calc();
  partonSystemsPtr->setSHat(iSysSel, shat);

  // dipEnd array may have expanded and been moved, so regenerate dipEndSel.
  dipEndSel = &dipEnd[iDipSel];

  // Update info on radiating dipole ends (QCD).
  for (int iDip = 0; iDip < int(dipEnd.size()); ++iDip)
  if ( dipEnd[iDip].system == iSysSel) {
    if (abs(dipEnd[iDip].side) == side) {
      dipEnd[iDip].iRadiator = iMother;
      dipEnd[iDip].iRecoiler = iNewRecoiler;
      if (dipEnd[iDip].colType  != 0)
        dipEnd[iDip].colType = mother.colType();

    // Update info on recoiling dipole ends (QCD or QED).
    } else {
      dipEnd[iDip].iRadiator = iNewRecoiler;
      dipEnd[iDip].iRecoiler = iMother;
      dipEnd[iDip].MEtype = 0;
    }
  }

  // Update info on beam remnants.
  BeamParticle& beamNow = (side == 1) ? *beamAPtr : *beamBPtr;
  double xNew = 2.*event[iMother].e()/event[0].m();
  beamNow[iSysSel].update( iMother, idMother, xNew);
  // Redo choice of companion kind whenever new flavour.
  if (idMother != idDaughterNow) {
    pdfScale2 = (useFixedFacScale) ? fixedFacScale2 : factorMultFac * pT2;
    beamNow.xfISR( iSysSel, idMother, xNew, pdfScale2);
    beamNow.pickValSeaComp();
  }
  BeamParticle& beamRec = (side == 1) ? *beamBPtr : *beamAPtr;
  beamRec[iSysSel].iPos( iNewRecoiler);

  // Store branching values of current dipole. (For rapidity ordering.)
  ++dipEndSel->nBranch;
  dipEndSel->pT2Old = pT2;
  dipEndSel->zOld   = z;

  // Update ISR dipoles.
  update(iSysSel,event);

  // Ungraceful exit for incorrect event.
  bool isHadronic = false;
  for (int i = 0; i < event.size(); ++i)
    if (event[i].statusAbs() > 60) isHadronic = true;
  if (!isHadronic && !validEvent(event)) {
    infoPtr->errorMsg("Error in DireSpace::branch_II: State after "
                      "branching not valid, thus reject.");
    return false;
  }

  // Set shower weight.
  weights->calcWeight(pT2);
  weights->reset();
  rejectProbability.clear();
  acceptProbability.clear();

  // Done without any errors.
  return true;

}

//--------------------------------------------------------------------------

// Kinematics of branching.
// Construct mother -> daughter + sister, with recoiler on other side.

bool DireSpace::branch_IF( Event& event) {

  // Side on which branching occured.
  int side          = abs(dipEndSel->side);

  // Read in flavour and colour variables.
  //int iDaughter     = partonSystemsPtr->getInA(iSysSel);
  //if (side == 2) iDaughter = partonSystemsPtr->getInB(iSysSel);
  int iDaughter     = dipEndSel->iRadiator;
  int iRecoiler     = dipEndSel->iRecoiler;
  int idDaughterNow = dipEndSel->idDaughter;
  int idMother      = dipEndSel->idMother;
  int idSister      = dipEndSel->idSister;
  int colDaughter   = event[iDaughter].col();
  int acolDaughter  = event[iDaughter].acol();
  int colRecBef     = event[iRecoiler].col();
  int acolRecBef    = event[iRecoiler].acol();
  bool colMatch     = (colRecBef  == colDaughter);
  bool acolMatch    = (acolRecBef == acolDaughter);

  // Read in kinematical variables.
  double pT2        = dipEndSel->pT2;
  double z          = dipEndSel->z;

  //BeamParticle& beamNow = (side == 1 && abs(beamAPtr->id()) == 2212)
  //                   ? *beamAPtr
  //                   : ((abs(beamBPtr->id()) == 2212) ? *beamBPtr
  //                                                    : *beamAPtr );
  BeamParticle& beamNow = (side == 1) ? *beamAPtr : *beamBPtr;

  // Get particle masses.
  double m2Bef = 0.0;
  double m2r   = 0.0;
  double m2s   = 0.0;
  int type     = (event[dipEndSel->iRecoiler].isFinal()) ? 1 : -1;
  if ( event[dipEndSel->iRecoiler].idAbs() < 6 && type == 1 )
    //m2s = pow2( max(0., beamNow.mQuarkPDF(event[dipEndSel->iRecoiler].id()) ) );
    m2s = getMass(event[dipEndSel->iRecoiler].id(),2);
  else if ( type == 1 )
    //m2s = event[dipEndSel->iRecoiler].m2Calc();
    m2s = getMass(event[dipEndSel->iRecoiler].id(),3,
          event[dipEndSel->iRecoiler].mCalc());

  // Emission.
  double m2e = (abs(idSister) < 6 ) ? getMass(idSister,2) : getMass(idSister,1);
  // Force emission massless by default.
  if (!forceMassiveMap) m2e = 0.0;

  // Adjust the dipole kinematical mass to accomodate masses after branching.
  double m2dipCorr  = dipEndSel->m2Dip - m2Bef + m2r + m2e;

  // Calculate CS variables.
  double xCS = z;
  double uCS = (pT2/m2dipCorr)/(1-z);

  // Current event and subsystem size.
  int eventSizeOld  = event.size();

  // Save properties to be restored in case of user-hook veto of emission.
  bool canMergeFirst = (mergingHooksPtr != 0)
                     ? mergingHooksPtr->canVetoEmission() : false;
  int iDauStatusV  = event[iDaughter].status();
  int iDauMot1V    = event[iDaughter].mother1();
  int iDauMot2V    = event[iDaughter].mother2();
  int iDauDau1V    = event[iDaughter].daughter1();
  int iDauDau2V    = event[iDaughter].daughter2();
  int iRecStatusV  = event[iRecoiler].status();
  int iRecMot1V    = event[iRecoiler].mother1();
  int iRecMot2V    = event[iRecoiler].mother2();
  int iRecDau1V    = event[iRecoiler].daughter1();
  int iRecDau2V    = event[iRecoiler].daughter2();
  int beamOff1 = 1 + beamOffset;
  int beamOff2 = 2 + beamOffset;

  // Define colour flow in branching.
  // Default corresponds to f -> f + gamma.
  int colMother     = colDaughter;
  int acolMother    = acolDaughter;
  int colSister     = 0;
  int acolSister    = 0;
  // q -> q + g and 50% of g -> g + g; need new colour.
  if (idSister == 21 && ( (idMother > 0 && idMother < 9)
    || ( idMother == 21 && colMatch && acolMatch
      && rndmPtr->flat() < 0.5) ) ) {
    colMother       = event.nextColTag();
    colSister       = colMother;
    acolSister      = colDaughter;
  // qbar -> qbar + g and other 50% of g -> g + g; need new colour.
  } else if (idSister == 21 && ( (idMother < 0 && idMother > -9)
    || ( idMother == 21 && colMatch && acolMatch) ) ) {
    acolMother      = event.nextColTag();
    acolSister      = acolMother;
    colSister       = acolDaughter;
  } else if (idMother == 21 && idSister == 21 && colMatch && !acolMatch) {
    colMother       = event.nextColTag();
    acolMother      = acolDaughter;
    colSister       = colMother;
    acolSister      = colDaughter;
  } else if (idMother == 21 && idSister == 21 && !colMatch && acolMatch) {
    colMother       = colDaughter;
    acolMother      = event.nextColTag();
    acolSister      = acolMother;
    colSister       = acolDaughter;
  // q -> g + q.
  } else if (idDaughterNow == 21 && idMother > 0) {
    colMother       = colDaughter;
    acolMother      = 0;
    colSister       = acolDaughter;
  // qbar -> g + qbar
  } else if (idDaughterNow == 21) {
    acolMother      = acolDaughter;
    colMother       = 0;
    acolSister      = colDaughter;
  // g -> q + qbar.
  } else if (idDaughterNow > 0 && idDaughterNow < 9) {
    acolMother      = event.nextColTag();
    acolSister      = acolMother;
  // g -> qbar + q.
  } else if (idDaughterNow < 0 && idDaughterNow > -9) {
    colMother       = event.nextColTag();
    colSister       = colMother;
  }

  // Swap colours if radiated gluon carries momentum fraction z
  int colSave = colSister, acolSave = acolSister;
  if ( splittingSelName.compare("isr_21->21&21b_CS") == 0) {
    colSister  = acolMother;
    acolSister = colMother;
    colMother  = acolSave;
    acolMother = colSave; 
  }

  // Indices of partons involved. Add new sister.
  int iMother       = event.append( idMother, -41, event[iDaughter].mother1(), 0, 0, 0,
     colMother, acolMother, Vec4(0.,0.,0.,0.), 0.0, sqrt(pT2) );
  int iNewRecoiler  = event.append( event[iRecoiler].id(), 48, iRecoiler, iRecoiler, 0, 0,
     event[iRecoiler].col(), event[iRecoiler].acol(),
     Vec4(0.,0.,0.,0.), 0.0, sqrt(pT2) );
  int iSister       = event.append( idSister, 43, iMother, 0, 0, 0,
     colSister, acolSister, Vec4(0.,0.,0.,0.), 0.0, sqrt(pT2) );

  // References to the partons involved.
  Particle& daughter    = event[iDaughter];
  Particle& mother      = event[iMother];
  Particle& newRecoiler = event[iNewRecoiler];
  Particle& sister      = event.back();

  // Replace old by new mother; update old recoiler.
  event[iRecoiler].statusNeg();
  event[iRecoiler].daughters( iNewRecoiler, iNewRecoiler);
  if (mother.idAbs() == 21 || mother.idAbs() == 22) mother.pol(9);

  // Update mother and daughter pointers; also for beams.
  daughter.mothers( iMother, 0);
  mother.daughters( iSister, iDaughter);
  if (iSysSel == 0) {
    if (side == 1) event[beamOff1].daughter1(iMother);
    else           event[beamOff2].daughter1(iMother);
  }

  // Massive kinematics, in two schemes.
  if (!useGlobalMapIF) {

    // Get dipole 4-momentum.
    Vec4 paj_tilde(event[iDaughter].p());
    Vec4 pk_tilde(event[iRecoiler].p());
    Vec4 pTk_tilde(event[iRecoiler].px(),event[iRecoiler].py(),0.,0.);
    Vec4 q(paj_tilde-pk_tilde);
    Vec4 qpar(q+pTk_tilde);

    // Calculate derived variables.
    double q2    = q.m2Calc();
    double q2par = qpar.m2Calc();
    double pT2k  = pTk_tilde.m2Calc();
    double sjk   = (1. - 1./xCS)*(q2 - m2r) + (m2e + m2s) / xCS;
    double zbar  = (q2-sjk-m2r) / bABC(q2,sjk,m2r)
                  *( uCS - m2r / gABC(q2,sjk,m2r)
                        * (sjk + m2e - m2s) / (q2 - sjk - m2r));
    double kT2   = zbar*(1.-zbar)*sjk - (1-zbar)*m2e - zbar*m2s;

    if (kT2 < 0.) {
      infoPtr->errorMsg("Warning in DireSpace::branch_IF: Reject state "
                        "with kinematically forbidden kT^2.");
      return false;
    }   

    // Now construct radiator in lab frame.
    Vec4 pRad( ( paj_tilde - q*paj_tilde/q2par * qpar )
               * sqrt( (lABC(q2,sjk,m2r) - 4.*m2r*pT2k)
                     / (lABC(q2,m2s,m2Bef) - 4.*m2Bef*pT2k))
              + qpar * 0.5 * (q2 + m2r - sjk) / q2par);

    // Now construct the transverse momentum vector in the dipole CM
    // frame.
    double RNphi = rndmPtr->flat();
    double phi_kt = 2.*M_PI*RNphi;
    Vec4 kTmom(sqrt(kT2)*cos(phi_kt), sqrt(kT2)*sin(phi_kt),0.,0.);
 
    // Boost the transverse momentum vector into the lab frame.
    RotBstMatrix fromDipCM;
    if (side == 1) fromDipCM.fromCMframe( paj_tilde, pk_tilde);
    else           fromDipCM.fromCMframe( pk_tilde, paj_tilde);
    kTmom.rotbst(fromDipCM);

    // Construct left-over dipole momentum by momentum conservation.
    Vec4 pjk(-q+pRad);

    // Construct new emission momentum.
    Vec4 pEmt( zbar * (gABC(q2,sjk,m2r)*pjk + sjk*pRad) / bABC(q2,sjk,m2r)
              - (m2e + kT2) / (zbar*bABC(q2,sjk,m2r))
              * (pRad + m2r/gABC(q2,sjk,m2r)*pjk)
              + kTmom);

    // Contruct recoiler momentum by momentum conservation.
    Vec4 pRec(-q+pRad-pEmt);

    // Ensure that radiator is on mass-shell
    double mTolErr = settingsPtr->parm("Check:mTolErr");
    double errMass = abs(pRad.mCalc() - sqrt(m2r)) / max( 1.0, pRad.e());
    if ( errMass > mTolErr*1e-2 ) {
      double deltam2 = pRad.m2Calc() - m2r;
      pRad.e(sqrtpos(pow2(pRad.e()) - deltam2));
    }
    // Ensure that emission is on mass-shell
    errMass = abs(pEmt.mCalc() - sqrt(m2e)) / max( 1.0, pEmt.e());
    if ( errMass > mTolErr*1e-2 ) {
      double deltam2 = pEmt.m2Calc() - m2e;
      pEmt.e(sqrtpos(pow2(pEmt.e()) - deltam2));
    }
    // Ensure that recoiler is on mass-shell
    errMass = abs(pRec.mCalc() - sqrt(m2s)) / max( 1.0, pRec.e());
    if ( errMass > mTolErr*1e-2 ) {
      double deltam2 = pRec.m2Calc() - m2s;
      pRec.e(sqrtpos(pow2(pRec.e()) - deltam2));
    }

    sister.p(pEmt);
    mother.p(pRad);
    newRecoiler.p(pRec);

  } else {

    // Get dipole 4-momentum.
    Vec4 paj_tilde(event[iDaughter].p());
    Vec4 pk_tilde(event[iRecoiler].p());
    Vec4 q(pk_tilde-paj_tilde);

    // Calculate derived variables.
    double q2 = q.m2Calc();
    double saj = uCS/xCS*(q2 - m2s) + (m2r+m2e) * (1-uCS)/xCS;
    double zbar = (q2-saj-m2s) / bABC(q2,saj,m2s)
                *( (xCS - 1)/(xCS-uCS)  - m2s / gABC(q2,saj,m2s)
                       * (saj + m2e - m2r) / (q2 - saj - m2s));
    double kT2  = zbar*(1.-zbar)*saj - (1-zbar)*m2e - zbar*m2r;

    // Now construct recoiler in lab frame.
    Vec4 pRec( (pk_tilde - q*pk_tilde/q2*q)
               *sqrt(lABC(q2,saj,m2s)/lABC(q2,m2Bef,m2s))
             + 0.5*(q2+m2s-saj)/q2*q );
    
    // Now construct the transverse momentum vector in the dipole CM
    // frame.
    double RNphi = rndmPtr->flat();
    double phi_kt = 2.*M_PI*RNphi;
    Vec4 kTmom(sqrt(kT2)*cos(phi_kt), sqrt(kT2)*sin(phi_kt),0.,0.);

    // Boost the transverse momentum vector into the lab frame.
    RotBstMatrix fromDipCM;
    if (side == 1) fromDipCM.fromCMframe( paj_tilde, pk_tilde);
    else           fromDipCM.fromCMframe( pk_tilde, paj_tilde);
    kTmom.rotbst(fromDipCM);

    // Construct left-over dipole momentum by momentum conservation.
    Vec4 paj(-q+pRec);

    // Construct new emission momentum.
    Vec4 pEmt( - zbar * (gABC(q2,saj,m2s)*paj + saj*pRec) / bABC(q2,saj,m2s)
              + (m2e + kT2) / (zbar*bABC(q2,saj,m2s))
              * (pRec + m2s/gABC(q2,saj,m2s)*paj)
              + kTmom);

    // Contruct radiator momentum by momentum conservation.
    Vec4 pRad(-q+pRec+pEmt);

    // Boost to realign the incoming radiator.
    int iOther  = partonSystemsPtr->getInB(iSysSel);
    if (side == 2) iOther = partonSystemsPtr->getInA(iSysSel);

    Vec4 pOther(event[iOther].p());

    // Boost to rest frame of incoming particles A and B.
    RotBstMatrix toABCM;
    if (side == 1) toABCM.toCMframe( pRad, pOther);
    else           toABCM.toCMframe( pOther, pRad);

    // After this, the radiator has vanishing pT.
    pRad.rotbst(toABCM);
    pOther.rotbst(toABCM);

    // Restore the momentum fraction of the incoming particle that
    // did not participate in the splitting.
    RotBstMatrix restoreB;
    restoreB.bst( pOther, event[iOther].p());

    // After this, the inactive beam returns to the correct energy fraction.
    pRad.rotbst(restoreB);
    pOther.rotbst(restoreB);

    // Set all momenta.
    sister.p(pEmt);
    mother.p(pRad);
    newRecoiler.p(pRec);

    // Rotate and boost all final state particles to absorb the pT of the
    // radiator.
    for ( int i = 0; i < event.size(); ++i) {
      if ( event[i].isFinal()) {
        event[i].rotbst(toABCM);
        event[i].rotbst(restoreB);
      }
    }

  }

  // Store masses.
  sister.m(sqrtpos(m2e));
  mother.m(sqrtpos(m2r));
  newRecoiler.m(sqrtpos(m2s));

  // Allow veto of branching. If so restore event record to before emission.
  if ( (canVetoEmission
    && userHooksPtr->doVetoISREmission(eventSizeOld, event, iSysSel))
    || (canMergeFirst
    && mergingHooksPtr->doVetoEmission( event )) ) {
    event.popBack( event.size() - eventSizeOld);
    event[iDaughter].status( iDauStatusV);
    event[iDaughter].mothers(iDauMot1V, iDauMot2V);
    event[iDaughter].daughters(iDauDau1V, iDauDau2V);
    event[iRecoiler].status( iRecStatusV);
    event[iRecoiler].mothers( iRecMot1V, iRecMot2V);
    event[iRecoiler].daughters( iRecDau1V, iRecDau2V);
    return false;
  }

  // Update the number of proposed emissions.
  if (dipEndSel != 0 && nProposedPT.find(dipEndSel->system) != nProposedPT.end())
    ++nProposedPT[dipEndSel->system];

  // Update list of partons in system; adding newly produced parton.
  if (side == 1) partonSystemsPtr->setInA(iSysSel, iMother);
  else partonSystemsPtr->setInB(iSysSel, iMother);

  // Find other initial state parton.
  int iOther = (side==1) ? partonSystemsPtr->getInB(iSysSel)
                         : partonSystemsPtr->getInA(iSysSel); 
  partonSystemsPtr->addOut(iSysSel, iSister);
  partonSystemsPtr->replace(iSysSel, iRecoiler, iNewRecoiler);
  // Get new center-of-mass energy
  double shat = (event[iMother].p() + event[iOther].p()).m2Calc();
  partonSystemsPtr->setSHat(iSysSel, shat);

  // dipEnd array may have expanded and been moved, so regenerate dipEndSel.
  dipEndSel = &dipEnd[iDipSel];

  // Update info on radiating dipole ends (QCD).
  for (int iDip = 0; iDip < int(dipEnd.size()); ++iDip)
  if ( dipEnd[iDip].system == iSysSel) {
    if (abs(dipEnd[iDip].side) == side) {
      dipEnd[iDip].iRadiator = iMother;
      dipEnd[iDip].iRecoiler = iNewRecoiler;
      if (dipEnd[iDip].colType  != 0)
        dipEnd[iDip].colType = mother.colType();

    // Update info on recoiling dipole ends (QCD or QED).
    } else {
      dipEnd[iDip].iRadiator = iNewRecoiler;
      dipEnd[iDip].iRecoiler = iMother;
      dipEnd[iDip].MEtype = 0;
    }
  }

  // Update info on beam remnants.
  double xNew = 2.*event[iMother].e() / event[0].m();
  beamNow[iSysSel].update( iMother, idMother, xNew);
  // Redo choice of companion kind whenever new flavour.
  if (idMother != idDaughterNow) {
    pdfScale2 = (useFixedFacScale) ? fixedFacScale2 : factorMultFac * pT2;
    beamNow.xfISR( iSysSel, idMother, xNew, pdfScale2);
    beamNow.pickValSeaComp();
  }

  // Store branching values of current dipole. (For rapidity ordering.)
  ++dipEndSel->nBranch;
  dipEndSel->pT2Old = pT2;
  dipEndSel->zOld   = z;

  // Update ISR dipoles.
  update(iSysSel,event);

  // Ungraceful exit for incorrect event.
  bool isHadronic = false;
  for (int i = 0; i < event.size(); ++i)
    if (event[i].statusAbs() > 60) isHadronic = true;
  if (!isHadronic && !validEvent(event)) {
    infoPtr->errorMsg("Error in DireSpace::branch_IF: State after "
                      "branching not valid, thus reject.");
    return false;
  }

  // Check that beam momentum not used up by rescattered-system boosts.
  if (beamAPtr->xMax(-1) < 0.0 || beamBPtr->xMax(-1) < 0.0) {
    infoPtr->errorMsg("Warning in DireSpace::branch: "
      "used up beam momentum; retrying parton level");
    rescatterFail = true;
    return false;
  }

  // Set shower weight.
  weights->calcWeight(pT2);
  weights->reset();
  rejectProbability.clear();
  acceptProbability.clear();

  // Done without any errors.
  return true;

}

//--------------------------------------------------------------------------

// Setup splitting kinematics.

Event DireSpace::branched( const Event& event, int iRadBef, int iRecBef,
  int idEmt, int idMother, double pT2, double z, double RN1, vector<double> aux) {

  // Construct the branched event
  Event NewEvent = Event();
  NewEvent.init("(hard process-modified)", particleDataPtr);
  // Copy all unchanged particles to NewEvent
  for (int i = 0; i < event.size(); ++i)
    NewEvent.append( event[i] );

  // Side on which branching occured.
  int side          = (NewEvent[iRadBef].pz() > 0) ? 1 : 2;

  // Read in flavour and colour variables.
  int idDaughterNow = NewEvent[iRadBef].id();
  int colDaughter   = NewEvent[iRadBef].col();
  int acolDaughter  = NewEvent[iRadBef].acol();

  // Current event size.
  int eventSizeOld  = NewEvent.size();

  double RN2 = aux[0];

  // Define colour flow in branching.
  // Default corresponds to f -> f + gamma.
  int colMother     = colDaughter;
  int acolMother    = acolDaughter;
  int colSister     = 0;
  int acolSister    = 0;
  // q -> q + g and 50% of g -> g + g; need new colour.
  if (idEmt == 21 && ( (idMother > 0 && idMother < 9)
  || (idMother == 21 && RN2 < 0.5) ) ) {
    colMother       = NewEvent.nextColTag();
    colSister       = colMother;
    acolSister      = colDaughter;
  // qbar -> qbar + g and other 50% of g -> g + g; need new colour.
  } else if (idEmt == 21) {
    acolMother      = NewEvent.nextColTag();
    acolSister      = acolMother;
    colSister       = acolDaughter;
  // q -> g + q.
  } else if (idDaughterNow == 21 && idMother > 0) {
    colMother       = colDaughter;
    acolMother      = 0;
    colSister       = acolDaughter;
  // qbar -> g + qbar
  } else if (idDaughterNow == 21) {
    acolMother      = acolDaughter;
    colMother       = 0;
    acolSister      = colDaughter;
  // g -> q + qbar.
  } else if (idDaughterNow > 0 && idDaughterNow < 9) {
    acolMother      = NewEvent.nextColTag();
    acolSister      = acolMother;
  // g -> qbar + q.
  } else if (idDaughterNow < 0 && idDaughterNow > -9) {
    colMother       = NewEvent.nextColTag();
    colSister       = colMother;
  }

  // Construct momenta
  int iMother, iSister, iNewRecoiler;
  iMother = iSister = iNewRecoiler = -1;
  vector<Vec4> momenta;
  if ( NewEvent[iRecBef].isFinal()) {
    momenta = split_IF(NewEvent,iRadBef,iRecBef,pT2,z,RN1);

    iSister       = NewEvent.size()-1;
    iNewRecoiler  = NewEvent.size()-1;
    iMother       = NewEvent.size()-3;
    NewEvent[iRecBef].statusNeg();

  } else {
    momenta = split_II(NewEvent,iRadBef,iRecBef,pT2,z,RN1);

    // Indices of partons involved. Add new sister.
    iMother       = eventSizeOld + side - 1;
    iNewRecoiler  = eventSizeOld + 2 - side;
    iSister       = NewEvent.size()-1;
    // Replace old by new mother; update new recoiler.
    NewEvent[iNewRecoiler].status(-42);

  }

  NewEvent[iSister].id(idEmt);
  NewEvent[iSister].cols(colSister, acolSister);

  NewEvent[iMother].id( idMother );
  NewEvent[iMother].status( -41);
  NewEvent[iMother].cols( colMother, acolMother);

  if ( NewEvent[iMother].idAbs() == 21
    || NewEvent[iMother].idAbs() == 22) NewEvent[iMother].pol(9);

  // Update mother and daughter pointers; also for beams.
  NewEvent[iRadBef].mothers( iMother, 0);
  NewEvent[iMother].daughters( iSister, iRadBef);
  if (iSysSel == 0) {
    NewEvent[1].daughter1( (side == 1) ? iMother : iNewRecoiler );
    NewEvent[2].daughter1( (side == 2) ? iMother : iNewRecoiler );
  }

  // Done.
  return NewEvent;

}

//--------------------------------------------------------------------------

vector<Vec4> DireSpace::split_II( Event& event, int iRadBef, int iRecBef,
  double pT2, double z, double RN) {

  // Side on which branching occured.
  int side          = (event[iRadBef].pz() > 0) ? 1 : 2;

  // Read in kinematical variables.
  double m2         = 2.*event[iRadBef].p()*event[iRecBef].p();

  double kappa2 = pT2 / m2;
  double xCS    = (z*(1-z)- kappa2)/(1-z);
  double vCS    = kappa2/(1-z); 

  // Current event and subsystem size.
  int eventSizeOld  = event.size();
  int systemSizeOld = partonSystemsPtr->sizeAll(iSysSel);

  // Take copy of existing system, to be given modified kinematics.
  // Incoming negative status. Rescattered also negative, but after copy.
  for ( int iCopy = 0; iCopy < systemSizeOld; ++iCopy) {
    int iOldCopy    = partonSystemsPtr->getAll(iSysSel, iCopy);
    int statusOld   = event[iOldCopy].status();
    int statusNew   = (iOldCopy == iRadBef
      || iOldCopy == iRecBef) ? statusOld : 44;
    int iNewCopy    = event.copy(iOldCopy, statusNew);
    if (statusOld < 0) event[iNewCopy].statusNeg();
  }

  // Indices of partons involved. Add new sister.
  int iMother       = eventSizeOld + side - 1;
  int iNewRecoiler  = eventSizeOld + 2 - side;
  int iSister       = event.append( 0, 43, iMother, 0, 0, 0, 0, 0,
     Vec4(0.,0.,0.,0.), 0.0, sqrt(pT2) );

  // Replace old by new mother; update new recoiler.
  Vec4 pMother(0.,0., event[iRadBef].pz()/xCS, event[iRadBef].e()/xCS);
  Vec4 pNewRecoiler(0.,0., event[iRecBef].pz(), event[iRecBef].e());

  // Construct CS kinematics.
  double a = 2. * M_PI * RN;

  Vec4 kTmom(sqrt(pT2/xCS)*cos(a), sqrt(pT2/xCS)*sin(a),0.,0.);
  Vec4 pSister( event[iRadBef].p()*(1.-xCS-vCS)/xCS + event[iRecBef].p()*vCS + kTmom);

  event[iMother].p(pMother);
  event[iSister].p(pSister);
  event[iNewRecoiler].p(pNewRecoiler);

  Vec4 kTilde(event[iRadBef].p() + event[iRecBef].p());
  Vec4 k(pMother + pNewRecoiler  - pSister);
  Vec4 kSum(kTilde + k);

  for ( int i = eventSizeOld + 2; i < eventSizeOld + systemSizeOld; ++i) {
    Vec4 pIn = event[i].p();
    double kSum2    = kSum.m2Calc();
    double k2       = k.m2Calc();
    double kTildeXp = kTilde*pIn;
    double kSumXp   = kSum*pIn;
    Vec4 res = pIn - kSum * 2.0*( kSumXp / kSum2 ) + k * 2.0 *( kTildeXp/k2);  
    event[i].p(res);
  }

  // Prepare outputs.
  vector<Vec4> ret;
  ret.push_back( pMother );
  ret.push_back( pSister );
  ret.push_back( pNewRecoiler );

  // Done.
  return ret;

}

//--------------------------------------------------------------------------

vector<Vec4> DireSpace::split_IF( Event& event, int iRadBef, int iRecBef,
  double pT2, double z, double RN) {

  // Side on which branching occured.
  int side          = (event[iRadBef].pz() > 0) ? 1 : 2;

  // Get particle masses.
  double m2Bef = 0.0;
  double m2r   = 0.0;
  double m2s   = 0.0;
  int type     = (event[iRecBef].isFinal()) ? 1 : -1;
  if ( event[iRecBef].idAbs() < 6 && type == 1 )
    m2s = getMass(event[iRecBef].id(),2);
  else if ( type == 1 )
    m2s = getMass(event[iRecBef].id(),3, event[iRecBef].mCalc());

  // Force emission massless for now.
  double m2e   = 0.0;

  // Adjust the dipole kinematical mass to accomodate masses after branching.
  double m2 = 2.*event[iRadBef].p()*event[iRecBef].p() - m2Bef + m2r + m2e;

  // Read in kinematical variables.
  double xCS = z;
  double uCS = (pT2/m2)/(1-z);
  
  // Indices of partons involved. Add new sister.
  int iMother       = event.append( 0, -41, event[iRadBef].mother1(), 0, 0, 0,
     0, 0, Vec4(0.,0.,0.,0.), 0.0, sqrt(pT2) );
  int iNewRecoiler  = event.append( event[iRecBef].id(), 48, iRecBef, iRecBef, 0, 0,
     event[iRecBef].col(), event[iRecBef].acol(),
     Vec4(0.,0.,0.,0.), 0.0, sqrt(pT2) );
  int iSister       = event.append( 0, 43, iMother, 0, 0, 0,
     0, 0, Vec4(0.,0.,0.,0.), 0.0, sqrt(pT2) );

  Vec4 pEmt, pRad, pRec;
  // Massive kinematics, in two schemes.
  if (!useGlobalMapIF) {

    // Get dipole 4-momentum.
    Vec4 paj_tilde(event[iRadBef].p());
    Vec4 pk_tilde(event[iRecBef].p());
    Vec4 pTk_tilde(event[iRecBef].px(),event[iRecBef].py(),0.,0.);
    Vec4 q(paj_tilde-pk_tilde);
    Vec4 qpar(q+pTk_tilde);

    // Calculate derived variables.
    double q2    = q.m2Calc();
    double q2par = qpar.m2Calc();
    double pT2k  = pTk_tilde.m2Calc();
    double sjk   = (1. - 1./xCS)*(q2 - m2r) + (m2e + m2s) / xCS;
    double zbar  = (q2-sjk-m2r) / bABC(q2,sjk,m2r)
                  *( uCS - m2r / gABC(q2,sjk,m2r)
                        * (sjk + m2e - m2s) / (q2 - sjk - m2r));
    double kT2   = zbar*(1.-zbar)*sjk - (1-zbar)*m2e - zbar*m2s;

    if (kT2 < 0.) {
      infoPtr->errorMsg("Warning in DireSpace::split_IF: Reject state "
                        "with kinematically forbidden kT^2.");
      return vector<Vec4>();
    }   

    // Now construct radiator in lab frame.
    pRad = ( paj_tilde - q*paj_tilde/q2par * qpar )
               * sqrt( (lABC(q2,sjk,m2r) - 4.*m2r*pT2k)
                     / (lABC(q2,m2s,m2Bef) - 4.*m2Bef*pT2k))
              + qpar * 0.5 * (q2 + m2r - sjk) / q2par;

    // Now construct the transverse momentum vector in the dipole CM
    // frame.
    double phi_kt = 2.*M_PI*RN;
    Vec4 kTmom(sqrt(kT2)*cos(phi_kt), sqrt(kT2)*sin(phi_kt),0.,0.);
 
    // Boost the transverse momentum vector into the lab frame.
    RotBstMatrix fromDipCM;
    if (side == 1) fromDipCM.fromCMframe( paj_tilde, pk_tilde);
    else           fromDipCM.fromCMframe( pk_tilde, paj_tilde);
    kTmom.rotbst(fromDipCM);

    // Construct left-over dipole momentum by momentum conservation.
    Vec4 pjk(-q+pRad);

    // Construct new emission momentum.
    pEmt = zbar * (gABC(q2,sjk,m2r)*pjk + sjk*pRad) / bABC(q2,sjk,m2r)
              - (m2e + kT2) / (zbar*bABC(q2,sjk,m2r))
              * (pRad + m2r/gABC(q2,sjk,m2r)*pjk)
              + kTmom;

    // Contruct recoiler momentum by momentum conservation.
    pRec = -q+pRad-pEmt;

    event[iSister].p(pEmt);
    event[iMother].p(pRad);
    //event[iNewRecoiler].p(pRec);
    event[iNewRecoiler].p(pk_tilde);
    event[iSister].m(sqrtpos(m2e));
    event[iMother].m(sqrtpos(m2r));
    //event[iNewRecoiler].m(sqrtpos(m2s));
    event[iNewRecoiler].m(0.0);

  } else {

    // Get dipole 4-momentum.
    Vec4 paj_tilde(event[iRadBef].p());
    Vec4 pk_tilde(event[iRecBef].p());
    Vec4 q(pk_tilde-paj_tilde);

    // Calculate derived variables.
    double q2 = q.m2Calc();
    double saj = uCS/xCS*(q2 - m2s) + (m2r+m2e) * (1-uCS)/xCS;
    double zbar = (q2-saj-m2s) / bABC(q2,saj,m2s)
                *( (xCS - 1)/(xCS-uCS)  - m2s / gABC(q2,saj,m2s)
                       * (saj + m2e - m2r) / (q2 - saj - m2s));
    double kT2  = zbar*(1.-zbar)*saj - (1-zbar)*m2e - zbar*m2r;

    // Now construct recoiler in lab frame.
    pRec = (pk_tilde - q*pk_tilde/q2*q)
               *sqrt(lABC(q2,saj,m2s)/lABC(q2,m2Bef,m2s))
             + 0.5*(q2+m2s-saj)/q2*q;
    
    // Now construct the transverse momentum vector in the dipole CM
    // frame.
    double phi_kt = 2.*M_PI*RN;
    Vec4 kTmom(sqrt(kT2)*cos(phi_kt), sqrt(kT2)*sin(phi_kt),0.,0.);

    // Boost the transverse momentum vector into the lab frame.
    RotBstMatrix fromDipCM;
    if (side == 1) fromDipCM.fromCMframe( paj_tilde, pk_tilde);
    else           fromDipCM.fromCMframe( pk_tilde, paj_tilde);
    kTmom.rotbst(fromDipCM);

    // Construct left-over dipole momentum by momentum conservation.
    Vec4 paj(-q+pRec);

    // Construct new emission momentum.
    pEmt=  - zbar * (gABC(q2,saj,m2s)*paj + saj*pRec) / bABC(q2,saj,m2s)
              + (m2e + kT2) / (zbar*bABC(q2,saj,m2s))
              * (pRec + m2s/gABC(q2,saj,m2s)*paj)
              + kTmom;

    // Contruct radiator momentum by momentum conservation.
    pRad = -q+pRec+pEmt;

    // Boost to realign the incoming radiator.
    int iOther  = partonSystemsPtr->getInB(iSysSel);
    if (side == 2) iOther = partonSystemsPtr->getInA(iSysSel);

    Vec4 pOther(event[iOther].p());

    // Boost to rest frame of incoming particles A and B.
    RotBstMatrix toABCM;
    if (side == 1) toABCM.toCMframe( pRad, pOther);
    else           toABCM.toCMframe( pOther, pRad);

    // After this, the radiator has vanishing pT.
    pRad.rotbst(toABCM);
    pOther.rotbst(toABCM);

    // Restore the momentum fraction of the incoming particle that
    // did not participate in the splitting.
    RotBstMatrix restoreB;
    restoreB.bst( pOther, event[iOther].p());

    // After this, the inactive beam returns to the correct energy fraction.
    pRad.rotbst(restoreB);
    pOther.rotbst(restoreB);

    // Set all momenta.
    event[iSister].p(pEmt);
    event[iMother].p(pRad);
    event[iNewRecoiler].p(pRec);
    event[iSister].m(sqrtpos(m2e));
    event[iMother].m(sqrtpos(m2r));
    event[iNewRecoiler].m(sqrtpos(m2s));

    // Rotate and boost all final state particles to absorb the pT of the
    // radiator.
    for ( int i = 0; i < event.size(); ++i) {
      if ( event[i].isFinal()) {
        event[i].rotbst(toABCM);
        event[i].rotbst(restoreB);
      }
    }

  }

  // Prepare outputs.
  vector<Vec4> ret;
  ret.push_back( pRad);
  ret.push_back( pRec);
  ret.push_back( pEmt);

  // Done.
  return ret;

}

//--------------------------------------------------------------------------

Event DireSpace::clustered( const Event& state, string name,
  int iRad, int iEmt, int iRecAft ) {

  // Flags for type of radiation
  int radType = state[iRad].isFinal() ? 1 : -1;
  int recType = state[iRecAft].isFinal() ? 1 : -1;

  // Construct the clustered event
  Event NewEvent = Event();
  NewEvent.init("(hard process-modified)", particleDataPtr);
  NewEvent.clear();
  // Copy all unchanged particles to NewEvent
  for (int i = 0; i < state.size(); ++i) {
    if ( i == iRad || i == iRecAft || i == iEmt ) continue;
    NewEvent.append( state[i] );
  }

  // Copy all the junctions one by one
  for (int i = 0; i < state.sizeJunction(); ++i)
    NewEvent.appendJunction( state.getJunction(i) );
  // Find an appropriate scale for the hard process
  double mu = infoPtr->QFac();
  // Initialise scales for new event
  NewEvent.saveSize();
  NewEvent.saveJunctionSize();
  NewEvent.scale(mu);
  NewEvent.scaleSecond(mu);

  // Set properties of radiator/recoiler after the clustering
  // Recoiler properties will be unchanged
  Particle RecBefore = Particle( state[iRecAft] );
  RecBefore.setEvtPtr(&NewEvent);
  RecBefore.daughters(0,0);
  // Find flavour of radiator before splitting
  int radID = (*splittingsPtr)[name]->radID(state[iRad].id(), state[iEmt].id());
  int recID = state[iRecAft].id();
  Particle RadBefore = Particle( state[iRad] );
  RadBefore.setEvtPtr(&NewEvent);
  RadBefore.id(radID);
  RadBefore.daughters(0,0);
  // Put dummy values for colours
  RadBefore.cols(RecBefore.acol(),RecBefore.col());

  // Reset status if the reclustered radiator is a resonance.
  if ( particleDataPtr->isResonance(radID) && radType == 1)
    RadBefore.status(state[iRad].status());

  // Put mass for radiator and recoiler
  double radMass = particleDataPtr->m0(radID);
  double recMass = particleDataPtr->m0(recID);
  if (radType == 1 ) RadBefore.m(radMass);
  else RadBefore.m(0.0);
  if (recType == 1 ) RecBefore.m(recMass);
  else RecBefore.m(0.0);

  // Construct momenta and  colours of clustered particles
  if ( state[iRecAft].isFinal())
    cluster_IF(state,iRad,iEmt,iRecAft,radID,RadBefore,RecBefore,NewEvent);
  else
    cluster_II(state,iRad,iEmt,iRecAft,radID,RadBefore,RecBefore,NewEvent);

  // Put some dummy production scales for RecBefore, RadBefore
  RecBefore.scale(mu);
  RadBefore.scale(mu);

  // Append new recoiler and find new radiator colour
  NewEvent.append(RecBefore);

  // Assign the correct colour to re-clustered radiator.
  pair<int,int> cols = (*splittingsPtr)[name]->radCols( state[iRad].col(),
    state[iRad].acol(), state[iEmt].col(), state[iEmt].acol());
  RadBefore.cols( cols.first, cols.second );

  // Build the clustered event
  Event outState = Event();
  outState.init("(hard process-modified)", particleDataPtr);
  outState.clear();

  // Copy system and incoming beam particles to outState
  for (int i = 0; i < 3; ++i)
    outState.append( NewEvent[i] );
  // Copy all the junctions one by one
  for (int i = 0; i < state.sizeJunction(); ++i)
    outState.appendJunction( state.getJunction(i) );
  // Initialise scales for new event
  outState.saveSize();
  outState.saveJunctionSize();
  outState.scale(mu);
  outState.scaleSecond(mu);
  bool radAppended = false;
  bool recAppended = false;
  int size = int(outState.size());
  // Save position of radiator in new event record
  int radPos = 0;

  // Append first incoming particle
  if ( RecBefore.mother1() == 1) {
    outState.append( RecBefore );
    recAppended = true;
  } else if ( RadBefore.mother1() == 1 ) {
    radPos = outState.append( RadBefore );
    radAppended = true;
  } else {
    // Find second incoming in input event
    int in1 = 0;
    for(int i=0; i < int(state.size()); ++i)
      if (state[i].mother1() == 1) in1 =i;
    outState.append( state[in1] );
    size++;
  }
  // Append second incoming particle
  if ( RecBefore.mother1() == 2) {
    outState.append( RecBefore );
    recAppended = true;
  } else if ( RadBefore.mother1() == 2 ) {
    radPos = outState.append( RadBefore );
    radAppended = true;
  } else {
    // Find second incoming in input event
    int in2 = 0;
    for(int i=0; i < int(state.size()); ++i)
      if (state[i].mother1() == 2) in2 =i;

    outState.append( state[in2] );
    size++;
  }

  // Append new recoiler if not done already
  if (!recAppended && !RecBefore.isFinal()) {
    recAppended = true;
    outState.append( RecBefore);
  }
  // Append new radiator if not done already
  if (!radAppended && !RadBefore.isFinal()) {
    radAppended = true;
    radPos = outState.append( RadBefore);
  }

  // Force incoming partons to have "hard event" status
  outState[3].status(-21);
  outState[4].status(-21);

  // Append intermediate particle
  // (careful not to append reclustered recoiler)
  // Append intermediate particle
  // (careful not to append reclustered recoiler)
  for (int i = 0; i < int(NewEvent.size()-1); ++i) {
    if (NewEvent[i].status() != -22) continue;
    if ( NewEvent[i].daughter1() == NewEvent[i].daughter2()
      && NewEvent[i].daughter1() > 0) continue;
    outState.append( NewEvent[i] );
  }
  // Append final state particles, resonances first
  for (int i = 0; i < int(NewEvent.size()-1); ++i)
    if (NewEvent[i].status() == 22) outState.append( NewEvent[i] );
  // Then start appending partons
  if (!radAppended && RadBefore.statusAbs() == 22)
    radPos = outState.append(RadBefore);
  if (!recAppended)
    outState.append(RecBefore);
  if (!radAppended && RadBefore.statusAbs() != 22)
    radPos = outState.append(RadBefore);
  // Then partons (not reclustered recoiler)
  for(int i = 0; i < int(NewEvent.size()-1); ++i)
    if ( NewEvent[i].status()  != 22
      && NewEvent[i].colType() != 0
      && NewEvent[i].isFinal()) {
      outState.append( NewEvent[i] );
      // Force partons to have "hard event" status
      int status = particleDataPtr->isResonance(NewEvent[i].id()) ? 22 : 23;
      outState.back().status(status);
      outState.back().mother1(3);
      outState.back().mother2(4);
    }
  // Then the rest
  for(int i = 0; i < int(NewEvent.size()-1); ++i)
    if ( NewEvent[i].status() != 22
      && NewEvent[i].colType() == 0
      && NewEvent[i].isFinal() ) {
      outState.append( NewEvent[i]);
      int status = particleDataPtr->isResonance(NewEvent[i].id()) ? 22 : 23;
      outState.back().status(status);
      outState.back().mother1(3);
      outState.back().mother2(4);
    }
 
  // Find intermediate and respective daughters
  vector<int> PosIntermediate;
  vector<int> PosDaughter1;
  vector<int> PosDaughter2;
  for(int i=0; i < int(outState.size()); ++i)
    if (outState[i].status() == -22) {
      PosIntermediate.push_back(i);
      int d1 = outState[i].daughter1();
      int d2 = outState[i].daughter2();
      // Find daughters in output state
      int daughter1 = FindParticle( state[d1], outState);
      int daughter2 = FindParticle( state[d2], outState);
      // If both daughters found, done
      // Else put first final particle as first daughter
      // and last final particle as second daughter
      if (daughter1 > 0)
        PosDaughter1.push_back( daughter1);
      else {
        daughter1 = 0;
        while(!outState[daughter1].isFinal() ) daughter1++;
        PosDaughter1.push_back( daughter1);
      }
      if (daughter2 > 0)
        PosDaughter2.push_back( daughter2);
      else {
        daughter2 = outState.size()-1;
        while(!outState[daughter2].isFinal() ) daughter2--;
        PosDaughter2.push_back( daughter2);
      }
    }
  // Set daughters and mothers
  for(int i=0; i < int(PosIntermediate.size()); ++i) {
    outState[PosIntermediate[i]].daughters(PosDaughter1[i],PosDaughter2[i]);
    outState[PosDaughter1[i]].mother1(PosIntermediate[i]);
    outState[PosDaughter2[i]].mother1(PosIntermediate[i]);
  }

  // Find range of final state partons
  int minParFinal = int(outState.size());
  int maxParFinal = 0;
  for(int i=0; i < int(outState.size()); ++i)
    if (outState[i].mother1() == 3 && outState[i].mother2() == 4) {
      minParFinal = min(i,minParFinal);
      maxParFinal = max(i,maxParFinal);
    }

  if (minParFinal == maxParFinal) maxParFinal = 0;
  outState[3].daughters(minParFinal,maxParFinal);
  outState[4].daughters(minParFinal,maxParFinal);

  // Update event properties
  outState.saveSize();
  outState.saveJunctionSize();

  // Almost there...
  // If an intermediate coloured parton exists which was directly
  // colour connected to the radiator before the splitting, and the
  // radiator before and after the splitting had only one colour, problems
  // will arise since the colour of the radiator will be changed, whereas
  // the intermediate parton still has the old colour. In effect, this
  // means that when setting up a event for trial showering, one colour will
  // be free.
  // Hence, check for an intermediate coloured triplet resonance has been
  // colour-connected to the "old" radiator.
  // Find resonance
  int iColRes = 0;
  if ( radType == -1 && state[iRad].colType() == 1) {
      // Find resonance connected to initial colour
      for(int i=0; i < int(state.size()); ++i)
        if ( i != iRad && i != iEmt && i != iRecAft
          && state[i].status() == -22
          && state[i].col() == state[iRad].col() )
          iColRes = i;
  } else if ( radType == -1 && state[iRad].colType() == -1) {
      // Find resonance connected to initial anticolour
      for(int i=0; i < int(state.size()); ++i)
        if ( i != iRad && i != iEmt && i != iRecAft
          && state[i].status() == -22
          && state[i].acol() == state[iRad].acol() )
          iColRes = i;
  } else if ( radType == 1 && state[iRad].colType() == 1) {
      // Find resonance connected to final state colour
      for(int i=0; i < int(state.size()); ++i)
        if ( i != iRad && i != iEmt && i != iRecAft
          && state[i].status() == -22
          && state[i].col() == state[iRad].col() )
          iColRes = i;
  } else if ( radType == 1 && state[iRad].colType() == -1) {
      // Find resonance connected to final state anticolour
      for(int i=0; i < int(state.size()); ++i)
        if ( i != iRad && i != iEmt && i != iRecAft
          && state[i].status() == -22
          && state[i].acol() == state[iRad].acol() )
          iColRes = i;
  }

  if (iColRes > 0) {
    // Now find this resonance in the reclustered state
    int iColResNow = FindParticle( state[iColRes], outState);

    // Find reclustered radiator colours
    int radCol = outState[radPos].col();
    int radAcl = outState[radPos].acol();
    // Find resonance radiator colours
    int resCol = outState[iColResNow].col();
    int resAcl = outState[iColResNow].acol();
    // Check if any of the reclustered radiators colours match the resonance
    bool matchesRes =  (radCol > 0
                          && ( radCol == resCol || radCol == resAcl))
                    || (radAcl > 0
                          && ( radAcl == resCol || radAcl == resAcl));

    // If a resonance has been found, but no colours match, change
    // the colour of the resonance
    if (!matchesRes && iColResNow > 0) {
      if ( radType == -1 && outState[radPos].colType() == 1)
        outState[iColResNow].col(radCol);
      else if ( radType ==-1 && outState[radPos].colType() ==-1)
        outState[iColResNow].acol(radAcl);
      else if ( radType == 1 && outState[radPos].colType() == 1)
        outState[iColResNow].col(radCol);
      else if ( radType == 1 && outState[radPos].colType() ==-1)
        outState[iColResNow].acol(radAcl);
    }


    // If a resonance has been found, but no colours match, and the position
    // of the resonance in the event record has been changed, update the
    // radiator mother
    if (!matchesRes && iColResNow > 0 && iColRes != iColResNow)
      outState[radPos].mother1(iColResNow);

  }

  // Done
  return outState;

}

bool DireSpace::cluster_II( const Event& state,
  int iRad, int iEmt, int iRecAft, int idRadBef, Particle& radBef,
  Particle& recBef, Event& partial ) {

  if (false) cout << idRadBef;

  // Calculate CS variables.
  double pT2    = pT2_II(state[iRad], state[iEmt], state[iRecAft]);
  double Q2     = 2.*state[iRad].p()*state[iRecAft].p()
                - 2.*state[iRad].p()*state[iEmt].p()
                - 2.*state[iEmt].p()*state[iRecAft].p();
  double z      = z_II(state[iRad], state[iEmt], state[iRecAft]);

  double kappa2 = pT2 / Q2;
  double xCS    = (z*(1-z)- kappa2)/(1-z);

  // Check phase space contraints.
  double xNew = 2.*state[iRad].e()/state[0].m();
  double xOld = xNew*xCS;
  if ( !inAllowedPhasespace( z, pT2, Q2, xOld, -1, 0.0, 0.0, 0.0,
    0.0) ) return false;

  // Set up kinematics.
  radBef.p( 0., 0., xCS * state[iRad].pz(), xCS * state[iRad].e());
  recBef.p( 0., 0., state[iRecAft].pz(), state[iRecAft].e() );

  // Set mass of initial recoiler to zero
  radBef.m( 0.0 );
  recBef.m( 0.0 );

  Vec4 kTilde(radBef.p() + recBef.p());
  Vec4 k(state[iRad].p() + state[iRecAft].p()  - state[iEmt].p());
  Vec4 kSum(kTilde + k);
  for ( int i = 0; i < partial.size(); ++i) {
    if ( !partial[i].isFinal() && partial[i].statusAbs() != 22 ) continue;
    Vec4 pIn = partial[i].p();
    double kSum2    = kSum.m2Calc();
    double k2       = k.m2Calc();
    double kXp      = k*pIn;
    double kSumXp   = kSum*pIn;
    Vec4 res = pIn - kSum * 2.0*( kSumXp / kSum2 ) + kTilde * 2.0 *( kXp/k2);  
    partial[i].p(res);
  }

  // Done
  return true;

}

bool DireSpace::cluster_IF( const Event& state, 
  int iRad, int iEmt, int iRecAft, int idRadBef, Particle& radBef,
  Particle& recBef, Event& partial ) {

  if (false) cout << idRadBef;

  // Calculate CS variables.
  double pT2 = pT2_IF(state[iRad], state[iEmt], state[iRecAft]);
  double z   = z_IF(state[iRad], state[iEmt], state[iRecAft]);
  double xCS = z;
  int side   = (state[iRad].pz() > 0.) ? 1 : -1;

  // Get particle masses.
  double m2Bef = 0.0, m2r = 0.0, m2e = 0.0;
  double m2s   = state[iRecAft].p().m2Calc();

  // Adjust the dipole kinematical mass to accomodate masses after branching.
  double Q2  = 2.*state[iRad].p()*state[iEmt].p()
             + 2.*state[iRad].p()*state[iRecAft].p()
             - 2.*state[iRecAft].p()*state[iEmt].p();
  double xNew = 2.*state[iRad].e()/state[0].m();
  double xOld = xNew*xCS; 
  // Check phase space contraints.
  if ( !inAllowedPhasespace( z, pT2, Q2, xOld, 2, m2Bef, m2r, m2s,
    m2e) ) return false;

  Vec4 pRadBef, pRecBef;

  // Massive kinematics, in two schemes.
  if (!useGlobalMapIF) {

    // Get dipole 4-momentum.
    Vec4 paj(state[iRad].p()-state[iEmt].p());
    Vec4 pk(state[iRecAft].p());
    Vec4 pTk(pk.px(),pk.py(),0.,0.);
    Vec4 q(paj-pk);
    Vec4 qpar(q+pTk);

    // Calculate derived variables.
    double q2    = q.m2Calc();
    double q2par = qpar.m2Calc();
    double pT2k  = pTk.m2Calc();
    double sjk   = 2.*state[iRecAft].p()*state[iEmt].p();

    // Now construct radiator in lab frame.
    pRadBef = ( paj - q*paj/q2par * qpar )
               * sqrt( (lABC(q2,m2Bef,m2r) - 4.*m2r*pT2k)
                     / (lABC(q2,m2s,sjk) - 4.*sjk*pT2k))
              + qpar * 0.5 * (q2 + m2r - m2Bef) / q2par;

    // Contruct recoiler momentum by momentum conservation.
    pRecBef = -q+pRadBef;

    pRadBef = xCS*state[iRad].p();
    pRecBef = state[iRecAft].p() + state[iEmt].p() - (1.-xCS)*state[iRad].p();

    radBef.p(pRadBef);
    recBef.p(pRecBef);
    radBef.m(sqrtpos(m2r));
    recBef.m(sqrtpos(m2s));

  } else {

    // Get dipole 4-momentum.
    Vec4 paj(state[iRad].p()-state[iEmt].p());
    Vec4 pk(state[iRecAft].p());
    Vec4 q(pk-paj);

    // Calculate derived variables.
    double q2 = q.m2Calc();
    double saj = 2.*state[iRad].p()*state[iEmt].p();

    // Now construct recoiler in lab frame.
    pRecBef = (pk - q*pk/q2*q)
               *sqrt(lABC(q2,m2Bef,m2s)/lABC(q2,saj,m2s))
             + 0.5*(q2+m2s-m2Bef)/q2*q;
    
    // Contruct radiator momentum by momentum conservation.
    pRadBef = -q+pRecBef;

    // Boost to realign the incoming radiator.
    int iOther  = partonSystemsPtr->getInB(iSysSel);
    if (side == 2) iOther = partonSystemsPtr->getInA(iSysSel);

    Vec4 pOther(state[iOther].p());

    // Boost to rest frame of incoming particles A and B.
    RotBstMatrix toABCM;
    if (side == 1) toABCM.toCMframe( pRadBef, pOther);
    else           toABCM.toCMframe( pOther, pRadBef);

    // After this, the radiator has vanishing pT.
    pRadBef.rotbst(toABCM);
    pOther.rotbst(toABCM);

    // Restore the momentum fraction of the incoming particle that
    // did not participate in the splitting.
    RotBstMatrix restoreB;
    restoreB.bst( pOther, state[iOther].p());

    // After this, the inactive beam returns to the correct energy fraction.
    pRadBef.rotbst(restoreB);
    pOther.rotbst(restoreB);

    // Set all momenta.
    radBef.p(pRadBef);
    recBef.p(pRecBef);
    radBef.m(sqrtpos(m2r));
    recBef.m(sqrtpos(m2s));

    // Rotate and boost all final state particles to absorb the pT of the
    // radiator.
    for ( int i = 0; i < partial.size(); ++i) {
      if ( !partial[i].isFinal() && partial[i].statusAbs() != 22 ) continue;
      partial[i].rotbst(toABCM);
      partial[i].rotbst(restoreB);
    }

  }

  // Done
  return true;
}

//--------------------------------------------------------------------------

// Function to in the input event find a particle with quantum
// numbers matching those of the input particle
// IN  Particle : Particle to be searched for
//     Event    : Event to be searched in
// OUT int      : > 0 : Position of matching particle in event
//                < 0 : No match in event

int DireSpace::FindParticle( const Particle& particle, const Event& event,
  bool checkStatus ) {

  int index = -1;

  for ( int i = int(event.size()) - 1; i > 0; --i )
    if ( event[i].id()         == particle.id()
      && event[i].colType()    == particle.colType()
      && event[i].chargeType() == particle.chargeType()
      && event[i].col()        == particle.col()
      && event[i].acol()       == particle.acol()
      && event[i].charge()     == particle.charge() ) {
      index = i;
      break;
    }

  if ( checkStatus && event[index].status() != particle.status() )
    index = -1;

  return index;
}

//--------------------------------------------------------------------------

double DireSpace::pT2_II ( const Particle& rad, const Particle& emt,
  const Particle& rec) {
  double sai = -2.*rad.p()*emt.p();
  double sbi = -2.*rec.p()*emt.p();
  double sab =  2.*rad.p()*rec.p();
  return sai*sbi / sab * ( sai+sbi+sab ) / sab;
}

//--------------------------------------------------------------------------

double DireSpace::pT2_IF ( const Particle& rad, const Particle& emt,
  const Particle& rec) {
  double sai = -2.*rad.p()*emt.p();
  double sik =  2.*rec.p()*emt.p();
  double sak = -2.*rad.p()*rec.p();
  return sai*sik / (sai+sak) * (sai+sik+sak) / (sai+sak);
}

//--------------------------------------------------------------------------

double DireSpace::z_II ( const Particle& rad, const Particle& emt,
  const Particle& rec) {
  double sbi = -2.*rec.p()*emt.p();
  double sab =  2.*rad.p()*rec.p();
  return 1. + sbi/sab;
}

//-------------------------------------------------------------------------

double DireSpace::z_IF ( const Particle& rad, const Particle& emt,
  const Particle& rec) {
  double sai = -2.*rad.p()*emt.p();
  double sik =  2.*rec.p()*emt.p();
  double sak = -2.*rad.p()*rec.p();
  return 1. + sik / (sai+sak);
}

//-------------------------------------------------------------------------

// Compute splitting probability.

double DireSpace::getSplittingProb( const Event& state, int iRad,
  int iEmt, int iRecAft) {

  string name  = getSplittingName( state, iRad, iEmt);
  double z     = zSpace(state[iRad], state[iEmt], state[iRecAft]);
  double pT2   = pT2Space(state[iRad], state[iEmt], state[iRecAft]);
  double m2D   = (state[iRecAft].isFinal())
               ? abs( 2.*state[iEmt].p()*state[iRad].p()
                     -2.*state[iEmt].p()*state[iRecAft].p()
                     +2.*state[iRad].p()*state[iRecAft].p()) 
               : abs(-2.*state[iEmt].p()*state[iRad].p()
                     -2.*state[iEmt].p()*state[iRecAft].p()
                     +2.*state[iRad].p()*state[iRecAft].p());
  int idRadBef = splittingsPtr->getSplittingRadBefID(state, iRad, iEmt);
  double m2Bef = pow2(particleDataPtr->m0(idRadBef));
  double m2r   = state[iRad].p().m2Calc();
  double m2e   = state[iEmt].p().m2Calc();
  double m2s   = state[iRecAft].p().m2Calc();
  int type     = (state[iRecAft].isFinal()) ? 1 : -1;

  // Adjust the dipole kinematical mass to accomodate masses after branching.
  double m2dipCorr  = m2D - m2Bef + m2r + m2e;
  double kappa2     = pT2 / m2dipCorr;
  double xDau       = (state[iRecAft].isFinal()
                      ? z : (z*(1-z) - kappa2) / (1 -z));

  // Check phase space contraints.
  if ( !inAllowedPhasespace( z, pT2, m2dipCorr, xDau, type, m2Bef, m2r, m2s,
    m2e) ) {
    return 0.0;
  }

  // Calculate splitting probability.
  // Sum over g->gg colour structures.
  double p = 0.;
  if ( name.compare("isr_21->21&21a") == 0 || name.compare("isr_21->21&21b") == 0) {
    p += (*splittingsPtr)["isr_21->21&21a"]->kernel(z,pT2,m2D,type,m2Bef,m2r,m2e,m2s);
    p += (*splittingsPtr)["isr_21->21&21b"]->kernel(z,pT2,m2D,type,m2Bef,m2r,m2e,m2s);
  } else
    p += (*splittingsPtr)[name]->kernel(z,pT2,m2D,type,m2Bef,m2r,m2e,m2s);

  return p;

}

// For new Pythia version superseeding 8.212
//double DireSpace::getSplittingProb( const Event& state, int iRad,
//  int iEmt, int iRecAft, string) {
//
//  string name  = getSplittingName( state, iRad, iEmt, iRecAft);
//  double z     = zSpace(state[iRad], state[iEmt], state[iRecAft]);
//  double pT2   = pT2Space(state[iRad], state[iEmt], state[iRecAft]);
//  double m2D   = (state[iRecAft].isFinal())
//               ? abs( 2.*state[iEmt].p()*state[iRad].p()
//                     -2.*state[iEmt].p()*state[iRecAft].p()
//                     +2.*state[iRad].p()*state[iRecAft].p()) 
//               : abs(-2.*state[iEmt].p()*state[iRad].p()
//                     -2.*state[iEmt].p()*state[iRecAft].p()
//                     +2.*state[iRad].p()*state[iRecAft].p());
//  int idRadBef = splittingsPtr->getSplittingRadBefID(state, iRad, iEmt);
//  double m2Bef = pow2(particleDataPtr->m0(idRadBef));
//  double m2r   = state[iRad].p().m2Calc();
//  double m2e   = state[iEmt].p().m2Calc();
//  double m2s   = state[iRecAft].p().m2Calc();
//  int type     = (state[iRecAft].isFinal()) ? 1 : -1;
//
//  // Adjust the dipole kinematical mass to accomodate masses after branching.
//  double m2dipCorr  = m2D - m2Bef + m2r + m2e;
//  double kappa2     = pT2 / m2dipCorr;
//  double xDau       = (state[iRecAft].isFinal()
//                      ? z : (z*(1-z) - kappa2) / (1 -z));
//
//  // Check phase space contraints.
//  if ( !inAllowedPhasespace( z, pT2, m2dipCorr, xDau, type, m2Bef, m2r, m2s,
//    m2e) ) {
//    return 0.0;
//  }
//
//  // Calculate splitting probability.
//  // Sum over g->gg colour structures.
//  double p = 0.;
//  if ( name.compare("isr_21->21&21a") == 0 || name.compare("isr_21->21&21b") == 0) {
//    p += (*splittingsPtr)["isr_21->21&21a"]->kernel(z,pT2,m2D,type,m2Bef,m2r,m2e,m2s);
//    p += (*splittingsPtr)["isr_21->21&21b"]->kernel(z,pT2,m2D,type,m2Bef,m2r,m2e,m2s);
//  } else
//    p += (*splittingsPtr)[name]->kernel(z,pT2,m2D,type,m2Bef,m2r,m2e,m2s);
//
//  return p;
//
//}

//-------------------------------------------------------------------------

// Check colour/flavour correctness of state.

bool DireSpace::validEvent( const Event& state ) {

  bool validColour  = true;
  bool validCharge  = true;
  bool validMomenta = true;

  bool hasSystems = partonSystemsPtr->sizeSys() > 0;
  int sizeSys     = (hasSystems) ? partonSystemsPtr->sizeSys() : 1;

    Event event = Event();

  for (int iSys = 0; iSys < sizeSys; ++iSys) {

    // Done if the state is already broken.
    if (!validColour || !validCharge ) break;

    event.clear();
    event.init( "(hard process-modified)", particleDataPtr );
    event.clear();

    int in1 = 0;
    for ( int i = state.size()-1; i > 0; --i)
      if ( state[i].mother1() == 1 && state[i].mother2() == 0
        && (!hasSystems || partonSystemsPtr->getSystemOf(i,true) == iSys))
        {in1 = i; break;}
    if (in1 == 0) in1 = partonSystemsPtr->getInA(iSys);
    int in2 = 0;
    for ( int i = state.size()-1; i > 0; --i)
      if ( state[i].mother1() == 2 && state[i].mother2() == 0
        && (!hasSystems || partonSystemsPtr->getSystemOf(i,true) == iSys))
        {in2 = i; break;}
    if (in2 == 0) in2 = partonSystemsPtr->getInB(iSys);

    event.append(state[0]);
    event.append(state[1]);
    event[1].daughters(3,0);
    event.append(state[2]);
    event[2].daughters(4,0);
    event.append(state[in1]);
    event[3].mothers(1,0);
    event[3].status(-21);
    event.append(state[in2]);
    event[4].mothers(2,0);
    event[4].status(-21);
    for ( int i = 0; i < state.size(); ++i)
      if ( state[i].isFinal() 
        && (!hasSystems || partonSystemsPtr->getSystemOf(i) == iSys)) {
         int iN = event.append(state[i]);
         event[iN].daughters(0,0);
         event[iN].mothers(3,4);
         event[iN].status(23);
      }
    event[3].daughters(5,event.size()-1);
    event[4].daughters(5,event.size()-1);

    // Check if event is coloured
    for ( int i = 0; i < event.size(); ++i)
     // Check colour of quarks
     if ( event[i].isFinal() && event[i].colType() == 1
            // No corresponding anticolour in final state
         && ( FindCol(event[i].col(),i,event,1) == 0
            // No corresponding colour in initial state
           && FindCol(event[i].col(),i,event,2) == 0 )) {
       validColour = false;
       break;
     // Check anticolour of antiquarks
     } else if ( event[i].isFinal() && event[i].colType() == -1
            // No corresponding colour in final state
         && ( FindCol(event[i].acol(),i,event,2) == 0
            // No corresponding anticolour in initial state
           && FindCol(event[i].acol(),i,event,1) == 0 )) {
       validColour = false;
       break;
     // No uncontracted colour (anticolour) charge of gluons
     } else if ( event[i].isFinal() && event[i].colType() == 2
            // No corresponding anticolour in final state
         && ( FindCol(event[i].col(),i,event,1) == 0
            // No corresponding colour in initial state
           && FindCol(event[i].col(),i,event,2) == 0 )
            // No corresponding colour in final state
         && ( FindCol(event[i].acol(),i,event,2) == 0
            // No corresponding anticolour in initial state
           && FindCol(event[i].acol(),i,event,1) == 0 )) {
       validColour = false;
       break;
     }

    // Check charge sum in initial and final state
    double initCharge = event[3].charge() + event[4].charge();
    double finalCharge = 0.0;
    for(int i = 0; i < event.size(); ++i)
      if (event[i].isFinal()) finalCharge += event[i].charge();
    if (abs(initCharge-finalCharge) > 1e-12) validCharge = false;

    double mTolErr = settingsPtr->parm("Check:mTolErr");
    // Check if particles are on mass shell
    for ( int i = 0; i < event.size(); ++i) {
      if (i==3 || i==4 || event[i].isFinal()) {
        int id = event[i].id();
        // Get mass. Zero for incoming.
        double mNow = (i==3 || i==4) ? 0.
                    : ((abs(id) < 6) ? getMass(id,2) : getMass(id,1));
        mNow = sqrt(mNow);
        // Do not check on-shell condition for massive intermediate (!)
        // resonances. Assuming all non-SM particles are heavy here! 
        if ( abs(id) == 6 || abs(id) > 22) mNow = event[i].mCalc();
        double errMass = abs(event[i].mCalc() - mNow)
                       / max( 1.0, event[i].e());
        if ( errMass > mTolErr ) validMomenta = false;
      }
    }

    // Check that overall pT is vanishing.
    Vec4 pSum(0.,0.,0.,0.);
    for ( int i = 0; i < event.size(); ++i)
      if ( i ==3 || i == 4 || event[i].isFinal()) pSum += event[i].p();
    if ( abs(pSum.px()) > mTolErr || abs(pSum.py()) > mTolErr)
      validMomenta = false;
    if ( abs(event[3].px()) > mTolErr || abs(event[3].py()) > mTolErr
      || abs(event[4].px()) > mTolErr || abs(event[4].py()) > mTolErr)
      validMomenta = false;

  } // Done with loop over systems.

  return (validColour && validCharge && validMomenta);

}

//-------------------------------------------------------------------------

// Find index colour partner for input colour.

int DireSpace::FindCol(int col, int iExclude, const Event& event, int type) {
  int index = 0;
  // Search event record for matching colour & anticolour
  for(int n = 0; n < event.size(); ++n) {
    if ( n != iExclude && event[n].colType() != 0
      &&(   event[n].status() > 0          // Check outgoing
         || event[n].mother1() == 1        // Check incoming
         || event[n].mother1() == 2) ) {   // Check incoming
       if ( event[n].acol() == col ) {
        index = -n;
        break;
      }
      if ( event[n].col()  == col ) {
        index =  n;
        break;
      }
    }
  }
  // if no matching colour / anticolour has been found, return false
  if ( type == 1 && index < 0) return abs(index);
  if ( type == 2 && index > 0) return abs(index);
  return 0;
}

//-------------------------------------------------------------------------

// Print the list of dipoles.

void DireSpace::list(ostream& os) const {

  // Header.
  os << "\n --------  PYTHIA DireSpace Dipole Listing  -------------- \n"
     << "\n    i  syst  side   rad   rec       pTmax  col  chg   ME rec \n"
     << fixed << setprecision(3);

  // Loop over dipole list and print it.
  for (int i = 0; i < int(dipEnd.size()); ++i)
  os << setw(5) << i << setw(6) << dipEnd[i].system
     << setw(6) << dipEnd[i].side << setw(6) << dipEnd[i].iRadiator
     << setw(6) << dipEnd[i].iRecoiler << setw(12) << dipEnd[i].pTmax
     << setw(5) << dipEnd[i].colType << setw(5) << dipEnd[i].chgType
     << setw(5) << dipEnd[i].MEtype << setw(4)
     << dipEnd[i].normalRecoil << "\n";

  // Done.
  os << "\n --------  End PYTHIA DireSpace Dipole Listing  ----------"
     << endl;


}

//==========================================================================

} // end namespace Pythia8
