// DireTimes.cc is a part of the PYTHIA event generator.
// Copyright (C) 2015 Torbjorn Sjostrand.
// PYTHIA is licenced under the GNU GPL version 2, see COPYING for details.
// Please respect the MCnet Guidelines, see GUIDELINES for details.

// Function definitions (not found in the header) for the DireTimes class.

#include "../include/DireTimes.h"

namespace Pythia8 {

//==========================================================================

// The DireTimes class.

//--------------------------------------------------------------------------

// Constants: could be changed here if desired, but normally should not.
// These are of technical nature, as described for each.

// Minimal allowed c and b quark masses, for flavour thresholds.
const double DireTimes::MCMIN        = 1.2;
const double DireTimes::MBMIN        = 4.0;

// For small x approximate 1 - sqrt(1 - x) by x/2.
const double DireTimes::SIMPLIFYROOT = 1e-8;

// Do not allow x too close to 0 or 1 in matrix element expressions.
// Warning: cuts into phase space for E_CM > 2 * pTmin * sqrt(1/XMARGIN),
// i.e. will become problem roughly for E_CM > 10^6 GeV.
const double DireTimes::XMARGIN      = 1e-12;
const double DireTimes::XMARGINCOMB  = 1e-4;

// Lower limit on PDF value in order to avoid division by zero.
//const double DireTimes::TINYPDF      = 1e-15;
const double DireTimes::TINYPDF      = 1e-6;

// Big starting value in search for smallest invariant-mass pair.
const double DireTimes::LARGEM2      = 1e20;

// In g -> q qbar or gamma -> f fbar require m2_pair > this * m2_q/f.
const double DireTimes::THRESHM2      = 4.004;

// Never pick pT so low that alphaS is evaluated too close to Lambda_3.
const double DireTimes::LAMBDA3MARGIN = 1.1;

// Extra overestimate of g -> q qbar branching rate for DGLAP comparison.
const double DireTimes::WG2QEXTRA = 20.;

const double DireTimes::TINYMASS = 1e-3;

//--------------------------------------------------------------------------

// Initialize alphaStrong, alphaEM and related pTmin parameters.

void DireTimes::init( BeamParticle* beamAPtrIn,
  BeamParticle* beamBPtrIn) {

  // Header.
  if (printBanner) {
  cout << "\n"
       << " *-----------------  Welcome to DIRE version " << DIRE_TIMES_VERSION
       << "  ----------------*\n"
       << " |                                                "
       << "                  |\n"
       << " | Please consider citing Eur.Phys.J. C75 (2015)"
       << " 9, 461             |\n"
       << " | if you use this program for scientific purposes."
       << "                 |\n";
  cout << " |                                                "
       << "                  |\n";
  cout << " | You are using DIRE as timelike shower.         "
       << "                  |\n";
  cout << " |                                                "
       << "                  |\n"
       << " *-----------------  Dire times end.  ------------"
       << "------------------*" << endl;
  printBanner = false;}

  // Store input pointers for future use.
  beamAPtr           = beamAPtrIn;
  beamBPtr           = beamBPtrIn;

  // Main flags.
  doQCDshower        = settingsPtr->flag("TimeShower:QCDshower");
  doMEcorrections    = settingsPtr->flag("TimeShower:MEcorrections");
  doMEafterFirst     = settingsPtr->flag("TimeShower:MEafterFirst");
  doPhiPolAsym       = settingsPtr->flag("TimeShower:phiPolAsym");
  doInterleave       = settingsPtr->flag("TimeShower:interleave");
  allowBeamRecoil    = settingsPtr->flag("TimeShower:allowBeamRecoil");
  dampenBeamRecoil   = settingsPtr->flag("TimeShower:dampenBeamRecoil");
  recoilToColoured   = settingsPtr->flag("TimeShower:recoilToColoured");

  // Matching in pT of hard interaction or MPI to shower evolution.
  pTmaxMatch         = settingsPtr->mode("TimeShower:pTmaxMatch");
  pTdampMatch        = settingsPtr->mode("TimeShower:pTdampMatch");
  pTmaxFudge         = settingsPtr->parm("TimeShower:pTmaxFudge");
  pTmaxFudgeMPI      = settingsPtr->parm("TimeShower:pTmaxFudgeMPI");
  pTdampFudge        = settingsPtr->parm("TimeShower:pTdampFudge");

  // Charm and bottom mass thresholds.
  mc                 = max( MCMIN, particleDataPtr->m0(4));
  mb                 = max( MBMIN, particleDataPtr->m0(5));
  m2c                = mc * mc;
  m2b                = mb * mb;

  // Parameters of scale choices.
  renormMultFac     = settingsPtr->parm("TimeShower:renormMultFac");
  factorMultFac     = settingsPtr->parm("TimeShower:factorMultFac");
  useFixedFacScale  = settingsPtr->flag("TimeShower:useFixedFacScale");
  fixedFacScale2    = pow2(settingsPtr->parm("TimeShower:fixedFacScale"));

  // Parameters of alphaStrong generation.
  alphaSvalue        = settingsPtr->parm("TimeShower:alphaSvalue");
  alphaSorder        = settingsPtr->mode("TimeShower:alphaSorder");
  alphaSnfmax        = settingsPtr->mode("StandardModel:alphaSnfmax");
  alphaSuseCMW       = settingsPtr->flag("TimeShower:alphaSuseCMW");
  alphaS2pi          = 0.5 * alphaSvalue / M_PI;

  // Initialize alphaStrong generation.
  alphaS.init( alphaSvalue, alphaSorder, alphaSnfmax, alphaSuseCMW);
  // Set flavour thresholds by default Pythia masses, unless zero.
  double mcpy = particleDataPtr->m0(4);
  double mbpy = particleDataPtr->m0(5);
  double mtpy = particleDataPtr->m0(6);
  if (mcpy > 0.0 && mbpy > 0.0 && mtpy > 0.0)
    alphaS.setThresholds(mcpy, mbpy, mtpy);

  // Lambda for 5, 4 and 3 flavours.
  Lambda3flav        = alphaS.Lambda3();
  Lambda4flav        = alphaS.Lambda4();
  Lambda5flav        = alphaS.Lambda5();
  Lambda5flav2       = pow2(Lambda5flav);
  Lambda4flav2       = pow2(Lambda4flav);
  Lambda3flav2       = pow2(Lambda3flav);

  // Parameters of QCD evolution. Warn if pTmin must be raised.
  nGluonToQuark      = settingsPtr->mode("TimeShower:nGluonToQuark");
  weightGluonToQuark = settingsPtr->mode("TimeShower:weightGluonToQuark");
  scaleGluonToQuark  = settingsPtr->parm("TimeShower:scaleGluonToQuark");
  extraGluonToQuark  = (weightGluonToQuark%4 == 3) ? WG2QEXTRA : 1.;
  pTcolCutMin        = settingsPtr->parm("TimeShower:pTmin");
  if (pTcolCutMin > LAMBDA3MARGIN * Lambda3flav / sqrt(renormMultFac))
    pTcolCut         = pTcolCutMin;
  else {
    pTcolCut         = LAMBDA3MARGIN * Lambda3flav / sqrt(renormMultFac);
    ostringstream newPTcolCut;
    newPTcolCut << fixed << setprecision(3) << pTcolCut;
    infoPtr->errorMsg("Warning in DireTimes::init: pTmin too low",
                      ", raised to " + newPTcolCut.str() );
    infoPtr->setTooLowPTmin(true);
  }
  pT2colCut          = pow2(pTcolCut);

  usePDFalphas       = settingsPtr->flag("ShowerPDF:usePDFalphas");
  useSummedPDF       = settingsPtr->flag("ShowerPDF:useSummedPDF");
  BeamParticle& beam = (abs(beamAPtr->id()) == 2212) ? *beamAPtr : *beamBPtr;
  alphaS2piOverestimate = (usePDFalphas) ? beam.alphaS(pT2colCut) * 0.5/M_PI
                        : (alphaSorder > 0) ? alphaS.alphaS(pT2colCut)*0.5/M_PI
                                            :  0.5 * 0.5/M_PI;

  nFinalMax          = settingsPtr->mode("DireTimes:nFinalMax");
  usePDFmasses       = settingsPtr->flag("ShowerPDF:usePDFmasses");

  rejectProbability.clear();
  acceptProbability.clear();
  splittingSelName="";
  splittingNowName="";

  // May have to fix up recoils related to rescattering.
  allowRescatter     = settingsPtr->flag("PartonLevel:MPI")
    && settingsPtr->flag("MultipartonInteractions:allowRescatter");

  // Possibility of two predetermined hard emissions in event.
  doSecondHard    = settingsPtr->flag("SecondHard:generate");

  // Possibility to allow user veto of emission step.
  hasUserHooks       = (userHooksPtr != 0);
  canVetoEmission    = (userHooksPtr != 0)
                     ? userHooksPtr->canVetoFSREmission() : false;

  // Set initial value, just in case.
  dopTdamp           = false;
  pT2damp            = 0.;

}

//--------------------------------------------------------------------------

// Find whether to limit maximum scale of emissions.
// Also allow for dampening at factorization or renormalization scale.

bool DireTimes::limitPTmax( Event& event, double Q2Fac, double Q2Ren) {

  // Find whether to limit pT. Begin by user-set cases.
  bool dopTlimit = false;
  dopTlimit1 = dopTlimit2 = false;
  int nHeavyCol = 0;
  if      (pTmaxMatch == 1) dopTlimit = dopTlimit1 = dopTlimit2 = true;
  else if (pTmaxMatch == 2) dopTlimit = dopTlimit1 = dopTlimit2 = false;

  // Always restrict SoftQCD processes.
  else if (infoPtr->isNonDiffractive() || infoPtr->isDiffractiveA()
    || infoPtr->isDiffractiveB() || infoPtr->isDiffractiveC() )
    dopTlimit = dopTlimit1 = dopTlimit2 = true;

  // Look if any quark (u, d, s, c, b), gluon or photon in final state.
  // Also count number of heavy coloured particles, like top. 
  else {
    int n21 = 0;
    for (int i = 5; i < event.size(); ++i) {
      if (event[i].status() == -21) ++n21;
      else if (n21 == 0) {
        int idAbs = event[i].idAbs();
        if (idAbs <= 5 || idAbs == 21 || idAbs == 22) dopTlimit1 = true;
        if ( (event[i].col() != 0 || event[i].acol() != 0)
          && idAbs > 5 && idAbs != 21 ) ++nHeavyCol;        
      } else if (n21 == 2) {
        int idAbs = event[i].idAbs();
        if (idAbs <= 5 || idAbs == 21 || idAbs == 22) dopTlimit2 = true;
      }
    }
    dopTlimit = (doSecondHard) ? (dopTlimit1 && dopTlimit2) : dopTlimit1;
  }

  // Dampening at factorization or renormalization scale; only for hardest.
  dopTdamp   = false;
  pT2damp    = 0.;
  if (!dopTlimit1 && (pTdampMatch == 1 || pTdampMatch == 2)) {
    dopTdamp = true;
    pT2damp  = pow2(pTdampFudge) * ((pTdampMatch == 1) ? Q2Fac : Q2Ren);
  }
  if (!dopTlimit1 && nHeavyCol > 1 && (pTdampMatch == 3 || pTdampMatch == 4)) {
    dopTdamp = true;
    pT2damp  = pow2(pTdampFudge) * ((pTdampMatch == 3) ? Q2Fac : Q2Ren);
  }

  // Done.
  return dopTlimit;

}

//--------------------------------------------------------------------------

// Top-level routine to do a full time-like shower in resonance decay.

int DireTimes::shower( int iBeg, int iEnd, Event& event, double pTmax,
  int nBranchMax) {

  // Add new system, automatically with two empty beam slots.
  int iSys = partonSystemsPtr->addSys();

  // Loop over allowed range to find all final-state particles.
  Vec4 pSum;
  for (int i = iBeg; i <= iEnd; ++i) if (event[i].isFinal()) {
    partonSystemsPtr->addOut( iSys, i);
    pSum += event[i].p();
  }
  partonSystemsPtr->setSHat( iSys, pSum.m2Calc() );

  // Let prepare routine do the setup.
  dopTlimit1        = true;
  dopTlimit2        = true;
  dopTdamp          = false;
  prepare( iSys, event, true);

  // Begin evolution down in pT from hard pT scale.
  int nBranch  = 0;
  pTLastBranch = 0.;
  do {
    double pTtimes = pTnext( event, pTmax, 0.);

    // Do a final-state emission (if allowed).
    if (pTtimes > 0.) {
      if (branch( event)) {
        ++nBranch;
        pTLastBranch = pTtimes;
      }
      pTmax = pTtimes;
    }

    // Keep on evolving until nothing is left to be done.
    else pTmax = 0.;
  } while (pTmax > 0. && (nBranchMax <= 0 || nBranch < nBranchMax));

  // Return number of emissions that were performed.
  return nBranch;

}

//--------------------------------------------------------------------------

// Global recoil: reset counters and store locations of outgoing partons.

void DireTimes::prepareGlobal( Event& ) {
  return;
}

//--------------------------------------------------------------------------

// Prepare system for evolution; identify ME.

void DireTimes::prepare( int iSys, Event& event, bool limitPTmaxIn) {

  // Reset dipole-ends list for first interaction and for resonance decays.
  int iInA = partonSystemsPtr->getInA(iSys);
  int iInB = partonSystemsPtr->getInB(iSys);
  if (iSys == 0 || iInA == 0) dipEnd.resize(0);
  int dipEndSizeBeg = dipEnd.size();

  // No dipoles for 2 -> 1 processes.
  if (partonSystemsPtr->sizeOut(iSys) < 2) return;

  // In case of DPS overwrite limitPTmaxIn by saved value.
  if (doSecondHard && iSys == 0) limitPTmaxIn = dopTlimit1;
  if (doSecondHard && iSys == 1) limitPTmaxIn = dopTlimit2;

  // Loop through final state of system to find possible dipole ends.
  for (int i = 0; i < partonSystemsPtr->sizeOut(iSys); ++i) {
    int iRad = partonSystemsPtr->getOut( iSys, i);

    if (event[iRad].isFinal() && event[iRad].scale() > 0.) {

      // Find dipole end formed by colour index.
      int colTag = event[iRad].col();
      if (doQCDshower && colTag > 0) setupQCDdip( iSys, i,  colTag,  1, event,
        false, limitPTmaxIn);

      // Find dipole end formed by anticolour index.
      int acolTag = event[iRad].acol();
      if (doQCDshower && acolTag > 0) setupQCDdip( iSys, i, acolTag, -1, event,
        false, limitPTmaxIn);

    // End loop over system final state. Have now found the dipole ends.
    }
  }

  // Loop through dipole ends and set matrix element correction flag.
  for (int iDip = dipEndSizeBeg; iDip < int(dipEnd.size()); ++iDip)
    dipEnd[iDip].MEtype = 0;

  // Update dipole list after a multiparton interactions rescattering.
  if (iSys > 0 && ( (iInA > 0 && event[iInA].status() == -34)
    || (iInB > 0 && event[iInB].status() == -34) ) )
    rescatterUpdate( iSys, event);

  // Counter of proposed emissions.
  nProposedPT.clear();
  if ( nProposedPT.find(iSys) == nProposedPT.end() )
    nProposedPT.insert(make_pair(iSys,0));

  splittingSelName="";
  splittingNowName="";
  rejectProbability.clear();
  acceptProbability.clear();

}

//--------------------------------------------------------------------------

// Update dipole list after a multiparton interactions rescattering.

void DireTimes::rescatterUpdate( int iSys, Event& event) {

  // Loop over two incoming partons in system; find their rescattering mother.
  // (iOut is outgoing from old system = incoming iIn of rescattering system.)
  for (int iResc = 0; iResc < 2; ++iResc) {
    int iIn = (iResc == 0) ? partonSystemsPtr->getInA(iSys)
                           : partonSystemsPtr->getInB(iSys);
    if (iIn == 0 || event[iIn].status() != -34) continue;
    int iOut = event[iIn].mother1();

    // Loop over all dipoles.
    int dipEndSize = dipEnd.size();
    for (int iDip = 0; iDip < dipEndSize; ++iDip) {
      DireTimesEnd& dipNow = dipEnd[iDip];

      // Kill dipoles where rescattered parton is radiator.
      if (dipNow.iRadiator == iOut) {
        dipNow.colType = 0;
        dipNow.chgType = 0;
        dipNow.gamType = 0;
        continue;
      }
      // No matrix element for dipoles between scatterings.
      if (dipNow.iMEpartner == iOut) {
        dipNow.MEtype     =  0;
        dipNow.iMEpartner = -1;
      }

      // Update dipoles where outgoing rescattered parton is recoiler.
      if (dipNow.iRecoiler == iOut) {
        int iRad = dipNow.iRadiator;

        // Colour dipole: recoil in final state, initial state or new.
        if (dipNow.colType > 0) {
          int colTag = event[iRad].col();
          bool done  = false;
          for (int i = 0; i < partonSystemsPtr->sizeOut(iSys); ++i) {
            int iRecNow = partonSystemsPtr->getOut( iSys, i);
            if (event[iRecNow].acol() == colTag) {
              dipNow.iRecoiler = iRecNow;
              dipNow.systemRec = iSys;
              dipNow.MEtype    = 0;
              done             = true;
              break;
            }
          }
          if (!done) {
            int iIn2 = (iResc == 0) ? partonSystemsPtr->getInB(iSys)
                                    : partonSystemsPtr->getInA(iSys);
            if (event[iIn2].col() == colTag) {
              dipNow.iRecoiler = iIn2;
              dipNow.systemRec = iSys;
              dipNow.MEtype    = 0;
              int isrType      = event[iIn2].mother1();
              // This line in case mother is a rescattered parton.
              while (isrType > 2 + beamOffset)
                isrType = event[isrType].mother1();
              if (isrType > 2) isrType -= beamOffset;
              dipNow.isrType   = isrType;
              done             = true;
            }
          }
          // If above options failed, then create new dipole.
          if (!done) {
            int iRadNow = partonSystemsPtr->getIndexOfOut(dipNow.system, iRad);
            if (iRadNow != -1)
              setupQCDdip(dipNow.system, iRadNow, event[iRad].col(), 1,
                          event, dipNow.isOctetOnium, true);
            else
              infoPtr->errorMsg("Warning in DireTimes::rescatterUpdate: "
              "failed to locate radiator in system");

            dipNow.colType = 0;
            dipNow.chgType = 0;
            dipNow.gamType = 0;

            infoPtr->errorMsg("Warning in DireTimes::rescatterUpdate: "
            "failed to locate new recoiling colour partner");
          }

        // Anticolour dipole: recoil in final state, initial state or new.
        } else if (dipNow.colType < 0) {
          int  acolTag = event[iRad].acol();
          bool done    = false;
          for (int i = 0; i < partonSystemsPtr->sizeOut(iSys); ++i) {
            int iRecNow = partonSystemsPtr->getOut( iSys, i);
            if (event[iRecNow].col() == acolTag) {
              dipNow.iRecoiler = iRecNow;
              dipNow.systemRec = iSys;
              dipNow.MEtype    = 0;
              done             = true;
              break;
            }
          }
          if (!done) {
            int iIn2 = (iResc == 0) ? partonSystemsPtr->getInB(iSys)
                                    : partonSystemsPtr->getInA(iSys);
            if (event[iIn2].acol() == acolTag) {
              dipNow.iRecoiler = iIn2;
              dipNow.systemRec = iSys;
              dipNow.MEtype    = 0;
              int isrType      = event[iIn2].mother1();
              // This line in case mother is a rescattered parton.
              while (isrType > 2 + beamOffset)
                isrType = event[isrType].mother1();
              if (isrType > 2) isrType -= beamOffset;
              dipNow.isrType   = isrType;
              done             = true;
            }
          }
          // If above options failed, then create new dipole.
          if (!done) {
            int iRadNow = partonSystemsPtr->getIndexOfOut(dipNow.system, iRad);
            if (iRadNow != -1)
              setupQCDdip(dipNow.system, iRadNow, event[iRad].acol(), -1,
                          event, dipNow.isOctetOnium, true);
            else
              infoPtr->errorMsg("Warning in DireTimes::rescatterUpdate: "
              "failed to locate radiator in system");

            dipNow.colType = 0;
            dipNow.chgType = 0;
            dipNow.gamType = 0;

            infoPtr->errorMsg("Warning in DireTimes::rescatterUpdate: "
            "failed to locate new recoiling colour partner");
          }
        }
      }

    // End of loop over dipoles and two incoming sides.
    }
  }

}

//--------------------------------------------------------------------------

// Update dipole list after each ISR emission (so not used for resonances).

void DireTimes::update( int iSys, Event& event, bool) {

  // Start list of rescatterers that gave further changed systems in ISR.
  vector<int> iRescatterer;

  // Find new and old positions of incoming partons in the system.
  vector<int> iNew, iOld;
  iNew.push_back( partonSystemsPtr->getInA(iSys) );
  iOld.push_back( event[iNew[0]].daughter2() );
  iNew.push_back( partonSystemsPtr->getInB(iSys) );
  iOld.push_back( event[iNew[1]].daughter2() );

  // Ditto for outgoing partons, except the newly created one.
  int sizeOut = partonSystemsPtr->sizeOut(iSys) - 1;
  for (int i = 0; i < sizeOut; ++i) {
    int iNow = partonSystemsPtr->getOut(iSys, i);
    iNew.push_back( iNow );
    iOld.push_back( event[iNow].mother1() );
    // Add non-final to list of rescatterers.
    if (!event[iNow].isFinal()) iRescatterer.push_back( iNow );
  }
  int iNewNew = partonSystemsPtr->getOut(iSys, sizeOut);

  // Swap beams to let 0 be side on which branching occured.
  if (event[iNew[0]].status() != -41) {
    swap( iNew[0], iNew[1]);
    swap( iOld[0], iOld[1]);
  }

  // Loop over all dipole ends belonging to the system
  // or to the recoil system, if different.
  for (int iDip = 0; iDip < int(dipEnd.size()); ++iDip)
  if (dipEnd[iDip].system == iSys || dipEnd[iDip].systemRec == iSys) {
    DireTimesEnd& dipNow = dipEnd[iDip];

    // Replace radiator (always in final state so simple).
    bool replacedRad = false;
    for (int i = 2; i < 2 + sizeOut; ++i)
    if (dipNow.iRadiator == iOld[i]) {
      dipNow.iRadiator = iNew[i];
      replacedRad = true;
      break;
    }

    // No need to update if both radiator and recoiler are still intact.
    if ( !replacedRad ) {
      bool oldRec = false;
      for ( int i=0; i < int(iNew.size()); ++i )
        if (dipNow.iRecoiler == iNew[i]) {
          oldRec = true;
          break;
        }
      if (oldRec) continue;
    }

    // Replace ME partner (always in final state, if exists, so simple).
    for (int i = 2; i < 2 + sizeOut; ++i)
    if (dipNow.iMEpartner == iOld[i]) {
      dipNow.iMEpartner = iNew[i];
      break;
    }

    // Check which colour line to follow to get new recoiler.
    int iR = dipNow.iRadiator;
    int iS = dipNow.iRecoiler;
    bool isInitial = (event[iS].status() == -21
                   || event[iS].status() == -41 || event[iS].status() == -42
                   || event[iS].status() == -53
                   || event[iS].status() == -31 || event[iS].status() == -34);
    bool colMatch  =  event[iR].col() > 0 && 
                     ((!isInitial && event[iR].col()  == event[iS].acol()) 
                   || (isInitial  && event[iR].col()  == event[iS].col()));
    bool acolMatch =  event[iR].acol() &&
                     ((!isInitial && event[iR].acol() == event[iS].col())
                   || (isInitial  && event[iR].acol() == event[iS].acol()));

    // Recoiler: Start by trying to follow colour/anticolour line.
    int iRec = 0;
    if (dipNow.systemRec == iSys) {

      if (colMatch && !acolMatch) {
        int acolF = FindCol(event[iR].col(),dipNow.iRadiator,event,1,iSys);
        int  colI = FindCol(event[iR].col(),dipNow.iRadiator,event,2,iSys);
        if (acolF  > 0 && colI == 0) iRec = acolF;
        if (acolF == 0 && colI >  0) iRec = colI;
      } else if (acolMatch && !colMatch) {
        int  colF = FindCol(event[iR].acol(),dipNow.iRadiator,event,2,iSys);
        int acolI = FindCol(event[iR].acol(),dipNow.iRadiator,event,1,iSys);
        if ( colF  > 0 && acolI == 0) iRec = colF;
        if ( colF == 0 && acolI >  0) iRec = acolI;
      }

      // Recoiler: by default pick old one, only moved.
      for (int i = 1; i < 2 + sizeOut; ++i)
      if (iRec == 0 && dipNow.iRecoiler == iOld[i]) {
        iRec = iNew[i];
        break;
      }

      // QCD recoiler: check if colour hooks up with new final parton.
      if ( iRec == 0 && dipNow.colType > 0
        && event[dipNow.iRadiator].col() == event[iNewNew].acol() ) {
        iRec = iNewNew;
        dipNow.isrType = 0;
      }
      if ( iRec == 0 && dipNow.colType < 0
        && event[dipNow.iRadiator].acol() == event[iNewNew].col() ) {
        iRec = iNewNew;
        dipNow.isrType = 0;
      }
      // QCD recoiler: check if colour hooks up with new beam parton.
      if ( iRec == 0 && dipNow.colType > 0
        && event[dipNow.iRadiator].col()  == event[iNew[0]].col() )
        iRec = iNew[0];
      if ( iRec == 0 && dipNow.colType < 0
        && event[dipNow.iRadiator].acol() == event[iNew[0]].acol() )
        iRec = iNew[0];

      // Careful with QCD recoiler after initial-final gluon emission.
      if ( iRec == 0 && dipNow.colType > 0
        && event[dipNow.iRadiator].status() == 48
        && event[dipNow.iRadiator].col()  == event[iNew[1]].col() )
        iRec = iNew[1];
      if ( iRec == 0 && dipNow.colType < 0
        && event[dipNow.iRadiator].status() == 48
        && event[dipNow.iRadiator].acol() == event[iNew[1]].acol() )
        iRec = iNew[1];
      // Initial-final gluon emission can repaint initial quark, so
      // that the recoiler is now a final state gluon.
      if ( iRec == 0 && dipNow.colType > 0
        && event[dipNow.iRadiator].status() == 48)
        for (int i = 1; i < 2 + sizeOut; ++i) {
          if (dipNow.iRadiator == iNew[i]) continue;
          if (event[dipNow.iRadiator].col() == event[iNew[i]].acol()) {
          iRec = iNew[i];
            break;
          }
        }
      if ( iRec == 0 && dipNow.colType < 0
        && event[dipNow.iRadiator].status() == 48)
        for (int i = 1; i < 2 + sizeOut; ++i) {
          if (dipNow.iRadiator == iNew[i]) continue;
          if (event[dipNow.iRadiator].acol() == event[iNew[i]].col()) {
            iRec = iNew[i];
            break;
          }
        }

      // QED/photon recoiler: either to new particle or remains to beam.
      if ( iRec == 0 && (dipNow.chgType != 0 || dipNow.gamType != 0) ) {
        if ( event[iNew[0]].chargeType() == 0 ) {
          iRec = iNewNew;
          dipNow.isrType = 0;
        } else {
          iRec = iNew[0];
        }
      }

    // Recoiler in another system: keep it as is.
    } else iRec = dipNow.iRecoiler;

    // Done. Kill dipole if failed to find new recoiler.
    dipNow.iRecoiler = iRec;
    if ( iRec == 0 && (dipNow.colType != 0 || dipNow.chgType != 0
      || dipNow.gamType != 0) ) {
      dipNow.colType = 0;
      dipNow.chgType = 0;
      dipNow.gamType = 0;
      infoPtr->errorMsg("Error in DireTimes::update: "
      "failed to locate new recoiling partner");
    }

  }

  // Find new dipole end formed by colour index.
  int colTag = event[iNewNew].col();
  if (doQCDshower && colTag > 0)
    setupQCDdip( iSys, sizeOut, colTag, 1, event, false, true);

  // Find new dipole end formed by anticolour index.
  int acolTag = event[iNewNew].acol();
  if (doQCDshower && acolTag > 0)
    setupQCDdip( iSys, sizeOut, acolTag, -1, event, false, true);

  // Check the dipoles are correctly set up.
  for (int iDip = 0; iDip < int(dipEnd.size()); ++iDip) {
    DireTimesEnd& dipNow = dipEnd[iDip];
    int iR      = dipNow.iRadiator;
    int iS      = dipNow.iRecoiler;
    int iSysNow = dipNow.system;
    bool isInitial = ( iS == partonSystemsPtr->getInA(iSysNow)
                    || iS == partonSystemsPtr->getInB(iSysNow));
    bool colMatch  =  (!isInitial && event[iR].col()  == event[iS].acol()) 
                   || (isInitial  && event[iR].col()  == event[iS].col());
    bool acolMatch =  (!isInitial && event[iR].acol() == event[iS].col())
                   || (isInitial  && event[iR].acol() == event[iS].acol());
    // Ungraceful exit if dipole ends to not share colour connection.
    if (!colMatch && !acolMatch) {
      infoPtr->errorMsg("Error in DireTimes::update: dipole ends",
                        "do not share colour connection");
      return;
    }
  }

}

//--------------------------------------------------------------------------

// Setup a dipole end for a QCD colour charge.

void DireTimes::setupQCDdip( int iSys, int i, int colTag, int colSign,
  Event& event, bool isOctetOnium, bool limitPTmaxIn) {

  // Initial values. Find if allowed to hook up beams.
  int iRad     = partonSystemsPtr->getOut(iSys, i);
  int iRec     = 0;
  int sizeAllA = partonSystemsPtr->sizeAll(iSys);
  int sizeOut  = partonSystemsPtr->sizeOut(iSys);
  int sizeAll  = ( allowBeamRecoil ) ? sizeAllA : sizeOut;
  int sizeIn   = sizeAll - sizeOut;
  int sizeInA  = sizeAllA - sizeIn - sizeOut;
  int iOffset  = i + sizeAllA - sizeOut;
  bool otherSystemRec = false;
  bool allowInitial   = (partonSystemsPtr->hasInAB(iSys)) ? true : false;
  vector<int> iRecVec(0);

  // Colour: other end by same index in beam or opposite in final state.
  // Exclude rescattered incoming and not final outgoing.
  if (colSign > 0)
  for (int j = 0; j < sizeAll; ++j) if (j + sizeInA != iOffset) {
    int iRecNow = partonSystemsPtr->getAll(iSys, j + sizeInA);
    if ( ( j <  sizeIn && event[iRecNow].col()  == colTag
      && !event[iRecNow].isRescatteredIncoming() )
      || ( j >= sizeIn && event[iRecNow].acol() == colTag
      && event[iRecNow].isFinal() ) ) {
      iRec = iRecNow;
      break;
    }
  }

  // Anticolour: other end by same index in beam or opposite in final state.
  // Exclude rescattered incoming and not final outgoing.
  if (colSign < 0)
  for (int j = 0; j < sizeAll; ++j) if (j + sizeInA != iOffset) {
    int iRecNow = partonSystemsPtr->getAll(iSys, j + sizeInA);
    if ( ( j <  sizeIn && event[iRecNow].acol()  == colTag
      && !event[iRecNow].isRescatteredIncoming() )
      || ( j >= sizeIn && event[iRecNow].col() == colTag
      && event[iRecNow].isFinal() ) ) {
      iRec = iRecNow;
      break;
    }
  }

  // Resonance decays (= no instate):
  // other end to nearest recoiler in same system final state,
  // by (p_i + p_j)^2 - (m_i + m_j)^2 = 2 (p_i p_j - m_i m_j).
  // (junction colours more involved, so keep track if junction colour)
  bool hasJunction = false;
  if (iRec == 0 && !allowInitial) {
    for (int iJun = 0; iJun < event.sizeJunction(); ++ iJun) {
      // For types 1&2, all legs in final state
      // For types 3&4, two legs in final state
      // For types 5&6, one leg in final state
      int iBeg = (event.kindJunction(iJun)-1)/2;
      for (int iLeg = iBeg; iLeg < 3; ++iLeg)
        if (event.endColJunction( iJun, iLeg) == colTag) hasJunction  = true;
    }
    double ppMin = LARGEM2;
    for (int j = 0; j < sizeOut; ++j) if (j != i) {
        int iRecNow  = partonSystemsPtr->getOut(iSys, j);
        if (!event[iRecNow].isFinal()) continue;
        double ppNow = event[iRecNow].p() * event[iRad].p()
          - event[iRecNow].m() * event[iRad].m();
        if (ppNow < ppMin) {
          iRec  = iRecNow;
          ppMin = ppNow;
        }
      }
  }

  // If no success then look for matching (anti)colour anywhere in final state.
  if ( iRec == 0 ) {
    for (int j = 0; j < event.size(); ++j) if (event[j].isFinal()) {
      if ( (colSign > 0 && event[j].acol() == colTag)
        || (colSign < 0 && event[j].col()  == colTag) ) {
        iRec = j;
        otherSystemRec = true;
        break;
      }
    }

    // If no success then look for match to non-rescattered in initial state.
    if (iRec == 0 && allowInitial) {
      for (int iSysR = 0; iSysR < partonSystemsPtr->sizeSys(); ++iSysR)
      if (iSysR != iSys) {
        int j = partonSystemsPtr->getInA(iSysR);
        if (j > 0 && event[j].isRescatteredIncoming()) j = 0;
        if (j > 0 && ( (colSign > 0 && event[j].col() == colTag)
          || (colSign < 0 && event[j].acol()  == colTag) ) ) {
          iRec = j;
          otherSystemRec = true;
          break;
        }
        j = partonSystemsPtr->getInB(iSysR);
        if (j > 0 && event[j].isRescatteredIncoming()) j = 0;
        if (j > 0 && ( (colSign > 0 && event[j].col() == colTag)
          || (colSign < 0 && event[j].acol()  == colTag) ) ) {
          iRec = j;
          otherSystemRec = true;
          break;
        }
      }
    }
  }

  // Junctions (PS&ND dec 2010)
  // For types 1&2: all legs in final state
  //                half-strength dipoles between all legs
  // For types 3&4, two legs in final state
  //                full-strength dipole between final-state legs
  // For types 5&6, one leg in final state
  //                no final-state dipole end

  if (hasJunction) {
    for (int iJun = 0; iJun < event.sizeJunction(); ++ iJun) {
      int kindJun = event.kindJunction(iJun);
      int iBeg = (kindJun-1)/2;
      for (int iLeg = iBeg; iLeg < 3; ++iLeg) {
        if (event.endColJunction( iJun, iLeg) == colTag) {
          // For types 5&6, no other leg to recoil against. Switch off if
          // no other particles at all, since radiation then handled by ISR.
          // Example: qq -> ~t* : no radiation off ~t*
          // Allow radiation + recoil if unconnected partners available
          // Example: qq -> ~t* -> tbar ~chi0 : allow radiation off tbar,
          //                                    with ~chi0 as recoiler
          if (kindJun >= 5) {
            if (sizeOut == 1) return;
            else break;
          }
          // For junction types 3 & 4, span one full-strength dipole
          // (only look inside same decay system)
          else if (kindJun >= 3) {
            int iLegRec = 3-iLeg;
            int colTagRec = event.endColJunction( iJun, iLegRec);
            for (int j = 0; j < sizeOut; ++j) if (j != i) {
                int iRecNow  = partonSystemsPtr->getOut(iSys, j);
                if (!event[iRecNow].isFinal()) continue;
                if ( (colSign > 0 && event[iRecNow].col()  == colTagRec)
                  || (colSign < 0 && event[iRecNow].acol() == colTagRec) ) {
                  // Only accept if staying inside same system
                  iRec = iRecNow;
                  break;
                }
              }
          }
          // For junction types 1 & 2, span two half-strength dipoles
          // (only look inside same decay system)
          else {
            // Loop over two half-strength dipole connections
            for (int jLeg = 1; jLeg <= 2; jLeg++) {
              int iLegRec = (iLeg + jLeg) % 3;
              int colTagRec = event.endColJunction( iJun, iLegRec);
              for (int j = 0; j < sizeOut; ++j) if (j != i) {
                  int iRecNow  = partonSystemsPtr->getOut(iSys, j);
                  if (!event[iRecNow].isFinal()) continue;
                  if ( (colSign > 0 && event[iRecNow].col()  == colTagRec)
                    || (colSign < 0 && event[iRecNow].acol() == colTagRec) ) {
                    // Store recoilers in temporary array
                    iRecVec.push_back(iRecNow);
                    // Set iRec != 0 for checks below
                    iRec = iRecNow;
                  }
                }
            }

          }     // End if-then-else of junction kinds

        }       // End if leg has right color tag
      }         // End of loop over junction legs
    }           // End loop over junctions

  }             // End main junction if

  // If fail, then other end to nearest recoiler in same system final state,
  // by (p_i + p_j)^2 - (m_i + m_j)^2 = 2 (p_i p_j - m_i m_j).
  if (iRec == 0) {
    double ppMin = LARGEM2;
    for (int j = 0; j < sizeOut; ++j) if (j != i) {
      int iRecNow  = partonSystemsPtr->getOut(iSys, j);
      if (!event[iRecNow].isFinal()) continue;
      double ppNow = event[iRecNow].p() * event[iRad].p()
                   - event[iRecNow].m() * event[iRad].m();
      if (ppNow < ppMin) {
        iRec  = iRecNow;
        ppMin = ppNow;
      }
    }
  }

  // If fail, then other end to nearest recoiler in any system final state,
  // by (p_i + p_j)^2 - (m_i + m_j)^2 = 2 (p_i p_j - m_i m_j).
  if (iRec == 0) {
    double ppMin = LARGEM2;
    for (int iRecNow = 0; iRecNow < event.size(); ++iRecNow)
    if (iRecNow != iRad && event[iRecNow].isFinal()) {
      double ppNow = event[iRecNow].p() * event[iRad].p()
                   - event[iRecNow].m() * event[iRad].m();
      if (ppNow < ppMin) {
        iRec  = iRecNow;
        otherSystemRec = true;
        ppMin = ppNow;
      }
    }
  }

  // PS dec 2010: make sure iRec is stored in iRecVec
  if (iRecVec.size() == 0 && iRec != 0) iRecVec.push_back(iRec);

  // Remove any zero recoilers from normalization
  int nRec = iRecVec.size();
  for (unsigned int mRec = 0; mRec < iRecVec.size(); ++mRec)
    if (iRecVec[mRec] <= 0) nRec--;

  // Check for failure to locate any recoiler
  if ( nRec <= 0 ) {
    infoPtr->errorMsg("Error in DireTimes::setupQCDdip: "
                      "failed to locate any recoiling partner");
    return;
  }

  // Store dipole colour end(s).
  for (unsigned int mRec = 0; mRec < iRecVec.size(); ++mRec) {
    iRec = iRecVec[mRec];
    if (iRec <= 0) continue;
    // Max scale either by parton scale or by half dipole mass.
    double pTmax = event[iRad].scale();
    if (limitPTmaxIn) {
      if (iSys == 0 || (iSys == 1 && doSecondHard)) pTmax *= pTmaxFudge;
      else if (sizeIn > 0) pTmax *= pTmaxFudgeMPI;
    } else pTmax = 0.5 * m( event[iRad], event[iRec]);
    int colType  = (event[iRad].id() == 21) ? 2 * colSign : colSign;
    int isrType  = (event[iRec].isFinal()) ? 0 : event[iRec].mother1();
    // This line in case mother is a rescattered parton.
    while (isrType > 2 + beamOffset) isrType = event[isrType].mother1();
    if (isrType > 2) isrType -= beamOffset;
    dipEnd.push_back( DireTimesEnd( iRad, iRec, pTmax,
      colType, 0, 0, 0, isrType, iSys, -1, -1, 0, isOctetOnium) );

    // If hooked up with other system then find which.
    if (otherSystemRec) {
      int systemRec = partonSystemsPtr->getSystemOf(iRec, true);
      if (systemRec >= 0) dipEnd.back().systemRec = systemRec;
      dipEnd.back().MEtype = 0;
    }

  }

}

//--------------------------------------------------------------------------

// Select next pT in downwards evolution of the existing dipoles.

double DireTimes::pTnext( Event& event, double pTbegAll, double pTendAll,
  bool, bool ) {

  // Begin loop over all possible radiating dipole ends.
  dipSel  = 0;
  iDipSel = -1;
  double pT2sel = pTendAll * pTendAll;
  splittingNowName="";
  splittingSelName="";

  for (int iDip = 0; iDip < int(dipEnd.size()); ++iDip) {
    DireTimesEnd& dip = dipEnd[iDip];

    // Limit final state multiplicity. For debugging only
    int nFinal = 0;
    for (int i=0; i < event.size(); ++i)
      if (event[i].isFinal()) nFinal++;
    if (nFinalMax > -10 && nFinal > nFinalMax) continue;

    // Dipole properties.
    dip.mRad   = event[dip.iRadiator].m();
    dip.mRec   = event[dip.iRecoiler].m();
    dip.mDip   =
      sqrt( abs(2. * event[dip.iRadiator].p() * event[dip.iRecoiler].p()) );
    dip.m2Rad  = pow2(dip.mRad);
    dip.m2Rec  = pow2(dip.mRec);
    dip.m2Dip  = pow2(dip.mDip);

    // Find maximum mass scale for dipole.
    dip.m2DipCorr    = dip.m2Dip;

//    double pTbegDip = min( pTbegAll, dip.pTmax );
//    double pT2begDip = min( pow2(pTbegDip), 0.25 * dip.m2DipCorr);

    double pTbegDip = pTbegAll;
    double pT2begDip = pow2(pTbegDip);
    dip.pT2 = 0.;

    // Do not try splitting if the corrected dipole mass is negative.
    if (dip.m2DipCorr < 0.) {
      infoPtr->errorMsg("Warning in DireTimes::pTnext: "
      "negative dipole mass.");
      continue;
    }

    // Do QCD evolution if it makes sense.
    if (pT2begDip > pT2sel) {

//      if (dip.colType != 0) pT2nextQCD(pT2begDip, pT2sel, dip, event);
      if (dip.colType != 0) pT2nextQCD(pT2begDip, pTendAll*pTendAll, dip, event);

      // Update if found pT larger than current maximum.
      if (dip.pT2 > pT2sel) {
        pT2sel  = dip.pT2;
        dipSel  = &dip;
        iDipSel = iDip;
        splittingSelName = splittingNowName;
      }
    }
  }

  // Update the number of proposed timelike emissions.
  if (dipSel != 0 && nProposedPT.find(dipSel->system) != nProposedPT.end())
    ++nProposedPT[dipSel->system];

  weights->insertWeights(acceptProbability, rejectProbability);
  rejectProbability.clear();
  acceptProbability.clear();

  // Return nonvanishing value if found pT bigger than already found.
  return (dipSel == 0) ? 0. : sqrt(pT2sel);

}

//--------------------------------------------------------------------------

double DireTimes::enhanceFactors( string, double, double ) {
 
  double factor = 1.;
  return factor;

}

//--------------------------------------------------------------------------

// Function to generate new user-defined overestimates to evolution.

void DireTimes::getNewOverestimates( int type, DireTimesEnd* dip,
  double pT2Old, double zMinAbs, double zMaxAbs,
  multimap<double,string>& newOverestimates) {

  double sum=0.;
  // Loop over splitting names and get overestimates.
  vector<Splitting* > splits = splittingsPtr->getSplittingNames();

  for (int i = 0; i < int(splits.size()); ++i) {

    int typeNow = splits[i]->type();
    if ( !(typeNow >= type && typeNow/100 == type/100) ) continue;
    if ( !(typeNow >= type && typeNow/1000 == type/1000) ) continue;
    string name          = splits[i]->name(); 

    double wt = splits[i]->overestimateInt(zMinAbs, zMaxAbs, pT2Old,
                                           dip->m2Dip);

    // Include artificial enhancements.
    wt *= enhanceFactors(name, dip->m2Dip, pT2Old);
 
    // Save this overestimate.
    // Do not include zeros (could lead to trouble with lower_bound?)
    if (wt != 0.) {
      sum += abs(wt);
      newOverestimates.insert(make_pair(sum,name));
    }
  }

}

//--------------------------------------------------------------------------

// Function to generate new user-defined overestimates to evolution.

void DireTimes::getNewSplitting( const Event& state, DireTimesEnd* dip,
  double pT2Old, double xOld, double pT2,
  double zMinAbs, double zMaxAbs, int idMother, string name,
  int& idDaughter, int& idSister, double& z, double& wt, double& full,
  double& over) {

  // Return auxiliary variable, mother and sister ids.
  z           = (*splittingsPtr)[name]->zSplit(zMinAbs, zMaxAbs, dip->m2Dip);
  idDaughter  = (*splittingsPtr)[name]->motherID(idMother);
  idSister    = (*splittingsPtr)[name]->sisterID(idMother);

  // Set flavours.
  if (idSister == 21 ) dip->flavour = 21;
  else dip->flavour = 0;
  dip->mFlavour = 0.;
  // Flavour choice for g -> q qbar.
  int sign = (idSister > 0) ? 1 : -1; 
  if (dip->flavour == 0) {
    idSister      = sign*min(5, 1 + int(nGluonToQuark * rndmPtr->flat()));
    //idSister      = sign*5;
    dip->flavour  = idSister;
    dip->mFlavour = sqrt(getMass(state[dip->iRadiator].id(),2));
    idDaughter    = -idSister;
  }

  // Get particle masses.
  // Radiator before splitting.
  double m2Bef = (state[dip->iRadiator].idAbs() < 6)
               ? getMass(state[dip->iRadiator].id(),2)
               : getMass(state[dip->iRadiator].id(),3,
                         state[dip->iRadiator].mCalc());
  // Radiator after splitting.
  double m2r = (abs(idDaughter) < 6) ? getMass(idDaughter,2)
             : getMass(idDaughter,1);
  // Recoiler. 
  double m2s = 0.0;
  int type   = (state[dip->iRecoiler].isFinal()) ? 1 : -1;
  if (type == 1) {
    m2s = (state[dip->iRecoiler].idAbs() < 6)
        ? getMass(state[dip->iRecoiler].id(),2)
        : getMass(state[dip->iRecoiler].id(),3,
                  state[dip->iRecoiler].mCalc());
  } 
  // Emission.
  double m2e = (abs(dip->flavour) < 6 ) ? getMass(dip->flavour,2)
             : getMass(dip->flavour,1);

  // Dipole invariant mass.
  double q2 = (state[dip->iRecoiler].p() +state[dip->iRadiator].p()).m2Calc();

  // For g-> qbar q, swap meaning of radiator an emission.
  if (dip->flavour < 0) swap(m2r,m2e);

  // Upate type if this is a massive splitting.
  if (type == 1 && (m2Bef > TINYMASS || m2r > TINYMASS || m2s > TINYMASS
    || m2e > TINYMASS) ) type = 2;
  if (type ==-1 && (m2Bef > TINYMASS || m2r > TINYMASS || m2s > TINYMASS
    || m2e > TINYMASS) ) type =-2;

  // Recalculate the kinematicaly available dipole mass.
  int massSign = (type > 0) ? 1 : -1;
  double Q2 = dip->m2Dip + massSign*(m2Bef - m2r - m2e);

  // Check phase space constraints.
  if ( !inAllowedPhasespace( z, pT2, Q2, q2, xOld, type, m2Bef, m2r, m2s,
        m2e) ) { 
    wt = full = over = 0.; return;
  }

  // Return overestimate.
  over        = (*splittingsPtr)[name]->overestimateDiff(z, dip->m2Dip);

  // Get complete kernel.
  full        = (*splittingsPtr)[name]->kernel( z, pT2, Q2, type, m2Bef, m2r,
                                                m2s, m2e);

  wt          = full/over;

  // Divide out artificial enhancements.
  double headRoom = enhanceFactors(name, dip->m2Dip, pT2Old);
  wt   /= headRoom;
  over *= headRoom;

  // Ensure positive weight.
  wt = abs(wt);

}

//--------------------------------------------------------------------------

// Check if variables are in allowed phase space.

bool DireTimes::inAllowedPhasespace( double z, double pT2, double m2dip, double q2,
  double xOld, int splitType, double m2RadBef, double m2r, double m2s,
  double m2e) {

  // splitType == 1 -> Massless FF
  if (splitType == 1) {

    // Calculate CS variables.
    double yCS = pT2/m2dip / (1.-z);
    double zCS = ( 1. - z - pT2/m2dip - pow2(1.-z) )
               / ( 1. - z - pT2/m2dip);

    // Forbidden emission if outside allowed z range for given pT2.
    if ( zCS < 0. || zCS > 1. || yCS < 0. || yCS > 1.) return false;

  // splitType == 2 -> Massive FF
  } else if (splitType == 2) {

    // Phase space limits - CS style.
    // Calculate CS variables.
    double yCS = pT2/m2dip / (1.-z);
    double zCS = ( 1. - z - pT2/m2dip - pow2(1.-z) )
               / ( 1. - z - pT2/m2dip);
    
    // Get yCS-boundaries.
    double mu2Rad = m2r/q2; 
    double mu2Emt = m2e/q2; 
    double mu2Rec = m2s/q2; 
    double yCSminMassive = 2.*sqrt(mu2Rad)*sqrt(mu2Emt)
                         / ( 1 - mu2Rad - mu2Emt - mu2Rec);   
    double yCSmaxMassive = 1.
                         - 2.*sqrt(mu2Rec)*( 1 - sqrt(mu2Rec) )
                         / ( 1 - mu2Rad - mu2Emt - mu2Rec);

    // Forbidden emission if outside allowed y range for given pT2.
    if ( yCS < yCSminMassive || yCS > yCSmaxMassive) return false;

    // Get zCS-boundaries.
    double nu2Rad = m2r/m2dip; 
    double nu2Emt = m2e/m2dip; 
    double nu2Rec = m2s/m2dip; 
    double vijk   = pow2(1.-yCS) - 4.*(yCS + nu2Rad + nu2Emt)*nu2Rec;
    double viji   = pow2(yCS) - 4.*nu2Rad*nu2Emt;
    if (vijk < 0. || viji < 0.) return false;
    vijk          = sqrt(vijk) / (1-yCS);
    viji          = sqrt(viji) / (yCS + 2.*nu2Rad);
    double prefac = (m2dip*yCS + 2.*m2r) / (2.*m2dip*yCS + 2.*m2r + 2.*m2e);
    double zCSminMassive = ( 1 - vijk*viji) * prefac;
    double zCSmaxMassive = ( 1 + vijk*viji) * prefac;
    
    // Forbidden emission if outside allowed z range for given pT2.
    if ( zCS < zCSminMassive || zCS > zCSmaxMassive) return false;

  // splitType ==-1 -> Massless FI
  } else if (splitType ==-1) {

    // Calculate CS variables.
    double kappa2 =  pT2/m2dip;
    double zCS = z;
    double xCS = 1 - kappa2/(1.-z);

    // Forbidden emission if outside allowed z range for given pT2.
    if ( zCS < 0. || zCS > 1. || xCS < xOld || xCS > 1.) return false;

  // splitType ==-2 -> Massive FI
  } else if (splitType ==-2) {

    // Phase space limits - CS style
    // Calculate CS variables.
    double kappa2 =  pT2/m2dip;
    double zCS = z;
    double xCS = 1 - kappa2/(1.-z);

    // Get xCS-boundaries.
    double xCDST = xCS / m2dip * (m2dip + m2RadBef-m2r-m2e);
    double pijpa_tilde = m2dip - m2r - m2e + m2RadBef;
    double pijpa     = pijpa_tilde/xCDST;
    double mu2RadBef = m2RadBef/pijpa;
    double muRad     = sqrt(m2r/pijpa);
    double muEmt     = sqrt(m2e/pijpa);
    double xCSmaxMassive = 1. + mu2RadBef - pow2(muRad + muEmt);

    // Forbidden emission if outside allowed x range for given pT2.
    if ( xCDST < xOld || xCDST > xCSmaxMassive) return false;

    // Get zCS-boundaries.
    double nu2Rad = m2r/m2dip; 
    double nu2Emt = m2e/m2dip; 
    double viji   = pow2(1.-xCS) - 4. * xCS*nu2Rad * xCS*nu2Emt; 
    if (viji < 0.) return false;
    viji   = sqrt( viji ) / (1.-xCS+2.*nu2Rad*xCS); 
    double vijk   = 1.; 
    double prefac = 0.5 * ( 1.-xCS + 2.*xCS*nu2Rad )
                  / ( 1.-xCS + xCS*nu2Rad + xCS*nu2Emt);
    double zCSminMassive = prefac * ( 1. - viji*vijk );
    double zCSmaxMassive = prefac * ( 1. + viji*vijk );

    // Forbidden emission if outside allowed z range for given pT2.
    if ( zCS < zCSminMassive || zCS > zCSmaxMassive) return false;

  }

  return true;

}

//--------------------------------------------------------------------------

// Function to add user-defined overestimates to old overestimate.

void DireTimes::addNewOverestimates( multimap<double,string> newOverestimates,
  double& oldOverestimate) {

  // No other tricks necessary at the moment.
  if (!newOverestimates.empty())
    oldOverestimate += newOverestimates.rbegin()->first;

  // Done.

}

//--------------------------------------------------------------------------

// Evolve a QCD dipole end.

void DireTimes::pT2nextQCD(double pT2begDip, double pT2sel,
  DireTimesEnd& dip, Event& event) {

  if (event[dip.iRecoiler].isFinal())
    pT2nextQCD_FF(pT2begDip, pT2sel, dip, event);
  else
    pT2nextQCD_FI(pT2begDip, pT2sel, dip, event);

  // Done.
  return;

}

//--------------------------------------------------------------------------

// Evolve a QCD dipole end.

void DireTimes::pT2nextQCD_FF(double pT2begDip, double pT2sel,
  DireTimesEnd& dip, Event& event) {

  // Lower cut for evolution. Return if no evolution range.
  double pT2endDip = max( pT2sel, pT2colCut );
  if (pT2begDip < pT2endDip) return;

  // Variables used inside evolution loop. (Mainly dummy start values.)
  dip.pT2              = pT2begDip;
  double zMinAbs       = 0.0;
  double pT2min        = pT2endDip;
  double b0            = 4.5;
  double Lambda2       = Lambda3flav2;
  double emitCoefTot   = 0.;
  double wt            = 0.;
  bool   mustFindRange = true;

  double xRecoiler = 0.;
  int idRadiator = event[dip.iRadiator].id();
  multimap<double,string> newOverestimates;
  int type = 1;
  if (event[dip.iRadiator].id() == 21) type += 200;
  else type += 100;

  double fullWeightNow = 0.;
  double overWeightNow = 0.;
  double auxWeightNow  = 0.;

  // Begin evolution loop towards smaller pT values.
  do {

    wt = 0.;

    splittingNowName="";

    // Update event weight after one step.  
    if ( fullWeightNow != 0. && overWeightNow != 0. ) {
      double w = auxWeightNow/overWeightNow
               * (overWeightNow-fullWeightNow)
              / (auxWeightNow - fullWeightNow);
      rejectProbability.insert( make_pair(dip.pT2,w));
    }
 
    fullWeightNow = overWeightNow = auxWeightNow = 0.;

    // Initialize evolution coefficients at the beginning and
    // reinitialize when crossing c and b flavour thresholds.
    if (mustFindRange) {

      newOverestimates.clear();
      emitCoefTot = 0.;

      // Determine overestimated z range; switch at c and b masses.
      if (dip.pT2 > m2b) {
        b0       = 23./6.;
        Lambda2  = Lambda5flav2;
      } else if (dip.pT2 > m2c) {
        b0       = 25./6.;
        Lambda2  = Lambda4flav2;
      } else {
        b0       = 27./6.;
        Lambda2  = Lambda3flav2;
      }
      // A change of renormalization scale expressed by a change of Lambda.
      Lambda2 /= renormMultFac;

      // Calculate and add user-defined overestimates.
      getNewOverestimates(type, &dip, dip.pT2, zMinAbs, 1.-zMinAbs,
        newOverestimates);
      addNewOverestimates(newOverestimates, emitCoefTot);

      // Initialization done for current range.
      mustFindRange = false;
    }

    double pT2Old = dip.pT2;

    // Fixed alpha_strong, reweighted later to PDF running alpha_s.
    if (usePDFalphas) {
      dip.pT2 = dip.pT2 * pow( rndmPtr->flat(),
          1. / (alphaS2piOverestimate * emitCoefTot));

    // Pick pT2 (in overestimated z range) for fixed alpha_strong.
    } else if (alphaSorder == 0) {
      dip.pT2 = dip.pT2 * pow( rndmPtr->flat(),
        1. / (alphaS2pi * emitCoefTot) );

    // Ditto for first-order alpha_strong.
    } else if (alphaSorder == 1) {
      dip.pT2 = Lambda2 * pow( dip.pT2 / Lambda2,
        pow( rndmPtr->flat(), b0 / emitCoefTot) );

      // For second order reject by second term in alpha_strong expression.
    } else {
      do dip.pT2 = Lambda2 * pow( dip.pT2 / Lambda2,
        pow( rndmPtr->flat(), b0 / emitCoefTot) );
      while (alphaS.alphaS2OrdCorr(renormMultFac * dip.pT2) < rndmPtr->flat()
        && dip.pT2 > pT2min);
    }

    // Abort evolution if below cutoff scale, or below another branching.
    if ( dip.pT2 < pT2endDip) { dip.pT2 = 0.; return; }

    // Try user-defined splittings first.
    double R = emitCoefTot*rndmPtr->flat();
    double z = 0.;
    int idDaughter, idSister;
    idDaughter = idSister = 0;
    if (!newOverestimates.empty()) {

      if (newOverestimates.lower_bound(R) == newOverestimates.end())
        splittingNowName = newOverestimates.rbegin()->second;
      else
        splittingNowName = newOverestimates.lower_bound(R)->second;
      getNewSplitting( event, &dip, pT2Old, xRecoiler, dip.pT2, zMinAbs,
        1-zMinAbs, idRadiator, splittingNowName, idDaughter, idSister, z, wt,
        fullWeightNow, overWeightNow);
      dip.z = z;

    }

    // Done for vanishing accept probability.
    if (wt == 0. ) {
      mustFindRange = true;
      wt = fullWeightNow = overWeightNow = auxWeightNow = 0.;
      continue;
    }

    // Get particle masses.
    // Radiator before splitting.
    double m2Bef = (event[dip.iRadiator].idAbs() < 6)
                 ? getMass(event[dip.iRadiator].id(),2)
                 : getMass(event[dip.iRadiator].id(),3,
                           event[dip.iRadiator].mCalc());
    // Radiator after splitting.
    double m2r = (abs(idDaughter) < 6) ? getMass(idDaughter,2)
               : getMass(idDaughter,1);
    // Recoiler. 
    double m2s = (event[dip.iRecoiler].idAbs() < 6)
          ? getMass(event[dip.iRecoiler].id(),2)
          : getMass(event[dip.iRecoiler].id(),3,
                    event[dip.iRecoiler].mCalc());
    // Emission.
    double m2e = (abs(dip.flavour) < 6 ) ? getMass(dip.flavour,2)
               : getMass(dip.flavour,1);

    // For g-> qbar q, swap meaning of radiator an emission.
    if (dip.flavour < 0) swap(m2r,m2e);

    // Recalculate the kinematicaly available dipole mass.
    double Q2 = dip.m2Dip + m2Bef - m2r - m2e;
    // Calculate CS variables and scaled masses.
    double yCS = dip.pT2/Q2 / (1.-dip.z);
    double q2 = (event[dip.iRadiator].p() +event[dip.iRecoiler].p()).m2Calc();
    double mu2RadBef = m2Bef/ q2;
    double mu2Rad    = m2r/ q2;
    double mu2Rec    = m2s/ q2;
    double mu2Emt    = m2e/ q2;
    // Calculate Jacobian.
    double jac1 = ( 1. - mu2Rad - mu2Rec - mu2Emt)
                / sqrt(lABC(1.,mu2RadBef,mu2Rec));
    double jac2 = 1. + ( mu2Rad + mu2Emt - mu2RadBef)
                      /( yCS*(1. - mu2Rad - mu2Rec - mu2Emt));

    // Multiply with Jacobian.
    wt            *= jac1/jac2;
    fullWeightNow *= jac1/jac2;

    // Reweight to PDF alpha_s, if necessary.
    BeamParticle& beam = (abs(beamAPtr->id()) == 2212) ? *beamAPtr : *beamBPtr;
    wt            *= (usePDFalphas)
                   ? beam.alphaS(dip.pT2)/(alphaS2piOverestimate*2.*M_PI)
                   : 1.; 
    fullWeightNow *= (usePDFalphas)
                   ? beam.alphaS(dip.pT2)/(2.*M_PI)
                   : (alphaSorder==0) ? alphaS2pi
                                      : alphaS.alphaS(dip.pT2)/(2.*M_PI); 
    overWeightNow *= (usePDFalphas)
                   ? alphaS2piOverestimate
                   : (alphaSorder==0) ? alphaS2pi
                                      : alphaS.alphaS(dip.pT2)/(2.*M_PI);
    auxWeightNow   = overWeightNow;

    // Ensure that accept probability is positive.
    if (fullWeightNow < 0.) auxWeightNow *= -1.;

    if (wt == 0. ) {
      mustFindRange = true;
      wt = fullWeightNow = overWeightNow = auxWeightNow = 0.;
      continue;
    }

    // Reset overestimate if necessary.
    if ( fullWeightNow/auxWeightNow > 1.) {
      mustFindRange = true;
      double rescale = fullWeightNow/auxWeightNow * 1.5;
      auxWeightNow *= rescale;
      wt /= rescale;
      infoPtr->errorMsg("Info in DireTimes::pT2nextQCD_FF: Found large "
                        "acceptance weight for " + splittingNowName);
    }

  // Iterate until acceptable pT (or have fallen below pTmin).
  } while (wt < rndmPtr->flat());

  // Update accepted event weight.
  if ( fullWeightNow != 0. && overWeightNow != 0. ) {
    acceptProbability.insert( make_pair(dip.pT2,auxWeightNow/overWeightNow));
    double w  = auxWeightNow/overWeightNow
              * (overWeightNow- fullWeightNow)
              / (auxWeightNow - fullWeightNow);
    rejectProbability.insert( make_pair(dip.pT2, w));
  }

}

//--------------------------------------------------------------------------

// Evolve a QCD dipole end.

void DireTimes::pT2nextQCD_FI(double pT2begDip, double pT2sel,
  DireTimesEnd& dip, Event& event) {

  // Lower cut for evolution. Return if no evolution range.
  double pT2endDip = max( pT2sel, pT2colCut );
  if (pT2begDip < pT2endDip) return;

  BeamParticle& beam = (dip.isrType == 1) ? *beamAPtr : *beamBPtr;
  double m2cPhys = (usePDFalphas) ? pow2(max(0.,beam.mQuarkPDF(4)))
                 : alphaS.muThres2(4);
  double m2bPhys = (usePDFalphas) ? pow2(max(0.,beam.mQuarkPDF(5)))
                 : alphaS.muThres2(5);

  // Variables used inside evolution loop. (Mainly dummy start values.)
  dip.pT2              = pT2begDip;
  int    nFlavour      = 3;
  double zMinAbs       = 0.5;
  double pT2min        = pT2endDip;
  double b0            = 4.5;
  double Lambda2       = Lambda3flav2;
  double emitCoefTot   = 0.;
  double wt            = 0.;
  bool   mustFindRange = true;

  int idRadiator = event[dip.iRadiator].id();

  int iSysRec = dip.systemRec;
  double xRecoiler = beam[iSysRec].x();

  multimap<double,string> newOverestimates;
  int type = 1;
  if (event[dip.iRadiator].id() == 21) type += 200;
  else type += 100;

  double fullWeightNow = 0.;
  double overWeightNow = 0.;
  double auxWeightNow = 0.;

  // Begin evolution loop towards smaller pT values.
  do {

    wt = 0.;
    splittingNowName="";

    // Update event weight after one step.  
    if ( fullWeightNow != 0. && overWeightNow != 0. ) {
      double w = auxWeightNow/overWeightNow
               * (overWeightNow-fullWeightNow)
               / (auxWeightNow - fullWeightNow);
      rejectProbability.insert( make_pair(dip.pT2,w));
    }
 
    fullWeightNow = overWeightNow = auxWeightNow = 0.;

    // Leave unconverted for now.
    if (event[dip.iRecoiler].idAbs() == 4 && dip.pT2 <= m2cPhys)
      { dip.pT2 = 0.; return; }
    if (event[dip.iRecoiler].idAbs() == 5 && dip.pT2 <= m2bPhys)
      { dip.pT2 = 0.; return; }

    // Initialize evolution coefficients at the beginning and
    // reinitialize when crossing c and b flavour thresholds.

    //mustFindRange = true;

    if (mustFindRange) {

      newOverestimates.clear();
      emitCoefTot = 0.;

      // Determine overestimated z range; switch at c and b masses.
      if (dip.pT2 > m2b) {
        nFlavour = 5;
        b0       = 23./6.;
        Lambda2  = Lambda5flav2;
      } else if (dip.pT2 > m2c) {
        nFlavour = 4;
        b0       = 25./6.;
        Lambda2  = Lambda4flav2;
      } else {
        nFlavour = 3;
        b0       = 27./6.;
        Lambda2  = Lambda3flav2;
      }
      // A change of renormalization scale expressed by a change of Lambda.
      Lambda2 /= renormMultFac;

      zMinAbs = 0;

      // Calculate and add user-defined overestimates.
      getNewOverestimates(type, &dip, dip.pT2, zMinAbs, 1.-zMinAbs,
        newOverestimates);
      addNewOverestimates(newOverestimates, emitCoefTot);

      // Initialization done for current range.
      mustFindRange = false;
    }

    double pT2Old = dip.pT2;

    // Fixed alpha_strong, reweighted later to PDF running alpha_s.
    if (usePDFalphas) {
      dip.pT2 = dip.pT2 * pow( rndmPtr->flat(),
          1. / (alphaS2piOverestimate * emitCoefTot));

    // Pick pT2 (in overestimated z range) for fixed alpha_strong.
    } else if (alphaSorder == 0) {
      dip.pT2 = dip.pT2 * pow( rndmPtr->flat(),
        1. / (alphaS2pi * emitCoefTot) );

    // Ditto for first-order alpha_strong.
    } else if (alphaSorder == 1) {
      dip.pT2 = Lambda2 * pow( dip.pT2 / Lambda2,
        pow( rndmPtr->flat(), b0 / emitCoefTot) );

      // For second order reject by second term in alpha_strong expression.
    } else {
      do dip.pT2 = Lambda2 * pow( dip.pT2 / Lambda2,
        pow( rndmPtr->flat(), b0 / emitCoefTot) );
      while (alphaS.alphaS2OrdCorr(renormMultFac * dip.pT2) < rndmPtr->flat()
        && dip.pT2 > pT2min);
    }
    wt = 0.;

    // If crossed c or b thresholds: continue evolution from threshold.
    if (nFlavour == 5 && dip.pT2 < m2b) {
      mustFindRange = true;
    } else if ( nFlavour == 4 && dip.pT2 < m2c) {
      mustFindRange = true;
    }

    // If crossed b threshold, continue evolution from this threshold.
    if (nFlavour == 5 && dip.pT2 < m2bPhys) {
      mustFindRange = true;
    // If crossed c threshold, continue evolution from this threshold.
    } else if (nFlavour == 4 && dip.pT2 < m2cPhys) {
      mustFindRange = true;
    }

    // Leave incoming heavy quarks below threshold unconverted for now.
    if (event[dip.iRecoiler].idAbs() == 5 && dip.pT2 <= m2bPhys) 
      { dip.pT2 = 0.; return; }
    if (event[dip.iRecoiler].idAbs() == 4 && dip.pT2 <= m2cPhys)
      { dip.pT2 = 0.; return; }

    // Abort evolution if below cutoff scale, or below another branching.
    if ( dip.pT2 < pT2endDip) { dip.pT2 = 0.; return; }

    // Try user-defined splittings first.
    double R = emitCoefTot*rndmPtr->flat();
    double z = 0.;
    int idDaughter, idSister;
    idDaughter = idSister = 0;
    if (!newOverestimates.empty()) {

      if (newOverestimates.lower_bound(R) == newOverestimates.end())
        splittingNowName = newOverestimates.rbegin()->second;
      else
        splittingNowName = newOverestimates.lower_bound(R)->second;
      getNewSplitting( event, &dip, pT2Old, xRecoiler, dip.pT2, zMinAbs,
        1-zMinAbs, idRadiator, splittingNowName, idDaughter, idSister, z, wt,
        fullWeightNow, overWeightNow);
      dip.z = z;

    }

    // Done for vanishing accept probability.
    if (wt == 0. ) {
      mustFindRange = true;
      wt = fullWeightNow = overWeightNow = auxWeightNow = 0.;
      continue;
    }

    // Get particle masses.
    // Radiator before splitting.
    double m2Bef = (event[dip.iRadiator].idAbs() < 6)
                 ? getMass(event[dip.iRadiator].id(),2)
                 : getMass(event[dip.iRadiator].id(),3,
                           event[dip.iRadiator].mCalc());
    // Radiator after splitting.
    double m2r = (abs(idDaughter) < 6) ? getMass(idDaughter,2)
               : getMass(idDaughter,1);
    // Emission.
    double m2e = (abs(dip.flavour) < 6 ) ? getMass(dip.flavour,2)
               : getMass(dip.flavour,1);

    //m2Bef = m2r = m2e = 0.;

    // For g-> qbar q, swap meaning of radiator an emission.
    if (dip.flavour < 0) swap(m2r,m2e);

    // Recalculate the kinematicaly available dipole mass.
    double Q2 = dip.m2Dip - m2Bef + m2r + m2e;
    // Calculate CS variables.
    double kappa2 =  dip.pT2/Q2;
    double xCS = 1 - kappa2/(1.-z);
    double xOld = beam[iSysRec].x();
    double xCDST = xCS*( 1. + (m2Bef-m2r-m2e)/Q2 );
    double xNew  = xOld / xCDST;

    // Firstly reduce by PDF ratio.
    double pdfRatio = 1.;
    pdfScale2 = (useFixedFacScale) ? fixedFacScale2
      : factorMultFac * dip.pT2;
    bool inD  = beam.insideBounds(xOld, pdfScale2); 
    bool inM  = beam.insideBounds(xNew, pdfScale2); 
    int idRec = event[dip.iRecoiler].id();
    double pdfOld = 
      (useSummedPDF) ? beam.xf(idRec, xOld, pdfScale2)
                     : beam.xfISR( iSysRec, idRec, xOld, pdfScale2);
    double pdfNew =
      (useSummedPDF) ? beam.xf(idRec, xNew, pdfScale2)
                     : beam.xfISR( iSysRec, idRec, xNew, pdfScale2);

    if ( abs(pdfOld) < TINYPDF ) {
      mustFindRange = true;
      wt = fullWeightNow = overWeightNow = auxWeightNow = 0.;
      continue;
    }

    // Set PDF ratio to zero if x-/pT-values are out-of-bounds.
    pdfRatio = (inD && inM) ? xCDST*pdfNew/pdfOld : 0.;

    wt             *= pdfRatio;
    fullWeightNow  *= pdfRatio;

    // Reweight to PDF alpha_s, if necessary.
    wt            *= (usePDFalphas)
                   ? beam.alphaS(dip.pT2)/(alphaS2piOverestimate*2.*M_PI)
                   : 1.; 
    fullWeightNow *= (usePDFalphas)
                   ? beam.alphaS(dip.pT2)/(2.*M_PI)
                   : (alphaSorder==0) ? alphaS2pi
                                      : alphaS.alphaS(dip.pT2)/(2.*M_PI); 
    overWeightNow *= (usePDFalphas)
                   ? alphaS2piOverestimate
                   : (alphaSorder==0) ? alphaS2pi
                                      : alphaS.alphaS(dip.pT2)/(2.*M_PI);

    auxWeightNow   = overWeightNow;

    // Ensure that accept probability is positive.
    if (fullWeightNow < 0.) auxWeightNow *= -1.;

    if (wt == 0. ) {
      mustFindRange = true;
      wt = fullWeightNow = overWeightNow = auxWeightNow = 0.;
      continue;
    }

    // Reset overestimate if necessary.
    if ( fullWeightNow/auxWeightNow > 1.) {
      mustFindRange = true;
      double rescale = fullWeightNow/auxWeightNow * 1.5;
      auxWeightNow *= rescale;
      wt /= rescale;
      infoPtr->errorMsg("Info in DireTimes::pT2nextQCD_FI: Found large "
                        "acceptance weight for " + splittingNowName);
    }

  // Iterate until acceptable pT (or have fallen below pTmin).
  } while (wt < rndmPtr->flat());

  // Update accepted event weight.
  if ( fullWeightNow != 0. && overWeightNow != 0. ) {
    acceptProbability.insert( make_pair(dip.pT2,auxWeightNow/overWeightNow));
    double w = auxWeightNow/overWeightNow
             * (overWeightNow- fullWeightNow)
             / (auxWeightNow - fullWeightNow);
    rejectProbability.insert( make_pair(dip.pT2, w));
  }

}

//--------------------------------------------------------------------------

bool DireTimes::branch( Event& event, bool ) {

  if ( event[dipSel->iRecoiler].isFinal())
    return branch_FF(event,false);
  else
    return branch_FI(event,false);

  return false;

}

//--------------------------------------------------------------------------

// ME corrections and kinematics that may give failure.
// Notation: radBef, recBef = radiator, recoiler before emission,
//           rad, rec, emt = radiator, recoiler, emitted efter emission.
//           (rad, emt distinguished by colour flow for g -> q qbar.)

bool DireTimes::branch_FF( Event& event, bool ) {

  // Check if the first emission should be studied for removal.
  bool canMergeFirst = (mergingHooksPtr != 0)
                     ? mergingHooksPtr->canVetoEmission() : false;

  // Find initial radiator and recoiler particles in dipole branching.
  int iRadBef      = dipSel->iRadiator;
  int iRecBef      = dipSel->iRecoiler;
  Particle& radBef = event[iRadBef];
  Particle& recBef = event[iRecBef];

  // Find their momenta, with special sum for global recoil.
  Vec4 pRadBef     = event[iRadBef].p();
  Vec4 pRecBef     = event[iRecBef].p();

  double pT2    = dipSel->pT2;
  double m2Dip  = dipSel->m2Dip;
  double z      = dipSel->z;
  // Calculate CS variables.
  double yCS    = pT2/m2Dip / (1.-z);
  double zCS    = ( 1. - z - pT2/m2Dip - pow2(1.-z) )
                / ( 1. - z - pT2/m2Dip);

  // Default flavours and colour tags for new particles in dipole branching.
  int idRad        = radBef.id();
  int idEmt        = abs(dipSel->flavour);
  int colRad       = radBef.col();
  int acolRad      = radBef.acol();
  int colEmt       = 0;
  int acolEmt      = 0;
  iSysSel          = dipSel->system;
  int iSysSelRec   = dipSel->systemRec;

  // New colour tag required for gluon emission.
  if (dipSel->flavour == 21 && dipSel->colType > 0) {
    colEmt  = colRad;
    colRad  = event.nextColTag();
    acolEmt = colRad;
  } else if (dipSel->flavour == 21) {
    acolEmt = acolRad;
    acolRad = event.nextColTag();
    colEmt  = acolRad;
  // New flavours for g -> q qbar; split colours.
  } else if (dipSel->colType > 0) {
    idEmt   = abs(dipSel->flavour);
    idRad   = -idEmt;
    colEmt  = colRad;
    colRad  = 0;
  } else if (dipSel->colType < 0) {
    idEmt   = -abs(dipSel->flavour);
    idRad   = -idEmt;
    acolEmt  = acolRad;
    acolRad  = 0;
  }

  // Get particle masses.
  // Radiator before splitting.
  double m2Bef = (event[dipSel->iRadiator].idAbs() < 6)
               ? getMass(event[dipSel->iRadiator].id(),2)
               : getMass(event[dipSel->iRadiator].id(),3,event[dipSel->iRadiator].mCalc());
  // Radiator after splitting.
  double m2r = (abs(idRad) < 6) ? getMass(idRad,2) : getMass(idRad,1);
  // Recoiler. 
  double m2s = (event[dipSel->iRecoiler].idAbs() < 6)
        ? getMass(event[dipSel->iRecoiler].id(),2)
        : getMass(event[dipSel->iRecoiler].id(),3,event[dipSel->iRecoiler].mCalc());
  // Emission.
  double m2e = (abs(idEmt) < 6 ) ? getMass(idEmt,2) : getMass(idEmt,1);

  //m2Bef = m2r = m2s = m2e = 0.;

  // For g-> qbar q, swap meaning of radiator an emission.
  if (dipSel->flavour < 0) swap(m2r,m2e);

  // Adjust the dipole kinematical mass to accomodate masses after branching.
  double Q2  = dipSel->m2Dip + m2Bef - m2r - m2e;

  // Calculate CS variables.
  double kappa2 = pT2/Q2;
  yCS    = kappa2 / (1.-z);
  zCS    = ( 1. - z - kappa2 - pow2(1.-z) ) / ( 1. - z - kappa2);

  // Get dipole 4-momentum.
  Vec4 q(pRadBef + pRecBef);
  double q2 = q.m2Calc();

  // Calculate derived variables.
  double sij  = yCS * (q2 - m2s) + (1.-yCS)*(m2r+m2e);
  double zbar = (q2-sij-m2s) / bABC(q2,sij,m2s)
              * (zCS - m2s/gABC(q2,sij,m2s)
                      *(sij + m2r - m2e)/(q2-sij-m2s));
  double kT2  = zbar*(1.-zbar)*sij - (1.-zbar)*m2r - zbar*m2e; 
  
  // Now construct the new recoiler momentum in the lab frame.
  Vec4 pRec( (pRecBef - (q*pRecBef)/q2 * q)
            * sqrt(lABC(q2,sij,m2s)/lABC(q2,m2Bef,m2s))
            + q * (q2+m2s-sij)/(2.*q2) );

  // Now construct the transverse momentum vector in the dipole CM
  // frame.
  double RNphi = rndmPtr->flat();
  double phi_kt = 2.*M_PI*RNphi;
  Vec4 kTmom(sqrt(kT2)*cos(phi_kt), sqrt(kT2)*sin(phi_kt),0.,0.);

  // Boost the transverse momentum vector into the lab frame.
  RotBstMatrix fromDipCM2;
  fromDipCM2.fromCMframe(pRadBef, pRecBef);
  kTmom.rotbst(fromDipCM2);

  // Construct left-over dipole momentum by momentum conservation.
  Vec4 pij(q-pRec);

  // Construct new radiator momentum.
  Vec4 pRad( zbar * (gABC(q2,sij,m2s)*pij - sij*pRec) / bABC(q2,sij,m2s)
            + (m2r+kT2) / (zbar*bABC(q2,sij,m2s))
              * (pRec - m2s/gABC(q2,sij,m2s)*pij)
            + kTmom);

  // Contruct the emission momentum by momentum conservation.
  Vec4 pEmt(q-pRad-pRec);

  // Ensure that radiator is on mass-shell
  double mTolErr = settingsPtr->parm("Check:mTolErr");
  double errMass = abs(pRad.mCalc() - sqrt(m2r)) / max( 1.0, pRad.e());
  if ( errMass > mTolErr*1e-2 ) {
    double deltam2 = pRad.m2Calc() - m2r;
    pRad.e(sqrtpos(pow2(pRad.e()) - deltam2));
  }
  // Ensure that emission is on mass-shell
  errMass = abs(pEmt.mCalc() - sqrt(m2e)) / max( 1.0, pEmt.e());
  if ( errMass > mTolErr*1e-2 ) {
    double deltam2 = pEmt.m2Calc() - m2e;
    pEmt.e(sqrtpos(pow2(pEmt.e()) - deltam2));
  }
  // Ensure that recoiler is on mass-shell
  errMass = abs(pRec.mCalc() - sqrt(m2s)) / max( 1.0, pRec.e());
  if ( errMass > mTolErr*1e-2 ) {
    double deltam2 = pRec.m2Calc() - m2s;
    pRec.e(sqrtpos(pow2(pRec.e()) - deltam2));
  }

  // For g-> qbar q, swap z->1-z, i.e. swam momenta of emission and radiator.
  if (dipSel->flavour < 0) swap(pRad,pEmt);

  // Define new particles from dipole branching.
  double pTsel = sqrt(dipSel->pT2);
  Particle rad = Particle(idRad, 51, iRadBef, 0, 0, 0,
    colRad, acolRad, pRad, sqrt(m2r), pTsel);
  Particle emt = Particle(idEmt, 51, iRadBef, 0, 0, 0,
    colEmt, acolEmt, pEmt, sqrt(m2e), pTsel);
  Particle rec = Particle(recBef.id(), 52, iRecBef, iRecBef, 0, 0,
    recBef.col(), recBef.acol(), pRec, sqrt(m2s), pTsel);

  // Special checks to set weak particles status equal to 56.
  // This is needed for decaying the particles. Also set polarisation.
  if (emt.idAbs() == 23 || emt.idAbs() == 24) {
    emt.status(56);
    event[iRadBef].pol( dipSel->weakPol );
    rad.pol( dipSel->weakPol );
  }

  // Save properties to be restored in case of user-hook veto of emission.
  int eventSizeOld = event.size();
  int iRadStatusV  = event[iRadBef].status();
  int iRadDau1V    = event[iRadBef].daughter1();
  int iRadDau2V    = event[iRadBef].daughter2();
  int iRecStatusV  = event[iRecBef].status();
  int iRecDau1V    = event[iRecBef].daughter1();
  int iRecDau2V    = event[iRecBef].daughter2();

  // Shower may occur at a displaced vertex.
  if (radBef.hasVertex()) {
    rad.vProd( radBef.vProd() );
    emt.vProd( radBef.vProd() );
  }
  if (recBef.hasVertex()) rec.vProd( recBef.vProd() );

  // Put new particles into the event record.
  // Mark original dipole partons as branched and set daughters/mothers.
  int iRad = event.append(rad);
  int iEmt = event.append(emt);
  event[iRadBef].statusNeg();
  event[iRadBef].daughters( iRad, iEmt);
  int iRec = event.append(rec);
  event[iRecBef].statusNeg();
  event[iRecBef].daughters( iRec, iRec);

  // Allow veto of branching. If so restore event record to before emission.
  bool inResonance = (partonSystemsPtr->getInA(iSysSel) == 0) ? true : false;
  if ( (canVetoEmission && userHooksPtr->doVetoFSREmission( eventSizeOld,
    event, iSysSel, inResonance))
    || (canMergeFirst && mergingHooksPtr->doVetoEmission( event )) ) {
    event.popBack( event.size() - eventSizeOld);
    event[iRadBef].status( iRadStatusV);
    event[iRadBef].daughters( iRadDau1V, iRadDau2V);
    event[iRecBef].status( iRecStatusV);
    event[iRecBef].daughters( iRecDau1V, iRecDau2V);
    return false;
  }

  // Gluon emission: update both dipole ends and add two new ones.
  if (dipSel->flavour == 21) {
    dipSel->iRadiator  = iRad;
    dipSel->iRecoiler  = iEmt;
    dipSel->systemRec  = iSysSel;
    dipSel->isrType    = 0;
    dipSel->pTmax      = pTsel;
    dipSel->MEtype = 0;
    for (int i = 0; i < int(dipEnd.size()); ++i) {
      if (dipEnd[i].iRadiator == iRecBef && dipEnd[i].iRecoiler == iRadBef
        && dipEnd[i].colType != 0) {
        dipEnd[i].iRadiator = iRec;
        dipEnd[i].iRecoiler = iEmt;
        dipEnd[i].MEtype = 0;
        // Strive to match colour to anticolour inside closed system.
        if ( dipEnd[i].colType * dipSel->colType > 0)
          dipEnd[i].iRecoiler = iRad;
        dipEnd[i].pTmax = pTsel;
      }
    }
    int colType = (dipSel->colType > 0) ? 2 : -2 ;
    // When recoiler was uncoloured particle, in resonance decays,
    // assign recoil to coloured particle.
    int iRecMod = iRec;
    if (recoilToColoured && inResonance && event[iRec].col() == 0
      && event[iRec].acol() == 0) iRecMod = iRad;
    dipEnd.push_back( DireTimesEnd(iEmt, iRecMod, pTsel,
       colType, 0, 0, 0, 0, iSysSel, 0));
    dipEnd.back().systemRec = iSysSelRec;
    dipEnd.push_back( DireTimesEnd(iEmt, iRad, pTsel,
      -colType, 0, 0, 0, 0, iSysSel, 0));

  // Gluon branching to q qbar: update current dipole and other of gluon.
  } else if (dipSel->colType != 0) {
    for (int i = 0; i < int(dipEnd.size()); ++i) {
      // Strive to match colour to anticolour inside closed system.
      if ( dipEnd[i].iRecoiler == iRadBef
        && dipEnd[i].colType * dipSel->colType < 0 )
        dipEnd[i].iRecoiler = iEmt;
      if (dipEnd[i].iRadiator == iRadBef && abs(dipEnd[i].colType) == 2) {
        dipEnd[i].colType /= 2;
        if (dipEnd[i].system != dipEnd[i].systemRec) continue;

        dipEnd[i].MEtype = 0;
        if (&dipEnd[i] == dipSel) dipEnd[i].iMEpartner = iRad;
        else                      dipEnd[i].iMEpartner = iEmt;
      }
    }

    dipSel->iRadiator = iEmt;
    dipSel->iRecoiler = iRec;
    dipSel->pTmax     = pTsel;

  }

  // Copy or set lifetime for new final state.
  if (event[iRad].id() == event[iRadBef].id()) {
    event[iRad].tau( event[iRadBef].tau() );
  } else {
    event[iRad].tau( event[iRad].tau0() * rndmPtr->exp() );
    event[iEmt].tau( event[iEmt].tau0() * rndmPtr->exp() );
  }
  event[iRec].tau( event[iRecBef].tau() );

  // Now update other dipoles that also involved the radiator or recoiler.
  for (int i = 0; i < int(dipEnd.size()); ++i) {
    if (dipEnd[i].iRadiator  == iRadBef) dipEnd[i].iRadiator  = iRad;
    if (dipEnd[i].iRecoiler  == iRadBef) dipEnd[i].iRecoiler  = iRad;
    if (dipEnd[i].iMEpartner == iRadBef) dipEnd[i].iMEpartner = iRad;
    if (dipEnd[i].iRadiator  == iRecBef) dipEnd[i].iRadiator  = iRec;
    if (dipEnd[i].iRecoiler  == iRecBef) dipEnd[i].iRecoiler  = iRec;
    if (dipEnd[i].iMEpartner == iRecBef) dipEnd[i].iMEpartner = iRec;
  }

  // Finally update the list of all partons in all systems.
  partonSystemsPtr->replace(iSysSel, iRadBef, iRad);
  partonSystemsPtr->addOut(iSysSel, iEmt);
  partonSystemsPtr->replace(iSysSelRec, iRecBef, iRec);

  // Ungraceful exit for incorrect event.
  bool isHadronic = false;
  for (int i = 0; i < event.size(); ++i)
    if (event[i].statusAbs() > 60) isHadronic = true;
  if ( !isHadronic && !validEvent(event)) {
    infoPtr->errorMsg("Error in DireTimes::branch_FF: State after "
                      "branching not valid, thus reject.");
    return false;
  }

  // Set shower weight.
  weights->calcWeight(pT2);
  weights->reset();
  rejectProbability.clear();
  acceptProbability.clear();

  // Done.
  return true;

}

//--------------------------------------------------------------------------

// ME corrections and kinematics that may give failure.
// Notation: radBef, recBef = radiator, recoiler before emission,
//           rad, rec, emt = radiator, recoiler, emitted efter emission.
//           (rad, emt distinguished by colour flow for g -> q qbar.)

bool DireTimes::branch_FI( Event& event, bool ) {

  // Check if the first emission should be studied for removal.
  bool canMergeFirst = (mergingHooksPtr != 0)
                     ? mergingHooksPtr->canVetoEmission() : false;

  // Find initial radiator and recoiler particles in dipole branching.
  int iRadBef      = dipSel->iRadiator;
  int iRecBef      = dipSel->iRecoiler;
  Particle& radBef = event[iRadBef];
  Particle& recBef = event[iRecBef];

  // Find their momenta, with special sum for global recoil.
  Vec4 pRadBef     = event[iRadBef].p();
  Vec4 pRecBef     = event[iRecBef].p();

  double pT2    = dipSel->pT2;
  double m2Dip  = dipSel->m2Dip;
  double z      = dipSel->z;

  // Calculate CS variables.
  double kappa2 =  pT2/m2Dip;
  double zCS = z;
  double xCS = 1 - kappa2/(1.-z);

  int side      = (pRecBef.pz() > 0.) ? 1 : -1;
  int isrType   = dipSel->isrType;

  // Default flavours and colour tags for new particles in dipole branching.
  int idRad        = radBef.id();
  int idEmt        = abs(dipSel->flavour);
  int colRad       = radBef.col();
  int acolRad      = radBef.acol();
  int colEmt       = 0;
  int acolEmt      = 0;
  iSysSel          = dipSel->system;
  int iSysSelRec   = dipSel->systemRec;

  // New colour tag required for gluon emission.
  if (dipSel->flavour == 21 && dipSel->colType > 0) {
    colEmt  = colRad;
    colRad  = event.nextColTag();
    acolEmt = colRad;
  } else if (dipSel->flavour == 21) {
    acolEmt = acolRad;
    acolRad = event.nextColTag();
    colEmt  = acolRad;
  // New flavours for g -> q qbar; split colours.
  } else if (dipSel->colType > 0) {
    idEmt   = abs(dipSel->flavour);
    idRad   = -idEmt;
    colEmt  = colRad;
    colRad  = 0;
  } else if (dipSel->colType < 0) {
    idEmt   = -abs(dipSel->flavour);
    idRad   = -idEmt;
    acolEmt = acolRad;
    acolRad = 0;
  }

  // Get particle masses.
  // Radiator before splitting.
  double m2Bef = (event[dipSel->iRadiator].idAbs() < 6)
               ? getMass(event[dipSel->iRadiator].id(),2)
               : getMass(event[dipSel->iRadiator].id(),3,
                         event[dipSel->iRadiator].mCalc());
  // Radiator after splitting.
  double m2r = (abs(idRad) < 6) ? getMass(idRad,2) : getMass(idRad,1);
  // Emission.
  double m2e = (abs(idEmt) < 6) ? getMass(idEmt,2) : getMass(idEmt,1);
  // Initial state recoiler always assumed massless. 
  double m2s = 0.0;

  //m2Bef = m2r = m2e = m2s = 0.;

  double mRad = sqrt(m2r);
  double mEmt = sqrt(m2e);

  // For g-> qbar q, swap meaning of radiator an emission.
  if (dipSel->flavour < 0) swap(m2r,m2e);

  // Recalculate the kinematicaly available dipole mass.
  double Q2 = dipSel->m2Dip - m2Bef + m2r + m2e;

  // Calculate CS variables.
  kappa2 =  pT2/Q2;
  xCS    = 1 - kappa2/(1.-z);

  // Get dipole 4-momentum.
  Vec4 q(pRecBef-pRadBef);
  double q2 = q.m2Calc();
  Vec4 qpar(q.px()+pRadBef.px(), q.py()+pRadBef.py(), q.pz(), q.e());
  double qpar2 = qpar.m2Calc();
  double pT2ijt = pow2(pRadBef.px()) + pow2(pRadBef.py());

  // Calculate derived variables.
  double sij  = (1.-1./xCS) * (q2 - m2s) + (m2r+m2e) / xCS;
  double zbar = (q2-sij-m2s) / bABC(q2,sij,m2s)
              * (zCS - m2s/gABC(q2,sij,m2s)
                      *(sij + m2r - m2e)/(q2-sij-m2s));
  double kT2  = zbar*(1.-zbar)*sij - (1.-zbar)*m2r - zbar*m2e; 

  // Now construct the new recoiler momentum in the lab frame.
  Vec4 pRec( (pRecBef - (qpar*pRecBef)/qpar2 * qpar)
            * sqrt( (lABC(q2,sij,m2s)   - 4.*m2s*pT2ijt)
                   /(lABC(q2,m2Bef,m2s) - 4.*m2s*pT2ijt))
            + qpar * (q2+m2s-sij)/(2.*qpar2) );

  // Now construct the transverse momentum vector in the dipole CM
  // frame.
  double RNphi = rndmPtr->flat();
  double phi_kt = 2.*M_PI*RNphi;
  Vec4 kTmom(sqrt(kT2)*cos(phi_kt), sqrt(kT2)*sin(phi_kt),0.,0.);

  // Boost the transverse momentum vector into the lab frame.
  RotBstMatrix fromDipCM;
  if (side == -1) fromDipCM.fromCMframe( pRadBef, pRecBef );
  else            fromDipCM.fromCMframe( pRecBef, pRadBef );
  kTmom.rotbst(fromDipCM);

  // Construct left-over dipole momentum by momentum conservation.
  Vec4 pij(-q+pRec);

  // Construct new radiator momentum.
  Vec4 pRad( zbar * (gABC(q2,sij,m2s)*pij + sij*pRec) / bABC(q2,sij,m2s)
           + (m2r+kT2) / (zbar*bABC(q2,sij,m2s))
             * (-pRec - m2s/gABC(q2,sij,m2s)*pij)
           + kTmom);

  // Contruct the emission momentum by momentum conservation.
  Vec4 pEmt(-q-pRad+pRec);

  // For g-> qbar q, swap z->1-z, i.e. swam momenta of emission and radiator.
  if (dipSel->flavour < 0) swap(pRad,pEmt);

  // Ensure that radiator is on mass-shell
  double mTolErr = settingsPtr->parm("Check:mTolErr");
  double errMass = abs(pRad.mCalc() - sqrt(m2r)) / max( 1.0, pRad.e());
  if ( errMass > mTolErr*1e-2 ) {
    double deltam2 = pRad.m2Calc() - m2r;
    pRad.e(sqrtpos(pow2(pRad.e()) - deltam2));
  }
  // Ensure that emission is on mass-shell
  errMass = abs(pEmt.mCalc() - sqrt(m2e)) / max( 1.0, pEmt.e());
  if ( errMass > mTolErr*1e-2 ) {
    double deltam2 = pEmt.m2Calc() - m2e;
    pEmt.e(sqrtpos(pow2(pEmt.e()) - deltam2));
  }
  // Ensure that recoiler is on mass-shell
  errMass = abs(pRec.mCalc() - sqrt(m2s)) / max( 1.0, pRec.e());
  if ( errMass > mTolErr*1e-2 ) {
    double deltam2 = pRec.m2Calc() - m2s;
    pRec.e(sqrtpos(pow2(pRec.e()) - deltam2));
  }

  // New: Return if the x-value for the incoming recoiler is nonsense.
  if ( 2.*pRec.e()/event[0].m() > 1. ) {
    infoPtr->errorMsg("Error in DireTimes::branch: "
            "Larger than unity Bjorken x value");
    return false;
  }

  // Define new particles from dipole branching.
  double pTsel = sqrt(dipSel->pT2);
  Particle rad = Particle(idRad, 51, iRadBef, 0, 0, 0,
    colRad, acolRad, pRad, mRad, pTsel);
  Particle emt = Particle(idEmt, 51, iRadBef, 0, 0, 0,
    colEmt, acolEmt, pEmt, mEmt, pTsel);
  Particle rec = Particle(recBef.id(), -53, 0, 0, iRecBef, iRecBef,
    recBef.col(), recBef.acol(), pRec, 0., pTsel);

  // Special checks to set weak particles status equal to 56.
  // This is needed for decaying the particles. Also set polarisation.
  if (emt.idAbs() == 23 || emt.idAbs() == 24) {
    emt.status(56);
    event[iRadBef].pol( dipSel->weakPol );
    rad.pol( dipSel->weakPol );
  }

  // Save properties to be restored in case of user-hook veto of emission.
  int eventSizeOld = event.size();
  int iRadStatusV  = event[iRadBef].status();
  int iRadDau1V    = event[iRadBef].daughter1();
  int iRadDau2V    = event[iRadBef].daughter2();
  int iRecMot1V    = event[iRecBef].mother1();
  int iRecMot2V    = event[iRecBef].mother2();
  int beamOff1     = 1 + beamOffset;
  int beamOff2     = 2 + beamOffset;
  int ev1Dau1V     = event[beamOff1].daughter1();
  int ev2Dau1V     = event[beamOff2].daughter1();

  // Shower may occur at a displaced vertex.
  if (radBef.hasVertex()) {
    rad.vProd( radBef.vProd() );
    emt.vProd( radBef.vProd() );
  }
  if (recBef.hasVertex()) rec.vProd( recBef.vProd() );

  // Put new particles into the event record.
  // Mark original dipole partons as branched and set daughters/mothers.
  int iRad = event.append(rad);
  int iEmt = event.append(emt);
  event[iRadBef].statusNeg();
  event[iRadBef].daughters( iRad, iEmt);
  int iRec = event.append(rec);
  event[iRecBef].mothers( iRec, iRec);
  event[iRec].mothers( iRecMot1V, iRecMot2V);
  if (iRecMot1V == beamOff1) event[beamOff1].daughter1( iRec);
  if (iRecMot1V == beamOff2) event[beamOff2].daughter1( iRec);

  // Allow veto of branching. If so restore event record to before emission.
  bool inResonance = (partonSystemsPtr->getInA(iSysSel) == 0) ? true : false;
  if ( (canVetoEmission && userHooksPtr->doVetoFSREmission( eventSizeOld,
    event, iSysSel, inResonance))
    || (canMergeFirst && mergingHooksPtr->doVetoEmission( event )) ) {
    event.popBack( event.size() - eventSizeOld);
    event[iRadBef].status( iRadStatusV);
    event[iRadBef].daughters( iRadDau1V, iRadDau2V);
    event[iRecBef].mothers( iRecMot1V, iRecMot2V);
    if (iRecMot1V == beamOff1) event[beamOff1].daughter1( ev1Dau1V);
    if (iRecMot1V == beamOff2) event[beamOff2].daughter1( ev2Dau1V);
    return false;
  }

  // For initial-state recoiler also update beam and sHat info.
  BeamParticle& beamRec = (isrType == 1) ? *beamAPtr : *beamBPtr;
  double xOld = beamRec[iSysSelRec].x();
  double xRec = 2. * pRec.e() / (beamAPtr->e() + beamBPtr->e());
  beamRec[iSysSelRec].iPos( iRec);
  beamRec[iSysSelRec].x( xRec);
  partonSystemsPtr->setSHat( iSysSelRec,
  partonSystemsPtr->getSHat(iSysSelRec) * xRec / xOld);

  // Gluon emission: update both dipole ends and add two new ones.
  if (dipSel->flavour == 21) {
    dipSel->iRadiator  = iRad;
    dipSel->iRecoiler  = iEmt;
    dipSel->systemRec  = iSysSel;
    dipSel->isrType    = 0;
    dipSel->pTmax      = pTsel;
    dipSel->MEtype = 0;
    for (int i = 0; i < int(dipEnd.size()); ++i) {
      if (dipEnd[i].iRadiator == iRecBef && dipEnd[i].iRecoiler == iRadBef
        && dipEnd[i].colType != 0) {
        dipEnd[i].iRadiator = iRec;
        dipEnd[i].iRecoiler = iEmt;
        dipEnd[i].MEtype = 0;
        // Strive to match colour to anticolour inside closed system.
        if ( dipEnd[i].colType * dipSel->colType > 0)
          dipEnd[i].iRecoiler = iRad;
        dipEnd[i].pTmax = pTsel;
      }
    }
    int colType = (dipSel->colType > 0) ? 2 : -2 ;
    // When recoiler was uncoloured particle, in resonance decays,
    // assign recoil to coloured particle.
    int iRecMod = iRec;
    if (recoilToColoured && inResonance && event[iRec].col() == 0
      && event[iRec].acol() == 0) iRecMod = iRad;
    dipEnd.push_back( DireTimesEnd(iEmt, iRecMod, pTsel,
       colType, 0, 0, 0, isrType, iSysSel, 0));
    dipEnd.back().systemRec = iSysSelRec;
    dipEnd.push_back( DireTimesEnd(iEmt, iRad, pTsel,
      -colType, 0, 0, 0, 0, iSysSel, 0));

  // Gluon branching to q qbar: update current dipole and other of gluon.
  } else if (dipSel->colType != 0) {
    for (int i = 0; i < int(dipEnd.size()); ++i) {
      // Strive to match colour to anticolour inside closed system.
      if ( dipEnd[i].iRecoiler == iRadBef
        && dipEnd[i].colType * dipSel->colType < 0 )
        dipEnd[i].iRecoiler = iEmt;
      if (dipEnd[i].iRadiator == iRadBef && abs(dipEnd[i].colType) == 2) {
        dipEnd[i].colType /= 2;
        if (dipEnd[i].system != dipEnd[i].systemRec) continue;

        dipEnd[i].MEtype = 0;
        if (&dipEnd[i] == dipSel) dipEnd[i].iMEpartner = iRad;
        else                      dipEnd[i].iMEpartner = iEmt;
      }
    }

    dipSel->iRadiator = iEmt;
    dipSel->iRecoiler = iRec;
    dipSel->pTmax     = pTsel;

  }

  // Copy or set lifetime for new final state.
  if (event[iRad].id() == event[iRadBef].id()) {
    event[iRad].tau( event[iRadBef].tau() );
  } else {
    event[iRad].tau( event[iRad].tau0() * rndmPtr->exp() );
    event[iEmt].tau( event[iEmt].tau0() * rndmPtr->exp() );
  }
  event[iRec].tau( event[iRecBef].tau() );

  // Now update other dipoles that also involved the radiator or recoiler.
  for (int i = 0; i < int(dipEnd.size()); ++i) {
    if (dipEnd[i].iRadiator  == iRadBef) dipEnd[i].iRadiator  = iRad;
    if (dipEnd[i].iRecoiler  == iRadBef) dipEnd[i].iRecoiler  = iRad;
    if (dipEnd[i].iMEpartner == iRadBef) dipEnd[i].iMEpartner = iRad;
    if (dipEnd[i].iRadiator  == iRecBef) dipEnd[i].iRadiator  = iRec;
    if (dipEnd[i].iRecoiler  == iRecBef) dipEnd[i].iRecoiler  = iRec;
    if (dipEnd[i].iMEpartner == iRecBef) dipEnd[i].iMEpartner = iRec;
  }

  // Finally update the list of all partons in all systems.
  partonSystemsPtr->replace(iSysSel, iRadBef, iRad);
  partonSystemsPtr->addOut(iSysSel, iEmt);
  partonSystemsPtr->replace(iSysSelRec, iRecBef, iRec);

  // Ungraceful exit for incorrect event.
  bool isHadronic = false;
  for (int i = 0; i < event.size(); ++i)
    if (event[i].statusAbs() > 60) isHadronic = true;
  if ( !isHadronic && !validEvent(event)) {
    infoPtr->errorMsg("Error in DireTimes::branch_FI: State after "
                      "branching not valid, thus reject.");
    return false;
  }

  // Set shower weight.
  weights->calcWeight(pT2);
  weights->reset();
  rejectProbability.clear();
  acceptProbability.clear();

  // Done.
  return true;

}

//--------------------------------------------------------------------------

Event DireTimes::branched( const Event& event, int iRadBef, int iRecBef,
  int idEmt, double pT2, double z, double RNphi, vector<double>) {

  // Construct the branched event
  Event NewEvent = Event();
  NewEvent.init("(hard process-modified)", particleDataPtr);
  // Copy all unchanged particles to NewEvent
  for (int i = 0; i < event.size(); ++i)
    NewEvent.append( event[i] );

  // Default flavours and colour tags for new particles in dipole branching.
  int idRad        = NewEvent[iRadBef].id();
  int colRad       = NewEvent[iRadBef].col();
  int acolRad      = NewEvent[iRadBef].acol();
  int colEmt       = 0;
  int acolEmt      = 0;

  // New colour tag required for gluon emission.
  if (idEmt == 21 && idRad > 0) {
    colEmt  = colRad;
    colRad  = NewEvent.nextColTag();
    acolEmt = colRad;
  } else if (idEmt == 21) {
    acolEmt = acolRad;
    acolRad = NewEvent.nextColTag();
    colEmt  = acolRad;
  // New flavours for g -> q qbar; split colours.
  } else if (idRad > 0) {
    idEmt   = idEmt;
    idRad   = -idEmt;
    colEmt  = colRad;
    colRad  = 0;
  } else if (idRad < 0) {
    idEmt   = -idEmt;
    idRad   = -idEmt;
    acolEmt = acolRad;
    acolRad = 0;
  }
  int flavour = (idEmt < 0) ? -1 : (idEmt == 21) ? 21 : -1;

  // Construct momenta
  vector<Vec4> momenta;
  int statusRec = 52;
  if ( event[iRecBef].isFinal())
    momenta = split_FF(NewEvent,iRadBef,iRecBef,pT2,z, RNphi, flavour, idRad, idEmt);
  else {
    momenta = split_FI(NewEvent,iRadBef,iRecBef,pT2,z, RNphi, flavour, idRad, idEmt);
    statusRec = -53;
  }

  double mRad = sqrtpos(momenta[0].m2Calc());
  double mEmt = sqrtpos(momenta[1].m2Calc());
  double mRec = sqrtpos(momenta[2].m2Calc());

  // Define new particles from dipole branching.
  double pTsel = sqrt(pT2);
  Particle rad = Particle(idRad, 51, iRadBef, 0, 0, 0,
    colRad, acolRad, momenta[0], mRad, pTsel);
  Particle emt = Particle(idEmt, 51, iRadBef, 0, 0, 0,
    colEmt, acolEmt, momenta[1], mEmt, pTsel);
  Particle rec = Particle(NewEvent[iRecBef].id(), statusRec, 0, 0,
    iRecBef, iRecBef, NewEvent[iRecBef].col(),
    NewEvent[iRecBef].acol(), momenta[2], mRec, 0.);

  int iRecMot1V    = NewEvent[iRecBef].mother1();
  int iRecMot2V    = NewEvent[iRecBef].mother2();

  // Mark original dipole partons as branched and set daughters/mothers.
  int iRad = NewEvent.append(rad);
  int iEmt = NewEvent.append(emt);
  NewEvent[iRadBef].statusNeg();
  if (statusRec == 52 ) NewEvent[iRecBef].statusNeg();
  NewEvent[iRadBef].daughters( iRad, iEmt);
  int iRec = NewEvent.append(rec);
  NewEvent[iRecBef].mothers( iRec, iRec);
  NewEvent[iRec].mothers( iRecMot1V, iRecMot2V);
  if (iRecMot1V == 1) NewEvent[1].daughter1( iRec);
  if (iRecMot1V == 2) NewEvent[2].daughter1( iRec);

  // Done.
  return NewEvent;

}

//--------------------------------------------------------------------------

vector<Vec4> DireTimes::split_FF( const Event& event, int iRadBef, int iRecBef,
  double pT2, double z, double RNphi, int flavour, int idRad, int idEmt) {

  // Find their momenta, with special sum for global recoil.
  Vec4 pRadBef     = event[iRadBef].p();
  Vec4 pRecBef     = event[iRecBef].p();

  double m2Dip     = 2.*event[iRadBef].p()*event[iRecBef].p();

  // Get particle masses.
  // Radiator before splitting.
  double m2Bef = (event[iRadBef].idAbs() < 6)
               ? getMass(event[iRadBef].id(),2)
               : getMass(event[iRadBef].id(),3,event[iRadBef].mCalc());
  // Radiator after splitting.
  double m2r = (abs(idRad) < 6) ? getMass(idRad,2) : getMass(idRad,1);
  // Recoiler. 
  double m2s = (event[iRecBef].idAbs() < 6)
        ? getMass(event[iRecBef].id(),2)
        : getMass(event[iRecBef].id(),3,event[iRecBef].mCalc());
  // Emission.
  double m2e = (abs(idEmt) < 6 ) ? getMass(idEmt,2) : getMass(idEmt,1);

  // For g-> qbar q, swap meaning of radiator an emission.
  if (flavour < 0) swap(m2r,m2e);

  // Adjust the dipole kinematical mass to accomodate masses after branching.
  double Q2  = m2Dip + m2Bef - m2r - m2e;

  // Calculate CS variables.
  double kappa2 = pT2/Q2;
  double yCS    = kappa2 / (1.-z);
  double zCS    = ( 1. - z - kappa2 - pow2(1.-z) ) / ( 1. - z - kappa2);

  // Get dipole 4-momentum.
  Vec4 q(pRadBef + pRecBef);
  double q2 = q.m2Calc();

  // Calculate derived variables.
  double sij  = yCS * (q2 - m2s) + (1.-yCS)*(m2r+m2e);
  double zbar = (q2-sij-m2s) / bABC(q2,sij,m2s)
              * (zCS - m2s/gABC(q2,sij,m2s)
                      *(sij + m2r - m2e)/(q2-sij-m2s));
  double kT2  = zbar*(1.-zbar)*sij - (1.-zbar)*m2r - zbar*m2e; 
  
  // Now construct the new recoiler momentum in the lab frame.
  Vec4 pRec( (pRecBef - (q*pRecBef)/q2 * q)
            * sqrt(lABC(q2,sij,m2s)/lABC(q2,m2Bef,m2s))
            + q * (q2+m2s-sij)/(2.*q2) );

  // Now construct the transverse momentum vector in the dipole CM
  // frame.
  double phi_kt = 2.*M_PI*RNphi;
  Vec4 kTmom(sqrt(kT2)*cos(phi_kt), sqrt(kT2)*sin(phi_kt),0.,0.);

  // Boost the transverse momentum vector into the lab frame.
  RotBstMatrix fromDipCM2;
  fromDipCM2.fromCMframe(pRadBef, pRecBef);
  kTmom.rotbst(fromDipCM2);

  // Construct left-over dipole momentum by momentum conservation.
  Vec4 pij(q-pRec);

  // Construct new radiator momentum.
  Vec4 pRad( zbar * (gABC(q2,sij,m2s)*pij - sij*pRec) / bABC(q2,sij,m2s)
            + (m2r+kT2) / (zbar*bABC(q2,sij,m2s))
              * (pRec - m2s/gABC(q2,sij,m2s)*pij)
            + kTmom);

  // Contruct the emission momentum by momentum conservation.
  Vec4 pEmt(q-pRad-pRec);

  // For g-> qbar q, swap z->1-z, i.e. swam momenta of emission and radiator.
  if (flavour < 0) swap(pRad,pEmt);

  // Prepare outputs.
  vector<Vec4> ret;
  ret.push_back( pRad );
  ret.push_back( pEmt );
  ret.push_back( pRec );

  // Done.
  return ret;

}

//--------------------------------------------------------------------------

vector<Vec4> DireTimes::split_FI( const Event& event, int iRadBef, int iRecBef,
  double pT2, double z, double RNphi, int flavour, int idRad, int idEmt) {

  // Find their momenta, with special sum for global recoil.
  Vec4 pRadBef     = event[iRadBef].p();
  Vec4 pRecBef     = event[iRecBef].p();
  int side         = (pRecBef.pz() > 0.) ? 1 : -1;

  double m2Dip     = abs(2.*event[iRadBef].p()*event[iRecBef].p());

  // Get particle masses.
  // Radiator before splitting.
  double m2Bef = (event[iRadBef].idAbs() < 6)
               ? getMass(event[iRadBef].id(),2)
               : getMass(event[iRadBef].id(),3,
                         event[iRadBef].mCalc());
  // Radiator after splitting.
  double m2r = (abs(idRad) < 6) ? getMass(idRad,2) : getMass(idRad,1);
  // Emission.
  double m2e = (abs(idEmt) < 6) ? getMass(idEmt,2) : getMass(idEmt,1);
  // Initial state recoiler always assumed massless. 
  double m2s = 0.0;

  // For g-> qbar q, swap meaning of radiator an emission.
  if (flavour < 0) swap(m2r,m2e);

  // Recalculate the kinematicaly available dipole mass.
  double Q2 = m2Dip - m2Bef + m2r + m2e;

  // Calculate CS variables.
  double kappa2 =  pT2/Q2;
  double xCS    = 1 - kappa2/(1.-z);
  double zCS = z;

  // Get dipole 4-momentum.
  Vec4 q(pRecBef-pRadBef);
  double q2 = q.m2Calc();
  Vec4 qpar(q.px()+pRadBef.px(), q.py()+pRadBef.py(), q.pz(), q.e());
  double qpar2 = qpar.m2Calc();
  double pT2ijt = pow2(pRadBef.px()) + pow2(pRadBef.py());

  // Calculate derived variables.
  double sij  = (1.-1./xCS) * (q2 - m2s) + (m2r+m2e) / xCS;
  double zbar = (q2-sij-m2s) / bABC(q2,sij,m2s)
              * (zCS - m2s/gABC(q2,sij,m2s)
                      *(sij + m2r - m2e)/(q2-sij-m2s));
  double kT2  = zbar*(1.-zbar)*sij - (1.-zbar)*m2r - zbar*m2e; 

  // Now construct the new recoiler momentum in the lab frame.
  Vec4 pRec( (pRecBef - (qpar*pRecBef)/qpar2 * qpar)
            * sqrt( (lABC(q2,sij,m2s)   - 4.*m2s*pT2ijt)
                   /(lABC(q2,m2Bef,m2s) - 4.*m2s*pT2ijt))
            + qpar * (q2+m2s-sij)/(2.*qpar2) );

  // Now construct the transverse momentum vector in the dipole CM
  // frame.
  double phi_kt = 2.*M_PI*RNphi;
  Vec4 kTmom(sqrt(kT2)*cos(phi_kt), sqrt(kT2)*sin(phi_kt),0.,0.);

  // Boost the transverse momentum vector into the lab frame.
  RotBstMatrix fromDipCM;
  if (side == -1) fromDipCM.fromCMframe( pRadBef, pRecBef );
  else            fromDipCM.fromCMframe( pRecBef, pRadBef );
  kTmom.rotbst(fromDipCM);

  // Construct left-over dipole momentum by momentum conservation.
  Vec4 pij(-q+pRec);

  // Construct new radiator momentum.
  Vec4 pRad( zbar * (gABC(q2,sij,m2s)*pij + sij*pRec) / bABC(q2,sij,m2s)
           + (m2r+kT2) / (zbar*bABC(q2,sij,m2s))
             * (-pRec - m2s/gABC(q2,sij,m2s)*pij)
           + kTmom);

  // Contruct the emission momentum by momentum conservation.
  Vec4 pEmt(-q-pRad+pRec);

  // For g-> qbar q, swap z->1-z, i.e. swam momenta of emission and radiator.
  if (flavour < 0) swap(pRad,pEmt);

  // Prepare outputs.
  vector<Vec4> ret;
  ret.push_back( pRad );
  ret.push_back( pEmt );
  ret.push_back( pRec );

  // Done.
  return ret;

}

//--------------------------------------------------------------------------

Event DireTimes::clustered( const Event& state, string name,
  int iRad, int iEmt, int iRec ) {

  // Flags for type of radiation
  int radType = state[iRad].isFinal() ? 1 : -1;
  int recType = state[iRec].isFinal() ? 1 : -1;

  // Construct the clustered event
  Event NewEvent = Event();
  NewEvent.init("(hard process-modified)", particleDataPtr);
  NewEvent.clear();
  // Copy all unchanged particles to NewEvent
  for (int i = 0; i < state.size(); ++i)
    if ( i != iRad && i != iRec && i != iEmt )
      NewEvent.append( state[i] );

  // Copy all the junctions one by one
  for (int i = 0; i < state.sizeJunction(); ++i)
    NewEvent.appendJunction( state.getJunction(i) );
  // Find an appropriate scale for the hard process
  double mu = infoPtr->QFac();
  // Initialise scales for new event
  NewEvent.saveSize();
  NewEvent.saveJunctionSize();
  NewEvent.scale(mu);
  NewEvent.scaleSecond(mu);

  // Set properties of radiator/recoiler after the clustering
  // Recoiler properties will be unchanged
  Particle RecBefore = Particle( state[iRec] );
  RecBefore.setEvtPtr(&NewEvent);
  RecBefore.daughters(0,0);

  // Find flavour of radiator before splitting
  int radID = (*splittingsPtr)[name]->radID(state[iRad].id(), state[iEmt].id());
  int recID = state[iRec].id();
  Particle RadBefore = Particle( state[iRad] );
  RadBefore.setEvtPtr(&NewEvent);
  RadBefore.id(radID);
  RadBefore.daughters(0,0);
  // Put dummy values for colours
  RadBefore.cols(RecBefore.acol(),RecBefore.col());

  // Reset status if the reclustered radiator is a resonance.
  if ( particleDataPtr->isResonance(radID) && radType == 1)
    RadBefore.status(state[iRad].status());

  // Put mass for radiator and recoiler
  double radMass = particleDataPtr->m0(radID);
  double recMass = particleDataPtr->m0(recID);
  if (radType == 1 ) RadBefore.m(radMass);
  else RadBefore.m(0.0);
  if (recType == 1 ) RecBefore.m(recMass);
  else RecBefore.m(0.0);

  // Construct momenta and  colours of clustered particles
  if ( state[iRec].isFinal())
    cluster_FF(state,iRad,iEmt,iRec,radID,RadBefore,RecBefore);
  else
    cluster_FI(state,iRad,iEmt,iRec,radID,RadBefore,RecBefore);

  // Put some dummy production scales for RecBefore, RadBefore
  RecBefore.scale(mu);
  RadBefore.scale(mu);

  // Append new recoiler and find new radiator colour
  NewEvent.append(RecBefore);

  // Assign the correct colour to re-clustered radiator.
  pair<int,int> cols = (*splittingsPtr)[name]->radCols( state[iRad].col(),
    state[iRad].acol(), state[iEmt].col(), state[iEmt].acol());
  RadBefore.cols( cols.first, cols.second );

  // Build the clustered event
  Event outState = Event();
  outState.init("(hard process-modified)", particleDataPtr);
  outState.clear();

  // Copy system and incoming beam particles to outState
  for (int i = 0; i < 3; ++i)
    outState.append( NewEvent[i] );
  // Copy all the junctions one by one
  for (int i = 0; i < state.sizeJunction(); ++i)
    outState.appendJunction( state.getJunction(i) );
  // Initialise scales for new event
  outState.saveSize();
  outState.saveJunctionSize();
  outState.scale(mu);
  outState.scaleSecond(mu);
  bool radAppended = false;
  bool recAppended = false;
  int size = int(outState.size());
  // Save position of radiator in new event record
  int radPos = 0;

  // Append first incoming particle
  if ( RecBefore.mother1() == 1) {
    outState.append( RecBefore );
    recAppended = true;
  } else if ( RadBefore.mother1() == 1 ) {
    radPos = outState.append( RadBefore );
    radAppended = true;
  } else {
    // Find second incoming in input event
    int in1 = 0;
    for(int i=0; i < int(state.size()); ++i)
      if (state[i].mother1() == 1) in1 =i;
    outState.append( state[in1] );
    size++;
  }
  // Append second incoming particle
  if ( RecBefore.mother1() == 2) {
    outState.append( RecBefore );
    recAppended = true;
  } else if ( RadBefore.mother1() == 2 ) {
    radPos = outState.append( RadBefore );
    radAppended = true;
  } else {
    // Find second incoming in input event
    int in2 = 0;
    for(int i=0; i < int(state.size()); ++i)
      if (state[i].mother1() == 2) in2 =i;

    outState.append( state[in2] );
    size++;
  }

  // Append new recoiler if not done already
  if (!recAppended && !RecBefore.isFinal()) {
    recAppended = true;
    outState.append( RecBefore);
  }
  // Append new radiator if not done already
  if (!radAppended && !RadBefore.isFinal()) {
    radAppended = true;
    radPos = outState.append( RadBefore);
  }

  // Append intermediate particle
  // (careful not to append reclustered recoiler)
  // Append intermediate particle
  // (careful not to append reclustered recoiler)
  for (int i = 0; i < int(NewEvent.size()-1); ++i)
    if (NewEvent[i].status() == -22) outState.append( NewEvent[i] );
  // Append final state particles, resonances first
  for (int i = 0; i < int(NewEvent.size()-1); ++i)
    if (NewEvent[i].status() == 22) outState.append( NewEvent[i] );
  // Then start appending partons
  if (!radAppended && RadBefore.statusAbs() == 22)
    radPos = outState.append(RadBefore);
  if (!recAppended)
    outState.append(RecBefore);
  if (!radAppended && RadBefore.statusAbs() != 22)
    radPos = outState.append(RadBefore);
  // Then partons (not reclustered recoiler)
  for(int i = 0; i < int(NewEvent.size()-1); ++i)
    if ( NewEvent[i].status()  != 22
      && NewEvent[i].colType() != 0
      && NewEvent[i].isFinal())
      outState.append( NewEvent[i] );
  // Then the rest
  for(int i = 0; i < int(NewEvent.size()-1); ++i)
    if ( NewEvent[i].status() != 22
      && NewEvent[i].colType() == 0
      && NewEvent[i].isFinal() )
      outState.append( NewEvent[i]);

  // Find intermediate and respective daughters
  vector<int> PosIntermediate;
  vector<int> PosDaughter1;
  vector<int> PosDaughter2;
  for(int i=0; i < int(outState.size()); ++i)
    if (outState[i].status() == -22) {
      PosIntermediate.push_back(i);
      int d1 = outState[i].daughter1();
      int d2 = outState[i].daughter2();
      // Find daughters in output state
      int daughter1 = FindParticle( state[d1], outState);
      int daughter2 = FindParticle( state[d2], outState);
      // If both daughters found, done
      // Else put first final particle as first daughter
      // and last final particle as second daughter
      if (daughter1 > 0)
        PosDaughter1.push_back( daughter1);
      else {
        daughter1 = 0;
        while(!outState[daughter1].isFinal() ) daughter1++;
        PosDaughter1.push_back( daughter1);
      }
      if (daughter2 > 0)
        PosDaughter2.push_back( daughter2);
      else {
        daughter2 = outState.size()-1;
        while(!outState[daughter2].isFinal() ) daughter2--;
        PosDaughter2.push_back( daughter2);
      }
    }
  // Set daughters and mothers
  for(int i=0; i < int(PosIntermediate.size()); ++i) {
    outState[PosIntermediate[i]].daughters(PosDaughter1[i],PosDaughter2[i]);
    outState[PosDaughter1[i]].mother1(PosIntermediate[i]);
    outState[PosDaughter2[i]].mother1(PosIntermediate[i]);
  }

  // Find range of final state partons
  int minParFinal = int(outState.size());
  int maxParFinal = 0;
  for(int i=0; i < int(outState.size()); ++i)
    if (outState[i].mother1() == 3 && outState[i].mother2() == 4) {
      minParFinal = min(i,minParFinal);
      maxParFinal = max(i,maxParFinal);
    }

  if (minParFinal == maxParFinal) maxParFinal = 0;
  outState[3].daughters(minParFinal,maxParFinal);
  outState[4].daughters(minParFinal,maxParFinal);

  // Update event properties
  outState.saveSize();
  outState.saveJunctionSize();

  // Almost there...
  // If an intermediate coloured parton exists which was directly
  // colour connected to the radiator before the splitting, and the
  // radiator before and after the splitting had only one colour, problems
  // will arise since the colour of the radiator will be changed, whereas
  // the intermediate parton still has the old colour. In effect, this
  // means that when setting up a event for trial showering, one colour will
  // be free.
  // Hence, check for an intermediate coloured triplet resonance has been
  // colour-connected to the "old" radiator.
  // Find resonance
  int iColRes = 0;
  if ( radType == -1 && state[iRad].colType() == 1) {
      // Find resonance connected to initial colour
      for(int i=0; i < int(state.size()); ++i)
        if ( i != iRad && i != iEmt && i != iRec
          && state[i].status() == -22
          && state[i].col() == state[iRad].col() )
          iColRes = i;
  } else if ( radType == -1 && state[iRad].colType() == -1) {
      // Find resonance connected to initial anticolour
      for(int i=0; i < int(state.size()); ++i)
        if ( i != iRad && i != iEmt && i != iRec
          && state[i].status() == -22
          && state[i].acol() == state[iRad].acol() )
          iColRes = i;
  } else if ( radType == 1 && state[iRad].colType() == 1) {
      // Find resonance connected to final state colour
      for(int i=0; i < int(state.size()); ++i)
        if ( i != iRad && i != iEmt && i != iRec
          && state[i].status() == -22
          && state[i].col() == state[iRad].col() )
          iColRes = i;
  } else if ( radType == 1 && state[iRad].colType() == -1) {
      // Find resonance connected to final state anticolour
      for(int i=0; i < int(state.size()); ++i)
        if ( i != iRad && i != iEmt && i != iRec
          && state[i].status() == -22
          && state[i].acol() == state[iRad].acol() )
          iColRes = i;
  }

  if (iColRes > 0) {
    // Now find this resonance in the reclustered state
    int iColResNow = FindParticle( state[iColRes], outState);

    // Find reclustered radiator colours
    int radCol = outState[radPos].col();
    int radAcl = outState[radPos].acol();
    // Find resonance radiator colours
    int resCol = outState[iColResNow].col();
    int resAcl = outState[iColResNow].acol();
    // Check if any of the reclustered radiators colours match the resonance
    bool matchesRes =  (radCol > 0
                          && ( radCol == resCol || radCol == resAcl))
                    || (radAcl > 0
                          && ( radAcl == resCol || radAcl == resAcl));

    // If a resonance has been found, but no colours match, change
    // the colour of the resonance
    if (!matchesRes && iColResNow > 0) {
      if ( radType == -1 && outState[radPos].colType() == 1)
        outState[iColResNow].col(radCol);
      else if ( radType ==-1 && outState[radPos].colType() ==-1)
        outState[iColResNow].acol(radAcl);
      else if ( radType == 1 && outState[radPos].colType() == 1)
        outState[iColResNow].col(radCol);
      else if ( radType == 1 && outState[radPos].colType() ==-1)
        outState[iColResNow].acol(radAcl);
    }


    // If a resonance has been found, but no colours match, and the position
    // of the resonance in the event record has been changed, update the
    // radiator mother
    if (!matchesRes && iColResNow > 0 && iColRes != iColResNow)
      outState[radPos].mother1(iColResNow);

  }

  // Force outgoing particles to be part of hard process.
  for ( int i=0; i < int(outState.size()); ++i)
    if (outState[i].isFinal()) outState[i].status(23);
  // Force incoming to be part of the hard process.
  outState[3].status(-21);
  outState[4].status(-21);

  // Done
  return outState;
}

bool DireTimes::cluster_FF( const Event& state,
  int iRad, int iEmt, int iRec, int idRadBef, Particle& radBef, Particle& recBef ) {

  // Calculate CS variables.
  double pT2    = pT2_FF(state[iRad], state[iEmt], state[iRec]);
  double z      = z_FF(state[iRad], state[iEmt], state[iRec]);

  // Get particle masses.
  double m2Bef = ( abs(idRadBef) < 6)
               ? getMass(idRadBef,2)
               : (idRadBef == state[iRad].id())
                  ? getMass(idRadBef,3,state[iRad].mCalc())
                  : getMass(idRadBef,2);
  double m2r   = state[iRad].p().m2Calc();
  double m2e   = state[iEmt].p().m2Calc();
  double m2s   = state[iRec].p().m2Calc();

  double m2D = 2.*state[iRad].p()*state[iRec].p()
             + 2.*state[iRad].p()*state[iEmt].p()
             + 2.*state[iRec].p()*state[iEmt].p();
  double Q2 = m2D + (m2Bef - m2r - m2e);

  // Get dipole 4-momentum.
  Vec4 q(state[iRad].p() + state[iEmt].p() + state[iRec].p());
  double q2 = q.m2Calc();

  int type     = 1;
  // Upate type if this is a massive splitting.
  if ( m2Bef > TINYMASS || m2r > TINYMASS || m2s > TINYMASS
    || m2e > TINYMASS ) type = 2;

  // Check phase space constraints.
  if ( !inAllowedPhasespace( z, pT2, Q2, q2, 0.0, type, m2Bef, m2r, m2s,
        m2e) ) return false; 

  // Calculate derived variables.
  double sij  = 2.*state[iRad].p()*state[iEmt].p();
  
  // Now construct the new recoiler momentum in the lab frame.
  Vec4 pRec(state[iRec].p());
  Vec4 pRecBef( (pRec - (q*pRec)/q2 * q)
            * sqrt(lABC(q2,m2Bef,m2s)/lABC(q2,sij,m2s))
            + q * (q2+m2s-m2Bef)/(2.*q2) );

  // Get momentum of radiator my momentum conservation.
  Vec4 pRadBef(q-pRecBef);

  radBef.p(pRadBef);
  recBef.p(pRecBef);
  radBef.m(sqrtpos(m2Bef));
  recBef.m(sqrtpos(m2s));

  // Done
  return true;
}

bool DireTimes::cluster_FI( const Event& state, 
  int iRad, int iEmt, int iRec, int idRadBef, Particle& radBef, Particle& recBef ) {

  // Calculate CS variables.
  double pT2    = pT2_FI(state[iRad], state[iEmt], state[iRec]);
  double z      = z_FI(state[iRad], state[iEmt], state[iRec]);

  // Get particle masses.
  double m2Bef = ( abs(idRadBef) < 6)
               ? getMass(idRadBef,2)
               : (idRadBef == state[iRad].id())
                  ? getMass(idRadBef,3,state[iRad].mCalc())
                  : getMass(idRadBef,2);
  double m2r   = state[iRad].p().m2Calc();
  double m2e   = state[iEmt].p().m2Calc();
  double m2s   = state[iRec].p().m2Calc();

  double m2D  = -2.*state[iRad].p()*state[iEmt].p()
             + 2.*state[iRad].p()*state[iRec].p()
             + 2.*state[iRec].p()*state[iEmt].p();
  double Q2 = m2D - (m2Bef - m2r - m2e);

  // Get dipole 4-momentum.
  //Vec4 q(state[iRad].p() + state[iEmt].p() - state[iRec].p());
  Vec4 q(-state[iRad].p() - state[iEmt].p() + state[iRec].p());
  double q2 = q.m2Calc();

  // Check phase space constraints.
  double xNew = 2. * state[iRec].e() / state[0].m();

  // Calculate CS variables.
  double kappa2 =  pT2/Q2;
  double xCS = 1 - kappa2/(1.-z);
  double xCDST = xCS*( 1. + (m2Bef-m2r-m2e)/Q2 );
  double xOld  = xNew * xCDST;

  int type     = -1;
  // Upate type if this is a massive splitting.
  if ( m2Bef > TINYMASS || m2r > TINYMASS || m2s > TINYMASS
    || m2e > TINYMASS ) type = -2;

  if ( !inAllowedPhasespace( z, pT2, Q2, q2, xOld, type, m2Bef, m2r, m2s,
        m2e) ) return false; 

  Vec4 pRad(state[iRad].p());
  Vec4 pRec(state[iRec].p());
  Vec4 qpar(q.px()+pRad.px(), q.py()+pRad.py(), q.pz(), q.e());
  double qpar2 = qpar.m2Calc();
  double pT2ijt = pow2(pRad.px()) + pow2(pRad.py());

  // Calculate derived variables.
  double sij  = 2.*state[iRad].p()*state[iEmt].p();

  // Now construct the new recoiler momentum in the lab frame.
  Vec4 pRecBef( (pRec - (qpar*pRec)/qpar2 * qpar)
            * sqrt( (lABC(q2,m2Bef,m2s) - 4.*m2s*pT2ijt)
                   /(lABC(q2,sij,m2s)   - 4.*m2s*pT2ijt))
            + qpar * (q2+m2s-m2Bef)/(2.*qpar2) );

  // Get momentum of radiator my momentum conservation.
  Vec4 pRadBef(-q+pRecBef);

  radBef.p(pRadBef);
  recBef.p(pRecBef);
  radBef.m(sqrtpos(m2Bef));
  recBef.m(sqrtpos(m2s));

  // Done
  return true;
}

//--------------------------------------------------------------------------

// Function to in the input event find a particle with quantum
// numbers matching those of the input particle
// IN  Particle : Particle to be searched for
//     Event    : Event to be searched in
// OUT int      : > 0 : Position of matching particle in event
//                < 0 : No match in event

int DireTimes::FindParticle( const Particle& particle, const Event& event,
  bool checkStatus ) {

  int index = -1;

  for ( int i = int(event.size()) - 1; i > 0; --i )
    if ( event[i].id()         == particle.id()
      && event[i].colType()    == particle.colType()
      && event[i].chargeType() == particle.chargeType()
      && event[i].col()        == particle.col()
      && event[i].acol()       == particle.acol()
      && event[i].charge()     == particle.charge() ) {
      index = i;
      break;
    }

  if ( checkStatus && event[index].status() != particle.status() )
    index = -1;

  return index;
}

//--------------------------------------------------------------------------

double DireTimes::pT2_FF ( const Particle& rad, const Particle& emt,
  const Particle& rec) {
  double sij = 2.*rad.p()*emt.p();
  double sik = 2.*rad.p()*rec.p();
  double sjk = 2.*rec.p()*emt.p();
  return sij*sjk / (sij+sik+sjk);
}

//--------------------------------------------------------------------------

double DireTimes::pT2_FI ( const Particle& rad, const Particle& emt,
  const Particle& rec) {
  double sij =  2.*rad.p()*emt.p();
  double sai = -2.*rec.p()*rad.p();
  double saj = -2.*rec.p()*emt.p();
  return sij*saj / (sai+saj) * (sij+saj+sai) / (sai+saj) ;
}

//--------------------------------------------------------------------------

double DireTimes::z_FF ( const Particle& rad, const Particle& emt,
  const Particle& rec) {
  double sij = 2.*rad.p()*emt.p();
  double sik = 2.*rad.p()*rec.p();
  double sjk = 2.*rec.p()*emt.p();
  return (sij + sik) / (sij+sik+sjk);
}

//--------------------------------------------------------------------------

double DireTimes::z_FI ( const Particle& rad, const Particle& emt,
  const Particle& rec) {
  double sai = -2.*rec.p()*rad.p();
  double saj = -2.*rec.p()*emt.p();
  return sai / (sai+saj);
}

//--------------------------------------------------------------------------

// Compute splitting probability.


double DireTimes::getSplittingProb( const Event& state, int iRad,
  int iEmt, int iRec) {

  string name  = getSplittingName( state, iRad, iEmt);
  double z     = zTimes(state[iRad], state[iEmt], state[iRec]);
  double pT2   = pT2Times(state[iRad], state[iEmt], state[iRec]);
  double m2D   = (state[iRec].isFinal())
               ? abs( 2.*state[iEmt].p()*state[iRad].p()
                    + 2.*state[iRec].p()*state[iRad].p()
                    + 2.*state[iEmt].p()*state[iRec].p()) 
               : abs( 2.*state[iEmt].p()*state[iRad].p()
                    - 2.*state[iRec].p()*state[iRad].p()
                    - 2.*state[iEmt].p()*state[iRec].p());
  int idRadBef = splittingsPtr->getSplittingRadBefID(state, iRad, iEmt);
  //double m2Bef = pow2(particleDataPtr->m0(idRadBef));
  double m2Bef = ( abs(idRadBef) < 6)
               ? getMass(idRadBef,2)
               : (idRadBef == state[iRad].id())
                  ? getMass(idRadBef,3,state[iRad].mCalc())
                  : getMass(idRadBef,2);
  double m2r   = state[iRad].p().m2Calc();
  double m2e   = state[iEmt].p().m2Calc();
  double m2s   = state[iRec].p().m2Calc();
  int type     = state[iRec].isFinal() ? 1 : -1;

  // Upate type if this is a massive splitting.
  if (type == 1 && (m2Bef > TINYMASS || m2r > TINYMASS || m2s > TINYMASS
    || m2e > TINYMASS) ) type = 2;
  if (type ==-1 && (m2Bef > TINYMASS || m2r > TINYMASS || m2s > TINYMASS
    || m2e > TINYMASS) ) type =-2;

  // Recalculate the kinematicaly available dipole mass.
  int massSign  = (type > 0) ? 1 : -1;
  double q2     = (state[iRec].p() + state[iRad].p()).m2Calc();
  double Q2     = m2D + massSign*(m2Bef - m2r - m2e);
  double kappa2 =  pT2/Q2;
  double xCS    = 1 - kappa2/(1.-z);
  double xOld   = (type > 0) ? 0.0 : xCS * 2.*state[iRec].e() / state[0].m();
  // Check phase space constraints.
  if ( !inAllowedPhasespace( z, pT2, Q2, q2, xOld, type, m2Bef, m2r, m2s,
        m2e) ) return 0.0; 

  return (*splittingsPtr)[name]->kernel(z,pT2,m2D,type,m2Bef,m2r,m2e,m2s);
}

// For new Pythia version superseeding 8.212
//double DireTimes::getSplittingProb( const Event& state, int iRad,
//  int iEmt, int iRec, string) {
//
//  string name  = getSplittingName( state, iRad, iEmt, iRec);
//  double z     = zTimes(state[iRad], state[iEmt], state[iRec]);
//  double pT2   = pT2Times(state[iRad], state[iEmt], state[iRec]);
//  double m2D   = (state[iRec].isFinal())
//               ? abs( 2.*state[iEmt].p()*state[iRad].p()
//                    + 2.*state[iRec].p()*state[iRad].p()
//                    + 2.*state[iEmt].p()*state[iRec].p()) 
//               : abs( 2.*state[iEmt].p()*state[iRad].p()
//                    - 2.*state[iRec].p()*state[iRad].p()
//                    - 2.*state[iEmt].p()*state[iRec].p());
//  int idRadBef = splittingsPtr->getSplittingRadBefID(state, iRad, iEmt);
//  double m2Bef = ( abs(idRadBef) < 6)
//               ? getMass(idRadBef,2)
//               : (idRadBef == state[iRad].id())
//                  ? getMass(idRadBef,3,state[iRad].mCalc())
//                  : getMass(idRadBef,2);
//  double m2r   = state[iRad].p().m2Calc();
//  double m2e   = state[iEmt].p().m2Calc();
//  double m2s   = state[iRec].p().m2Calc();
//  int type     = state[iRec].isFinal() ? 1 : -1;
//
//  // Upate type if this is a massive splitting.
//  if (type == 1 && (m2Bef > TINYMASS || m2r > TINYMASS || m2s > TINYMASS
//    || m2e > TINYMASS) ) type = 2;
//  if (type ==-1 && (m2Bef > TINYMASS || m2r > TINYMASS || m2s > TINYMASS
//    || m2e > TINYMASS) ) type =-2;
//
//  // Recalculate the kinematicaly available dipole mass.
//  int massSign  = (type > 0) ? 1 : -1;
//  double q2     = (state[iRec].p() + state[iRad].p()).m2Calc();
//  double Q2     = m2D + massSign*(m2Bef - m2r - m2e);
//  double kappa2 =  pT2/Q2;
//  double xCS    = 1 - kappa2/(1.-z);
//  double xOld   = (type > 0) ? 0.0 : xCS * 2.*state[iRec].e() / state[0].m();
//  // Check phase space constraints.
//  if ( !inAllowedPhasespace( z, pT2, Q2, q2, xOld, type, m2Bef, m2r, m2s,
//        m2e) ) return 0.0; 
//
//  return (*splittingsPtr)[name]->kernel(z,pT2,m2D,type,m2Bef,m2r,m2e,m2s);
//}

//--------------------------------------------------------------------------

// Check colour/flavour correctness of state.
bool DireTimes::validEvent( const Event& state ) {

  bool validColour  = true;
  bool validCharge  = true;
  bool validMomenta = true;

  bool hasSystems = partonSystemsPtr->sizeSys() > 0;
  int sizeSys     = (hasSystems) ? partonSystemsPtr->sizeSys() : 1;

  Event event = Event();
  event.clear();
  event.init( "(hard process-modified)", particleDataPtr );

  for (int iSys = 0; iSys < sizeSys; ++iSys) {

    // Done if the state is already broken.
    if (!validColour || !validCharge ) break;

    event.clear();
    event.init( "(hard process-modified)", particleDataPtr );
    event.clear();

    int in1 = 0;
    for ( int i = state.size()-1; i > 0; --i)
      if ( state[i].mother1() == 1 && state[i].mother2() == 0
        && (!hasSystems || partonSystemsPtr->getSystemOf(i,true) == iSys))
        {in1 = i; break;}
    if (in1 == 0) in1 = partonSystemsPtr->getInA(iSys);
    int in2 = 0;
    for ( int i = state.size()-1; i > 0; --i)
      if ( state[i].mother1() == 2 && state[i].mother2() == 0
        && (!hasSystems || partonSystemsPtr->getSystemOf(i,true) == iSys))
        {in2 = i; break;}
    if (in2 == 0) in2 = partonSystemsPtr->getInB(iSys);

    event.append(state[0]);
    event.append(state[1]);
    event[1].daughters(3,0);
    event.append(state[2]);
    event[2].daughters(4,0);
    event.append(state[in1]);
    event[3].mothers(1,0);
    event[3].status(-21);
    event.append(state[in2]);
    event[4].mothers(2,0);
    event[4].status(-21);
    for ( int i = 0; i < state.size(); ++i)
      if ( state[i].isFinal() 
        && (!hasSystems || partonSystemsPtr->getSystemOf(i) == iSys)) {
       int iN = event.append(state[i]);
       event[iN].daughters(0,0);
       event[iN].mothers(3,4);
       event[iN].status(23);
     }
    event[3].daughters(5,event.size()-1);
    event[4].daughters(5,event.size()-1);

    // Check if event is coloured
    for ( int i = 0; i < event.size(); ++i)
     // Check colour of quarks
     if ( event[i].isFinal() && event[i].colType() == 1
            // No corresponding anticolour in final state
         && ( FindCol(event[i].col(),i,event,1) == 0
            // No corresponding colour in initial state
           && FindCol(event[i].col(),i,event,2) == 0 )) {
       validColour = false;
       break;
     // Check anticolour of antiquarks
     } else if ( event[i].isFinal() && event[i].colType() == -1
            // No corresponding colour in final state
         && ( FindCol(event[i].acol(),i,event,2) == 0
            // No corresponding anticolour in initial state
           && FindCol(event[i].acol(),i,event,1) == 0 )) {
       validColour = false;
       break;
     // No uncontracted colour (anticolour) charge of gluons
     } else if ( event[i].isFinal() && event[i].colType() == 2
            // No corresponding anticolour in final state
         && ( FindCol(event[i].col(),i,event,1) == 0
            // No corresponding colour in initial state
           && FindCol(event[i].col(),i,event,2) == 0 )
            // No corresponding colour in final state
         && ( FindCol(event[i].acol(),i,event,2) == 0
            // No corresponding anticolour in initial state
           && FindCol(event[i].acol(),i,event,1) == 0 )) {
       validColour = false;
       break;
     }

    // Check charge sum in initial and final state
    double initCharge = event[3].charge() + event[4].charge();
    double finalCharge = 0.0;
    for(int i = 0; i < event.size(); ++i)
      if (event[i].isFinal()) finalCharge += event[i].charge();
    if (abs(initCharge-finalCharge) > 1e-12) validCharge = false;

    double mTolErr = settingsPtr->parm("Check:mTolErr");
    // Check if particles are on mass shell
    for ( int i = 0; i < event.size(); ++i) {
      if (i==3 || i==4 || event[i].isFinal()) {
        int id = event[i].id();
        // Get mass. Zero for incoming.
        double mNow = (i==3 || i==4) ? 0.
                    : ((abs(id) < 6) ? getMass(id,2) : getMass(id,1));
        mNow = sqrt(mNow);
        // Do not check on-shell condition for massive intermediate (!)
        // resonances. Assuming all non-SM particles are heavy here! 
        if ( abs(id) == 6 || abs(id) > 22) mNow = event[i].mCalc();
        double errMass = abs(event[i].mCalc() - mNow)
                       / max( 1.0, event[i].e());
        if ( errMass > mTolErr ) validMomenta = false;
      }
    }

    // Check that overall pT is vanishing.
    Vec4 pSum(0.,0.,0.,0.);
    for ( int i = 0; i < event.size(); ++i)
      if ( i ==3 || i == 4 || event[i].isFinal()) pSum += event[i].p();
    if ( abs(pSum.px()) > mTolErr || abs(pSum.py()) > mTolErr)
      validMomenta = false;

    if ( abs(event[3].px()) > mTolErr || abs(event[3].py()) > mTolErr
      || abs(event[4].px()) > mTolErr || abs(event[4].py()) > mTolErr)
      validMomenta = false;

  } // Done with loop over systems.

  return (validColour && validCharge && validMomenta);

}

//--------------------------------------------------------------------------

// Find index of colour partner for input colour.

int DireTimes::FindCol(int col, int iExclude, const Event& event, int type,
    int iSys) {

  int index = 0;

  int inA = 0, inB = 0;
  for (int i=event.size()-1; i > 0; --i) {
    if ( event[i].mother1() == 1 && event[i].status() != -31
      && event[i].status() != -34) { if (inA == 0) inA = i; }
    if ( event[i].mother1() == 2 && event[i].status() != -31
      && event[i].status() != -34) { if (inB == 0) inB = i; }
  }
  if (iSys >= 0) inA = partonSystemsPtr->getInA(iSys);
  if (iSys >= 0) inB = partonSystemsPtr->getInB(iSys);

  // Search event record for matching colour & anticolour
  for(int n = 0; n < event.size(); ++n) {
    if ( n != iExclude && event[n].colType() != 0
      &&  event[n].status() > 0 ) {
       if ( event[n].acol() == col ) {
        index = -n;
        break;
      }
      if ( event[n].col()  == col ) {
        index =  n;
        break;
      }
    }
  }
  // Search event record for matching colour & anticolour
  for(int n = event.size()-1; n > 0; --n) {
    if ( index == 0 && n != iExclude && event[n].colType() != 0
      && ( n == inA || n == inB) ) {  // Check incoming
       if ( event[n].acol() == col ) {
        index = -n;
        break;
      }
      if ( event[n].col()  == col ) {
        index =  n;
        break;
      }
    }
  }
  // if no matching colour / anticolour has been found, return false
  if ( type == 1 && index < 0) return abs(index);
  if ( type == 2 && index > 0) return abs(index);
  return 0;
}

//--------------------------------------------------------------------------

// Rescatter: If a dipole stretches between two different systems, those
//            systems will no longer locally conserve momentum. These
//            imbalances become problematic when ISR or primordial kT
//            is switched on as these steps involve Lorentz boosts.
//
//            'rescatterPropagateRecoil' tries to fix momentum in all
//            systems by propogating recoil momentum through all
//            intermediate systems. As the momentum transfer is already
//            defined, this can lead to internal lines gaining a
//            virtuality.

// Useful definitions for a pair of integers and a vector of pairs
typedef pair < int, int >  pairInt;
typedef vector < pairInt > vectorPairInt;

//--------------------------------------------------------------------------

// findParentSystems
//  Utility routine to find all parent systems of a given system
//  Returns a vector of pairs of integers with:
//   a) The system index, including the starting system (negative
//      if (b) points to a parent system, positive if (b) points
//      to a daughter system
//   b) The event record index that is the path out of the system
//      (if forwards == false, this is an incoming parton to the
//      system, and is +ve if side A or -ve if side B,
//      if forwards == true, this is an outgoing parton from the
//      system).
//  Returns as empty vector on failure
//  Note: this assumes single rescattering only and therefore only
//        one possible parent system

inline vectorPairInt findParentSystems(const int sys,
  Event& event, PartonSystems* partonSystemsPtr, bool forwards) {

  vectorPairInt parentSystems;
  parentSystems.reserve(10);

  int iSysCur = sys;
  while (true) {
    // Get two incoming partons
    int iInA = partonSystemsPtr->getInA(iSysCur);
    int iInB = partonSystemsPtr->getInB(iSysCur);

    // Check if either of these links to another system
    int iIn = 0;
    if (event[iInA].isRescatteredIncoming()) iIn =  iInA;
    if (event[iInB].isRescatteredIncoming()) iIn = -iInB;

    // Save the current system to the vector
    parentSystems.push_back( pairInt(-iSysCur, iIn) );
    if (iIn == 0) break;

    int iInAbs  = abs(iIn);
    int iMother = event[iInAbs].mother1();
    iSysCur     = partonSystemsPtr->getSystemOf(iMother);
    if (iSysCur == -1) {
      parentSystems.clear();
      break;
    }
  } // while (true)

  // If forwards is set, change all event record indices to go to daughter
  // systems rather than parent systems
  if (forwards) {
    vectorPairInt::reverse_iterator rit;
    for (rit = parentSystems.rbegin(); rit < (parentSystems.rend() - 1);
         ++rit) {
      pairInt &cur  = *rit;
      pairInt &next = *(rit + 1);
      cur.first     = -cur.first;
      cur.second    = (next.second < 0) ? -event[abs(next.second)].mother1() :
                                           event[abs(next.second)].mother1();
    }
  }

  return parentSystems;
}

//--------------------------------------------------------------------------

// Print the list of dipoles.

void DireTimes::list(ostream& os) const {

  // Header.
  os << "\n --------  PYTHIA DireTimes Dipole Listing  ----------------"
     << "------------------------------------------------------- \n \n  "
     << "  i    rad    rec       pTmax  col  chg  gam weak  oni   hv  is"
     << "r  sys sysR type  MErec     mix  ord  spl  ~gR  pol \n"
     << fixed << setprecision(3);

  // Loop over dipole list and print it.
  for (int i = 0; i < int(dipEnd.size()); ++i)
  os << setw(5) << i                     << setw(7) << dipEnd[i].iRadiator
     << setw(7) << dipEnd[i].iRecoiler   << setw(12) << dipEnd[i].pTmax
     << setw(5) << dipEnd[i].colType     << setw(5) << dipEnd[i].chgType
     << setw(5) << dipEnd[i].gamType     << setw(5) << dipEnd[i].weakType
     << setw(5) << dipEnd[i].isOctetOnium
     << setw(5) << dipEnd[i].isHiddenValley << setw(5) << dipEnd[i].isrType
     << setw(5) << dipEnd[i].system      << setw(5) << dipEnd[i].systemRec
     << setw(5) << dipEnd[i].MEtype      << setw(7) << dipEnd[i].iMEpartner
     << setw(8) << dipEnd[i].MEmix       << setw(5) << dipEnd[i].MEorder
     << setw(5) << dipEnd[i].MEsplit     << setw(5) << dipEnd[i].MEgluinoRec
     << setw(5) << dipEnd[i].weakPol << "\n";

  // Done.
  os << "\n --------  End PYTHIA DireTimes Dipole Listing  ------------"
     << "-------------------------------------------------------" << endl;

}

//==========================================================================

} // end namespace Pythia8
