
#include "../include/SplittingLibrary.h"

namespace Pythia8 {
 
//==========================================================================

// class SplittingLibrary

//--------------------------------------------------------------------------

// Clean up
void SplittingLibrary::clear() {
  for (int i=0; i < int(splittings.size()); ++i)
    if (splittings[i]) delete splittings[i];
  splittings.resize(0);
}

//--------------------------------------------------------------------------

// Initialisation.
void SplittingLibrary::init(Settings* settings, ParticleData* particleData,
  Rndm* rndm, BeamParticle* beamA, BeamParticle* beamB) {

  // Store infrastructure pointers.
  settingsPtr     = settings;
  particleDataPtr = particleData;
  rndmPtr         = rndm;
  beamAPtr        = beamA;
  beamBPtr        = beamB;

  // Initialise splitting names.
  splittings.resize(0);
  initISR();
  initFSR();

  // Done.
}

//--------------------------------------------------------------------------

void SplittingLibrary::initFSR() {

  /********************* FSR ******************************************/

  // Add corrections to the LO splitting kernel.
  // softRescale = 0 -> A1, B1
  // softRescale = 1 -> A1, A2, B1
  // softRescale = 2 -> A1, A2, A3, B1
  // softRescale = 3 -> A1, A2, A3, B1, B2
  int softRescaleGluon = 1;
  int softRescaleQuark = 1;

  // Q -> Q G, soft part + collinear
  string name="fsr_1->1&21_CS";
  splittings.push_back( new fsr_Q2QG( name, softRescaleGluon, settingsPtr,
    particleDataPtr, rndmPtr, beamAPtr, beamBPtr) );

  // G -> G G
  name="fsr_21->21&21_CS";
  splittings.push_back( new fsr_G2GG( name, softRescaleGluon, settingsPtr,
    particleDataPtr, rndmPtr, beamAPtr, beamBPtr) );

  // G -> Q Q (regular DGLAP kernel)
  name="fsr_21->1&1a_CS";
  splittings.push_back( new fsr_G2QQ1( name, softRescaleQuark, settingsPtr,
    particleDataPtr, rndmPtr, beamAPtr, beamBPtr) );

  // G -> Q Q (regular DGLAP kernel)
  name="fsr_21->1&1b_CS";
  splittings.push_back( new fsr_G2QQ2( name, softRescaleQuark, settingsPtr,
    particleDataPtr, rndmPtr, beamAPtr, beamBPtr) );

  // Done.
}

//--------------------------------------------------------------------------

void SplittingLibrary::initISR() {

  /********************* ISR off Drell-Yan ****************************/

  // Add corrections to the LO splitting kernel.
  // softRescale = 0 -> A1, B1
  // softRescale = 1 -> A1, A2, B1
  // softRescale = 2 -> A1, A2, A3, B1
  // softRescale = 3 -> A1, A2, A3, B1, B2
  int softRescaleGluon = 1;
  int softRescaleQuark = 1;

  // Q -> Q G, soft and collinear part.
  string name="isr_1->1&21_CS";
  splittings.push_back( new isr_Q2QG( name, softRescaleGluon, settingsPtr,
    particleDataPtr, rndmPtr, beamAPtr, beamBPtr) );

  // G -> Q Q~ (regular DGLAP kernel)
  name="isr_21->1&1_CS";
  splittings.push_back( new isr_G2QQ( name, softRescaleQuark, settingsPtr,
    particleDataPtr, rndmPtr, beamAPtr, beamBPtr) );

  /********************* ISR off Higgs ********************************/

  // G -> G G, soft part.
  name="isr_21->21&21a_CS";
  splittings.push_back( new isr_G2GG1( name, softRescaleGluon, settingsPtr,
    particleDataPtr, rndmPtr, beamAPtr, beamBPtr) );

  name="isr_21->21&21b_CS";
  splittings.push_back( new isr_G2GG2( name, softRescaleGluon, settingsPtr,
    particleDataPtr, rndmPtr, beamAPtr, beamBPtr) );

  // Q -> G Q (regular DGLAP kernel)
  name="isr_1->21&1_CS";
  splittings.push_back( new isr_Q2GQ( name, softRescaleQuark, settingsPtr,
    particleDataPtr, rndmPtr, beamAPtr, beamBPtr) );

  // Done.
}

//--------------------------------------------------------------------------

// Overload index operator to access element of splitting vector.

Splitting* SplittingLibrary::operator[](string id) {
  int iPos = -1;
  for ( int i = 0; i < int(splittings.size()); ++i) {
    if (splittings[i]->name().compare(id) == 0) {
      iPos = i;
        break;
    }
  }
  return (iPos > -1) ? splittings[iPos] : NULL;
}

const Splitting* SplittingLibrary::operator[](string id) const {
  int iPos = -1;
  for ( int i = 0; i < int(splittings.size()); ++i) {
    if (splittings[i]->name().compare(id) == 0) {
      iPos = i;
      break;
    }
  }
  return (iPos > -1) ? splittings[iPos] : NULL;
}

//--------------------------------------------------------------------------

// Generate identifier of radiator before the splitting

int SplittingLibrary::getSplittingRadBefID(const Event& event, int rad, int emt) {

  // Get ID of radiator before branching.
  int type = event[rad].isFinal() ? 1 :-1;
  int emtID  = event[emt].id();
  int radID  = event[rad].id();
  int emtCOL = event[emt].col();
  int radCOL = event[rad].col();
  int emtACL = event[emt].acol();
  int radACL = event[rad].acol();

  bool colConnected = ((type == 1) && ( (emtCOL !=0 && (emtCOL ==radACL))
                                     || (emtACL !=0 && (emtACL ==radCOL)) ))
                    ||((type ==-1) && ( (emtCOL !=0 && (emtCOL ==radCOL))
                                     || (emtACL !=0 && (emtACL ==radACL)) ));

  // Electroweak splittings splittings
  // Photon / Z radiation: Calculate invariant mass of system
  double m2final = (event[rad].p()+ event[emt].p()).m2Calc();

  // SQCD bookkeeping
  int radSign = (radID < 0) ? -1 : 1;
  int offsetL = 1000000;
  int offsetR = 2000000;
  int emtSign = (emtID < 0) ? -1 : 1;
  // Get PDG numbering offsets.
  int emtOffset = 0;
  if ( abs(emtID) > offsetL && abs(emtID) < offsetL+10 )
    emtOffset = offsetL;
  if ( abs(emtID) > offsetR && abs(emtID) < offsetR+10 )
    emtOffset = offsetR;
  int radOffset = 0;
  if ( abs(radID) > offsetL && abs(radID) < offsetL+10 )
    radOffset = offsetL;
  if ( abs(radID) > offsetR && abs(radID) < offsetR+10 )
    radOffset = offsetR;

  int radBefID = 0;
  // QCD splittings
  // Gluon radiation
  if ( emtID == 21 ) {
    radBefID = radID;
  // Final state gluon splitting
  } else if ( type == 1 && emtID == -radID && !colConnected ) {
    radBefID = 21;
  // Initial state s-channel gluon splitting
  } else if ( type ==-1 && radID == 21 ) {
    radBefID = -emtID;
  // Initial state t-channel gluon splitting
  } else if ( type ==-1 && !colConnected
  && emtID != 21 && radID != 21 && abs(emtID) < 10 && abs(radID) < 10) {
    radBefID = 21;

  // Gluino radiation
  } else if ( emtID == 1000021 ) {
    // Gluino radiation combined with quark yields squark.
    if (abs(radID) < 10 ) {
      int offset = offsetL;
      // Check if righthanded squark present. If so, make the reclustered
      // squark match. Works for squark pair production + gluino emission.
      for (int i=0; i < int(event.size()); ++i)
        if ( event[i].isFinal()
          && event[i].idAbs() < offsetR+10 && event[i].idAbs() > offsetR)
          offset = offsetR;
      radBefID = radSign*(abs(radID)+offset);
    // Gluino radiation combined with squark yields quark.
    } else if (abs(radID) > offsetL && abs(radID) < offsetL+10 ) {
      radBefID = radSign*(abs(radID)-offsetL);
    } else if (abs(radID) > offsetR && abs(radID) < offsetR+10 ) {
      radBefID = radSign*(abs(radID)-offsetR);
    // Gluino radiation off gluon yields gluino.
    } else if (radID == 21 ) {
      radBefID = emtID;
    }

  // Final state gluino splitting
  } else if ( type == 1 && !colConnected ) {
    // Emitted squark, radiating quark.
    if ( emtOffset > 0 && radOffset == 0
      && emtSign*(abs(emtID) - emtOffset) == -radID ) {
      radBefID = 1000021;
    // Emitted quark, radiating squark.
    } else if ( emtOffset == 0 && radOffset > 0
      && emtID == -radSign*(abs(radID) - radOffset) ) {
      radBefID = 1000021;
    }
  // Initial state s-channel gluino splitting
  } else if ( type ==-1 && radID == 1000021 ) {
    // Quark entering underlying hard process.
    if ( emtOffset > 0 ) {
      radBefID = -emtSign*(abs(emtID) - emtOffset);
    // Squark entering underlying hard process.
    } else {
      radBefID = -emtSign*(abs(emtID) + emtOffset);
    }
  // Initial state t-channel gluino splitting.
  } else if ( type ==-1
    && ( (abs(emtID) > offsetL && abs(emtID) < offsetL+10)
      || (abs(emtID) > offsetR && abs(emtID) < offsetR+10))
    && ( (abs(radID) > offsetL && abs(radID) < offsetL+10)
      || (abs(radID) > offsetR && abs(radID) < offsetR+10))
    && emtSign*(abs(emtID)+emtOffset) == radSign*(abs(radID) - radOffset)
    && !colConnected ) {
    radBefID = 1000021;
  // Photon/Z radiation
  } else if ( emtID == 22 || emtID == 23 ) {
    radBefID = radID;
  // Final state Photon splitting
  } else if ( type == 1 && emtID == -radID && colConnected && sqrt(m2final) <= 10. ){
    radBefID = 22;
  // Final state Z splitting
  } else if ( type == 1 && emtID == -radID && colConnected && sqrt(m2final)  > 10. ){
    radBefID = 23;
  // Initial state s-channel photon/ Z splitting
  } else if ( type ==-1 && (radID == 22 || radID == 23) ){
    radBefID = -emtID;
  // Initial state t-channel photon / Z splitting: Always bookkeep as photon
  } else if ( type ==-1 && abs(emtID) < 10 && abs(radID) < 10 && colConnected ){
    radBefID = 22;
  // W+ radiation
  // Final state W+ splitting
  } else if ( emtID == 24 && radID < 0 ) {
    radBefID = radID + 1;
  } else if ( emtID == 24 && radID > 0 ) {
    radBefID = radID + 1;
  // W- radiation
  // Final state W- splitting
  } else if ( emtID ==-24 && radID < 0 ) {
    radBefID = radID - 1;
  } else if ( emtID ==-24 && radID > 0 ) {
    radBefID = radID - 1;
  } else
    radBefID = 0;

  // Done.
  return radBefID;

}

//--------------------------------------------------------------------------

// Generate name for a splitting

string SplittingLibrary::getSplittingName(const Event& event, int rad, int emt) {

  // Get ID of radiator before branching.
  int type = event[rad].isFinal() ? 1 :-1;
  int radBefID = getSplittingRadBefID(event, rad, emt);
  if (radBefID == 0) return "";

  // Construct name.
  string name = "";
  if (type ==  1) name +="fsr_";
  else name +="isr_";
  int radAftID = event[rad].id();
  if (type == -1) swap(radBefID,radAftID);
  int emtAftID = event[emt].id();

  int rbID = (abs(radBefID) < 10) ? 1 : radBefID; 
  stringstream ss; ss << rbID;
  name += ss.str(); name +="->";

  int rID = (abs(radAftID) < 10) ? 1 : radAftID; 
  ss.str(""); ss << rID;
  name += ss.str(); name +="&";

  int eID = (abs(emtAftID) < 10) ? 1 : emtAftID; 
  ss.str(""); ss << eID;
  name += ss.str();


  // Distinguish FSR g->qq~ and g->q~q
  if ( name.compare("fsr_21->1&1") == 0 && event[emt].id() > 0)
    name += "a";
  if ( name.compare("fsr_21->1&1") == 0 && event[emt].id() < 0)
    name += "b";

  // Two colour structures for ISR g->gg --> Sum later.
  if ( name.compare("isr_21->21&21") == 0 )
    name += "a";

  name += "_CS";

  // Loop through splittings to check if the name exists.
  bool foundName = false;
  for (int i = 0; i < int(splittings.size()); ++i)
    if ( splittings[i]->name().compare(name) == 0)
      foundName = true;

  // Done.
  return (foundName ? name : "");

}

//==========================================================================

} // end namespace Pythia8
