
#include "../include/Splittings.h"
#include "../include/DireSpace.h"
#include "../include/DireTimes.h"

namespace Pythia8 {

// Function to turn off splittings - for debugging only!

bool doSplit(string name) {

  if (      name.compare("fsr_Q2QG_CS")  == 0) return true;
  else if ( name.compare("fsr_G2GG_CS")  == 0) return true;
  else if ( name.compare("fsr_G2QQ1_CS") == 0) return true;
  else if ( name.compare("fsr_G2QQ2_CS") == 0) return true;

  else if ( name.compare("isr_Q2QG_CS")  == 0) return true;
  else if ( name.compare("isr_G2GG1_CS") == 0) return true;
  else if ( name.compare("isr_G2GG2_CS") == 0) return true;
  else if ( name.compare("isr_G2QQ_CS")  == 0) return true;
  else if ( name.compare("isr_Q2GQ_CS")  == 0) return true;
  else return false;

}

//==========================================================================

// The Splitting class.

//--------------------------------------------------------------------------

void Splitting::init() {

  CA = 3.;
  TR = 0.5;
  CF = 4./3.;

  NF_fsr = settingsPtr->mode("TimeShower:nGluonToQuark");
  //NF_fsr = 1;

  // Parameters of alphaS.
  double alphaSvalue = settingsPtr->parm("SpaceShower:alphaSvalue");
  int alphaSorder    = settingsPtr->mode("SpaceShower:alphaSorder");
  int alphaSnfmax    = settingsPtr->mode("StandardModel:alphaSnfmax");
  bool alphaSuseCMW  = settingsPtr->flag("SpaceShower:alphaSuseCMW");
  // Initialize alphaS.
  alphaS.init( alphaSvalue, alphaSorder, alphaSnfmax, alphaSuseCMW);

  // Parameters of alphaEM.
  int alphaEMorder = settingsPtr->mode("SpaceShower:alphaEMorder");
  // Initialize alphaiEM.
  alphaEM.init( alphaEMorder, settingsPtr);

  double pTmin       = settingsPtr->parm("SpaceShower:pTmin");
  usePDFalphaS       = settingsPtr->flag("ShowerPDF:usePDFalphas");
  alphaS2pi          = (alphaSorder > 0)
                     ? alphaS.alphaS(0.5*pow2(pTmin)) * 0.5 / M_PI
                     : 0.5 * 0.5 / M_PI;

}

//--------------------------------------------------------------------------

double Splitting::Gamma2(double NF) {
  return (67./18.-pow2(M_PI)/6.)*CA - 10./9.*NF*TR;
}

//--------------------------------------------------------------------------

double Splitting::Gamma3(double NF) {
  return 1./4.* (CA*CA*(245./6.-134./27.*pow2(M_PI)+11./45.*pow(M_PI,4)
                        +22./3.*ZETA3)
                +CA*NF*TR*(-418./27.+40./27.*pow2(M_PI)-56./3.*ZETA3)
                +CF*NF*TR*(-55./3.+16.*ZETA3)-16./27.*pow2(NF*TR));
}

//--------------------------------------------------------------------------

double Splitting::beta0(double NF)
  { return 11./6.*CA - 2./3.*NF*TR;}

//--------------------------------------------------------------------------

double Splitting::beta1(double NF)
  { return 17./3.*pow2(CA) - (10./3.*CA+2.*CF)*NF*TR;}

//--------------------------------------------------------------------------

double Splitting::softRescaleInt(int order) {
  double rescale = 1.;
  if (order > 0) rescale += alphaS2pi*Gamma2(3.);
  if (order > 1) rescale += pow2(alphaS2pi)*Gamma3(3.);
  return rescale;
}

//--------------------------------------------------------------------------

double Splitting::softRescaleDiff(int order, double pT2) {
  double rescale = 1.;
  double nQuarks = 6.;

  if ( !usePDFalphaS
    || (abs(beamAPtr->id()) < 30 && abs(beamBPtr->id()) < 30) ) {
    if ( pT2 > pow2( max(0., particleDataPtr->m0(5) ) )
      && pT2 < pow2( particleDataPtr->m0(6)) )             nQuarks = 5.;
    else if ( pT2 > pow2( max( 0., particleDataPtr->m0(4)) ) )  nQuarks = 4.; 
    else if ( pT2 > pow2( max( 0., particleDataPtr->m0(3)) ) )  nQuarks = 3.; 
    if (order > 0) rescale += 0.5/M_PI * alphaS.alphaS(pT2) * Gamma2(nQuarks);
    if (order > 1) rescale += pow2(0.5/M_PI * alphaS.alphaS(pT2)) *Gamma3(nQuarks);
  } else {
    BeamParticle& beam = (abs(beamAPtr->id()) == 2212) ? *beamAPtr : *beamBPtr;
    if ( pT2 > pow2( max(0., beam.mQuarkPDF(5) ) )
      && pT2 < pow2( particleDataPtr->m0(6)) )             nQuarks = 5.;
    else if ( pT2 > pow2( max( 0., beam.mQuarkPDF(4)) ) )  nQuarks = 4.; 
    else if ( pT2 > pow2( max( 0., beam.mQuarkPDF(3)) ) )  nQuarks = 3.; 
    if (order > 0) rescale += 0.5/M_PI * beam.alphaS(pT2) * Gamma2(nQuarks);
    if (order > 1) rescale += pow2(0.5/M_PI * beam.alphaS(pT2)) *Gamma3(nQuarks);
  }

  return rescale;

}

//==========================================================================

// Class inheriting from Splitting class.

// Splitting function Q->QG (FSR)

string fsr_Q2QG::name()                { return (id.empty()) ? "fsr_Q2QG_CS" : id;}
int fsr_Q2QG::type()                   { return 101;}
int fsr_Q2QG::motherID(int idDaughter) { return idDaughter;}
int fsr_Q2QG::sisterID(int)            { return 21;}
double fsr_Q2QG::gaugeFactor ()        { return CF;}
double fsr_Q2QG::symmetryFactor ()     { return 1.;}

int fsr_Q2QG::radID(int idRA, int){ return idRA;}
pair<int,int> fsr_Q2QG::radCols(
  int colRadAfter, int, 
  int colEmtAfter, int acolEmtAfter) {
  bool isQuark = (colRadAfter > 0);
  if (isQuark) return make_pair(colEmtAfter,0);
  return make_pair(0,acolEmtAfter);
}

// Pick z for new splitting.
double fsr_Q2QG::zSplit(double zMinAbs, double, double m2dip) {
  double Rz = rndmPtr->flat();
  double kappa2 = pow2(settingsPtr->parm("TimeShower:pTmin"))/m2dip;
  double p = pow( 1. + pow2(1-zMinAbs)/kappa2, Rz );
  double res = 1. - sqrt( p - 1. )*sqrt(kappa2);
  return res;
}

// New overestimates, z-integrated versions.
double fsr_Q2QG::overestimateInt(double zMinAbs, double,
  double, double m2dip) {
  double wt = 0.;
  double preFac = symmetryFactor() * gaugeFactor();
  // Q -> QG, soft part (currently also used for collinear part).
  double kappa2 = pow2(settingsPtr->parm("TimeShower:pTmin"))/m2dip;
  wt  = preFac * softRescaleInt(correctionOrder)
      * 2. * 0.5 * log( 1. + pow2(1.-zMinAbs)/kappa2);

  if (!doSplit("fsr_Q2QG_CS")) wt = 0.;

  return wt;
}

// Return overestimate for new splitting.
double fsr_Q2QG::overestimateDiff(double z, double m2dip) {
  double wt = 0.;
  double preFac = symmetryFactor() * gaugeFactor();
  double kappaOld2 = pow2(settingsPtr->parm("TimeShower:pTmin"))/m2dip;
  wt  = preFac * softRescaleInt(correctionOrder)
      * 2.* (1.-z) / ( pow2(1.-z) + kappaOld2);
  return wt;
}

// Return kernel for new splitting.
double fsr_Q2QG::kernel(double z, double pT2, double m2dip,
  int splitType, double m2RadBef, double m2Rad, double m2Rec,
  double m2Emt ) {

  double wt = 0.;
  double preFac = symmetryFactor() * gaugeFactor();
  double kappa2 = pT2/m2dip;
  wt   = preFac * softRescaleDiff(correctionOrder, pT2)
       * ( 2.* (1.-z) / ( pow2(1.-z) + kappa2) );

  // Correction for massive splittings.
  bool doMassive = (abs(splitType) == 2);

  // Add collinear term for massless splittings.
  if (!doMassive) wt  += -preFac * ( 1.+z );

  // Add collinear term for massive splittings.
  else {

    double pipj = 0., vijkt = 1., vijk = 1.;

    // splitType == 2 -> Massive FF
    if (splitType == 2) {

      // Calculate CS variables.
      double yCS = kappa2 / (1.-z);
      double nu2RadBef = m2RadBef/m2dip; 
      double nu2Rad = m2Rad/m2dip; 
      double nu2Emt = m2Emt/m2dip; 
      double nu2Rec = m2Rec/m2dip; 
      vijk          = pow2(1.-yCS) - 4.*(yCS+nu2Rad+nu2Emt)*nu2Rec;
      double Q2mass = m2dip + m2Rad + m2Rec + m2Emt;
      vijkt         = pow2(Q2mass/m2dip - nu2RadBef - nu2Rec)
                    - 4.*nu2RadBef*nu2Rec;
      vijk          = sqrt(vijk) / (1-yCS);
      vijkt         = sqrt(vijkt)/ (Q2mass/m2dip - nu2RadBef - nu2Rec);
      pipj          = m2dip * yCS/2.;
    // splitType ==-2 -> Massive FI
    } else if (splitType ==-2) {

      // Calculate CS variables.
      double xCS = 1 - kappa2/(1.-z);
      vijk   = 1.; 
      vijkt  = 1.;
      pipj   = m2dip/2. * (1-xCS)/xCS;
    }

    // Add B1 for massive splittings.
    double massCorr = -1.*vijkt/vijk*( 1. + z + m2RadBef/pipj);
    wt += preFac*massCorr;

  }

  if (!doSplit("fsr_Q2QG_CS")) wt = 0.;

  return wt;
}

//==========================================================================

// Class inheriting from Splitting class.

// Splitting function G->GG (FSR)

string fsr_G2GG::name()                { return (id.empty()) ? "fsr_G2GG_CS" : id;}
int fsr_G2GG::type()                   { return 201;}
int fsr_G2GG::motherID(int)            { return 21;}
int fsr_G2GG::sisterID(int)            { return 21;}
double fsr_G2GG::gaugeFactor ()        { return 2.*CA;}
double fsr_G2GG::symmetryFactor ()     { return 0.5;}

int fsr_G2GG::radID(int, int){ return 21;}
pair<int,int> fsr_G2GG::radCols(
  int colRadAfter, int acolRadAfter, 
  int colEmtAfter, int acolEmtAfter) {
  int colRemove = (colRadAfter == acolEmtAfter)
                ? colRadAfter : acolRadAfter;
  int col       = (colRadAfter == colRemove)
                ? colEmtAfter : colRadAfter;
  int acol      = (acolRadAfter == colRemove)
                ? acolEmtAfter : acolRadAfter;
  return make_pair(col,acol);
}

// Pick z for new splitting.
double fsr_G2GG::zSplit(double zMinAbs, double, double m2dip) {
  double R = rndmPtr->flat();
  double kappa2 = pow2(settingsPtr->parm("TimeShower:pTmin"))/m2dip;
  // Just pick according to soft.
  double p = pow( 1. + pow2(1-zMinAbs)/kappa2, R );
  double res = 1. - sqrt( p - 1. )*sqrt(kappa2);
  return res;
}

// New overestimates, z-integrated versions.
double fsr_G2GG::overestimateInt(double zMinAbs, double,
  double, double m2dip) {
  double wt = 0.;
  double preFac = symmetryFactor() * gaugeFactor();
  double kappa2 = pow2(settingsPtr->parm("TimeShower:pTmin"))/m2dip;
  // Overestimate by soft
  wt   = preFac * softRescaleInt(correctionOrder)
       * 0.5 * log( 1. + pow2(1.-zMinAbs)/kappa2);
  if (!doSplit("fsr_G2GG_CS")) wt = 0.;

  return wt;
}

// Return overestimate for new splitting.
double fsr_G2GG::overestimateDiff(double z, double m2dip) {
  double wt = 0.;
  double preFac = symmetryFactor() * gaugeFactor();
  double kappaOld2 = pow2(settingsPtr->parm("TimeShower:pTmin"))/m2dip;
  // Overestimate by soft
  wt  = preFac * softRescaleInt(correctionOrder)
      * (1.-z) / ( pow2(1.-z) + kappaOld2);
  return wt;
}

// Return kernel for new splitting.
double fsr_G2GG::kernel(double z, double pT2, double m2dip,
  int splitType, double m2RadBef, double m2Rad, double m2Rec,
  double m2Emt) {

  // Dummy statement to avoid compiler warnings.
  if (false) cout << m2RadBef << endl;

  double wt = 0.;
  double preFac = symmetryFactor() * gaugeFactor();
  double kappa2 = pT2/m2dip;

  wt   = preFac * softRescaleDiff(correctionOrder, pT2)
       * (1.-z) / ( pow2(1.-z) + kappa2);

  // Correction for massive splittings.
  bool doMassive = (abs(splitType) == 2);

  // Add collinear term for massless splittings.
  if (!doMassive) wt += preFac * ( -1. + 0.5 * z*(1.-z) );

  // Add collinear term for massive splittings.
  else {

    double vijk = 1.;

    // splitType == 2 -> Massive FF
    if (splitType == 2) {
      // Calculate CS variables.
      double yCS = kappa2 / (1.-z);
      double nu2Rad = m2Rad/m2dip; 
      double nu2Emt = m2Emt/m2dip; 
      double nu2Rec = m2Rec/m2dip; 
      vijk          = pow2(1.-yCS) - 4.*(yCS+nu2Rad+nu2Emt)*nu2Rec;
      vijk          = sqrt(vijk) / (1-yCS);

    // splitType ==-2 -> Massive FI
    } else if (splitType ==-2) {
      // No changes, as initial recoiler is massless!
      vijk          = 1.; 
    }

    // Add correction for massive splittings.
    wt += preFac * 1./ vijk * ( -1. + 0.5 * z*(1.-z) );
  }

  if (!doSplit("fsr_G2GG_CS")) wt = 0.;

  return wt;
}

//==========================================================================

// Class inheriting from Splitting class.

// Splitting function G->QQ (FSR)

string fsr_G2QQ1::name()                { return (id.empty()) ? "fsr_G2QQ1_CS" : id;}
int fsr_G2QQ1::type()                   { return 201;}
int fsr_G2QQ1::motherID(int)            { return 1;} // Use 1 as dummy variable.
int fsr_G2QQ1::sisterID(int)            { return 1;} // Use 1 as dummy variable.
double fsr_G2QQ1::gaugeFactor ()        { return NF_fsr*TR;}
double fsr_G2QQ1::symmetryFactor ()     { return 0.5;}

int fsr_G2QQ1::radID(int, int){ return 21;}
pair<int,int> fsr_G2QQ1::radCols(
  int colRadAfter, int acolRadAfter, 
  int colEmtAfter, int acolEmtAfter) {
  int col  = (colRadAfter  > 0) ? colRadAfter  : colEmtAfter;
  int acol = (acolRadAfter > 0) ? acolRadAfter : acolEmtAfter;
  return make_pair(col,acol);
}

// Pick z for new splitting.
double fsr_G2QQ1::zSplit(double zMinAbs, double zMaxAbs, double) {
  return (zMinAbs + rndmPtr->flat() * (zMaxAbs - zMinAbs));
}

// New overestimates, z-integrated versions.
double fsr_G2QQ1::overestimateInt(double zMinAbs,double zMaxAbs,
  double, double) {
  double wt = 0.;
  double preFac = symmetryFactor() * gaugeFactor();
  wt  = preFac * 0.5 * ( zMaxAbs - zMinAbs);
  if (!doSplit("fsr_G2QQ1_CS")) wt = 0.;

  return wt;
}

// Return overestimate for new splitting.
double fsr_G2QQ1::overestimateDiff(double, double) {
  double wt = 0.;
  double preFac = symmetryFactor() * gaugeFactor();
  wt  = preFac * 0.5;
  return wt;
}

// Return kernel for new splitting.
double fsr_G2QQ1::kernel(double z, double pT2, double m2dip,
  int splitType, double m2RadBef, double m2Rad, double m2Rec,
  double m2Emt) {

  // Dummy statement to avoid compiler warnings.
  if (false) cout << m2RadBef << endl;

  double wt = 0.;
  double preFac = symmetryFactor() * gaugeFactor();
  wt  = preFac * 0.5
      * (pow(1.-z,2.) + pow(z,2.));

  // Correction for massive splittings.
  bool doMassive = (abs(splitType) == 2);

  if (doMassive) {

    double kappa2 = pT2/m2dip;
    double vijk = 1., pipj = 0.;

    // splitType == 2 -> Massive FF
    if (splitType == 2) {
      // Calculate CS variables.
      double yCS = kappa2 / (1.-z);
      double nu2Rad = m2Rad/m2dip; 
      double nu2Emt = m2Emt/m2dip; 
      double nu2Rec = m2Rec/m2dip; 
      vijk          = pow2(1.-yCS) - 4.*(yCS+nu2Rad+nu2Emt)*nu2Rec;
      vijk          = sqrt(vijk) / (1-yCS);
      pipj          = m2dip * yCS /2.;

    // splitType ==-2 -> Massive FI
    } else if (splitType ==-2) {
      // Calculate CS variables.
      double xCS = 1 - kappa2/(1.-z);
      vijk   = 1.; 
      pipj   = m2dip/2. * (1-xCS)/xCS;
    }

    // Reset kernel for massive splittings.
    wt = preFac * 0.5 * 1. / vijk * ( pow2(1.-z) + pow2(z)
                                    + m2Emt / ( pipj + m2Emt) );  
  }

  if (!doSplit("fsr_G2QQ1_CS")) wt = 0.;

  return wt;
}

//==========================================================================

// Class inheriting from Splitting class.

// Splitting function G->QQ (FSR)

string fsr_G2QQ2::name()                { return (id.empty()) ? "fsr_G2QQ2_CS" : id;}
int fsr_G2QQ2::type()                   { return 201;}
int fsr_G2QQ2::motherID(int)            { return -1;} // Use -1 as dummy variable.
int fsr_G2QQ2::sisterID(int)            { return -1;} // Use -1 as dummy variable.
double fsr_G2QQ2::gaugeFactor ()        { return NF_fsr*TR;}
double fsr_G2QQ2::symmetryFactor ()     { return 0.5;}

int fsr_G2QQ2::radID(int, int){ return 21;}
pair<int,int> fsr_G2QQ2::radCols(
  int colRadAfter, int acolRadAfter, 
  int colEmtAfter, int acolEmtAfter) {
  int col  = (colRadAfter  > 0) ? colRadAfter  : colEmtAfter;
  int acol = (acolRadAfter > 0) ? acolRadAfter : acolEmtAfter;
  return make_pair(col,acol);
}

// Pick z for new splitting.
double fsr_G2QQ2::zSplit(double zMinAbs, double zMaxAbs, double) {
  return (zMinAbs + rndmPtr->flat() * (zMaxAbs - zMinAbs));
}

// New overestimates, z-integrated versions.
double fsr_G2QQ2::overestimateInt(double zMinAbs,double zMaxAbs,
  double, double) {
  double wt = 0.;
  double preFac = symmetryFactor() * gaugeFactor();
  wt  = preFac * 0.5 * ( zMaxAbs - zMinAbs);
  if (!doSplit("fsr_G2QQ2_CS")) wt = 0.;

  return wt;
}

// Return overestimate for new splitting.
double fsr_G2QQ2::overestimateDiff(double, double) {
  double wt = 0.;
  double preFac = symmetryFactor() * gaugeFactor();
  wt  = preFac * 0.5;
  return wt;
}

// Return kernel for new splitting.
double fsr_G2QQ2::kernel(double z, double pT2, double m2dip,
  int splitType, double m2RadBef, double m2Rad, double m2Rec,
  double m2Emt) {

  // Dummy statement to avoid compiler warnings.
  if (false) cout << m2RadBef << endl;

  double wt = 0.;
  double preFac = symmetryFactor() * gaugeFactor();
  wt  = preFac * 0.5
      * (pow(1.-z,2.) + pow(z,2.));

  // Correction for massive splittings.
  bool doMassive = (abs(splitType) == 2);

  if (doMassive) {

    double kappa2 = pT2/m2dip;
    double vijk = 1., pipj = 0.;

    // splitType == 2 -> Massive FF
    if (splitType == 2) {
      // Calculate CS variables.
      double yCS = kappa2 / (1.-z);
      double nu2Rad = m2Rad/m2dip; 
      double nu2Emt = m2Emt/m2dip; 
      double nu2Rec = m2Rec/m2dip; 
      vijk          = pow2(1.-yCS) - 4.*(yCS+nu2Rad+nu2Emt)*nu2Rec;
      vijk          = sqrt(vijk) / (1-yCS);
      pipj          = m2dip * yCS /2.;

    // splitType ==-2 -> Massive FI
    } else if (splitType ==-2) {
      // Calculate CS variables.
      double xCS = 1 - kappa2/(1.-z);
      vijk   = 1.; 
      pipj   = m2dip/2. * (1-xCS)/xCS;
    }

    // Reset kernel for massive splittings.
    wt = preFac * 0.5 * 1. / vijk * ( pow2(1.-z) + pow2(z)
                                    + m2Emt / ( pipj + m2Emt) );  
  }

  if (!doSplit("fsr_G2QQ2_CS")) wt = 0.;

  return wt;
}

//==========================================================================

// Class inheriting from Splitting class.

// Splitting function Q->QG (ISR)

string isr_Q2QG::name()                { return (id.empty()) ? "isr_Q2QG_CS" : id;}
int isr_Q2QG::type()                   { return -101;}
int isr_Q2QG::motherID(int idDaughter) { return idDaughter;} 
int isr_Q2QG::sisterID(int)            { return 21;} 
double isr_Q2QG::gaugeFactor ()        { return CF;}
double isr_Q2QG::symmetryFactor ()     { return 1.;}

int isr_Q2QG::radID(int idRA, int){ return idRA;}
pair<int,int> isr_Q2QG::radCols(
  int colRadAfter, int acolRadAfter, 
  int colEmtAfter, int acolEmtAfter) {
  bool isQuark  = (colRadAfter > 0);
  int colRemove = (colRadAfter == colEmtAfter)
                ? colRadAfter : 0;
  int col       = (colRadAfter  == colRemove)
                ? acolEmtAfter : colRadAfter;
  if (isQuark) return make_pair(col,0); 
  colRemove = (acolRadAfter == acolEmtAfter)
                ? acolRadAfter : 0;
  int acol      = (acolRadAfter  == colRemove)
                ? colEmtAfter : acolRadAfter;
  return make_pair(0,acol); 
}

// Pick z for new splitting.
double isr_Q2QG::zSplit(double zMinAbs, double, double m2dip) {
  double Rz = rndmPtr->flat();
  double kappa2 = pow2(settingsPtr->parm("SpaceShower:pTmin"))/m2dip;
  double p = pow( 1. + pow2(1-zMinAbs)/kappa2, Rz );
  double res = 1. - sqrt( p - 1. )*sqrt(kappa2);
  return res;
}

// New overestimates, z-integrated versions.
double isr_Q2QG::overestimateInt(double zMinAbs, double,
  double, double m2dip) {
  double wt = 0.;
  double preFac = symmetryFactor() * gaugeFactor();
  double kappa2 = pow2(settingsPtr->parm("SpaceShower:pTmin"))/m2dip;
  wt  = preFac * softRescaleInt(correctionOrder)
      * 2. * 0.5 * log( 1. + pow2(1.-zMinAbs)/kappa2);
  if (!doSplit("isr_Q2QG_CS")) wt = 0.;

  return wt;
}

// Return overestimate for new splitting.
double isr_Q2QG::overestimateDiff(double z, double m2dip) {
  double wt = 0.;
  double preFac = symmetryFactor() * gaugeFactor();
  double kappaOld2 = pow2(settingsPtr->parm("SpaceShower:pTmin"))/m2dip;
  wt  = preFac * softRescaleInt(correctionOrder)
      * 2.* (1.-z) / ( pow2(1.-z) + kappaOld2);
  return wt;
}

// Return kernel for new splitting.
double isr_Q2QG::kernel(double z, double pT2, double m2dip,
  int, double, double, double, double) {

  double wt = 0.;
  double preFac = symmetryFactor() * gaugeFactor();
  double kappa2 = pT2/m2dip;
  wt   =  preFac * softRescaleDiff(correctionOrder, pT2)
       * ( 2.* (1.-z) / ( pow2(1.-z) + kappa2) );
  wt  += -preFac * (1.+z);

  // Add NLO term, subtracted by ~ 1/(1-z)*Gamma2,
  // since latter already present in soft rescaling term.
  if (correctionOrder == 3) {
    double NF = 6.;
    BeamParticle& beam = (abs(beamAPtr->id()) == 2212) ? *beamAPtr : *beamBPtr;
    if ( pT2 > pow2( max( 0., beam.mQuarkPDF(5) ) )
      && pT2 < pow2(particleDataPtr->m0(6)))               NF = 5.;
    else if ( pT2 > pow2( max( 0., beam.mQuarkPDF(4)) ) )  NF = 4.; 
    else if ( pT2 > pow2( max( 0., beam.mQuarkPDF(3)) ) )  NF = 3.; 
    double TF = TR*NF;
    double alphasPT2pi = 0.5/M_PI * beam.alphaS(pT2);
    double pqq1   = preFac * 1 / ( 18*z*(z-1) ) * (
      (-1 + z)*(-8*TF*(-5 + (-1 + z)*z*(-5 + 14*z)) 
                + z*(90*CF*(-1 + z) + CA*(53 - 187*z + 3*(1 + z)*pow2(M_PI))))
      +3*z*log(z)*(-2*(TF + CF*(-9 + 6*(-1 + z)*z) + TF*z*(12 - z*(9 + 8*z)))
                  + 12*CF*log(1 - z)*(1 + pow2(z)) - CA*(17 + 5*pow2(z)))
      -9*z*(CA - CF - 2*TF + (CA + CF + 2*TF)*pow2(z))*pow2(log(z)));
    // replace 1/z term in NLO kernel with z/(z^2+kappa^2)
    pqq1 += preFac * 20./9.*TF * ( z/(pow2(z)+kappa2) - 1./z); 
    wt  += alphasPT2pi*pqq1;
  }

  if (!doSplit("isr_Q2QG_CS")) wt = 0.;

  return wt;
}

//==========================================================================

// Class inheriting from Splitting class.

// Splitting function G->GG (ISR)

string isr_G2GG1::name()                { return (id.empty()) ? "isr_G2GG1_CS" : id;}
int isr_G2GG1::type()                   { return -201;}
int isr_G2GG1::motherID(int)            { return 21;} 
int isr_G2GG1::sisterID(int)            { return 21;} 
double isr_G2GG1::gaugeFactor ()        { return 2.*CA;}
double isr_G2GG1::symmetryFactor ()     { return 0.5;}

int isr_G2GG1::radID(int, int){ return 21;}
pair<int,int> isr_G2GG1::radCols(
  int colRadAfter, int acolRadAfter, 
  int colEmtAfter, int acolEmtAfter) {
  int colRemove = (colRadAfter == colEmtAfter)
                ? colRadAfter : acolRadAfter;
  int col       = (colRadAfter  == colRemove)
                ? acolEmtAfter : colRadAfter;
  int acol      = (acolRadAfter == colRemove)
                ? colEmtAfter : acolRadAfter;
  return make_pair(col,acol);
}

// Pick z for new splitting.
double isr_G2GG1::zSplit(double zMinAbs, double, double m2dip) {
  double R = rndmPtr->flat();
  double kappa2 = pow2(settingsPtr->parm("SpaceShower:pTmin"))/m2dip;
  // Pick according to soft + 1/z
  double res = (-2.*pow(kappa2,R)*pow(zMinAbs,2.*R) +
             sqrt(4.*pow(kappa2,2.*R)
                   *pow(zMinAbs,4.*R)
                + 4.*(pow(kappa2,R) + pow(kappa2,1. + R))
                   *pow(zMinAbs,2.*R)
                   *(-(pow(kappa2,R)*pow(zMinAbs,2.*R))
                     + kappa2
                       *pow(1. + kappa2 - 2.*zMinAbs + pow(zMinAbs,2.),R))))
          / (2.*(-(pow(kappa2,R)*pow(zMinAbs,2.*R))
                 + kappa2
                   *pow(1. + kappa2 - 2.*zMinAbs + pow(zMinAbs,2.),R)));
  return res;
}

// New overestimates, z-integrated versions.
double isr_G2GG1::overestimateInt(double zMinAbs, double,
  double, double m2dip) {
  double wt = 0.;
  double preFac = symmetryFactor() * gaugeFactor();
  double kappa2 = pow2(settingsPtr->parm("SpaceShower:pTmin"))/m2dip;
  // Overestimate by soft + 1/z
  wt   = preFac * softRescaleInt(correctionOrder)
       * 0.5 * ( log( 1./pow2(zMinAbs) + pow2(1.-zMinAbs)/(kappa2*pow2(zMinAbs)) ) );
  if (!doSplit("isr_G2GG1_CS")) wt = 0.;

  return wt;
}

// Return overestimate for new splitting.
double isr_G2GG1::overestimateDiff(double z, double m2dip) {
  double wt = 0.;
  double preFac = symmetryFactor() * gaugeFactor();
  double kappaOld2 = pow2(settingsPtr->parm("SpaceShower:pTmin"))/m2dip;
  // Overestimate by soft + 1/z
  wt  = preFac * softRescaleInt(correctionOrder)
      * ((1.-z) / ( pow2(1.-z) + kappaOld2) + 1./z);
  return wt;
}

// Return kernel for new splitting.
double isr_G2GG1::kernel(double z, double pT2, double m2dip,
  int splitType, double m2RadBef, double m2Rad, double m2Rec,
  double m2Emt) {

  // Dummy statement to avoid compiler warnings.
  if (false) cout << m2RadBef << m2Rad << m2Emt << endl;

  double wt = 0.;
  double preFac = symmetryFactor() * gaugeFactor();
  double kappa2 = pT2/m2dip;
  wt   = preFac * softRescaleDiff(correctionOrder, pT2)
       * (1.-z) / ( pow2(1.-z) + kappa2);
  wt  += preFac 
       * 0.5 * ( z / ( pow2(z) + kappa2) - 1. );
  wt  += preFac
       * -1.;

  // Correction for massive IF splittings.
  bool doMassive = ( m2Rec > 0. && splitType == 2);

  if (doMassive) {
    // Construct CS variables.
    double uCS = kappa2 / (1-z);
    double massCorr = - m2Rec / m2dip * uCS / (1.-uCS);
    // Mass correction shared in equal parts between both g->gg kernels.
    wt += preFac * 0.5 * massCorr;
  }

  // Add NLO term, subtracted by 1/(1-z)*(Gamma2+beta0*log(z)),
  // since latter already present in soft rescaling term.
  if (!doMassive && correctionOrder == 3) {
    double NF = 6.;
    BeamParticle& beam = (abs(beamAPtr->id()) == 2212) ? *beamAPtr : *beamBPtr;
    if ( pT2 > pow2( max( 0., beam.mQuarkPDF(5) ) )
      && pT2 < pow2(particleDataPtr->m0(6)))               NF = 5.;
    else if ( pT2 > pow2( max( 0., beam.mQuarkPDF(4)) ) )  NF = 4.; 
    else if ( pT2 > pow2( max( 0., beam.mQuarkPDF(3)) ) )  NF = 3.; 

    double alphasPT2pi = 0.5/M_PI * beam.alphaS(pT2);
    double TF = NF*TR;
    // Splitting function directly taken from Mathematica file.
    // Note: After removal of the cusp anomalous dimensions, the NLO kernel
    // is a purely collinear term. As such, it should not distinguish between
    // colour structures, and hence should contribute equally to isr_G2GG1
    // and isr_G2GG2. Hence one factor of 0.5 . Then, another factor of 0.5
    // is necessary, since the NLO kernel in the Mathematica file is normalised
    // to CA, and not 2*CA.
    double pgg1   = preFac * 0.5 / ( 18*z*(pow2(z)-1) ) * 0.5 * (
      TF*(-1 + pow2(z))*((4*(-1 + z)*(-23 + z*(6 + z*(10 + z*(4 + 23*z)))))/(-1 + pow2(z))
                        +(24*(1 - z)*z*log(z)*pow2(1 + z))/(-1 + pow2(z)))
     +(CF*TF*(-1 + pow2(z))*((36*(1 - z)*z*(1 + z)*(3 + 5*z)*log(z))/(-1 + pow2(z))
                            +(24*(1 + z)*(-1 + z*(11 + 5*z))*pow2(-1 + z))/(-1 + pow2(z))
                            -(36*(-1 + z)*z*pow2(1 + z)*pow2(log(z)))/(-1 + pow2(z))))/CA
     -72*CA*(-1 + z)*polylog(2,1/(1 + z))*pow2(1 + z + pow2(z))
     +CA*(-1 + pow2(z))*((6*(1 - z)*z*(1 + z)*(25 + 11*z*(-1 + 4*z))*log(z))/(-1 + pow2(z))
                        +((1 - z)*(z*(1 + z)*(25 + 109*z) + 6*(2 + z*(1 + 2*z*(1 + z)))*pow2(M_PI)))/(-1 + pow2(z))
                        +(72*(1 + z)*log(1 - z)*log(z)*pow2(1 + (-1 + z)*z))/(-1 + pow2(z))
                        -(36*z*pow2(log(z))*pow2(1 + z - pow2(z)))/(-1 + pow2(z))
                        +(144*polylog(2,1/(1 + z))*pow2(1 + z + pow2(z)))/(1 + z)
                        +(36*(-1 + z)*pow2(log(1 + z))*pow2(1 + z + pow2(z)))/(-1 + pow2(z))) );

    // replace 1/x term in NLO kernel with x/(x^2+kappa^2)
    pgg1 += -preFac * 0.5 * 40./9.*TF * 0.5 * ( z/(pow2(z)+kappa2) - 1./z); 
    // Add NLO term.
    wt  += alphasPT2pi*pgg1;
  }

  if (!doSplit("isr_G2GG1_CS")) wt = 0.;

  return wt;
}

//==========================================================================

// Class inheriting from Splitting class.

// Splitting function G->GG (ISR)

string isr_G2GG2::name()                { return (id.empty()) ? "isr_G2GG2_CS" : id;}
int isr_G2GG2::type()                   { return -201;}
int isr_G2GG2::motherID(int)            { return 21;} 
int isr_G2GG2::sisterID(int)            { return 21;} 
double isr_G2GG2::gaugeFactor ()        { return 2.*CA;}
double isr_G2GG2::symmetryFactor ()     { return 0.5;}

int isr_G2GG2::radID(int, int){ return 21;}
pair<int,int> isr_G2GG2::radCols(
  int colRadAfter, int acolRadAfter, 
  int colEmtAfter, int acolEmtAfter) {
  int colRemove = (colRadAfter == colEmtAfter)
                ? colRadAfter : acolRadAfter;
  int col       = (colRadAfter  == colRemove)
                ? acolEmtAfter : colRadAfter;
  int acol      = (acolRadAfter == colRemove)
                ? colEmtAfter : acolRadAfter;
  return make_pair(col,acol);
}

// Pick z for new splitting.
double isr_G2GG2::zSplit(double zMinAbs, double, double m2dip) {
  double R = rndmPtr->flat();
  double kappa2 = pow2(settingsPtr->parm("SpaceShower:pTmin"))/m2dip;
  // Pick according to soft + 1/z
  double res = (-2.*pow(kappa2,R)*pow(zMinAbs,2.*R) +
             sqrt(4.*pow(kappa2,2.*R)
                   *pow(zMinAbs,4.*R)
                + 4.*(pow(kappa2,R) + pow(kappa2,1. + R))
                   *pow(zMinAbs,2.*R)
                   *(-(pow(kappa2,R)*pow(zMinAbs,2.*R))
                     + kappa2
                       *pow(1. + kappa2 - 2.*zMinAbs + pow(zMinAbs,2.),R))))
          / (2.*(-(pow(kappa2,R)*pow(zMinAbs,2.*R))
                 + kappa2
                   *pow(1. + kappa2 - 2.*zMinAbs + pow(zMinAbs,2.),R)));
  return res;
}

// New overestimates, z-integrated versions.
double isr_G2GG2::overestimateInt(double zMinAbs, double,
  double, double m2dip) {
  double wt = 0.;
  double preFac = symmetryFactor() * gaugeFactor();
  double kappa2 = pow2(settingsPtr->parm("SpaceShower:pTmin"))/m2dip;
  // Overestimate by soft + 1/z
  wt   = preFac
       * 0.5 * ( log( 1./pow2(zMinAbs) + pow2(1.-zMinAbs)/(kappa2*pow2(zMinAbs)) ) );
  if (!doSplit("isr_G2GG2_CS")) wt = 0.;

  return wt;
}

// Return overestimate for new splitting.
double isr_G2GG2::overestimateDiff(double z, double m2dip) {
  double wt = 0.;
  double preFac = symmetryFactor() * gaugeFactor();
  double kappaOld2 = pow2(settingsPtr->parm("SpaceShower:pTmin"))/m2dip;
  // Overestimate by soft + 1/z
  wt  = preFac
      * ((1.-z) / ( pow2(1.-z) + kappaOld2) + 1./z);
  return wt;
}

// Return kernel for new splitting.
double isr_G2GG2::kernel(double z, double pT2, double m2dip,
  int splitType, double m2RadBef, double m2Rad, double m2Rec,
  double m2Emt) {

  // Dummy statement to avoid compiler warnings.
  if (false) cout << m2RadBef << m2Rad << m2Emt << endl;

  double wt = 0.;
  double preFac = symmetryFactor() * gaugeFactor();
  double kappa2 = pT2/m2dip;
  wt   = preFac 
       * 0.5 * ( z / ( pow2(z) + kappa2) - 1. );
  wt  += preFac * z*(1.-z);

  // Correction for massive IF splittings.
  bool doMassive = ( m2Rec > 0. && splitType == 2);

  if (doMassive) {
    // Construct CS variables.
    double uCS = kappa2 / (1-z);
    double massCorr = - m2Rec / m2dip * uCS / (1.-uCS);
    // Mass correction shared in equal parts between both g->gg kernels.
    wt += preFac * 0.5 * massCorr;
  }

  // Add NLO term, subtracted by 1/(1-z)*(Gamma2+beta0*log(z)),
  // since latter already present in soft rescaling term.
  if (!doMassive && correctionOrder == 3) {
    double NF = 6.;
    BeamParticle& beam = (abs(beamAPtr->id()) == 2212) ? *beamAPtr : *beamBPtr;
    if ( pT2 > pow2( max( 0., beam.mQuarkPDF(5) ) )
      && pT2 < pow2(particleDataPtr->m0(6)))               NF = 5.;
    else if ( pT2 > pow2( max( 0., beam.mQuarkPDF(4)) ) )  NF = 4.; 
    else if ( pT2 > pow2( max( 0., beam.mQuarkPDF(3)) ) )  NF = 3.; 

    double alphasPT2pi = 0.5/M_PI * beam.alphaS(pT2);
    double TF = NF*TR;
    // Splitting function directly taken from Mathematica file.
    // Note: After removal of the cusp anomalous dimensions, the NLO kernel
    // is a purely collinear term. As such, it should not distinguish between
    // colour structures, and hence should contribute equally to isr_G2GG1
    // and isr_G2GG2. Hence one factor of 0.5 . Then, another factor of 0.5
    // is necessary, since the NLO kernel in the Mathematica file is normalised
    // to CA, and not 2*CA.
    double pgg1   = preFac * 0.5 / ( 18*z*(pow2(z)-1) ) * 0.5 * (
      TF*(-1 + pow2(z))*((4*(-1 + z)*(-23 + z*(6 + z*(10 + z*(4 + 23*z)))))/(-1 + pow2(z))
                         +(24*(1 - z)*z*log(z)*pow2(1 + z))/(-1 + pow2(z)))
     +(CF*TF*(-1 + pow2(z))*((36*(1 - z)*z*(1 + z)*(3 + 5*z)*log(z))/(-1 + pow2(z))
                            +(24*(1 + z)*(-1 + z*(11 + 5*z))*pow2(-1 + z))/(-1 + pow2(z))
                            -(36*(-1 + z)*z*pow2(1 + z)*pow2(log(z)))/(-1 + pow2(z))))/CA
     -72*CA*(-1 + z)*polylog(2,1/(1 + z))*pow2(1 + z + pow2(z))
     +CA*(-1 + pow2(z))*((6*(1 - z)*z*(1 + z)*(25 + 11*z*(-1 + 4*z))*log(z))/(-1 + pow2(z))
                        +((1 - z)*(z*(1 + z)*(25 + 109*z) + 6*(2 + z*(1 + 2*z*(1 + z)))*pow2(M_PI)))/(-1 + pow2(z))
                        +(72*(1 + z)*log(1 - z)*log(z)*pow2(1 + (-1 + z)*z))/(-1 + pow2(z))
                        -(36*z*pow2(log(z))*pow2(1 + z - pow2(z)))/(-1 + pow2(z))
                        +(144*polylog(2,1/(1 + z))*pow2(1 + z + pow2(z)))/(1 + z)
                        +(36*(-1 + z)*pow2(log(1 + z))*pow2(1 + z + pow2(z)))/(-1 + pow2(z))) );
    // replace 1/z term in NLO kernel with z/(z^2+kappa^2)
    pgg1 += -preFac * 0.5 * 40./9.*TF * 0.5 * ( z/(pow2(z)+kappa2) - 1./z); 

    // Add NLO term.
    wt  += alphasPT2pi*pgg1;
  }

  if (!doSplit("isr_G2GG2_CS")) wt = 0.;

  return wt;
}

//==========================================================================

// Class inheriting from Splitting class.

// Splitting function G->QQ (ISR)

string isr_G2QQ::name()                { return (id.empty()) ? "isr_G2QQ_CS" : id;}
int isr_G2QQ::type()                   { return -101;}
int isr_G2QQ::motherID(int)            { return 21;} 
int isr_G2QQ::sisterID(int idDaughter) { return -idDaughter;} 
double isr_G2QQ::gaugeFactor ()        { return TR;}
double isr_G2QQ::symmetryFactor ()     { return 1.0;}

int isr_G2QQ::radID(int, int idEA){ return -idEA;}
pair<int,int> isr_G2QQ::radCols(
  int colRadAfter, int acolRadAfter, 
  int colEmtAfter, int acolEmtAfter) {
  bool isQuark  = (acolEmtAfter > 0);
  int colRemove = (colRadAfter  == colEmtAfter)
                ? colRadAfter : 0;
  int col       = (colRadAfter  == colRemove)
                ? acolEmtAfter : colRadAfter;
  if (isQuark) return make_pair(col,0);
  colRemove     = (acolRadAfter == acolEmtAfter)
                ? acolRadAfter : 0;
  int acol      = (acolRadAfter == colRemove)
                ? colEmtAfter : acolRadAfter;
  return make_pair(0,acol);
}

// Pick z for new splitting.
double isr_G2QQ::zSplit(double zMinAbs, double zMaxAbs, double) {
  // Note: Combined with PDF ratio, flat overestimate performs
  // better than using the full splitting kernel as overestimate. 
  double res = zMinAbs + rndmPtr->flat() * (zMaxAbs - zMinAbs);
  //double R   = rndmPtr->flat();
  //double res = (
  //(1 + pow(-2 + 4*R - 6*R*zMinAbs + 6*R*pow(zMinAbs,2) - 4*R*pow(zMinAbs,3) + 
  //     sqrt(1 + 4*pow(1 + R*(-1 + zMinAbs)*(2 + zMinAbs*(-1 + 2*zMinAbs)),2)),1./3.) - 
  //   pow(-2 + 4*R - 6*R*zMinAbs + 6*R*pow(zMinAbs,2) - 4*R*pow(zMinAbs,3) + 
  //     sqrt(1 + 4*pow(1 + R*(-1 + zMinAbs)*(2 + zMinAbs*(-1 + 2*zMinAbs)),2)),2./3.))/
  //(2.*pow(-2 + 4*R - 6*R*zMinAbs + 6*R*pow(zMinAbs,2) - 4*R*pow(zMinAbs,3) + 
  //     sqrt(1 + 4*pow(1 + R*(-1 + zMinAbs)*(2 + zMinAbs*(-1 + 2*zMinAbs)),2)),1./3.))
  //);
  return res;
}

// New overestimates, z-integrated versions.
double isr_G2QQ::overestimateInt(double zMinAbs, double zMaxAbs,
  double, double) {
  double wt = 0.;
  double preFac = symmetryFactor() * gaugeFactor();
  // Note: Combined with PDF ratio, flat overestimate performs
  // better than using the full splitting kernel as overestimate. 
  wt  = preFac 
      * 2. * ( zMaxAbs - zMinAbs);
  //wt  = preFac
  //    * ( 2./3. - zMinAbs + pow(zMinAbs,2) - (2*pow(zMinAbs,3))/3.);
  if (!doSplit("isr_G2QQ_CS")) wt = 0.;

  return wt;
}

// Return overestimate for new splitting.
double isr_G2QQ::overestimateDiff(double, double) {
  double wt = 0.;
  double preFac = symmetryFactor() * gaugeFactor();
  // Note: Combined with PDF ratio, flat overestimate performs
  // better than using the full splitting kernel as overestimate. 
  wt = preFac 
     * 2.;
  //wt = preFac
  //   * (pow(1.-z,2.) + pow(z,2.));
  return wt;
}

// Return kernel for new splitting.
double isr_G2QQ::kernel(double z, double pT2, double m2dip,
  int, double, double, double, double) {

  double wt = 0.;
  double preFac = symmetryFactor() * gaugeFactor();
  double kappa2 = pT2 / m2dip;
  wt  = preFac
      * (pow(1.-z,2.) + pow(z,2.));

  if (correctionOrder == 3) {
    BeamParticle& beam = (abs(beamAPtr->id()) == 2212) ? *beamAPtr : *beamBPtr;
    double alphasPT2pi = 0.5/M_PI * beam.alphaS(pT2);
    double pgq1 = preFac * (
      (CF*(4 - 9*z + 4*log(1 - z) + (-1 + 4*z)*log(z)
          -(2*(1 + 2*(-1 + z)*z)*(-15 - 3*(-2 + log(-1 + 1/z))*log(-1 + 1/z) + pow2(M_PI)))/3.
          +(-1 + 2*z)*pow2(log(z)))
      +(2*CA*(20 - 18*z*(1 + 2*z*(1 + z))*polylog(2,1/(1 + z))
             +z*(-18 + (225 - 218*z)*z + pow2(M_PI)*(3 + 6*pow2(z)))
             +3*z*(12*(-1 + z)*z*log(1 - z)
                  +log(z)*(3 + 4*z*(6 + 11*z) - 3*(1 + 2*z)*log(z))
                  +(-3 - 6*(-1 + z)*z)*pow2(log(1 - z))
                  -3*(1 + 2*z*(1 + z))*pow2(log(1 + z)))))/(9.*z))/2. );
    // replace 1/z term in NLO kernel with z/(z^2+kappa^2)
    pgq1 += preFac * 20./9.*CA * ( z/(pow2(z)+kappa2) - 1./z); 
    // Add NLO term.
    wt  += alphasPT2pi*pgq1;
  }

  if (!doSplit("isr_G2QQ_CS")) wt = 0.;

  return wt;
}

//==========================================================================

// Class inheriting from Splitting class.

// Splitting function Q->GQ (ISR)

string isr_Q2GQ::name()                { return (id.empty()) ? "isr_Q2GQ_CS" : id;}
int isr_Q2GQ::type()                   { return -201;}
int isr_Q2GQ::motherID(int)            { return 1;} // Use 1 as dummy 
int isr_Q2GQ::sisterID(int)            { return 1;} // Use 1 as dummy
double isr_Q2GQ::gaugeFactor ()        { return CF;}
double isr_Q2GQ::symmetryFactor ()     { return 0.5;}

int isr_Q2GQ::radID(int, int){ return 21;}
pair<int,int> isr_Q2GQ::radCols(
  int colRadAfter, int acolRadAfter, 
  int colEmtAfter, int acolEmtAfter) {
  int col  = (colRadAfter  > 0) ? colRadAfter  : acolEmtAfter;
  int acol = (acolRadAfter > 0) ? acolRadAfter : colEmtAfter;
  return make_pair(col,acol);
}

// Pick z for new splitting.
double isr_Q2GQ::zSplit(double zMinAbs, double, double) {
  double R = rndmPtr->flat();
  double res = pow(zMinAbs,3./4.)
          / ( pow(1. + R*(-1. + pow(zMinAbs,-3./8.)),2./3.)
             *pow(R - (-1. + R)*pow(zMinAbs,3./8.),2.));
//  double kappaOld2 = pow2(settingsPtr->parm("SpaceShower:pTmin"))/m2dip;
//  double res = sqrt( pow( pow2(zMinAbs) + kappaOld2, R)
//                   * pow( 1.            + kappaOld2, 1.- R) - kappaOld2); 
  return res;
}

// New overestimates, z-integrated versions.
double isr_Q2GQ::overestimateInt(double zMinAbs, double,
  double, double) {
  double wt = 0.;
  double preFac = symmetryFactor() * gaugeFactor();
  wt  = preFac * 2./3. * (8.*(-1. + pow(zMinAbs,-3./8.)));
//  double kappaOld2 = pow2(settingsPtr->parm("SpaceShower:pTmin"))/m2dip;
//  wt = preFac * log( (1+kappaOld2) / (pow2(zMinAbs) + kappaOld2) );
  if (!doSplit("isr_Q2GQ_CS")) wt = 0.;

  return wt;
}

// Return overestimate for new splitting.
double isr_Q2GQ::overestimateDiff(double z, double) {
  double wt = 0.;
  double preFac = symmetryFactor() * gaugeFactor();
  wt  = preFac * 2. / pow(z,11./8.);
//  double kappaOld2 = pow2(settingsPtr->parm("SpaceShower:pTmin"))/m2dip;
//  wt  = preFac * 2. * z / (pow2(z) + kappaOld2);
  return wt;
}

// Return kernel for new splitting.
double isr_Q2GQ::kernel(double z, double pT2, double m2dip,
  int splitType, double m2RadBef, double m2Rad, double m2Rec,
  double m2Emt) {

  // Dummy statement to avoid compiler warnings.
  if (false) cout << m2RadBef << m2Rad << m2Emt << endl;

  double wt = 0.;
  double preFac = symmetryFactor() * gaugeFactor();
  double kappa2 = pT2 / m2dip;
  wt   = preFac
       * ( z + 2.*z / (pow2(z)+kappa2) - 2. );

  // Correction for massive IF splittings.
  bool doMassive = ( m2Rec > 0. && splitType == 2);

  if (doMassive) {
    // Construct CS variables.
    double uCS = kappa2 / (1-z);

    double massCorr = -2. * m2Rec / m2dip * uCS / (1.-uCS);
    // Add correction.
    wt += preFac * massCorr;

  }

  if (!doMassive && correctionOrder == 3) {
    double NF = 6.;
    BeamParticle& beam = (abs(beamAPtr->id()) == 2212) ? *beamAPtr : *beamBPtr;
    if ( pT2 > pow2( max( 0., beam.mQuarkPDF(5) ) )
      && pT2 < pow2(particleDataPtr->m0(6)))               NF = 5.;
    else if ( pT2 > pow2( max( 0., beam.mQuarkPDF(4)) ) )  NF = 4.; 
    else if ( pT2 > pow2( max( 0., beam.mQuarkPDF(3)) ) )  NF = 3.; 
    double alphasPT2pi = 0.5/M_PI * beam.alphaS(pT2);
    double TF = NF*TR;
    double pqg1 = preFac * (
      (-9*CF*z*(5 + 7*z) - 16*TF*(5 + z*(-5 + 4*z))
       +36*CA*(2 + z*(2 + z))*polylog(2,1/(1 + z))
       +2*CA*(9 + z*(19 + z*(37 + 44*z)) - 3*pow2(M_PI)*(2 + pow2(z)))
       +3*(-2*log(1 - z)*(CA*(-22 + (22 - 17*z)*z)
                         +4*TF*(2 + (-2 + z)*z) + 3*CF*(6 + z*(-6 + 5*z))
                         +6*CA*(2 + (-2 + z)*z)*log(z))
       +z*log(z)*(3*CF*(4 + 7*z) - 2*CA*(36 + z*(15 + 8*z))
                 +3*(CF*(-2 + z) + 2*CA*(2 + z))*log(z))
       +6*(CA - CF)*(2 + (-2 + z)*z)*pow2(log(1 - z))
       +6*CA*(2 + z*(2 + z))*pow2(log(1 + z))))/(18.*z) );
    // replace 1/z term in NLO kernel with z/(z^2+kappa^2)
    pqg1 +=  - preFac * 40./9.*TF * ( z/(pow2(z)+kappa2) - 1./z); 
    // Add NLO term.
    wt  += alphasPT2pi*pqg1;
  }

  if (!doSplit("isr_Q2GQ_CS")) wt = 0.;

  return wt;
}

// Splitting function state->state+photon (FSR)
// This is currently just a proxy illustrating how a full post-branching
// state could be used to calculate the splitting kernel.
// All overestimates are recycled from the soft piece of q->qg.

string X2XA::name()                { return (id.empty()) ? "X2XA" : id;}
int X2XA::type()                   { return 101;}
int X2XA::motherID(int idDaughter) { return idDaughter;}
int X2XA::sisterID(int)            { return 22;}
double X2XA::gaugeFactor ()        { return 1.;}
double X2XA::symmetryFactor ()     { return 1.;}

int X2XA::radID(int idRA, int){ return idRA;}
pair<int,int> X2XA::radCols(
  int colRadAfter, int acolRadAfter, 
  int, int) {
  bool isQuark = (colRadAfter > 0);
  if (isQuark) return make_pair(colRadAfter,0);
  return make_pair(0,acolRadAfter);
}

// Pick z for new splitting.
double X2XA::zSplit(double zMinAbs, double, double m2dip) {
  double Rz = rndmPtr->flat();
  double kappa2 = pow2(settingsPtr->parm("TimeShower:pTmin"))/m2dip;
  double p = pow( 1. + pow2(1-zMinAbs)/kappa2, Rz );
  double res = 1. - sqrt( p - 1. )*sqrt(kappa2);
  return res;
}

// New overestimates, z-integrated versions.
double X2XA::overestimateInt(double zMinAbs, double,
  double, double m2dip) {
  double wt = 0.;
  double preFac = symmetryFactor() * gaugeFactor();
  double kappa2 = pow2(settingsPtr->parm("TimeShower:pTmin"))/m2dip;
  wt  = preFac * softRescaleInt(correctionOrder)
      * 2. * 0.5 * log( 1. + pow2(1.-zMinAbs)/kappa2);
  return wt;
}

// Return overestimate for new splitting.
double X2XA::overestimateDiff(double z, double m2dip) {
  double wt = 0.;
  double preFac = symmetryFactor() * gaugeFactor();
  double kappaOld2 = pow2(settingsPtr->parm("TimeShower:pTmin"))/m2dip;
  wt  = preFac * softRescaleInt(correctionOrder)
      * 2.* (1.-z) / ( pow2(1.-z) + kappaOld2);
  return wt;
}

// Return kernel for new splitting.
double X2XA::kernelME( const Event& state, int iRadBef, int iRecBef,
  double z, double pT2, double m2dip, vector<double> aux) {

  double preFac = symmetryFactor() * gaugeFactor();

  // Calculate CS variables.
  //double kappa2 = pT2/m2dip;
  //double yCS    = pT2/m2dip / (1.-z);
  //double zCS    = ( 1. - z - pT2/m2dip - pow2(1.-z) )
  //              / ( 1. - z - pT2/m2dip);
  double RNphi  = aux[0];
  aux.erase(aux.begin());

  // Set up full kinematics after branching.
  Event branched = Event();
  branched.init("(hard process-modified)", particleDataPtr);
  branched.clear();
  branched = fsr->branched( state, iRadBef, iRecBef,
    sisterID(iRadBef), pT2, z, RNphi, aux);

  // We now have access to all pre-branching particles (via state)
  // and to all post-branching particles (via branched)

  // This allows to get the Jacobian of the phase space factorsation.
  double jacobian = 1.;
  if ( state[iRecBef].isFinal())
    jacobian = fsr->jacobian_FF(z,pT2,m2dip,m2dip);
  else
    jacobian = fsr->jacobian_FI(z,pT2,m2dip,m2dip);

  // We can now in principle calculate the kernel.
  double real = 1.;
  double born = 1.;
  double wt = jacobian*real/born;
  // Shower will multiply a alphaS factor, but this is a QED splitting.
  wt       *= preFac * alphaEM.alphaEM(pT2) / alphaS.alphaS(pT2);

  // Done.
  return wt;

}

} // end namespace Pythia8
